(***********************************************************************)
(*                                                                     *)
(*            The Symbolic Communicating Machines Library              *)
(*                                                                     *)
(*                Bertrand Jeannet and Tristan Le Gall                 *)
(*                                                                     *)
(*  Copyright 2008 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file LICENSE.        *)
(*                                                                     *)
(***********************************************************************)

(** SCM Library without Apron: abstract manager *)

open Scm_syn

type environment = unit 

type 'abstract manager = 
    {
      env_var : environment ;
      env_param : environment ;
      env_global : environment ;
      get_env : 'abstract -> environment ;
      
      (* lattice part *)
      bottom:environment  -> 'abstract ;
      top: environment -> 'abstract ;
      is_bottom : 'abstract -> bool ;
      is_top:  'abstract -> bool ;
      is_leq: 'abstract -> 'abstract -> bool ;
      join: 'abstract -> 'abstract -> 'abstract ;
      join_list: 'abstract list  -> 'abstract ;
      meet: 'abstract -> 'abstract -> 'abstract ;
      widening: 'abstract -> 'abstract -> 'abstract ;
      print_abs: Format.formatter -> 'abstract -> unit ;
      print_env :  Format.formatter -> 'abstract -> unit ;
 
  (* change in the environment 
     global warning : save the earth
  *)
      param_to_glob: 'abstract -> 'abstract ;
      var_to_glob: 'abstract -> 'abstract ;
      glob_to_param: 'abstract -> 'abstract ;
      glob_to_var : 'abstract -> 'abstract ;

(* for the non-standard semantics 
   - glob_to_globprime and globprime_to_primebis change the environment
   - p_to_pprime (resp primebis_to_glob) rename the parameters (resp the variables)
*)
      p_to_pprime : 'abstract -> 'abstract ;
      glob_to_globprime :  'abstract -> 'abstract ;
      globprime_to_primebis :  'abstract -> 'abstract ;
      primebis_to_glob :  'abstract -> 'abstract ;

(* depending on the scm analyzed *)
      from_guard : ?nonstd:bool -> Scm_syn.bexpr -> 'abstract ;
      assign_list : ?nonstd:bool -> 'abstract -> (Scm_syn.var * Scm_syn.iexpr) list -> 'abstract ;
      rev_assign_list : ?nonstd:bool -> 'abstract -> (Scm_syn.var * Scm_syn.iexpr) list -> 'abstract ;
      mutable nbqueues : int    ;
      mutable list_param : string list ;
      mutable list_variables: string list ;
      mutable k_forward : int ; (* parameter of the standard widening *)
      mutable k_backward : int ; (* parameter of the standard widening *)
    }
      
type bidon = Bot | Top

module type Bidon =
sig
  type t = bidon
      
  val bottom : t

  val top : t

  val is_bottom : t -> bool

  val is_top : t -> bool
        
  val is_leq : t -> t -> bool
    
  val join  : t -> t -> t

  val join_list : t list -> t 

  val meet : t -> t -> t

  val widening   : t -> t -> t

  val  print : Format.formatter -> t -> unit
   
end

module Bidon  : Bidon =
struct
  type t = bidon

  let bottom = Bot

  let top = Top

  let is_bottom x = (x = Bot)

  let is_top x = (x = Top)

  let is_leq (x1:t) (x2:t) = (x1 <= x2)

  let join (x1:t) (x2:t) = if (x1=Bot) && (x2=Bot) then Bot else Top

  let join_list lx = List.fold_left join Bot lx

  let meet (x1:t) (x2:t) = if (x1=Bot) || (x2=Bot) then Bot else Top

  let widening = join

  let print fmt x = 
    match x with
	Bot -> Format.pp_print_string fmt "Bot"
      | Top -> Format.pp_print_string fmt "Top"

end


let finite_manager ?(k_backward=0)  ?(k_forward=0) () (scm : Scm_syn.global_scm) : bidon manager =
  let env = () in
  {    
    env_var = env ;
    env_param = env ;
    env_global = env ;
    get_env = (fun x -> env);


    param_to_glob =(fun x -> x) ;
    var_to_glob = (fun x -> x);
    glob_to_param =(fun x -> x) ;
    glob_to_var = (fun x -> x);

    p_to_pprime = (fun x -> x);
    glob_to_globprime =  (fun x -> x);
    globprime_to_primebis =  (fun x -> x);
    primebis_to_glob = (fun x -> x);


    bottom = (fun _ ->  Bidon.bottom) ;
    top = (fun _ ->Bidon.top) ;
    is_bottom = Bidon.is_bottom ;
    is_top = Bidon.is_top ;

    is_leq = Bidon.is_leq;
    join = Bidon.join ;
    join_list = Bidon.join_list ;
    meet = Bidon.meet;
    widening = Bidon.widening;
    print_abs = Bidon.print ;
    print_env = (fun fmt e -> ()) ;

    from_guard = (fun ?(nonstd=false) x ->  Bidon.top) ;
    assign_list = (fun  ?(nonstd=false) x l -> x) ;
    rev_assign_list = (fun  ?(nonstd=false) x l -> x) ;
    nbqueues = scm.gb_nb_channels ;
    list_param = List.map (fun (x,t,e) -> x) scm.gb_params;
    list_variables = List.map (fun (x,t,e) -> x) scm.gb_vars;
    k_forward = k_forward ;
    k_backward = k_backward;
  }

