(***********************************************************************)
(*                                                                     *)
(*            The Symbolic Communicating Machines Library              *)
(*                                                                     *)
(*                Bertrand Jeannet and Tristan Le Gall                 *)
(*                                                                     *)
(*  Copyright 2008 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file LICENSE.        *)
(*                                                                     *)
(***********************************************************************)

(** SCM analyzer: main file *)

let version = "0.3 beta"


open Apron
open Format
open StandardSem

(* differents analyseurs *)

type apronPolka = Polka.loose Polka.t
type apronBox = Box.t
type apronOct = Oct.t

module CustomBidon =
  struct
    type tabstract = Scm_manager.bidon

  end

module CustomPolka =
  struct
    type tabstract = apronPolka Abstract1.t 

  end

module CustomBox =
  struct
    type tabstract = apronBox Abstract1.t 

  end

module CustomOct =
  struct
    type tabstract = apronOct Abstract1.t 

  end

type topt_analysis =
    {
      mutable tabstract : string ;
      mutable forward : int ;
      mutable backward : int ;
      mutable threshold : bool ;
      mutable obs : string ;
      mutable widening_start : int ;
      mutable widening_freq : int ;
      mutable print_step : bool ;
      mutable print_state : bool ;
      mutable print_postpre : bool ;
      mutable forward_analysis: bool;
    }

module AnalyzerBidon = Scm_analyzer.Make(CustomBidon)

module AnalyzerPolka = Scm_analyzer.Make(CustomPolka)

module AnalyzerBox = Scm_analyzer.Make(CustomBox)

module AnalyzerOct = Scm_analyzer.Make(CustomOct)

exception Bad_option of string

exception Help

exception Version

let _ =
  let opt_analysis =
    {
      tabstract = "" ;
      forward = 0 ;
      backward = 0 ;
      threshold = false ;
      obs = "" ;
      widening_start = 1 ;
      widening_freq = 1 ;
      print_step = false ;
      print_state = false ;
      print_postpre = false ;
      forward_analysis = true;
    }
  in
  let i = ref 1 in
    try
      begin
	while !i < Array.length Sys.argv do
	  (match Sys.argv.(!i) with
	      "-v" -> raise Version
	    | "-help" -> raise Help
	    | "-threshold" -> opt_analysis.threshold <- true
	    | "-print_step" -> opt_analysis.print_step <- true
	    | "-print_state" -> opt_analysis.print_state <- true
	    | "-print_postpre" -> opt_analysis.print_postpre <- true
	    | "-backward_analysis" -> opt_analysis.forward_analysis <- false
	    | "-forward" ->
		begin
		  if !i < (Array.length Sys.argv) -1
		  then (incr i ; try opt_analysis.forward <- int_of_string Sys.argv.(!i) with Failure "int_of_string" -> failwith "-forward : integer expected")
		  else failwith "-forward : integer expected"
		end
	    | "-backward" ->
		begin
		  if !i < (Array.length Sys.argv) -1
		  then (incr i ; try opt_analysis.backward <- int_of_string Sys.argv.(!i)with Failure "int_of_string" -> failwith "-backward : integer expected")
		  else failwith "-backward : integer expected"
		end
	    | "-widening_start" ->
		begin
		  if !i < (Array.length Sys.argv) -1
		  then (incr i ; try opt_analysis.widening_start <- int_of_string Sys.argv.(!i) with Failure "int_of_string" -> failwith "-widening_start: integer expected")
		  else failwith "-widening_start : integer expected"
		end
	    | "-widening_freq" ->
		begin
		  if !i < (Array.length Sys.argv) -1
		  then (incr i ; try opt_analysis.widening_freq <- int_of_string Sys.argv.(!i) with Failure "int_of_string" -> failwith "-widening_freq : integer expected")
		  else failwith "-widening_freq : integer expected"
		end

(*	    | "-obs" ->
		begin
		  if (!i < (Array.length Sys.argv) -1) 
		  then 
		    begin 
		      incr i ; 
		      let name = Sys.argv.(!i) in
			if Sys.file_exists name
			then
			  opt_analysis.obs <- name
			else
			  failwith ("-obs : unknown file ="^name)
		       end
		  else failwith "-obs : file name is expected expected"
		end
*)
	    | "-apron" ->
		begin
		  if !i < (Array.length Sys.argv) -1
		  then (incr i ; 
			match Sys.argv.(!i) with
			    "polka" -> opt_analysis.tabstract <- "polka"
			  | "oct" -> opt_analysis.tabstract <- "oct"
			  | "box" -> opt_analysis.tabstract <- "box"
			  | _ -> failwith "-apron : 'polka', 'box' or 'oct' expected"
		       )
		  else failwith "-apron : 'polka', 'box' or 'oct' expected"
		end
	    | s -> raise (Bad_option s))
	  ;
	  incr i
	done ;
	Format.printf "Lecture du SCM :@." ;

	let scm_lu = Frontend.parse_lexbuf Format.std_formatter (Lexing.from_channel stdin) in

	let scm_glob = Scm_prod.from_syntree scm_lu in 
	  Scm_syn.check_global_scm scm_glob ;
	  let scm_glob = 
	    if opt_analysis.obs = ""
	    then scm_glob 
	    else 
	      begin
		let ch_fichier = open_in  opt_analysis.obs in (* fichier obs *)
		let obs_lu = Frontend.parse_lexbuf Format.std_formatter (Lexing.from_channel ch_fichier) in
		  close_in ch_fichier ;
		  Format.printf "Observateur lu @.";
		  let obs_glob = Scm_prod.from_syntree obs_lu in 
	
		  let res =  
		    Scm_prod.synchronous_product
		      ~same_variables:true
		      obs_glob 
		      scm_glob
		  in
		    Scm_print.print_global_scm Format.std_formatter res ;
		    res
	      end
	  in
	  let result : bool =
	  match opt_analysis.tabstract with
	      "" -> 
		let manager = 
		  Scm_manager.finite_manager 
		    ~k_forward:opt_analysis.forward 
		    ~k_backward:opt_analysis.backward 
		    () 
		    scm_glob 
		in
		let partition = AnalyzerBidon.AbsSem.create_partition manager
		and lattice = AnalyzerBidon.AbsSem.create_lattice manager

		in
		let reach_or_coreach =
		  AnalyzerBidon.analyze 
		    ~forward:opt_analysis.forward_analysis
		    ~widening_start:opt_analysis.widening_start
		    ~widening_freq:opt_analysis.widening_freq
		    ~print_step:opt_analysis.print_step
		    ~print_state:opt_analysis.print_state
		    ~print_postpre:opt_analysis.print_postpre
		    manager 
		    partition
		    lattice
		    scm_glob
		 in
		  AnalyzerBidon.check_bad_states
		    ~forward:opt_analysis.forward_analysis
		    manager
		    partition
		    lattice
		    scm_glob
		    reach_or_coreach
 



	    | "polka" ->
		let manager = 
		  Scm_manager.apron_manager 		    
		    ~k_forward:opt_analysis.forward 
		    ~k_backward:opt_analysis.backward 
		    ~threshold:opt_analysis.threshold
		    (Polka.manager_alloc_loose ())
		    scm_glob 
		in
		let partition = AnalyzerPolka.AbsSem.create_partition manager
		and lattice = AnalyzerPolka.AbsSem.create_lattice manager

		in
		let reach_or_coreach =
		  AnalyzerPolka.analyze 
		    ~forward:opt_analysis.forward_analysis
		    ~widening_start:opt_analysis.widening_start
		    ~widening_freq:opt_analysis.widening_freq
		    ~print_step:opt_analysis.print_step
		    ~print_state:opt_analysis.print_state
		    ~print_postpre:opt_analysis.print_postpre
		    manager
		    partition
		    lattice
		    scm_glob 
		in
		  AnalyzerPolka.check_bad_states
		    ~forward:opt_analysis.forward_analysis
		    manager
		    partition
		    lattice
		    scm_glob
		    reach_or_coreach



	    | "box" ->
		let manager = 
		  Scm_manager.apron_manager 
		    ~k_forward:opt_analysis.forward 
		    ~k_backward:opt_analysis.backward 
		    ~threshold:opt_analysis.threshold
		    (Box.manager_alloc ())
		    scm_glob 
		in
		let partition = AnalyzerBox.AbsSem.create_partition manager
		and lattice = AnalyzerBox.AbsSem.create_lattice manager
		  
		in
		let reach_or_coreach =
		  AnalyzerBox.analyze 
		    ~forward:opt_analysis.forward_analysis
		    ~widening_start:opt_analysis.widening_start
		    ~widening_freq:opt_analysis.widening_freq
		    ~print_step:opt_analysis.print_step
		    ~print_state:opt_analysis.print_state
		    ~print_postpre:opt_analysis.print_postpre
		    manager
		    partition
		    lattice
		    scm_glob 
		in
		  AnalyzerBox.check_bad_states
		    ~forward:opt_analysis.forward_analysis
		    manager
		    partition
		    lattice
		    scm_glob
		    reach_or_coreach


	    | "oct" ->
		let manager = 
		  Scm_manager.apron_manager 
		    ~k_forward:opt_analysis.forward 
		    ~k_backward:opt_analysis.backward 
		    ~threshold:opt_analysis.threshold
		    (Oct.manager_alloc ())
		    scm_glob 
		in
		let partition = AnalyzerOct.AbsSem.create_partition manager
		and lattice = AnalyzerOct.AbsSem.create_lattice manager

		in
		let reach_or_coreach =
		  AnalyzerOct.analyze 
		    ~forward:opt_analysis.forward_analysis
		    ~widening_start:opt_analysis.widening_start
		    ~widening_freq:opt_analysis.widening_freq
		    ~print_step:opt_analysis.print_step
		    ~print_state:opt_analysis.print_state
		    ~print_postpre:opt_analysis.print_postpre
		    manager
		    partition
		    lattice
		    scm_glob
		in
		  AnalyzerOct.check_bad_states
		    ~forward:opt_analysis.forward_analysis
		    manager
		    partition
		    lattice
		    scm_glob
		    reach_or_coreach
	    | _ -> failwith "cas impossible"
	  in

	    if result then 
	      begin
	      Format.printf "@. Bad states MAY BE REACHABLE - verdict DON'T KNOW @.";
		Format.printf "You should try to increase the number given to -forward and -backward options. @."
	      end
	    else
	      Format.printf "@. Bad states are NOT REACHABLE - verdict OK @." ;
	    Format.printf "@. END of analysis @."
      end
    with
	Bad_option s -> Format.fprintf Format.std_formatter "Unknown option : %s" s
      |	Help -> 
	  Format.pp_print_string
	    Format.std_formatter 
	    "SCM analyzer : option -v, -apron type, -help, -forward k, -backward k, -threshold -backward_analysis @."
      | Version ->
	  Format.printf
	    "SCM analyzer - version %s @." 
	    version
