(***********************************************************************)
(*                                                                     *)
(*            The Symbolic Communicating Machines Library              *)
(*                                                                     *)
(*                Bertrand Jeannet and Tristan Le Gall                 *)
(*                                                                     *)
(*  Copyright 2008 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file LICENSE.        *)
(*                                                                     *)
(***********************************************************************)


(** SCM analyzer:  Non-Standard Semantics *)



open Format
open Scm_syn
open Scm_manager

(*  ********************************************************************** *)
(** 2 Module type for the alphabet *)
(*  ********************************************************************** *)

module type Alphabet = StandardSem.Alphabet

module type S = StandardSem.S


(*  ********************************************************************** *)
(** 2 Auxiliary functions *)
(*  ********************************************************************** *)

(* renaming parameters *)

let rename_var (pref : string) (lparam : string list) (v : string) =  
  if List.mem v lparam
  then v
  else (pref^v)

let rec rename_iexpr (pref : string)(lparam : string list) (expi : iexpr) =
  match expi with
      Scm_syn.Cst _ -> expi
    | Scm_syn.Var v ->  Scm_syn.Var (rename_var pref lparam v)
    | Scm_syn.Unop (op,exp,typ,rd) -> 
	Scm_syn.Unop (op, rename_iexpr pref lparam exp,typ,rd)
    | Scm_syn.Binop(op,exp1,exp2,typ,rd) ->  
	Scm_syn.Binop(op, rename_iexpr pref lparam exp1,rename_iexpr pref lparam exp2,typ,rd)
 

let rename_cons (pref : string) (lparam : string list) (exp1,comp,exp2) = 
  (rename_iexpr pref lparam exp1, comp, rename_iexpr pref lparam exp2)

let rec rename_bexpr (pref : string) (lparam : string list) (expb : bexpr) =
  match expb with
    | TRUE -> TRUE
    | FALSE -> FALSE
    | BRANDOM -> BRANDOM 
    | CONS(c) ->  CONS(rename_cons pref lparam c)
    | AND(b1,b2) -> AND(rename_bexpr pref lparam b1, rename_bexpr pref lparam b2)
    | OR(b1,b2) -> OR(rename_bexpr pref lparam b1, rename_bexpr pref lparam b2)
    | NOT(b) -> NOT(rename_bexpr pref lparam b)


let rename_iexpr_opt  (pref : string) (lparam : string list) (exp1 : iexpr option) =
  match exp1 with
      None -> None
    | Some expr1 -> Some (rename_iexpr pref lparam expr1)



let rename_guard (man: 'a Scm_manager.manager) (guard :  bexpr) =
  rename_bexpr "Prime_" man.list_param guard
    (* transforme p -> Prime_p dans les gardes pour tt parametre p *)


(* rename assignment transforms an assignement to something compatible with the non-standard semantics *)
let rename_assignment  (man: 'a Scm_manager.manager) (assignment : (var *iexpr) list) =
  let assignement_prime =
    List.map (fun (v,exp) -> (rename_var "Prime_" [] v, rename_iexpr "Prime_" man.list_param exp)) assignment
      (* transforme v -> Prime_v dans les affectations, et  p -> Prime_p *)
  and list_var_assigned = 
    List.map (fun (v,exp) -> v) assignment
  in
    List.fold_left 
      (fun l_res v -> if not (List.mem v list_var_assigned) then ("Prime_"^v,Scm_syn.Var v)::l_res else l_res) 
      assignement_prime 
      man.list_variables
    (* ajoute des assignations Prime_v = v pour toutes les variables non assignes *)


(*  ********************************************************************** *)
(** 2 Functor *)
(*  ********************************************************************** *)


module Make 
  (Alpha : Alphabet)
  : (S with module Alpha = Alpha)
  = 
struct

  module Alpha = Alpha


  type talphabet = Sep | AlphaLetter of Alpha.t


  let cmp_talphabet x y = 
    match (x,y) with
	(Sep,Sep) -> 0
      | (Sep,_) -> -1
      | (_,Sep) -> 1
      | (AlphaLetter ax, AlphaLetter ay) -> Alpha.compare ax ay      

  module AlphaSep =
    PLattice.MakeAlphabet (struct type t = talphabet let compare = cmp_talphabet  let sep = Sep end)


  let print_alphasep fmt x =
    match x with
	Sep -> Format.pp_print_string fmt "##" 
      | AlphaLetter c -> Alpha.print fmt c



  module PL = PLattice.Make (AlphaSep)

  module Langage = LAutomaton.Make(PL)

  module NQueues = StackQueueNRel.Make(Langage)


  type 'abstract t = Bottom | Top | Value of 'abstract * ('abstract NQueues.t)

  let create_partition 
      (man : 'abstract manager)
      : 'abstract PL.t 
      =
    let param_list = man.list_param in
    List.fold_left
      (fun map_res s -> PL.A.Map.add (AlphaLetter (Alpha.from_string s)) (man.top man.env_global) map_res)
      (PL.A.Map.add Sep (man.top man.env_global) PL.A.Map.empty)
      param_list

  let create_lattice (man : 'abstract manager) : 'abstract PL.lattice = 
    { 
      PLattice.sep = man.top man.env_global;
      PLattice.is_bottom = (fun c a -> man.is_bottom a) ;
      PLattice.is_leq = (fun c a1 a2 -> man.is_leq a1 a2) ;
      PLattice.is_eq = (fun c a1 a2 -> (man.is_leq a1 a2)&& (man.is_leq a2 a1)) ;
      PLattice.join = (fun c a1 a2 -> man.join a1 a2) ;
      PLattice.meet = (fun c a1 a2 -> man.meet a1 a2) ;
      PLattice.widening = (fun c a1 a2 -> man.widening a1 a2) ;
    }

  let get_partition auto =
    failwith ""

  let get_lattice auto =
    failwith ""

  let bottom man partition lattice = Bottom
    
  let top man partition lattice = Top

  let is_bottom (man : 'abstract manager) x = 
    match x with
	Bottom -> true
      | _ -> false

  let is_top (man : 'abstract manager) x = 
    match x with
	Top -> true
      | _ -> false


  let is_leq (man : 'abstract manager) (abs1 :'abstract t) (abs2 :'abstract t) : bool =
    match (abs1,abs2) with
	(Bottom, _) -> true
      | (_,Top) -> true
      | (Value(absv1,lang1),Value(absv2,lang2)) -> 
	  begin
	    let b1 = (man.is_leq absv1 absv2) 
	    and b2 =(NQueues.is_leq lang1 lang2)
	    in
(* DEBUG	      Format.printf "abs_leq : %b lang_leq : %b @." b1 b2 ; *)
	      b1 && b2
	    end
      |_ -> false

  let is_prefix (man : 'abstract manager) (w :string list) (abs :'abstract t) (i:int) : bool =
    failwith "NonstandardSemantics : is_prefix is NOT IMPLEMENTED"


  let init_abstr (man : 'abstract manager) partition lattice aff_list =
    let abst = man.glob_to_var (man.assign_list (man.top man.env_global) aff_list) in
    let lang = NQueues.allempty lattice partition man.nbqueues in
(* DEBUG      NQueues.print (PL.print print_alphasep man.print_env) Format.std_formatter lang ; *)
      Value (abst,lang )

 let init_bad_state (man : 'abstract manager) partition lattice bexpr regexpr =

    let abst = man.glob_to_var (man.from_guard bexpr)  in
    let lang = NQueues.top lattice partition man.nbqueues in
      Value (abst,lang) 

 let elim_queues (abs: 'abstract t) : 'abstract =
   match abs with
       Value (a,_) -> a
     | _ -> failwith "undefined"
 

 let complementary  (man : 'abstract manager) (abs :'abstract t) : 'abstract t =
    failwith "NonstandardSemantics : complementary is NOT IMPLEMENTED"

 let determinise  (man : 'abstract manager) (abs :'abstract t) : 'abstract t =
    failwith "NonstandardSemantics : determinise is NOT IMPLEMENTED"

 let minimise  (man : 'abstract manager) (abs :'abstract t) : 'abstract t =
    failwith "NonstandardSemantics : minimise is NOT IMPLEMENTED"


  let join (man : 'abstract manager) (abs1 :'abstract t) (abs2 :'abstract t) :'abstract t =
    match (abs1,abs2) with
	(Bottom, _) -> abs2
      | (_,Bottom) -> abs1
      | (_,Top) -> Top
      | (Top,_) -> Top
      | (Value(absv1,lang1),Value(absv2,lang2)) -> 
	  begin
	    assert ((man.get_env absv1) = man.env_var) ;
	    assert ((man.get_env absv2) = man.env_var) ;
	    let lang_new =  NQueues.join lang1 lang2 in
	      (* DEBUG
		 Format.printf 
		 "avant union @. L1 = @[%a@] @. L2 = @[%a@]@." 
		 (NQueues.print (PL.print  print_alphasep man.print_abs)) lang1
		 (NQueues.print (PL.print  print_alphasep man.print_abs)) lang2;
		 Format.printf 
		 "apres union @[%a@] @." 
		(NQueues.print (PL.print  print_alphasep man.print_abs)) lang_new;
	      *)
	      Value (man.join absv1 absv2, lang_new)
	  end

  let join_list (man : 'abstract manager)lx =
    List.fold_left 
      (join man)
      (List.hd lx) 
      (List.tl lx)

  let meet (man : 'abstract manager)(abs1 :'abstract t) (abs2 :'abstract t) :'abstract t =
    match (abs1,abs2) with
	(Bottom, _) -> Bottom
      | (_,Bottom) -> Bottom
      | (_,Top) -> abs1
      | (Top,_) -> abs2
      | (Value(absv1,lang1),Value(absv2,lang2)) -> 
	  begin
	    let absv_new = man.meet absv1 absv2
	    and lang_new = NQueues.meet lang1 lang2
	    in
	      if (man.is_bottom absv_new) || (NQueues.is_bottom lang_new)
	      then Bottom
	      else Value (absv_new,lang_new)
	  end


  let widening  ?(initial=true)  ?(final=true) (man : 'abstract manager)(abs1 :'abstract t) (abs2 :'abstract t) :'abstract t =
    if is_leq man abs1 abs2
    then 
      begin
	match (abs1,abs2) with
	    (Bottom, _) -> abs2
	  | (_,Top) -> Top
	  | (Value(absv1,lang1),Value(absv2,lang2)) -> 
	      Value (man.widening absv1 absv2, 
		    NQueues.widening_shape 
		      (Langage.Abstract.standard  ~initial:initial ~final:final ~backward:man.k_backward ~forward:man.k_forward) 
		      lang1
		      lang2
		    )
	  | _ -> failwith "Widening NonStandardSem : impossible case"
      end
    else  failwith "Widening NonStandardSem : inclusion test failed"

  let extrapolation ?(initial=true)  ?(final=true) ~(k_forward:int) ~(k_backward:int) (abs :'abstract t) :'abstract t =
    failwith "Non-standart semantics: Extrapolation is NON IMPLEMENTED"

  let extrapolation_alt ?(initial=true)  ?(final=true) ~(k_forward:int) ~(k_backward:int) (abs :'abstract t) :'abstract t =
    failwith "Non-standart semantics: Extrapolation_alt is NON IMPLEMENTED"


 let widening_alt  ?(initial=true)  ?(final=true) (man : 'abstract manager)(abs1 :'abstract t) (abs2 :'abstract t) :'abstract t =
    if is_leq man abs1 abs2
    then 
      begin
	match (abs1,abs2) with
	    (Bottom, _) -> abs2
	  | (_,Top) -> Top
	  | (Value(absv1,lang1),Value(absv2,lang2)) -> 
	      Value (man.widening absv1 absv2, 
		    NQueues.widening_shape 
		      (Langage.Abstract.bounded_languages ~backward:man.k_backward ~forward:man.k_forward) 
		      lang1
		      lang2
		    )
	  | _ -> failwith "Widening NonStandardSem : impossible case"
      end
    else  failwith "Widening NonStandardSem : inclusion test failed"


(* output et input *)


  let output_trans 
      (man : 'abstract manager)
      (garde : bexpr) 
      (message : Alpha.t)
      (channel : int)
      (affectations : (var * iexpr) list)
      (ancienne_valeur :'abstract t) 
      :'abstract t 
      =

    match ancienne_valeur with
	Bottom -> Bottom
      | Top -> Top
      | Value (absv_old,lang_old) -> 
	begin
	  let absv_g = 
	    man.meet 
	      (man.glob_to_globprime (man.var_to_glob absv_old)) 
	      (man.glob_to_globprime (man.from_guard ~nonstd:true (rename_guard man garde)))
	      (* on renomme la garde et on travaille dans l'environement qui contient tout *) 
	  in
	    match man.is_bottom absv_g with
		true -> Bottom
	      | _ -> 
		  begin	  
		    let absv_param = man.meet (man.var_to_glob absv_old) (man.from_guard garde) in
		      (* pour le parametre, on refait le meet mais avec les "vrais" parametres *)

		    let plItem = 
		      AlphaSep.Map.add 
			(AlphaLetter message)  
			absv_param
			AlphaSep.Map.empty
		    in
		      (* partition contenant le message + la valeur des parametres *)

		    let absv_new =
		      man.glob_to_var (man.assign_list absv_param affectations)
			(* valeur apres les affectations, avec les "vraies" gardes *)
		    and lang_new = NQueues.Right.queue lang_old channel plItem
		    in	
		      (* reetiquetage de toutes les transitions *)
		    let absv_aff = 
		      man.assign_list ~nonstd:true absv_g (rename_assignment man affectations)
			(* absv_aff : predicat v'= A(v,p') ^ G(v,p') *)
		    in
		    let lang_new = 
		      NQueues.map_trans 
			(fun pl -> 
			  PL.map 
			    (fun absv -> 
			      begin
				let phi = man.meet (man.glob_to_globprime absv) absv_aff in
				  (* phi defini comme dans le papier SAS *)
				let phi_prime = man.globprime_to_primebis phi in
				  (* quantification existentielle *)
				  man.primebis_to_glob phi_prime
				    (* renommage des variables *)
			      end
			    )
			    pl
			)
			lang_new
			
		    in
		      if NQueues.is_bottom lang_new
		      then Bottom
		      else Value(absv_new, lang_new)
		  end
	end

  let pre_output_trans 
      (man : 'abstract manager)
      (garde : bexpr) 
      (message : Alpha.t)
      (channel : int)
      (affectations : (var * iexpr) list)
      (ancienne_valeur :'abstract t) 
      :'abstract t 
      =
    failwith "PRE IS NOT IMPLEMENTED "





  let input_trans 
     (man : 'abstract manager)
     (garde : bexpr) 
     (message : Alpha.t)
     (channel : int)
     (affectations : (var * iexpr) list)
     (ancienne_valeur :'abstract t) 
     :'abstract t 
      =

    match ancienne_valeur with
	Bottom -> Bottom
      | Top -> Top
      | Value (absv_old,lang_old) -> 
	begin
	  let absv_g = man.meet (man.var_to_glob absv_old) (man.from_guard garde) in
	  let (lang_new,param) = 
	    begin
	      try AlphaSep.Map.find (AlphaLetter message) (NQueues.Right.dequeue_partitioned lang_old channel) 
	      with Not_found -> (NQueues.bottom (NQueues.get_lattice lang_old) (NQueues.get_partition lang_old) man.nbqueues, man.bottom man.env_param)
		(* cherche le message dans la table, bottom sinon *)
	    end 
	  in
	  let absv_param = 
	    man.meet absv_g (man.param_to_glob param) 
	      (* intersection garde et valeur du paramettre *)
	  in 
	    match man.is_bottom absv_param with
		true -> Bottom
	      | _ -> 
		  begin
		    let absv_new =
		      man.glob_to_var (man.assign_list absv_param affectations)
			(* valeur apres les affectations *)
		    in

		    let absv_aff = 
		      let absv_prime = man.glob_to_globprime (man.p_to_pprime absv_param) in
		      man.assign_list ~nonstd:true absv_prime (rename_assignment man affectations)
			(* absv_aff : predicat v'= A(v,p') ^ G(v,p') *)
		    in
		    let lang_new = 
		      NQueues.map_trans 
			(fun pl -> 
			  PL.map 
			    (fun absv -> 
			      begin
				let phi = man.meet (man.glob_to_globprime absv) absv_aff in
				  (* phi defini comme dans le papier SAS *)
				let phi_prime = man.globprime_to_primebis phi in
				  (* quantification existentielle *)
				  man.primebis_to_glob phi_prime
				    (* renommage des variables *)
			      end
			    )
			    pl
			)
			lang_new
			
		    in


		      Value(absv_new, lang_new)		    		    
		  end
	end

  let pre_input_trans 
      (man : 'abstract manager)
      (garde : bexpr) 
      (message : Alpha.t)
      (channel : int)
      (affectations : (var * iexpr) list)
      (ancienne_valeur :'abstract t) 
      :'abstract t 
      =
    failwith "PRE IS NOT IMPLEMENTED "


  let epsilon_trans 
      (man : 'abstract manager)
      (garde : bexpr) 
      (affectations : (var * iexpr) list)
      (ancienne_valeur :'abstract t) 
      :'abstract t 
      = 
    match ancienne_valeur with
	Bottom -> Bottom
      | Top -> Top
      | Value (absv_old,lang_old) -> 
	begin
	  let absv_g = man.meet (man.var_to_glob absv_old) (man.from_guard garde) in

	    match man.is_bottom absv_g with
		true -> Bottom
	      | _ -> 
		  begin	  
		    let absv_new =
		      man.glob_to_var (man.assign_list absv_g affectations)
			(* valeur apres les affectations *)
		    in
		    let absv_aff = 
		      let absv_prime = man.glob_to_globprime (man.p_to_pprime absv_g) in
		      man.assign_list ~nonstd:true absv_prime (rename_assignment man affectations)
			(* absv_aff : predicat v'= A(v,p') ^ G(v,p') *)

		    in
		    let lang_new = 
		      NQueues.map_trans 
			(fun pl -> 
			  PL.map 
			    (fun absv -> 
			      begin
				let phi = man.meet (man.glob_to_globprime absv) absv_aff in
				  (* phi defini comme dans le papier SAS *)
				let phi_prime = man.globprime_to_primebis phi in
				  (* quantification existentielle *)
				  man.primebis_to_glob phi_prime
				    (* renommage des variables *)
			      end
			    )
			    pl
			)
			lang_old
		    in	

		      Value(absv_new, lang_new)
		  end
	end

 let pre_epsilon_trans 
      (man : 'abstract manager)
      (garde : bexpr) 
      (affectations : (var * iexpr) list)
      (ancienne_valeur :'abstract t) 
      :'abstract t 
      = 
   failwith "PRE IS NOT IMPLEMENTED "

(* print function *)

  let print(man : 'abstract manager) fmt x =
    match x with 
	Bottom ->
	  begin
	    pp_print_newline fmt ();
	    pp_print_string fmt "########## Empty State ########### " ;
	    pp_print_newline fmt ()
	  end
      | Top ->
	  begin
	    pp_print_newline fmt ();
	    pp_print_string fmt "########## Top tOp toP ########### " ;
	    pp_print_newline fmt ()
	  end
      | Value(abs,lang) ->
	  begin	    
	    pp_print_string fmt "Constaints on Variables : " ;
	    man.print_abs fmt abs ;
	    pp_print_newline fmt ();
	    man.print_env fmt abs ;
	    pp_print_newline fmt ();
	    pp_print_string fmt "Queues : " ;
	    NQueues.print (PL.print  print_alphasep man.print_abs) fmt lang
	  end

  let print_dot ?margin ?titlestyle ?vertexstyle ?edgestyle ?title (man : 'abstract manager) fmt x =
    match x with
      | Bottom -> ()
      | Top -> ()
      | Value(abs,lang) ->
	  NQueues.print_dot 
	    ?margin 
	    ?titlestyle 
	    ?vertexstyle
	    ?edgestyle
	    ?title
	    (PL.print  print_alphasep man.print_abs) 
	    fmt 
	    lang


  let print_regexp (man : 'abstract manager) fmt x = failwith "not implemented"

  let get_automata_size x =
      failwith "not implemented"

  
 
end

