(***********************************************************************)
(*                                                                     *)
(*            The Symbolic Communicating Machines Library              *)
(*                                                                     *)
(*                Bertrand Jeannet and Tristan Le Gall                 *)
(*                                                                     *)
(*  Copyright 2008 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file LICENSE.        *)
(*                                                                     *)
(***********************************************************************)

(** SCM analyzer: analyzer module *)


open Format
open Scm_syn

let debugdotfmt =
  let file = open_out "t.dot" in
    Format.formatter_of_out_channel file

module type Alpha =
sig
  type t
  val compare : t -> t -> int
  val from_string : string -> t
  val print : Format.formatter -> t -> unit
end

module type Custom =
sig
  type tabstract

end

module type S = sig
  module Custom : Custom
  module AbsSem : StandardSem.S

  val analyze : 
    ?forward:bool->
    ?widening_start:int ->
    ?widening_freq:int ->
    ?print_step:bool ->
    ?print_state:bool ->
    ?print_postpre:bool ->
    Custom.tabstract Scm_manager.manager ->
    Custom.tabstract AbsSem.PL.t ->
    Custom.tabstract AbsSem.PL.lattice->
    Scm_syn.global_scm  ->  
    (string,Custom.tabstract AbsSem.t) Mappe.t 

  val check_bad_states :
    ?forward:bool->
    Custom.tabstract Scm_manager.manager ->
    Custom.tabstract AbsSem.PL.t ->
    Custom.tabstract AbsSem.PL.lattice->
    Scm_syn.global_scm  -> 
    (string,Custom.tabstract AbsSem.t) Mappe.t  ->
    bool
 
  end

module Make (Custom:Custom) =
  struct

    module Custom = Custom
    module Alpha =
    struct
      type t = string
      let compare = compare
      let from_string x = x
      let print = Format.pp_print_string
    end 
  
  
    module AbsSem = 
	StandardSem.Make (Alpha) 

    module SetteI = Sette.Make(struct 
      type t=int
      let compare x y = 
	if x<y then -1
	else if x>y then 1
	else 0
    end)

    module HashheI = Hashhe.Make(struct
      type t=int
      let equal = (==)
      let hash x = x
    end)

(* ancienne implmentation
    module T = struct
      type vertex=int
      type hedge=int
      let vertex_dummy=min_int
      let hedge_dummy=min_int
      module SetV=SetteI
      module SetH=SetteI
      module HashV=HashheI
      module HashH=HashheI
    end

    module Graph = SHGraph.Make(T)

*)
    let pp_print_unit fmt () = pp_print_string fmt "()"


	     
    (* SUPPOSITION: le SCM en entre doit satisfaire un certains
       nombre de conditions, qui sont vrifies par la fonction
       Scm_syn.check_gscm *)

      
    let analyze 
	?(forward=true)
	?(widening_start=2)
	?(widening_freq=1)
	?(print_step=false)
	?(print_state=false)
	?(print_postpre=false)
	(manager_abstr : Custom.tabstract Scm_manager.manager) 
	(partition :Custom.tabstract AbsSem.PL.t)
	(lattice:Custom.tabstract AbsSem.PL.lattice)
	(gscm : Scm_syn.global_scm) 
	= 
      (* Verifie le SCM *)
      Scm_syn.check_global_scm gscm;


      let nb_states = List.length gscm.gb_lstates in
	Format.printf "NB STATES: %d @."  nb_states  ;
	let (g:(int,int,unit,unit,unit) PSHGraph.t) = 
	  PSHGraph.create 
	    {
	      PSHGraph.hashv = Hashhe.stdcompare;
	      PSHGraph.hashh = Hashhe.stdcompare;
	      PSHGraph.comparev = compare;
	      PSHGraph.compareh = compare;
	    } 
	    nb_states 
	    () 
	in
	let empty_set : int PSette.t = PSette.empty g.PSHGraph.compare.PSHGraph.comparev in

	let map_states : (int,string) Mappe.t = 
	  let cmpt = ref (-1) in
	  let f_fold res (s:Scm_syn.state) : (int,string) Mappe.t =
	    incr cmpt ; 
	    PSHGraph.add_vertex g !cmpt ();
	    (* IMPORTANT : cre aussi le noeud dans le graphe *)
	    Mappe.add !cmpt s.id_state res 
	  in
	    List.fold_left f_fold Mappe.empty gscm.gb_lstates
	in

	  (* gives the number of in a map *)
	let find_number (s:string) =
	  let res = ref (-1) in
	    Mappe.iter (fun nx sx -> if sx = s then res := nx) map_states;
	    assert (!res >=0);
	    !res
	in

	let map_trans : (int,int*int) Mappe.t =
	  begin	  
	    let cmpt = ref (-1) 
	    and res = ref Mappe.empty
	    in
	    let f_iter_trans (idx:int) (t:Scm_syn.transition) =
	      let idy = find_number t.new_state in
		if forward 
		then (* ***forward analysis*** *)
		  begin
		    if Mappe.fold (fun  n (x,y) b -> b || (idx=x && idy=y)) !res false
		      (*search whether (idx,idy) is already in the map *)
		    then
		      () (* nothing to do *)
		    else
		      begin
			incr cmpt ;
			PSHGraph.add_hedge g !cmpt () ~pred:[|idx|] ~succ:[|idy|];
			(* IMPORTANT : cre aussi l'arrte dans le graphe *)
			res := Mappe.add !cmpt (idx,idy) !res;
			Format.printf 
			  "new_transition: %d -> [%d,@[%a@],%d]@.@." 
			  !cmpt idx Scm_print.print_action t.action idy
		      end
		  end
		else (* ***backward analysis*** *)
		  begin
		    if Mappe.fold (fun  n (y,x) b -> b || (idx=x && idy=y)) !res false
		      (*search whether (idy,idx) is already in the map *)
		    then
		      () (* do nothing *)
		    else
		      begin
			incr cmpt ;
			PSHGraph.add_hedge g !cmpt () ~pred:[|idy|] ~succ:[|idx|];
			(* IMPORTANT : cre aussi l'arrte dans le graphe *)
			res := Mappe.add !cmpt (idy,idx) !res;
			Format.printf 
			  "new_transition: %d -> [%d,pre(@[%a@]),%d]@.@." 
			  !cmpt idy Scm_print.print_action t.action idx
		      end
		  end

	    in
	    let f_iter_state (s:Scm_syn.state) =
	      let idx = find_number s.id_state in
		List.iter (f_iter_trans idx) s.ltrans

	    in
	      List.iter f_iter_state gscm.gb_lstates;
	      !res
	  end
	in
	  (* ICI : tous les etats et toutes les transitions du graphe sont crees *)



	(* fonctions associants les numeros aux etats/transitions *)
	let get_state (i:int) : string = 
	  try Mappe.find i map_states 
	  with Not_found -> failwith "get_state : unknown state" 


	in 


	  (* DEBUG : impression de la table des tats *)
	  Format.printf 
	    "corresp etats :  @. @[ %a @] @." 
	    (Mappe.print Format.pp_print_int Format.pp_print_string)
	    map_states
	  ;

	  (*Format.printf "Table transisions :@. @[%a@]@."
	    (Mappe.print  
	       Format.pp_print_int 
	       (fun fmt (x,y) -> Format.fprintf fmt "(%d,@[%a@], %d)" x y)
	    ) 
	    map_trans;*)
	  
	  (* let get_trans (i :int) : (string * string) = try begin
	     match Mappe.find i map_trans with (x,y) -> (get_state x,
	     get_state y) end with Not_found -> failwith "get_trans : unknown
	     trans" in *)

	  (* tab_hedge_trans associe a chaque hyperedge les transitions tirables *)
	  
	  let tab_hedge_trans : Scm_syn.transition list array =
	    Array.make (nb_states*nb_states) []

	  in
	    (* rajoute la transition dans la bonne case du tableau *)
	  let f_iter_trans  (i:int)(name2:string) (trans : Scm_syn.transition) =
		if trans.new_state = name2
		then
		  tab_hedge_trans.(i) <- trans::tab_hedge_trans.(i)

	  in

	  let f_iter_state (i:int) (name1:string) (name2:string) (state:Scm_syn.state) =
		if state.id_state = name1
		then
		  List.iter (f_iter_trans i name2) state.ltrans

	  in
	    for i=0 to (nb_states*nb_states) -1 do
	      try
		begin
		  let (id1,id2) = Mappe.find i map_trans in
		  let name1 = 
		    try Mappe.find id1 map_states 
		    with Not_found -> failwith "state unknown in map_apply"
		  and name2 = 
		    try Mappe.find id2 map_states 
		    with Not_found -> failwith "state unknown in map_apply"
		  in
		    if forward
		    then
		      List.iter (f_iter_state i name1 name2) gscm.gb_lstates
		    else
		      List.iter (f_iter_state i name2 name1) gscm.gb_lstates
		end
	      with Not_found -> ()
	    done ;
	    (* tab_hedge_trans contient maintenant toutes les transitions associees a chaque hedge *)
	    
	    (* choix des fonctions de transitions*)
	    let epsilon_trans = if forward then AbsSem.epsilon_trans else AbsSem.pre_epsilon_trans
	    and output_trans = if forward then AbsSem.output_trans else  AbsSem.pre_output_trans
	    and input_trans = if forward then AbsSem.input_trans else AbsSem.pre_input_trans
	    in

	    (* fonction auxiliaire pour dfinir le Manager.apply *)
	    let f_fold_trans (x : Custom.tabstract AbsSem.t) absv_res trans =
	      begin
		let absv_add =
		  match trans.action with
		      None -> AbsSem.epsilon_trans manager_abstr trans.guard trans.assignement x
		    | Some (nbq,act,p) ->
			begin
			  match act with
			      SEND -> 
				begin
				  if List.mem nbq gscm.gb_lossy_list
				  then
				    AbsSem.join 
				      manager_abstr	
				      (epsilon_trans  
					 manager_abstr 
					 trans.guard 
					 trans.assignement 
					 x)
				      (output_trans  
					 manager_abstr 
					 trans.guard 
					 (Alpha.from_string p)
					 nbq 
					 trans.assignement 
					 x)
				      (* if nbq is a lossy channel we
					 consider both cases : the message
					 is sent or not *)
				  else
				    output_trans  
				      manager_abstr 
				      trans.guard 
				      (Alpha.from_string p)
				      nbq 
				      trans.assignement 
				      x
				end
			    | RECEIVE -> 
				input_trans  
				  manager_abstr 
				  trans.guard 
				  (Alpha.from_string p) 
				  nbq 
				  trans.assignement 
				  x
			end
		in 
		  AbsSem.join manager_abstr absv_res absv_add

	      end


	    in

	    let aff_init_list = 
	      let rec f l =
		match l with
		    [] -> []
		  | (v,t,e)::reste ->
		      begin
			match e with
			    None -> f reste
			  | Some expr -> (v,expr)::(f reste)
		      end
	      in f gscm.gb_vars
	    in

	    let absv_allempty =  AbsSem.init_abstr manager_abstr partition lattice aff_init_list in


	    (* MANAGER *)
	    let (manager:(int,int,Custom.tabstract AbsSem.t,unit) Fixpoint.manager) = {
	      Fixpoint.bottom = (fun v -> AbsSem.bottom manager_abstr partition lattice) ;
	      Fixpoint.canonical = begin fun v abs ->
		if print_postpre then
		  AbsSem.print_dot
		    ~title:(Format.sprintf "vertex %i" v)
		    manager_abstr
		    debugdotfmt
		    abs
		;
	      end;
	      Fixpoint.is_bottom = (fun v a ->  AbsSem.is_bottom manager_abstr a);
	      Fixpoint.is_leq = (fun v a1 a2 -> AbsSem.is_leq manager_abstr a1 a2) ;
	      Fixpoint.join = (fun v a1 a2 -> AbsSem.join manager_abstr a1 a2) ;
	      Fixpoint.join_list = begin fun v la -> 
		match la with
		  | [] -> AbsSem.bottom manager_abstr partition lattice
		  | x::l ->
		      List.fold_left 
			(fun res va -> AbsSem.join manager_abstr va res)
			x l
	      end;
	      Fixpoint.widening = 
		(begin fun vertex x y -> 
		   if print_postpre then begin
		     AbsSem.print_dot ~title:"widening auto1" manager_abstr debugdotfmt x;
		     AbsSem.print_dot ~title:"widening auto2" manager_abstr debugdotfmt y;
		   end;
		   if AbsSem.is_leq manager_abstr x y
		   then
		     begin
		       let res = AbsSem.widening ~initial:true ~final:true manager_abstr x y in
			 if print_postpre then
			   AbsSem.print_dot ~title:"widening res" manager_abstr debugdotfmt res;
			 AbsSem.minimise manager_abstr res 
		     end
		   else 
		     begin
		       Format.printf 
			 "vertex %i \n abs1 = @. @[ %a @] @."
			 vertex
			 (AbsSem.print manager_abstr) x ;
		       Format.printf 
			 "abs2 = @. @[ %a @] @."
			 (AbsSem.print manager_abstr) y ;
		       
		       raise (Invalid_argument "Fixpoint.widening")
		     end
		 end);
	      Fixpoint.apply=(begin fun hedge tx ->
				  (* debug	    Format.printf " hedge : %i @.  " hedge ; *)
				  assert ((Array.length tx) = 1);
				  let x = tx.(0) in
				    ((), 
				     begin
				       let liste_trans = tab_hedge_trans.(hedge) in
					 Print.list Scm_print.print_transition Format.std_formatter liste_trans ;
					 List.fold_left (f_fold_trans x) (AbsSem.bottom manager_abstr partition lattice) liste_trans
				     end)
				end);

	      Fixpoint.arc_init=(fun v -> ());
	      Fixpoint.abstract_init=(begin fun v -> 
				     if forward 
				     then (* *** forward analysis *** *)
				       begin
					 if List.mem (get_state v) gscm.gb_init_states
					 then absv_allempty
					 else failwith ""
				       end
				     else (* *** backward analysis *** *)
				       begin
					 let x= get_state v in  
					 let rec f_init_bad_states l =
					   match l with
					       [] -> failwith ""
					     | (s_bad,abs_bad,regexp_bad)::reste when s_bad = x -> 
						 AbsSem.init_bad_state manager_abstr  partition lattice abs_bad regexp_bad
					     | _::reste -> f_init_bad_states reste   
					 in f_init_bad_states
					      gscm.gb_bad_states
				       end
				   end);
	      Fixpoint.accumulate=true;

	      Fixpoint.print_abstract= (fun fmt x -> AbsSem.print manager_abstr fmt (AbsSem.minimise manager_abstr x));
	      Fixpoint.print_arc =(fun fmt _ -> ()) ;
	      
	      Fixpoint.print_vertex=pp_print_int;
	      Fixpoint.print_hedge=pp_print_int;

	      Fixpoint.widening_first=false;
	      Fixpoint.widening_start= widening_start ;
	      Fixpoint.widening_freq= widening_freq ;
	      Fixpoint.widening_descend=2;
	      
	      Fixpoint.print_fmt = Format.std_formatter;
	      Fixpoint.print_analysis=true;
	      Fixpoint.print_component = false;
	      Fixpoint.print_step= print_step ;
	      Fixpoint.print_state= print_state ;
	      Fixpoint.print_postpre= print_postpre;
	      Fixpoint.print_workingsets= false;

	      Fixpoint.dot_fmt = None;
	      Fixpoint.dot_vertex = Format.pp_print_int; 
	      Fixpoint.dot_hedge = Format.pp_print_int;
	      
	      Fixpoint.dot_attrvertex = (fun x y -> ());
	      Fixpoint.dot_attrhedge = (fun x y -> ());

	       
	    }

	    in
	    let sette_init = (* ensemble des sommets initiaux du graphe *)
		if forward
		then 
		  begin (* ***forward analysis*** *)
		    Mappe.fold 
		      (fun i name sette_res ->
			 if List.mem name gscm.gb_init_states
			 then PSette.add i sette_res
			 else sette_res
		      )
		      map_states
		      empty_set
		  end
		else 
		  begin (* ***backward analysis*** *)
		    Mappe.fold 
		      (fun i name sette_res ->
			 if List.fold_left (fun b (sb,ab,rb)-> b||(name=sb)) false gscm.gb_bad_states
			 then 
			   begin
			     Format.printf "bad state @[%d@]@." i;
			     PSette.add i sette_res
			   end
			 else sette_res
		      )
		      map_states
		      (PSette.empty compare)
		  end
	    in
	      assert (not (PSette.is_empty sette_init));
	      Format.printf "initial nodes of the graph: @[%a@]@." 
		(PSette.print Format.pp_print_int) sette_init;

	      
	      let strat = Fixpoint.make_strategy_default (-1) (-1) g sette_init in
	      printf "strategy=%a@." (Fixpoint.print_strategy manager) strat;


	      let output = 
		Fixpoint.analysis_std
		  manager 
		  g 
		  sette_init
		  strat
	      in

(*

_guided 
		  ~depth:0 
		  ~priority:(fun _->0) 
		  ~modify_strategy:(fun x -> strat)
*)

		printf "postfixpt=%a@." (Fixpoint.print_output manager) output;
		
		PSHGraph.iter_vertex output 
		  (begin fun vertex reach ~pred ~succ ->
		     AbsSem.print_dot 
		       ~title:(Format.sprintf "Final result minimised: vertex %i" vertex)
		       manager_abstr 
		       debugdotfmt 
		       (AbsSem.minimise manager_abstr reach)
		   end)
		;
		PSHGraph.iter_vertex output 
		  (begin fun vertex reach ~pred ~succ ->
		     AbsSem.print_dot 
		       ~title:(Format.sprintf "Final result: vertex %i" vertex)
		       manager_abstr 
		       debugdotfmt 
		       reach
		     
		   end)
		;

		let map_output = ref (Mappe.empty) in
		let f_iter_output_to_mappe vertex reach ~pred ~succ =
		  map_output := Mappe.add (get_state vertex) reach !map_output
		in
		  PSHGraph.iter_vertex output f_iter_output_to_mappe ;
		  !map_output










    let check_bad_states 
	?(forward=true)
	(man:Custom.tabstract Scm_manager.manager)
	(partition :Custom.tabstract AbsSem.PL.t)
	(lattice:Custom.tabstract AbsSem.PL.lattice)
	(scm:Scm_syn.global_scm)

	(map_output : (string,Custom.tabstract AbsSem.t) Mappe.t)
	:
	bool
	=
      (* get the lattice and partition of the bindings preent in map_output *)
      assert (not (Mappe.is_empty map_output));
      let (name, abs)= Mappe.choose map_output in
      let lattice  = AbsSem.get_lattice abs
      and partition = AbsSem.get_partition abs
      in

      if forward
      then 
	begin
	  (* construction of a list of bad states *)
	  let list_bad_states = 
	    let map_bad_states  (s_bad,abs_bad,regexp_bad) =
	      (s_bad,(AbsSem.init_bad_state man  partition lattice abs_bad regexp_bad) )
			
	    in List.map  map_bad_states scm.gb_bad_states
	  in
	    (* checking whether the bad states are reachable *)
	  List.fold_left 
	    (fun b (name,bad) -> 
	       b ||
		 not 
		 (AbsSem.is_bottom
		    man
		    (AbsSem.meet 
		       man
		       bad 
		       (try Mappe.find name map_output 
			with Not_found -> failwith "Error in check_bad_states")
		    )
		 )
	    )
	    false
	    list_bad_states			  
	end
      else
	begin
	  (* construction of a list of initial states *)
	  let aff_init_list = 
	      let rec f l =
		match l with
		    [] -> []
		  | (v,t,e)::reste ->
		      begin
			match e with
			    None -> f reste
			  | Some expr -> (v,expr)::(f reste)
		      end
	      in f scm.gb_vars
	    in

	  let absv_allempty =  AbsSem.init_abstr man partition lattice aff_init_list in


	  let list_initial_states =
	    List.map 
	      (fun s -> (s,absv_allempty))
	      scm.gb_init_states
	  in

	  (* checking whether the initial states are coreachable *)
	    List.fold_left 
	      (fun b (name,bad) -> 
		 b ||
		   not 
		   (AbsSem.is_bottom
		      man
		      (AbsSem.meet
			 man
			 bad 
			 (try Mappe.find name map_output 
			  with Not_found -> failwith "Error in check_bad_states")
		      )
		   )
	      )
	      false
	      list_initial_states	   
	  
	end


	  
  end
