(***********************************************************************)
(*                                                                     *)
(*            The Symbolic Communicating Machines Library              *)
(*                                                                     *)
(*                Bertrand Jeannet and Tristan Le Gall                 *)
(*                                                                     *)
(*  Copyright 2008 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file LICENSE.        *)
(*                                                                     *)
(***********************************************************************)

(** SCM analyzer: abstract manager module *)


open Apron
open Scm_syn

type environment = Environment.t

type 'abstract manager = 
    {
      env_var : environment ;
      env_param : environment ;
      env_global : environment ;
      get_env : 'abstract -> environment ;
      
      (* lattice part *)
      bottom:environment  -> 'abstract ;
      top: environment -> 'abstract ;
      is_bottom : 'abstract -> bool ;
      is_top:  'abstract -> bool ;
      is_leq: 'abstract -> 'abstract -> bool ;
      join: 'abstract -> 'abstract -> 'abstract ;
      join_list: 'abstract list  -> 'abstract ;
      meet: 'abstract -> 'abstract -> 'abstract ;
      widening: 'abstract -> 'abstract -> 'abstract ;
      print_abs: Format.formatter -> 'abstract -> unit ;
      print_env :  Format.formatter -> 'abstract -> unit ;
 
  (* change in the environment 
     global warning : save the earth
  *)
      param_to_glob: 'abstract -> 'abstract ;
      var_to_glob: 'abstract -> 'abstract ;
      glob_to_param: 'abstract -> 'abstract ;
      glob_to_var : 'abstract -> 'abstract ;

(* for the non-standard semantics 
   - glob_to_globprime and globprime_to_primebis change the environment
   - p_to_pprime (resp primebis_to_glob) rename the parameters (resp the variables)
*)
      p_to_pprime : 'abstract -> 'abstract ;
      glob_to_globprime :  'abstract -> 'abstract ;
      globprime_to_primebis :  'abstract -> 'abstract ;
      primebis_to_glob :  'abstract -> 'abstract ;

(* depending on the scm analyzed *)
      from_guard : ?nonstd:bool -> Scm_syn.bexpr -> 'abstract ;
      assign_list : ?nonstd:bool -> 'abstract -> (Scm_syn.var * Scm_syn.iexpr) list -> 'abstract ;
      rev_assign_list : ?nonstd:bool -> 'abstract -> (Scm_syn.var * Scm_syn.iexpr) list -> 'abstract ;
      mutable nbqueues : int    ;
      mutable list_param : string list ;
      mutable list_variables: string list ;
      mutable k_forward : int ; (* parameter of the standard widening *)
      mutable k_backward : int ; (* parameter of the standard widening *)
    }
      
type bidon = Bot | Top

module type Bidon =
sig
  type t = bidon
      
  val bottom : t

  val top : t

  val is_bottom : t -> bool

  val is_top : t -> bool
        
  val is_leq : t -> t -> bool
    
  val join  : t -> t -> t

  val join_list : t list -> t 

  val meet : t -> t -> t

  val widening   : t -> t -> t

  val  print : Format.formatter -> t -> unit
   
end

module Bidon  : Bidon =
struct
  type t = bidon

  let bottom = Bot

  let top = Top

  let is_bottom x = (x = Bot)

  let is_top x = (x = Top)

  let is_leq (x1:t) (x2:t) = (x1 <= x2)

  let join (x1:t) (x2:t) = if (x1=Bot) && (x2=Bot) then Bot else Top

  let join_list lx = List.fold_left join Bot lx

  let meet (x1:t) (x2:t) = if (x1=Bot) || (x2=Bot) then Bot else Top

  let widening = join

  let print fmt x = 
    match x with
	Bot -> Format.pp_print_string fmt "Bot"
      | Top -> Format.pp_print_string fmt "Top"

end


let finite_manager ?(k_backward=0)  ?(k_forward=0) () (scm : Scm_syn.global_scm) : bidon manager =
  let env = Environment.make [||] [||] in
  {    
    env_var = env ;
    env_param = env ;
    env_global = env ;
    get_env = (fun x -> env);


    param_to_glob =(fun x -> x) ;
    var_to_glob = (fun x -> x);
    glob_to_param =(fun x -> x) ;
    glob_to_var = (fun x -> x);

    p_to_pprime = (fun x -> x);
    glob_to_globprime =  (fun x -> x);
    globprime_to_primebis =  (fun x -> x);
    primebis_to_glob = (fun x -> x);


    bottom = (fun _ ->  Bidon.bottom) ;
    top = (fun _ ->Bidon.top) ;
    is_bottom = Bidon.is_bottom ;
    is_top = Bidon.is_top ;

    is_leq = Bidon.is_leq;
    join = Bidon.join ;
    join_list = Bidon.join_list ;
    meet = Bidon.meet;
    widening = Bidon.widening;
    print_abs = Bidon.print ;
    print_env = (fun fmt e -> ()) ;

    from_guard = (fun ?(nonstd=false) x ->  Bidon.top) ;
    assign_list = (fun  ?(nonstd=false) x l -> x) ;
    rev_assign_list = (fun  ?(nonstd=false) x l -> x) ;
    nbqueues = scm.gb_nb_channels ;
    list_param = List.map (fun (x,t,e) -> x) scm.gb_params;
    list_variables = List.map (fun (x,t,e) -> x) scm.gb_vars;
    k_forward = k_forward ;
    k_backward = k_backward;
  }



(* ******************************************************************************** *)
(* fonction de conversion vers les types APRON *)

let convert_var (x:Scm_syn.var): Var.t =
  Var.of_string x


let convert_typ (x:Scm_syn.typ):  Texpr1.typ  =
  match x with
    | Real -> Texpr1.Real
    | Int -> Texpr1.Int
    | Single -> Texpr1.Single
    | Double -> Texpr1.Double
    | Extended -> Texpr1.Extended
    | Quad -> Texpr1.Quad

let convert_unop (x:Scm_syn.unop):  Texpr1.unop =
  match x with 
    | Neg -> Texpr1.Neg
    | Cast -> Texpr1.Cast
    | Sqrt -> Texpr1.Sqrt

let convert_binop (x:Scm_syn.binop):  Texpr1.binop =
  match x with  
    | Add -> Texpr1.Add
    | Sub -> Texpr1.Sub 
    | Mul -> Texpr1.Mul 
    | Div -> Texpr1.Div
    | Mod -> Texpr1.Mod

let convert_round (x:Scm_syn.round) : Texpr1.round = 
  match x with  
| Near -> Texpr1.Near
| Zero -> Texpr1.Zero
| Up -> Texpr1.Up
| Down -> Texpr1.Down
| Rnd -> Texpr1.Rnd

let convert_scalar (x:Scm_syn.scalar) : Scalar.t =
  match x with  
  Float f -> Scalar.Float f
| Mpqf () -> failwith  "Mpqf not supported"

let convert_interval (x:Scm_syn.interval) : Interval.t = 
  {
    Interval.inf = convert_scalar x.inf;
    Interval.sup = convert_scalar x.sup;
  }

let convert_coeff (x:Scm_syn.coeff) : Coeff.t = 
  match x with  
    |Scalar s -> Coeff.Scalar (convert_scalar s)
    |Interval i -> Coeff.Interval (convert_interval i)

let rec convert_iexpr (x:Scm_syn.iexpr) : Texpr1.expr = 
  match x with
    | Cst c -> Texpr1.Cst (convert_coeff c) 
    | Var v -> Texpr1.Var (convert_var v)
    | Unop(un,ie,ty,ro) ->  Texpr1.Unop (convert_unop un, convert_iexpr ie, convert_typ ty, convert_round ro)
    | Binop(bi,ie1,ie2,ty,ro) ->  Texpr1.Binop (convert_binop bi, convert_iexpr ie1,  convert_iexpr ie2, convert_typ ty, convert_round ro) 

let convert_cons env (exp1,rel,exp2) : Tcons1.t =
  let apron_exp1 = convert_iexpr exp1
  and apron_exp2 = convert_iexpr exp2
  in
  let (t_res,r_res) = (Texpr1.Real,Texpr1.Rnd)      
  in
  match rel with
      EQ -> 
	Tcons1.make 
	  (Texpr1.of_expr
	      env
	      (Texpr1.Binop(Texpr1.Sub, apron_exp1, apron_exp2,t_res,r_res))
	  )
	  Tcons1.EQ
    | NEQ -> 
 	Tcons1.make 
	  (Texpr1.of_expr
	      env
	      (Texpr1.Binop(Texpr1.Sub, apron_exp1, apron_exp2,t_res,r_res))
	  )
	  Tcons1.DISEQ      
    | GT ->
 	Tcons1.make 
	  (Texpr1.of_expr
	      env
	      (Texpr1.Binop(Texpr1.Sub, apron_exp1, apron_exp2,t_res,r_res))
	  )
	  Tcons1.SUP
    | GEQ ->
 	Tcons1.make 
	  (Texpr1.of_expr
	      env
	      (Texpr1.Binop(Texpr1.Sub, apron_exp1, apron_exp2,t_res,r_res))
	  )
	  Tcons1.SUPEQ
    | LEQ ->
 	Tcons1.make 
	  (Texpr1.of_expr
	      env
	      (Texpr1.Binop(Texpr1.Sub, apron_exp2, apron_exp1,t_res,r_res))
	  )
	  Tcons1.SUPEQ
    | LT ->
 	Tcons1.make 
	  (Texpr1.of_expr
	      env
	      (Texpr1.Binop(Texpr1.Sub, apron_exp2, apron_exp1,t_res,r_res))
	  )
	  Tcons1.SUP
	

(* ******************************************************************* *)
(* Ngation des conditions. Reprise du package interproc *)

let negate_texpr (texpr:Apron.Texpr1.t) : Apron.Texpr1.t
  =
  let expr = Apron.Texpr1.to_expr texpr in
  let nexpr = match expr with
    | Apron.Texpr1.Unop(Apron.Texpr1.Neg,e,typ,round) -> 
	e
    | _ ->
	Apron.Texpr1.Unop(
	  Apron.Texpr1.Neg, expr,
	  Apron.Texpr1.Real, Apron.Texpr1.Rnd
	)
  in
  let env = Apron.Texpr1.get_env texpr in
  Apron.Texpr1.of_expr env nexpr


let negate_tcons (tcons:Apron.Tcons1.t) : Apron.Tcons1.t
  =
  let texpr = Apron.Tcons1.get_texpr1 tcons in
  let (ntyp,ntexpr) = match Apron.Tcons1.get_typ tcons with
    | Apron.Tcons1.EQ -> (Apron.Tcons1.DISEQ,texpr)
    | Apron.Tcons1.DISEQ -> (Apron.Tcons1.EQ,texpr)
    | Apron.Tcons1.SUPEQ -> (Apron.Tcons1.SUP, negate_texpr texpr)
    | Apron.Tcons1.SUP -> (Apron.Tcons1.SUPEQ, negate_texpr texpr)
    | Apron.Tcons1.EQMOD _ -> failwith "EQMOD not supported now"
  in
  Apron.Tcons1.make ntexpr ntyp


(* ******************************************************************** *)
(* Fonction de conversion des gardes *)

let rec boolexpr_of_bexpr (man: 'a Manager.t) (env: Environment.t) (bexpr:Scm_syn.bexpr) 
    : 
    'a Abstract1.t
    =
  match bexpr with
    | TRUE ->
	Abstract1.top man env 
    | FALSE ->
	Abstract1.bottom man env      
    | BRANDOM ->
	failwith "Random not supported"
    | CONS(cons) ->
	let tcons = convert_cons env cons in
	let tcons_array =  Tcons1.array_make env 1 in
	  Tcons1.array_set tcons_array 0 tcons ;
	  Abstract1.of_tcons_array man env tcons_array

    | AND(e1,e2) ->
	let abs1 = boolexpr_of_bexpr man env e1
	and abs2 = boolexpr_of_bexpr man env e2
	in Abstract1.meet man abs1 abs2

    | OR(e1,e2) ->
	let abs1 = boolexpr_of_bexpr man env e1
	and abs2 = boolexpr_of_bexpr man env e2
	in Abstract1.join man abs1 abs2

    | NOT(e) ->
	begin
	  match e with
	      TRUE -> 
		Abstract1.bottom man env 
	    | FALSE -> 
		Abstract1.top man env 
	    | BRANDOM ->
		failwith "BRandom not supported"
	    | CONS(cons) ->
		begin
		  let tcons = negate_tcons (convert_cons env cons) in
		  let tcons_array =  Tcons1.array_make env 1 in
		    Tcons1.array_set tcons_array 0 tcons ;
		    Abstract1.of_tcons_array man env tcons_array
		end
	    | AND(e1,e2) ->
		let abs1 = boolexpr_of_bexpr man env (NOT e1)
		and abs2 = boolexpr_of_bexpr man env (NOT e2)
		in Abstract1.join man abs1 abs2
	    | OR(e1,e2) ->
		let abs1 = boolexpr_of_bexpr man env (NOT e1)
		and abs2 = boolexpr_of_bexpr man env (NOT e2)
		in Abstract1.meet man abs1 abs2
	    | NOT(e) -> boolexpr_of_bexpr man env e	
	end

(* ******************************************************************************** *)
(* fonction de conversion *)

let concat_earray (env : Environment.t) (list_earray : Lincons1.earray list) : Lincons1.earray =
  let longueur =
    List.fold_left (fun res tab -> res + Lincons1.array_length tab) 0 list_earray
  in
  let tab_res = Lincons1.array_make env longueur 
  and i = ref 0 
  in
    List.iter 
      (fun tab ->
	for j=0 to (Lincons1.array_length tab)-1 do
	  Lincons1.array_set tab_res !i (Lincons1.array_get tab j);
	  incr i
	done)
       list_earray ;
    tab_res

(* ******************************************************************************** *)

let apron_manager 
    ?(k_backward=0)  
    ?(k_forward=0) 
    ?(threshold=false)
    (man : 'a Manager.t) 
    (scm : Scm_syn.global_scm) 
    : ('a Abstract1.t) manager 
    =
  (* creating environments for variables and parameters *)
  let tab_of_param_int = 
    let list_param_int = List.filter (fun (x,t,e) -> t = Scm_syn.Int) scm.gb_params
    in Array.of_list (List.map (fun (x,t,e) -> Var.of_string x)  list_param_int)
  and tab_of_var_int = 
    let list_var_int = List.filter (fun (x,t,e) -> t = Scm_syn.Int) scm.gb_vars
    in Array.of_list (List.map (fun (x,t,e) -> Var.of_string x) list_var_int)
  and tab_of_param_real = 
    let list_param_real = List.filter (fun (x,t,e) -> t = Scm_syn.Real) scm.gb_params
    in Array.of_list (List.map (fun (x,t,e) -> Var.of_string x)  list_param_real)
  and tab_of_var_real =
    let list_var_real = List.filter (fun (x,t,e) -> t = Scm_syn.Real) scm.gb_vars
    in Array.of_list (List.map (fun (x,t,e) -> Var.of_string x)  list_var_real)
  in

  let tab_of_param_int_prime =
    Array.map (fun v -> Var.of_string( "Prime_"^(Var.to_string v))) tab_of_param_int
  and tab_of_var_int_prime =
    Array.map (fun v -> Var.of_string( "Prime_"^(Var.to_string v))) tab_of_var_int
  and tab_of_param_real_prime =
    Array.map (fun v -> Var.of_string( "Prime_"^(Var.to_string v))) tab_of_param_real
  and tab_of_var_real_prime =
    Array.map (fun v -> Var.of_string( "Prime_"^(Var.to_string v))) tab_of_var_real

  in
  let tab_glob_int = Array.append tab_of_var_int tab_of_param_int
  and tab_glob_real = Array.append tab_of_var_real tab_of_param_real
  in
  let tab_glob_int_prime = 
    Array.append 
      tab_glob_int
      (Array.append tab_of_var_int_prime tab_of_param_int_prime)
  and tab_glob_real_prime = 
    Array.append 
      tab_glob_real
      (Array.append tab_of_var_real_prime tab_of_param_real_prime)
  and tab_int_primebis = 
    Array.append tab_of_var_int_prime tab_of_param_int

  and tab_real_primebis = 
    Array.append tab_of_var_real_prime tab_of_param_real
      
      
  in

  let env_var = Environment.make tab_of_var_int tab_of_var_real
  and env_param = Environment.make tab_of_param_int tab_of_param_real
  and env_global = 
    Environment.make 
      (Array.append tab_of_var_int tab_of_param_int) 
      (Array.append tab_of_var_real tab_of_param_real) 

  and env_globprime = Environment.make tab_glob_int_prime tab_glob_real_prime

  and env_primebis = Environment.make tab_int_primebis tab_real_primebis
  in
    (* Environmnent created *)

  let f_apply_ltrans ?(nonstd=false)  x l =
    let env_normal = 
      if nonstd
      then
	env_globprime
      else
	env_global
    in
    let tab_affectations =
      Array.of_list 
	(List.map 
	    (fun (v,a) -> 
	      Texpr1.of_expr 
		env_normal
		(convert_iexpr a)
		
	    )
	    l
	)
    in
      Abstract1.assign_texpr_array
	man
	x
	(Array.of_list (List.map (fun (v,a) -> Var.of_string v) l))
	tab_affectations
	None
	(* function transforming an abstract value according to assignments *)
  in

  let f_apply_ltrans_rev ?(nonstd=false) x l_assign =
    let env_normal = 
      if nonstd then env_globprime else env_global
    in

    let liste_var = List.map (fun (x,e) -> Var.of_string x) l_assign 
    and liste_var_prime_type = List.map (fun (x,e) -> Var.of_string (x^"_prime"), Environment.typ_of_var env_normal (Var.of_string x)) l_assign  
      (* construit la liste des variables "primes" ainsi que leur type *)
    in
    let liste_var_prime_int = List.filter (fun (x,t) -> t = Environment.INT) liste_var_prime_type
    and liste_var_prime_real = List.filter (fun (x,t) -> t = Environment.REAL) liste_var_prime_type
      (* spare la liste prcdente en 2, selon le type des variables *)
    in
    let tab_var = Array.of_list liste_var 
    and tab_var_prime_int = Array.of_list (List.map fst liste_var_prime_int) 
    and tab_var_prime_real = Array.of_list (List.map fst liste_var_prime_real)
      (* oublie le type et construit deux tableaux de variables *)
    in 
    let env_total = Environment.add env_normal tab_var_prime_int tab_var_prime_real
      (* env_total = env_normal + variables prime *)
    in
    let tab_affectations =
      Array.of_list 
	(List.map 
	    (fun (v,a) -> 
	      Texpr1.of_expr 
		env_total
		(convert_iexpr a)
		
	    )
	    l_assign
	)
    in
    let abs_assign =  
      Abstract1.assign_texpr_array
	man
	x
	(Array.of_list (List.map (fun (v,a) -> Var.of_string (v^"_prime")) l_assign))
	tab_affectations
	None
    and abs_prime =
      Abstract1.change_environment
	man
	(Abstract1.rename_array man x  tab_var (Array.append tab_var_prime_int tab_var_prime_real))
	env_total
	false
	(* renomme v en v' dans la valeur abstraite initiale, puis tend son environnement *)     
    in
      Abstract1.change_environment
	man
	(Abstract1.meet man abs_assign  abs_prime)
	env_normal
	false
  in

  let  trans_to_earray (env: Environment.t)(lres : Lincons1.earray list) (trans:Scm_syn.transition) =
    (* transforme une contrainte (differente de top) en un tableau de contraintes lineaires *)

    (* lres = accumulateur *)
    let abstr_g = 
      boolexpr_of_bexpr man env_global trans.guard
    in
      
    let abstr_post =
      if  trans.assignement = []
      then
	abstr_g
      else
	f_apply_ltrans abstr_g trans.assignement

    in

    let abstr_post =
      if Environment.equal env env_var
      then
	Abstract1.change_environment 
	  man 
	  abstr_post
	  env_var 
	  false 
      else 
	abstr_post
    in
      if Abstract1.is_top man abstr_post
      then
	lres
      else
	(Abstract1.to_lincons_array man abstr_post)::lres
  in

  let state_to_earray (env: Environment.t) (lres : Lincons1.earray list) (s:Scm_syn.state) =
    List.fold_left (trans_to_earray env) lres s.ltrans
  in

  let contr_abstr : Lincons1.earray =
    let list_contraintes =
      List.map 
	(fun c -> 
	  Abstract1.to_lincons_array 
	    man 
	    ( 
	      let tcons = convert_cons env_global c in
	      let tcons_array =  Tcons1.array_make env_global 1 in
		Tcons1.array_set tcons_array 0 tcons ;
		let abstr = Abstract1.of_tcons_array 
		  man 
		  env_global 
		  tcons_array
		in
		  Abstract1.change_environment 
		    man
		    abstr
		    env_var
		    false
	    )
	)
	scm.gb_threshold
    in
      concat_earray env_var (list_contraintes@(List.fold_left (state_to_earray env_var) [] scm.gb_lstates)	)      
	(* transforme toutes les postconditions des transitions en
	   contraintes lineaires, rajoute les contraintes spcifiees par l'utilisateur
	   et concatene le tout *)

  and contr_queues : Lincons1.earray =
    let list_contraintes =
      List.map 
	(fun c -> 
	  Abstract1.to_lincons_array 
	    man 
	    ( 
	      let tcons = convert_cons env_global c in
	      let tcons_array =  Tcons1.array_make env_global 1 in
		Tcons1.array_set tcons_array 0 tcons ;
		Abstract1.of_tcons_array man env_global tcons_array
	    )
	)
	scm.gb_threshold
    in
      concat_earray env_global (list_contraintes@(List.fold_left (state_to_earray env_global) [] scm.gb_lstates)	)      
	(* transforme toutes les postconditions des transitions en
	   contraintes lineaires, rajoute les contraintes spcifiees par l'utilisateur
	   et concatene le tout *)

  in
    (* we now have an array [tab_contraintes] containing all the constraints of the guards *)
    Format.printf 
      "Thresholds : %a @." 
      (Lincons1.array_print ~first:"@[" ~sep:" ; " ~last:"@]") 
      contr_queues ; 

    (* global manager *)

    {
      env_var = env_var ;
      env_param = env_param ;
      env_global = env_global ;
      get_env = (fun x -> Abstract1.env x );

      param_to_glob = (fun x -> Abstract1.change_environment man x env_global false) ;
      var_to_glob = (fun x -> Abstract1.change_environment man x env_global false)  ;
      glob_to_param = (fun x -> Abstract1.change_environment man x env_param false) ;
      glob_to_var = (fun x -> Abstract1.change_environment man x env_var false) ;

      (* for the non-standard semantics *)

      p_to_pprime =
	(fun x -> Abstract1.rename_array 
	  man 
	  x 
	  (Array.append tab_of_param_int tab_of_param_real)
	  (Array.append tab_of_param_int_prime tab_of_param_real_prime)
	)
      ;

      glob_to_globprime = (fun x -> Abstract1.change_environment man x env_globprime false)  ;
      globprime_to_primebis =  (fun x -> Abstract1.change_environment man x env_primebis false);
      (* il faut renommer les variables v' -> v *)
      primebis_to_glob = 
	(fun x -> Abstract1.rename_array 
	  man 
	  x 
	  (Array.append tab_of_var_int_prime tab_of_var_real_prime)
	  (Array.append tab_of_var_int tab_of_var_real)
	)
      ;


      bottom = (fun env -> Abstract1.bottom man env) ;
      top = (fun env -> Abstract1.top man env) ;
      is_bottom = 
	begin 
	  fun x -> 
	    Abstract1.is_bottom man x 
	end
      ;

     is_top = 
	begin 
	  fun x -> 
	    Abstract1.is_top man x 
	end
      ;

      is_leq =
	begin 
	  fun x y -> 
	    Abstract1.is_leq man x y 
	      
	end
      ;

      join = Abstract1.join man ;
      join_list = (fun lx -> Abstract1.join_array man (Array.of_list lx)) ;
      meet= Abstract1.meet man;

      widening  = 
	(fun  x y ->
	  if threshold
	  then 
	    begin
	      assert (Environment.equal (Abstract1.env x) (Abstract1.env y) );
	      let env = Abstract1.env x in
		if Environment.equal env env_var(* cas ou on largit les abstractions des variables *)
		then Abstract1.widening_threshold man x y contr_abstr
		else
		  (* cas ou on largit les files, sem non standard *)
		  begin
		    assert (Environment.equal env env_global);
		    Abstract1.widening_threshold man x y contr_queues
		  end
	    end	      
	  else
	    Abstract1.widening man x y
	);



      print_abs = Abstract1.print ;
      print_env = (fun fmt x -> Environment.print fmt (Abstract1.env x)) ;
      
      from_guard = (fun ?(nonstd=false) x -> 
	if nonstd
	then boolexpr_of_bexpr man env_globprime x
	else boolexpr_of_bexpr man env_global x) ;


      assign_list = f_apply_ltrans ;
      rev_assign_list = f_apply_ltrans_rev ;

      nbqueues = scm.gb_nb_channels ;
      list_param = List.map (fun (x,t,e) -> x) scm.gb_params;
      list_variables = List.map (fun (x,t,e) -> x) scm.gb_vars;
      k_forward = k_forward ;
      k_backward = k_backward;
    }

