(***********************************************************************)
(*                                                                     *)
(*            The Symbolic Communicating Machines Library              *)
(*                                                                     *)
(*                Bertrand Jeannet and Tristan Le Gall                 *)
(*                                                                     *)
(*  Copyright 2008 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file LICENSE.        *)
(*                                                                     *)
(***********************************************************************)

(** SCM analyzer: products of SCMs module *)

open Scm_syn

(* ************************************************************************************ *)
(* renaming functions *)
(* ************************************************************************************ *)

(* lparam : liste des paramtres qu'il ne faut PAS renommer *)

(* ----------------------------------------------------------------- *)
(* variables and expressions *)

let rename_var (pref : string) (lparam : string list) (v : string) =  
  if List.mem v lparam
  then v
  else (pref^v)

let rec rename_iexpr (pref : string)(lparam : string list) (expi : iexpr) =
  match expi with
      Scm_syn.Cst _ -> expi
    | Scm_syn.Var v ->  Scm_syn.Var (rename_var pref lparam v)
    | Scm_syn.Unop (op,exp,typ,rd) -> 
	Scm_syn.Unop (op, rename_iexpr pref lparam exp,typ,rd)
    | Scm_syn.Binop(op,exp1,exp2,typ,rd) ->  
	Scm_syn.Binop(op, rename_iexpr pref lparam exp1,rename_iexpr pref lparam exp2,typ,rd)
 

let rename_cons (pref : string) (lparam : string list) (exp1,comp,exp2) = 
  (rename_iexpr pref lparam exp1, comp, rename_iexpr pref lparam exp2)

let rec rename_bexpr (pref : string) (lparam : string list) (expb : bexpr) =
  match expb with
    | TRUE -> TRUE
    | FALSE -> FALSE
    | BRANDOM -> BRANDOM 
    | CONS(c) ->  CONS(rename_cons pref lparam c)
    | AND(b1,b2) -> AND(rename_bexpr pref lparam b1, rename_bexpr pref lparam b2)
    | OR(b1,b2) -> OR(rename_bexpr pref lparam b1, rename_bexpr pref lparam b2)
    | NOT(b) -> NOT(rename_bexpr pref lparam b)


let rename_iexpr_opt  (pref : string) (lparam : string list) (exp1 : iexpr option) =
  match exp1 with
      None -> None
    | Some expr1 -> Some (rename_iexpr pref lparam expr1)

(* ----------------------------------------------------------------- *)
(* declaration, states and transitions *)


let rec dec_rename (pref : string) (lparam : string list)(dec_list :  declaration list) :  declaration list =
  match dec_list with
      [] -> []
    | (n1,t1,exp1)::reste -> 
	(rename_var pref lparam n1, 
	t1,
	rename_iexpr_opt pref lparam exp1) ::(dec_rename pref lparam reste)


let rename_assignement (pref : string) (lparam : string list) (v,iexp) : (var * iexpr) =
  (rename_var pref lparam v, rename_iexpr pref lparam iexp)
  

let rename_assignement_list  (pref : string) (lparam : string list) (li : (var * iexpr) list ) : (var * iexpr) list  =
  List.map (rename_assignement pref lparam) li

(* ----------------------------------------------------------------- *)
(* functions for renaming a single automaton *)

let trans_rename_single 
    ?(add_channel=0)
    (pref : string)
    (lparam : string list) 
    (trans : transition)
    : transition
    =
  {
    owner = trans.owner;
    controllable = trans.controllable;
    new_state = pref^trans.new_state ;
    guard = rename_bexpr pref lparam trans.guard ;
    action = 
      begin
	match trans.action with
	    None -> None
	  | Some (q,act,v) -> Some (q+add_channel,act,v) 
      end;
    assignement = rename_assignement_list pref lparam trans.assignement ;
  }  

let state_rename_single 
    ?(add_channel=0)
    (pref : string)
    (lparam : string list) 
    (s : state)
    : state
    =
    {
      id_state= pref^s.id_state ;
      ltrans = List.map (trans_rename_single  ~add_channel:add_channel pref lparam) s.ltrans ;
    }


(* ----------------------------------------------------------------- *)
(* functions renaming a product of two automata *)

(* rename trans1 : transitions of the first automaton *)
let rename_trans1 
    ?(add_channel=0)
    (pref1 : string) 
    (pref2 : string) 
    (lparam : string list) 
    (state2 : string)
    (trans : transition) 
    : transition 
    =
  {
    owner = trans.owner;
    controllable = trans.controllable;
    new_state = pref1^trans.new_state^"x"^pref2^state2 ;
    guard = rename_bexpr pref1 lparam trans.guard ;
    action = 
      begin
	match trans.action with
	    None -> None
	  | Some (q,act,v)-> Some (q+add_channel,act,v) 
      end;
    assignement = rename_assignement_list pref1 lparam trans.assignement ;
  }
    
(* rename trans2 : transitions of the second automaton *)
let rename_trans2
    ?(add_channel=0)
    (pref1 : string) 
    (pref2 : string) 
    (lparam : string list)
    (state1 : string)
    (trans : transition) 
    =
  {
    owner = trans.owner;
    controllable = trans.controllable;
    new_state = pref1^state1^"x"^pref2^trans.new_state ;
    guard = rename_bexpr pref2 lparam trans.guard ;
    action = 
      begin
	match  trans.action with
	    None -> None
	  | Some (q,act,v) ->
	  Some (q+add_channel,act,v) 
      end;
    assignement = rename_assignement_list pref2 lparam trans.assignement ;
  }
    

let rename_ltrans1 
    ?(add_channel=0)
    (pref1 : string) (* name of the first automaton *)
    (pref2 : string) (* name of the second automaton *) 
    (lparam : string list) (* list of parameters *)
    (s1 : state) (* state of the first automaton *)
    (s2 : state) (*  state of the second automata *)
    : transition list
  =
  List.map (rename_trans1 ~add_channel:add_channel pref1 pref2 lparam s2.id_state) s1.ltrans

let rename_ltrans2 
    ?(add_channel=0)
    (pref1 : string) (* name of the first automaton *)
    (pref2 : string) (* name of the second automaton *) 
    (lparam : string list) (* list of parameters *)
    (s1 : state) (* state of the first automaton *)
    (s2 : state) (*  state of the second automaton *)
    : transition list
  =
  List.map (rename_trans2 ~add_channel:add_channel pref1 pref2 lparam s1.id_state) s2.ltrans
 
(*  [rename_state] also performs the asynchronous product *)
let  rename_state
    ?(add_channel=0)
    (pref1 : string) (* name of the first automaton *)
    (pref2 : string) (* name of the second automaton *) 
    (lparam : string list) (* list of parameters *)
    (s1 : state) (* state of the first automaton *)
    (s2 : state) (*  state of the second automaton *)
    : state
  =
  {
    id_state = pref1^s1.id_state^"x"^pref2^s2.id_state ;
    ltrans = 
      (rename_ltrans1 ~add_channel:add_channel pref1 pref2 lparam s1 s2)
      @(rename_ltrans2 ~add_channel:0 pref1 pref2 lparam s1 s2);
  }

(* ----------------------------------------------------------------- *)
(* list of states and automata *)


(*  [rename_lstate] also performs the asynchronous product *)
let  rename_lstate
    ?(add_channel=0)
    (pref1 : string) (* name of the first automaton *)
    (pref2 : string) (* name of the second automaton *)
    (lparam : string list) (* list of parameters *)
    (ls1 : state list) (* state of the first automaton *)
    (ls2 : state list) (*  state of the second automaton *)    
    : state list
    =
    List.fold_left  
      (fun acc x -> List.fold_left (fun acc2 y -> (rename_state  ~add_channel:add_channel pref1 pref2 lparam x y)::acc2) acc (List.rev ls2))
      []
      (List.rev ls1)

let automaton_rename_single 
    ?(add_channel=0)
    (lparam : string list) (* list of parameters *)
    (auto : automaton) 
    : automaton =
  {
    id_auto = auto.id_auto ;
    ldec = dec_rename (auto.id_auto^"_") lparam auto.ldec ;
    init_states = List.map (fun s -> auto.id_auto^"_"^s) auto.init_states ;
    lstate = List.map (state_rename_single  ~add_channel:add_channel (auto.id_auto^"_") lparam) auto.lstate
  }

let rename_init_states pref1 pref2 list1 list2 : string list =
  let f_fold_u st1 st2 =
    pref1^st1^"x"^pref2^st2
  in
  let f_fold_l2 st1 lres2 list2  : string list =
    List.fold_left (fun lres st2 -> (f_fold_u st1 st2)::lres) lres2 (List.rev list2)
  in 
    List.fold_left (fun lres st1 -> f_fold_l2 st1 lres list2) [] (List.rev list1)

(* fait le produit des deux listes et renome les tats *)

let rename_automata
    ?(add_channel=0)
    (lparam : string list) (* list of parameters *)
    (auto1 : automaton) (* first automaton *)
    (auto2 : automaton) (*  result of the product of automata  *)    
    : automaton
    =  
  {
    id_auto = auto1.id_auto^"x"^auto2.id_auto ;
    ldec = (dec_rename (auto1.id_auto^"_") lparam auto1.ldec)@auto2.ldec ;
    init_states = rename_init_states (auto1.id_auto^"_") "" auto1.init_states  auto2.init_states ;
    lstate = rename_lstate  ~add_channel:add_channel (auto1.id_auto^"_") "" lparam auto1.lstate auto2.lstate
  }


(* ----------------------------------------------------------------- *)
(* bad states *)
(* renomme et fait le produit d'une liste de mauvais tats. Si un
   automate n'est pas spcifi, tous ses tats sont mauvais pour la liste
   en cours *)
let rename_bad_states
    (lauto : automaton list)
    (lbad_states : (string*(string*bexpr) list) list)
    : (string*bexpr) list
    =
  let rec find_auto_name (name:string) (lauto) =
    match lauto with
	[] -> false
      | tt::qq -> if tt.id_auto = name then true else find_auto_name name qq
  in
  let check_auto (aut,lstates) : unit =
    if not (find_auto_name aut lauto) then failwith ("Unknown automaton name in bad states declaration: "^aut)
  in
    List.iter check_auto lbad_states;

  (* f_find_auto donne, pour chaque automate, la liste des mauvais
     tats. Si un automate n'est pas dans la liste, alors tous ses tats
     sont mauvais *)
  let rec f_find_auto 
      (auto: automaton) 
      (l_states : (string*(string*bexpr) list) list) 
      : (string*bexpr) list 
      =
    match l_states with
	[] -> 
	  List.map (fun s -> (auto.id_auto^"_"^s.id_state,TRUE)) auto.lstate
	    (* si on n'a pas trouv le nom dans la liste, toutes les localits sont mauvaises *)
      | (nom,liste)::reste -> 
	  if nom = auto.id_auto
	  then 
	    List.map 
	      (fun (loc_name,bexp) -> 
		(auto.id_auto^"_"^loc_name, 
		rename_bexpr (auto.id_auto^"_") [] bexp
		)
	      )
	      liste
	  else 
	    f_find_auto auto reste
  in

  let lbad_states : (string*bexpr) list list =
    List.map 
      (fun auto -> f_find_auto auto lbad_states)
      lauto
      (* met les mauvais tats "dans l'ordre" et ajoute les automates n'tant pas dans la liste *)
  in
(*    Format.printf "@.DEBUG: LBAD_STATES: @[ %a @] @."
    (Print.list 
      (Print.list (fun fmt (s,b) -> Format.pp_print_string fmt s)))
      lbad_states ; *)

    (* fait le produit asynchrone des mauvais tats *)
  let bad_stat_prod (list1: (string*bexpr) list) (list2: (string*bexpr) list) : (string*bexpr) list =
    let f_fold2 (name1,bexp1) lres (name2,bexp2) =
      if (bexp1 = TRUE)
      then
	(name1^"x"^name2,bexp2)::lres
      else 
	if (bexp2 = TRUE)
	then
	(name1^"x"^name2,bexp1)::lres
	else 
	  (name1^"x"^name2,AND(bexp1,bexp2))::lres
    in
    let f_fold1 lres (name1,bexp1) =
      List.fold_left (f_fold2 (name1,bexp1)) lres (List.rev list2)
    in
      List.fold_left  f_fold1 [] (List.rev list1) 
  in 
  let rec rename_list (liste: (string*bexpr) list list) =
    match liste with
	[] -> []
      | [liste1] -> liste1
      | tt::qq -> bad_stat_prod tt (rename_list qq)
  in 
    (rename_list lbad_states)

(* ************************************************************************************ *)
(* transformation functions *)
(* ************************************************************************************ *)


let from_syntree (syntree : Scm_syn.scm) : global_scm =
  let lparam =
    List.map (fun (v,t,e) -> v) syntree.params
  in

  let rec rename_auto_list (lauto : automaton list) : automaton = 
    match lauto with
	[] -> 
	  {
	    id_auto = "" ;
	    ldec = [] ; 
	    init_states = [] ;
	    lstate = []
	  }
      | auto::[] -> automaton_rename_single lparam auto
      | tt::qq -> rename_automata lparam tt (rename_auto_list qq)
  in

  let auto_res = rename_auto_list syntree.lauto
  in
    {
      gb_id_scm = syntree.id_scm ;
      gb_nb_channels = syntree.nb_channels ;
      gb_lossy_list = syntree.lossy_list ;
      gb_vars = auto_res.ldec ;
      gb_params = syntree.params ;
      gb_init_states = auto_res.init_states;
      gb_lstates = auto_res.lstate ;
      gb_threshold = syntree.threshold ;
      gb_bad_states =
	List.fold_left 
	  (fun lres (lbad,r) -> (List.map (fun (x,bx) -> (x,bx,r))(rename_bad_states syntree.lauto lbad))@lres) 
	  [] 
	  (List.rev syntree.bad_states) ;
    }























(* ****************************************************************************** *)

let asynchronous_product 
    ?(disjoint_channels=false) 
    (scm1 : global_scm)  
    (scm2 : global_scm)   
    : global_scm 
    = 
  assert (List.fold_left (fun b x -> b && (List.mem x scm2.gb_lossy_list)) true scm1.gb_lossy_list) ;
  (* check that the two automata have the same set of lossy channels *)

  let lparam = [] in (* les parametres sont renommes *)
  {
    gb_id_scm = scm1.gb_id_scm^"x"^scm2.gb_id_scm ;
    gb_nb_channels =
      if disjoint_channels
      then scm1.gb_nb_channels + scm2.gb_nb_channels 
      else max scm1.gb_nb_channels scm2.gb_nb_channels ;
    gb_lossy_list = scm1.gb_lossy_list ;

    gb_vars = (dec_rename (scm1.gb_id_scm^"_") lparam scm1.gb_vars)@(dec_rename (scm2.gb_id_scm^"_") lparam scm2.gb_vars) ;
    gb_params =  (dec_rename (scm1.gb_id_scm^"_") lparam scm1.gb_params)@(dec_rename (scm2.gb_id_scm^"_") lparam scm2.gb_params) ;
    gb_init_states = rename_init_states (scm1.gb_id_scm^"_") (scm2.gb_id_scm^"_") scm1.gb_init_states scm1.gb_init_states ;
    gb_lstates =
      begin
      let add_channel =
	if disjoint_channels
	then  scm2.gb_nb_channels	
	else 0 
      in
      let pref1 = scm1.gb_id_scm^"_"
      and pref2 = scm2.gb_id_scm^"_"
      in
	List.fold_left  
	  (fun acc x -> List.fold_left (fun acc2 y -> (rename_state  ~add_channel:add_channel pref1 pref2 lparam x y)::acc2) acc scm2.gb_lstates)
	  []
	  scm1.gb_lstates
      end	;
    gb_threshold = scm1.gb_threshold@scm2.gb_threshold

(* BUG ICI : il faut renommer les variables *)  ;
      gb_bad_states = [] (* A MODIFIER !!!!!!! *)
  }


(* ************************************************************************************ *)
(* synchronous product *)
(* ************************************************************************************ *)


let rename_trans_sync 
    ?(same_variables=true)
    (pref1 : string)
    (pref2 : string) 
    (lparam : string list) (* list of parameters *)
    (trans1 : transition) (* transition of the first automaton *)
    (trans2 : transition) (*  transition of the second automaton *)
    : transition
    =
  if same_variables (* if the two automata have the same variables, we do not rename the variables *)
  then
    {
      owner =  pref1^trans1.owner^"x"^pref2^trans2.owner;
      controllable = trans1.controllable && trans2.controllable;
      new_state = pref1^trans1.new_state^"x"^pref2^trans2.new_state ;
      guard = 
	Scm_syn.AND(trans1.guard,trans2.guard) ; 

      action = trans1.action;
      assignement = 
	List.append
	  trans1.assignement
	  trans2.assignement
      ;
    }
  else
    {
      owner =  pref1^trans1.owner^"x"^pref2^trans2.owner;
      controllable = trans1.controllable && trans2.controllable;
      new_state = pref1^trans1.new_state^"x"^pref2^trans2.new_state ;
      guard = 
	Scm_syn.AND(rename_bexpr pref1 lparam trans1.guard,rename_bexpr pref1 lparam trans2.guard) 
      ;
      action = trans1.action;
      assignement = 
	List.append
	  (rename_assignement_list pref1 lparam trans1.assignement)
	  (rename_assignement_list pref2 lparam trans2.assignement)
      ;
    }


let rename_list_trans_sync 
    ?(same_variables=true)
    (pref1 : string)
    (pref2 : string) 
    (lparam : string list) (* list of parameters *)
    (ltrans1 : transition list) (* transition of the first automaton *)
    (ltrans2 : transition list) (*  transition of the second automaton *)
    : transition list
    =
  let f_apply2 pref1 pref2 lparam ltrans2 lres_apply trans1=
    List.fold_left
      (fun l_res trans2 ->
	if trans1.action = trans2.action
	then (rename_trans_sync ~same_variables:same_variables pref1 pref2 lparam trans1 trans2)::l_res
	else l_res
      )
      lres_apply
      ltrans2
  in
    List.fold_left 
      (fun l_res trans1 ->
	f_apply2 pref1 pref2 lparam ltrans2 l_res trans1
      )
      []
      ltrans1

let rename_state_sync 
    ?(same_variables=true)
    (pref1 : string)
    (pref2 : string) 
    (lparam : string list) (* list of parameters *)
    (state1 : state) (* state of the first automaton *)
    (state2 : state) (* state of the second automaton *)
    : state
    =
  {
    id_state = pref1^state1.id_state^"x"^pref2^state2.id_state ;
    ltrans = 
      rename_list_trans_sync 
	~same_variables:same_variables 
	pref1 
	pref2 
	lparam 
	state1.ltrans 
	state2.ltrans
  }

let rename_list_state_sync 
    ?(same_variables=true)
    (pref1 : string)
    (pref2 : string) 
    (lparam : string list) (* list of parameters *)
    (lstate1 : state list) (* states of the first automaton *)
    (lstate2 : state list) (* states of the second automaton *)
    : state list
    =
  let f_apply2 pref1 pref2 lparam lstate2 lres_apply state1=
    List.fold_left
      (fun l_res state2 ->
	(rename_state_sync 
	    ~same_variables:same_variables 
	    pref1 
	    pref2 
	    lparam 
	    state1 
	    state2
	)::l_res

      )
      lres_apply
      lstate2
  in
    List.fold_left 
      (fun l_res state1 ->
	f_apply2 pref1 pref2 lparam lstate2 l_res state1
      )
      []
      lstate1


(* do not employ ths one *)
let synchronous_product 
    ?(same_variables=true)
    (scm1 : global_scm)  
    (scm2 : global_scm)   
    : global_scm 
    = 
  assert (scm1.gb_nb_channels = scm2.gb_nb_channels);
  assert (List.fold_left (fun b x -> b && (List.mem x scm2.gb_params)) true scm1.gb_params) ;
  (* check that the two automata have the same parameters and the same number of channel *)
  assert (List.fold_left (fun b x -> b && (List.mem x scm2.gb_lossy_list)) true scm1.gb_lossy_list) ;
  (* check that the two automata have the same set of lossy channels *)

  let lparam =     List.map (fun (v,t,e) -> v) (scm1.gb_params@scm2.gb_params) in
  {
    gb_id_scm = scm1.gb_id_scm^"x"^scm2.gb_id_scm ;
    gb_nb_channels = scm1.gb_nb_channels ;
    gb_lossy_list = scm1.gb_lossy_list ;
    gb_vars = 
      if same_variables
      then 
	List.append scm1.gb_vars scm2.gb_vars
      else
	List.append
	  (dec_rename (scm1.gb_id_scm^"_") lparam scm1.gb_vars)
	  (dec_rename (scm2.gb_id_scm^"_") lparam scm2.gb_vars) 
    ;
    gb_params =  scm1.gb_params;
    gb_init_states = 
      rename_init_states 
	(scm1.gb_id_scm^"_") 
	(scm2.gb_id_scm^"_") 
	scm1.gb_init_states 
	scm1.gb_init_states ;
    gb_lstates =
      begin
      let pref1 = scm1.gb_id_scm^"_"
      and pref2 = scm2.gb_id_scm^"_"
      in
	rename_list_state_sync 
	  ~same_variables:same_variables 
	  pref1 
	  pref2 
	  lparam 
	  scm1.gb_lstates 
	  scm2.gb_lstates 
      end	;
    gb_threshold = scm1.gb_threshold@scm2.gb_threshold
(* BUG ICI *)  ;
      gb_bad_states = [] (* A MODIFIER !!!!!!! *)
  }

