(***********************************************************************)
(*                                                                     *)
(*            The Symbolic Communicating Machines Library              *)
(*                                                                     *)
(*                Bertrand Jeannet and Tristan Le Gall                 *)
(*                                                                     *)
(*  Copyright 2008 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file LICENSE.        *)
(*                                                                     *)
(***********************************************************************)

(** SCM analyzer: Standard Semantics module *)


(* ####################################################################
######################## Module Standard Sem ##########################
#################################################################### *)




open Format
open Scm_syn
open Scm_manager

(*  ********************************************************************** *)
(** 2 Module type for the alphabet *)
(*  ********************************************************************** *)

module type Alphabet = sig
  type t
  val from_string: string -> t
    (** creates a letter from a string *)
  val compare: t -> t -> int
(** [compare x y] returns 0 if [x] is equal to [y], a negative integer if [x] 
   is less than [y], and a positive integer if [x] is greater than [y] *)
  val print: Format.formatter -> t -> unit
  end


(*  ********************************************************************** *)
(** 2 Functor *)
(*  ********************************************************************** *)

module type S = sig

  module Alpha : Alphabet
    (** module of the alphabet/finite partition of the lattice *)

  module AlphaSep : PLattice.Alphabet
    (** same thing, with the separation letter *)

  module PL : (PLattice.S with module A=AlphaSep)
    (** partitionned lattice *)

  module Langage : (LAutomaton.S  with module A=AlphaSep
				   and module PL=PL )
    (** QDD-like automata *)

  module NQueues : 
    (StackQueueNRel.S  
      with module PL=PL
      and module LAuto=Langage)
    (** abstraction of the queue-contents *)

  type 'abstract t
      (** abstract value *)

 (*  ********************************************************************** *)
  (** 3 Print *)
  (*  ********************************************************************** *)

  val print : 'abstract manager -> Format.formatter -> 'abstract t -> unit

  val print_dot : 
   ?margin:int ->
    ?titlestyle:string ->
    ?vertexstyle:string ->
    ?edgestyle:string ->
    ?title:string ->
    'abstract manager -> 
    Format.formatter -> 
    'abstract t -> 
    unit
    (** prints only the automaton part of the abstract value *)

  val print_regexp : 'abstract manager -> Format.formatter -> 'abstract t -> unit

  val get_automata_size : 'abstract t -> int


  (*  ********************************************************************** *)
  (** 3 Constructors *)
  (*  ********************************************************************** *)

  val bottom : 'abstract manager -> 'abstract PL.t -> 'abstract PL.lattice -> 'abstract t
  val top : 'abstract manager -> 'abstract PL.t -> 'abstract PL.lattice -> 'abstract t
    (** contructors.  NB: the same lattice (physical equality ==) and
       the same partition must be used in order to have compatible
       automata *)

  val create_partition : 'abstract manager -> 'abstract PL.t 

  val create_lattice : 'abstract manager -> 'abstract PL.lattice
    (** create a partition and a lattice.  NB: the same lattice
       (physical equality ==) and the same partition must be used in
       order to have compatible automata *)

  val get_partition : 'abstract t -> 'abstract PL.t 

  val get_lattice : 'abstract t -> 'abstract PL.lattice
    (** returns the partition and lattice of the automaton part.  NB: the same lattice
       (physical equality ==) and the same partition must be used in
       order to have compatible automata *)


  val init_abstr : 'abstract manager -> 'abstract PL.t -> 'abstract PL.lattice -> (var * iexpr) list -> 'abstract t
    (** [init_abstr partition lattice assignment_list] creates an abstract value, all
       the [n] queues being empty.  The initial values of the variables
       are defined by the assignement list.  *)

  val init_bad_state : 'abstract manager -> 'abstract PL.t -> 'abstract PL.lattice -> bexpr -> string Regexp.t -> 'abstract t
    (** [init_bad_State partition lattice bexpr regexpr] creates an
	abstract value, the content of the queue is given by
	[regexpr]. This function helps to create a set of bad
	configurations. *)

  val elim_queues : 'abstract t -> 'abstract

  (*  ********************************************************************** *)
  (** 3 Comparison operators *)
  (*  ********************************************************************** *)

  val is_bottom :  'abstract manager -> 'abstract t -> bool

  val is_top :  'abstract manager -> 'abstract t -> bool

  val is_leq : 'abstract manager -> 'abstract t  -> 'abstract t -> bool
    (** comparision operators: NB: is_bottom is a test on the physical
       representation, others are tests involving language inclusion *)

  val is_prefix : 'abstract manager -> string list -> 'abstract t -> int -> bool 
    (** [is_prefix man w auto i] checks whether the word [w] is
	a prefix of a word in the queue [i] of [auto]. *)

  (*  ********************************************************************** *)
  (** 3 Operations on the queue content *)
  (*  ********************************************************************** *)

  (* operations on the queue only *)
  val complementary :  'abstract manager -> 'abstract t  -> 'abstract t
    (** complements the lattice automaton : ONLY WORKS WITH FINITE AUTOMATA *)

  val determinise : 'abstract manager -> 'abstract t -> 'abstract t
    (** determinise the lattice automaton *)
  val minimise : 'abstract manager -> 'abstract t -> 'abstract t
    (** minimise the lattice automaton *)

  (*  ********************************************************************** *)
  (** 3 Lattice operations *)
  (*  ********************************************************************** *)

  val join : 'abstract manager -> 'abstract t -> 'abstract t -> 'abstract t 

  val meet : 'abstract manager -> 'abstract t -> 'abstract t -> 'abstract t 

  val widening : ?initial:bool ->  ?final:bool -> 'abstract manager ->  'abstract t -> 'abstract t -> 'abstract t 

  val widening_alt : ?initial:bool ->  ?final:bool -> 'abstract manager ->  'abstract t -> 'abstract t -> 'abstract t

  (*  ********************************************************************** *)
  (** 3 extrapolation operations *)
  (*  ********************************************************************** *) 

  val extrapolation : ?initial:bool ->  ?final:bool -> k_forward:int -> k_backward:int -> 'abstract t -> 'abstract t

  val extrapolation_alt : ?initial:bool ->  ?final:bool -> k_forward:int -> k_backward:int -> 'abstract t -> 'abstract t

  (*  ********************************************************************** *)
  (** 3 Operational Semantics *)
  (*  ********************************************************************** *)

  val output_trans : 
    'abstract manager -> 
    bexpr -> 
    Alpha.t -> 
    int -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [ouput_trans guard message channel assignment_list
	 old_abstract_value] computes the new abstract value after an output *)

 val pre_output_trans : 
    'abstract manager -> 
    bexpr -> 
    Alpha.t -> 
    int -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [pre_ouput_trans guard message channel assignment_list
	 old_abstract_value] computes the pre image of an output *)

  val input_trans : 
    'abstract manager -> 
    bexpr -> 
    Alpha.t -> 
    int -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [input_trans guard message channel assignment_list
	 old_abstract_value] computes the new abstract value after an input *)

  val pre_input_trans : 
    'abstract manager -> 
    bexpr -> 
    Alpha.t -> 
    int -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [pre_input_trans guard message channel assignment_list
	 old_abstract_value] computes the pre image of an input *)

  val epsilon_trans : 
    'abstract manager -> 
    bexpr -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [epsilon_trans guard assignment_list old_abstract_value]
	 computes the new abstract value after a transition without input nor output *)

 val pre_epsilon_trans : 
    'abstract manager -> 
    bexpr -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [pre_epsilon_trans guard assignment_list old_abstract_value]
	 computes the pre image of an epsilon transition *)

 

  end


module Make 
  (Alpha : Alphabet)
  : (S with module Alpha = Alpha)
  = 
struct

  module Alpha = Alpha


  type talphabet = Sep | AlphaLetter of Alpha.t


  let cmp_talphabet x y = 
    match (x,y) with
	(Sep,Sep) -> 0
      | (Sep,_) -> -1
      | (_,Sep) -> 1
      | (AlphaLetter ax, AlphaLetter ay) -> Alpha.compare ax ay      

  module AlphaSep =
    PLattice.MakeAlphabet (struct type t = talphabet let compare = cmp_talphabet  let sep = Sep end)


  let print_alphasep fmt x =
    match x with
	Sep -> Format.pp_print_string fmt "##" 
      | AlphaLetter c -> Alpha.print fmt c



  module PL = PLattice.Make (AlphaSep)

  module Langage = LAutomaton.Make(PL)

  module NQueues = StackQueueNRel.Make(Langage)


  type 'abstract t = 
      {
	vars : 'abstract ; (* abstraction of the values of the variables *)
	queues : 'abstract NQueues.t ; (* abstraction of the queues *)
	partition : 'abstract PL.t; (* partition of lattice automata *)
	lattice : 'abstract PL.lattice; (* ground lattice of lattice automata *)
      }

(* print function *)

  let print(man : 'abstract manager) fmt x =  
    (*    Format.fprintf fmt "vars= %a @, queues = @[%a @] @. "
	  man.print_abs x.vars
	  (NQueues.print (PL.print  print_alphasep man.print_abs)) x.queues *)
    begin
      fprintf fmt
        "{ @[<hov>variables = @[%a@]@,queues = @[%a@]@] }"
        man.print_abs x.vars
        (NQueues.print (PL.print  print_alphasep man.print_abs)) (x.queues)
    end


  let print_dot ?margin ?titlestyle ?vertexstyle ?edgestyle ?title (man : 'abstract manager) fmt x =
	  NQueues.print_dot 
	    ?margin 
	    ?titlestyle 
	    ?vertexstyle
	    ?edgestyle
	    ?title
	    (PL.print  print_alphasep man.print_abs) 
	    fmt 
	    x.queues


  let print_regexp (man : 'abstract manager) fmt x =  
    (*    WARNING : TO BE DONE *)
    begin
      fprintf fmt
        "{ @[<hov>queues = @[<hov>%a@]@] }"
        (NQueues.print_regexp (PL.print  print_alphasep man.print_abs)) (x.queues)
    end

  let get_automata_size x =
      NQueues.get_size x.queues



(* constructors *)

  let create_partition 
      (man : 'abstract manager)
      : 'abstract PL.t 
      =
    let param_list = man.list_param in
    List.fold_left
      (fun map_res s -> PL.A.Map.add (AlphaLetter (Alpha.from_string s)) (man.top man.env_param) map_res)
      (PL.A.Map.add Sep (man.top man.env_param) PL.A.Map.empty)
      param_list
      
  let create_lattice (man : 'abstract manager) : 'abstract PL.lattice = 
    { 
      PLattice.sep = man.top man.env_param;
      PLattice.is_bottom = (fun c a -> man.is_bottom a) ;
      PLattice.is_leq = (fun c a1 a2 -> man.is_leq a1 a2) ;
      PLattice.is_eq = (fun c a1 a2 -> (man.is_leq a1 a2)&& (man.is_leq a2 a1)) ;
      PLattice.join = (fun c a1 a2 -> man.join a1 a2) ;
      PLattice.meet = (fun c a1 a2 -> man.meet a1 a2) ;
      PLattice.widening = (fun c a1 a2 -> man.widening a1 a2) ;
    }

  let get_partition auto =
    auto.partition

  let get_lattice auto =
    auto.lattice


  let bottom (man : 'abstract manager) (partition : 'abstract PL.t) (lattice : 'abstract PL.lattice) =
    {
      vars = man.bottom man.env_var ;
      queues = NQueues.bottom  lattice partition man.nbqueues ;
      partition = partition ;
      lattice = lattice 
    }
    
  let top (man : 'abstract manager) (partition : 'abstract PL.t) (lattice : 'abstract PL.lattice) = 
    {
      vars = man.top man.env_var ;
      queues = NQueues.top lattice partition man.nbqueues;
      partition = partition ;
      lattice = lattice 
    }

  let is_bottom (man : 'abstract manager) x = 
    (man.is_bottom x.vars) || (NQueues.is_bottom x.queues )

  let is_top (man : 'abstract manager) x = 
    (man.is_top x.vars) && (NQueues.is_top x.queues) 


  let is_leq (man : 'abstract manager) (abs1 :'abstract t) (abs2 :'abstract t) : bool =
    let absv1 = abs1.vars
    and absv2 = abs2.vars
    and lang1 = abs1.queues
    and lang2 = abs2.queues
    in 
     (man.is_leq absv1 absv2) && (NQueues.is_leq lang1 lang2)

  let is_prefix (man : 'abstract manager) (w :string list) (abs :'abstract t) (i:int) : bool =
    let lang = abs.queues
    in
    let projection (s:string) (letter,abs) = 
      if cmp_talphabet (AlphaLetter (Alpha.from_string s)) letter = 0
      then abs
      else man.bottom man.env_param 
    in
      NQueues.is_prefix ~projection:projection w lang i

  let init_abstr (man : 'abstract manager) partition lattice aff_list =

    let abst = man.glob_to_var (man.assign_list (man.top man.env_global) aff_list) in
    let lang = NQueues.allempty lattice partition man.nbqueues in
      {
	vars = abst ;
	queues = lang ;
	partition = partition ;
	lattice = lattice
      }

	
  let init_bad_state (man : 'abstract manager) partition lattice bexpr regexpr =

    let regexpr = Regexp.map 
      (fun s -> 
	if s="#"
	then
	   AlphaSep.Map.add 
	     Sep
	     lattice.PLattice.sep
	     AlphaSep.Map.empty
	else
	   AlphaSep.Map.add 
	     (AlphaLetter (Alpha.from_string s))
	     lattice.PLattice.sep
	     AlphaSep.Map.empty
      ) 
      regexpr 
    in
    let abst = man.glob_to_var (man.from_guard bexpr)  in
    let lang = 
      match regexpr with
	  Regexp.Empty ->  NQueues.top lattice partition man.nbqueues
	    (* warning : here, Empty means all posible words, not the empty language *)
	| _ ->
	    NQueues.of_regexp lattice partition man.nbqueues regexpr
    in
      {
	vars = abst ;
	queues = lang ;
	partition = partition ;
	lattice = lattice
      }

  let elim_queues (abs :'abstract t) : 'abstract =
    abs.vars


  let complementary  (man : 'abstract manager) (abs :'abstract t) : 'abstract t =

    (* [GS:] Without this test, we may get a result where vars is bottom whereas
       queues is a non empty language.  E.g. when the input is such that vars is
       bottom and queues is also bottom. *)
    if (man.is_bottom abs.vars) || (NQueues.is_bottom abs.queues)
    then
      top man abs.partition abs.lattice
    else

    let absv = abs.vars
    and lang = NQueues.complementary abs.queues
    and partition = abs.partition
    and lattice = abs.lattice
    in
      if NQueues.is_bottom lang
      then
	bottom man partition lattice
      else
      {
	vars = absv ;
	queues = lang ;
	partition = partition ;
	lattice = lattice
      }

 let determinise  (man : 'abstract manager) (abs :'abstract t) : 'abstract t =
    let absv = abs.vars
    and lang = NQueues.f_determinise abs.queues
    and partition = abs.partition
    and lattice = abs.lattice
    in
      if NQueues.is_bottom lang
      then
	bottom man partition lattice
      else
      {
	vars = absv ;
	queues = lang ;
	partition = partition ;
	lattice = lattice
      }

 let minimise  (man : 'abstract manager) (abs :'abstract t) : 'abstract t =
    let absv = abs.vars
    and lang = NQueues.f_minimise abs.queues
    and partition = abs.partition
    and lattice = abs.lattice
    in
      if NQueues.is_bottom lang
      then
	bottom man partition lattice
      else
      {
	vars = absv ;
	queues = lang ;
	partition = partition ;
	lattice = lattice
      }

  let join (man : 'abstract manager) (abs1 :'abstract t) (abs2 :'abstract t) :'abstract t =
    let absv1 = abs1.vars
    and absv2 = abs2.vars
    and lang1 = abs1.queues
    and lang2 = abs2.queues
    in
(*      assert (abs1.partition = abs2.partition) ;
      assert (abs1.lattice = abs2.lattice) ; *)
      let lang_new =  NQueues.join lang1 lang2 in
	    
      let abst_new =  man.join absv1 absv2 in
	if 
	  (man.is_bottom abst_new) || (NQueues.is_bottom lang_new)
	then
	  {
	    vars = man.bottom man.env_var ;
	    queues = NQueues.bottom abs1.lattice abs1.partition man.nbqueues ;
	    partition = abs1.partition ;
	    lattice = abs1.lattice 
	  }
	else
	  {
	    vars = man.join absv1 absv2 ;
	    queues = lang_new ;
	    partition = abs1.partition ;
	    lattice = abs1.lattice 
	  }


  let join_list (man : 'abstract manager)lx =
    List.fold_left 
      (join man)
      (List.hd lx) 
      (List.tl lx)

  let meet (man : 'abstract manager)(abs1 :'abstract t) (abs2 :'abstract t) :'abstract t =
    let absv1 = abs1.vars
    and absv2 = abs2.vars
    and lang1 = abs1.queues
    and lang2 = abs2.queues
    in
(*      assert (abs1.partition = abs2.partition) ;
      assert (abs1.lattice = abs2.lattice) ; *)
      let absv_new = man.meet absv1 absv2
      and lang_new = NQueues.meet lang1 lang2
      in
	if (man.is_bottom absv_new) || (NQueues.is_bottom lang_new)
	then
	  {
	    vars = man.bottom man.env_var ;
	    queues = NQueues.bottom  abs1.lattice abs1.partition man.nbqueues ;
	    partition = abs1.partition ;
	    lattice = abs1.lattice 
	  }
	else
	  {
	    vars = man.join absv1 absv2 ;
	    queues = lang_new ;
	    partition = abs1.partition ;
	    lattice = abs1.lattice 
	  }

  let widening ?(initial=true)  ?(final=true) (man : 'abstract manager)(abs1 :'abstract t) (abs2 :'abstract t) :'abstract t =
(*    assert (abs1.partition = abs2.partition) ;
    assert (abs1.lattice = abs2.lattice) ; *)
    if is_leq man abs1 abs2
    then 
      {
	vars = man.widening abs1.vars abs2.vars; 
	queues = NQueues.widening_shape 
	  (Langage.Abstract.standard ~initial:initial ~final:final ~backward:man.k_backward ~forward:man.k_forward) 
	  abs1.queues
	  abs2.queues ;
	partition = abs1.partition ;
	lattice = abs1.lattice ;
      }
    else  failwith "Widening StandardSem : inclusion test failed"

 let extrapolation ?(initial=true)  ?(final=true) ~(k_forward:int) ~(k_backward:int) (abs :'abstract t) :'abstract t =
      {
	vars = abs.vars;
	queues = ref (Langage.Abstract.standard
			~initial:initial
			~final:final
			~backward:k_backward
			~forward:k_forward
			!(abs.queues)) ;
	partition = abs.partition ;
	lattice = abs.lattice ;
      }



  let widening_alt ?(initial=true)  ?(final=true) (man : 'abstract manager)(abs1 :'abstract t) (abs2 :'abstract t) :'abstract t =
(*    assert (abs1.partition = abs2.partition) ;
    assert (abs1.lattice = abs2.lattice) ; *)
    if is_leq man abs1 abs2
    then 
      {
	vars = man.widening abs1.vars abs2.vars; 
	queues = NQueues.widening_shape 
	  (Langage.Abstract.bounded_languages 
	     ~initial:initial
	     ~final:final
	     ~backward:man.k_backward 
	     ~forward:man.k_forward) 
	  abs1.queues
	  abs2.queues ;
	partition = abs1.partition ;
	lattice = abs1.lattice ;
      }
    else  failwith "Widening StandardSem : inclusion test failed"

  let extrapolation_alt ?(initial=true)  ?(final=true) ~(k_forward:int) ~(k_backward:int) (abs :'abstract t) :'abstract t =
      {
	vars = abs.vars;
	queues = ref (Langage.Abstract.bounded_languages
			~initial:initial
			~final:final
			~backward:k_backward
			~forward:k_forward
			!(abs.queues)) ;
	partition = abs.partition ;
	lattice = abs.lattice ;
      }

(* output et input *)

(* ------------------------------------------------------------------------- *)
(* output-transitions *)
(* ------------------------------------------------------------------------- *)

  let rec output_trans 
      (man : 'abstract manager)
      (garde : bexpr) 
      (message : Alpha.t)
      (channel : int)
      (affectations : (var * iexpr) list)
      (ancienne_valeur :'abstract t) 
      :'abstract t 
      =
    let absv_old = ancienne_valeur.vars
    and lang_old = ancienne_valeur.queues
    and partition = ancienne_valeur.partition
    and lattice = ancienne_valeur.lattice
    in 
    let absv_g = man.meet (man.var_to_glob absv_old) (man.from_guard garde) in
      if man.is_bottom absv_g 
      then
	 bottom man partition lattice
      else
	begin	  
	  let absv_param = man.glob_to_param absv_g in
	    (* oubli de la valeur des variables *)
	  let plItem = 
	    AlphaSep.Map.add 
	      (AlphaLetter message)  
	      absv_param
	      AlphaSep.Map.empty
	  in
	    (* partition contenant le message + la valeur des parametres *)
	  (*  print_alphasep Format.std_formatter (AlphaLetter message) ; *)
	    let absv_new =
	      man.glob_to_var (man.assign_list absv_g affectations)
		(* valeur apres les affectations *)
	    and lang_new = NQueues.Right.queue lang_old channel plItem
	    in	
	      if NQueues.is_bottom lang_new
	      then  bottom man partition lattice
	      else 
		{
		  vars = absv_new ;
		  queues = lang_new ;
		  partition= partition ;
		  lattice = lattice ;
		}
	end
			


  let pre_output_trans 
      (man : 'abstract manager)
      (garde : bexpr) 
      (message : Alpha.t)
      (channel : int)
      (affectations : (var * iexpr) list)
      (ancienne_valeur :'abstract t) 
      :'abstract t 
      =
    let absv_old = ancienne_valeur.vars
    and lang_old = ancienne_valeur.queues
    and partition = ancienne_valeur.partition
    and lattice = ancienne_valeur.lattice
    in 
    let absv_pre = man.rev_assign_list absv_old affectations in
    let (lang_new,param) = 
      begin
	try 
	  AlphaSep.Map.find 
	    (AlphaLetter message) 
	    (NQueues.Left.dequeue_partitioned 
		~cond:(PL.A.Map.add 
			  (AlphaLetter message) 
			  (man.top man.env_param) 
			  PL.A.Map.empty) 
		lang_old channel) 
	with 
	    Not_found -> 
	      (NQueues.bottom (NQueues.get_lattice lang_old) (NQueues.get_partition lang_old) man.nbqueues, man.bottom man.env_param)
		(* cherche le message dans la table, bottom sinon *)
      end 
    in
    let absv_param = man.meet (man.param_to_glob param) absv_pre in

    let absv_g = man.meet absv_param (man.from_guard garde) in

      if (man.is_bottom absv_g)  || (NQueues.is_bottom lang_new)
      then bottom man partition lattice
      else
	{
	  vars = absv_g ;
	  queues = lang_new ;
	  partition= partition ;
	  lattice = lattice ;
	}



(* ------------------------------------------------------------------------- *)
(* input-transitions *)
(* ------------------------------------------------------------------------- *)

  let input_trans 
      (man : 'abstract manager)
      (garde : bexpr) 
      (message : Alpha.t)
      (channel : int)
      (affectations : (var * iexpr) list)
      (ancienne_valeur :'abstract t) 
      :'abstract t 
      =
    let absv_old = ancienne_valeur.vars
    and lang_old = ancienne_valeur.queues
    and partition = ancienne_valeur.partition
    and lattice = ancienne_valeur.lattice
    in 
    let absv_g = man.meet (man.var_to_glob absv_old) (man.from_guard garde) in
    let (lang_new,param) = 
      begin
	try AlphaSep.Map.find (AlphaLetter message) (NQueues.Right.dequeue_partitioned lang_old channel) 
	with Not_found -> (NQueues.bottom (NQueues.get_lattice lang_old) (NQueues.get_partition lang_old) man.nbqueues, man.bottom man.env_param)
	  (* cherche le message dans la table, bottom sinon *)
      end 
    in
    let absv_param = 
      man.meet absv_g (man.param_to_glob param) 
	(* intersection garde et valeur du paramettre *)
    in 
      if (man.is_bottom absv_param) || (NQueues.is_bottom lang_new)
      then 
	bottom man partition lattice
      else
	let absv_new =
	  man.glob_to_var (man.assign_list absv_param affectations)
	    (* valeur apres les affectations *)
	in
	  {
	    vars = absv_new ;
	    queues = lang_new ;
	    partition= partition ;
	    lattice = lattice ;
	  }		    		    
			   


 let pre_input_trans 
     (man : 'abstract manager)
     (garde : bexpr) 
     (message : Alpha.t)
     (channel : int)
     (affectations : (var * iexpr) list)
     (ancienne_valeur :'abstract t) 
     :'abstract t 
     =
   let absv_old = ancienne_valeur.vars
   and lang_old = ancienne_valeur.queues
   and partition = ancienne_valeur.partition
   and lattice = ancienne_valeur.lattice
   in 
   let absv_pre = man.rev_assign_list absv_old affectations in
   let absv_g = man.meet absv_pre (man.from_guard garde) in
     (* intersection G cap A^-1(absv_old) *) 
     if man.is_bottom absv_g
     then 
    
	 bottom man partition lattice
      
     else
       begin
	 let absv_param = man.glob_to_param absv_g 
	 in
	   (*debug Format.printf "avant: @[%a@]@." man.print_abs absv_param ;*) 
	 let absv_new = man.glob_to_var absv_g 
	 in
	   (*debug Format.printf "apres: @[%a@]@." man.print_abs absv_new ;*) 
	 let plItem = 
	   AlphaSep.Map.add 
	     (AlphaLetter message)  
	     absv_param
	     AlphaSep.Map.empty
	 in
	 let lang_new = NQueues.Left.queue lang_old channel plItem in
	   {
	     vars = absv_new ;
	     queues = lang_new ;
	     partition= partition ;
	     lattice = lattice ;
	   }			    		    
       end
			  




(* ------------------------------------------------------------------------- *)
(* epsilon-transitions *)
(* ------------------------------------------------------------------------- *)

 let epsilon_trans 
     (man : 'abstract manager)
     (garde : bexpr) 
     (affectations : (var * iexpr) list)
     (ancienne_valeur :'abstract t) 
     :'abstract t 
     = 
   let absv_old = ancienne_valeur.vars
   and lang_old = ancienne_valeur.queues
   and partition = ancienne_valeur.partition
   and lattice = ancienne_valeur.lattice
   in 
   let absv_g = man.meet (man.var_to_glob absv_old) (man.from_guard garde) in
     if man.is_bottom absv_g 
     then
       bottom man partition lattice
     else     
       let absv_new =
	 man.glob_to_var (man.assign_list absv_g affectations)
	   (* valeur apres les affectations *)
       in
	 {
	   vars = absv_new ;
	   queues = lang_old ;
	   partition= partition ;
	   lattice = lattice ;
	 }	

  let pre_epsilon_trans 
      (man : 'abstract manager)
      (garde : bexpr) 
      (affectations : (var * iexpr) list)
      (ancienne_valeur :'abstract t) 
      :'abstract t 
      = 
    let absv_old = ancienne_valeur.vars
    and lang_old = ancienne_valeur.queues
    and partition = ancienne_valeur.partition
    and lattice = ancienne_valeur.lattice
    in 
    let absv_pre = man.rev_assign_list absv_old affectations in
    let absv_new = man.glob_to_var (man.meet absv_pre (man.from_guard garde)) in
      if man.is_bottom absv_new 
      then
	bottom man partition lattice
      else     
	{
	  vars = absv_new ;
	  queues = lang_old ;
	  partition= partition ;
	  lattice = lattice ;
	}	



end

