(***********************************************************************)
(*                                                                     *)
(*            The Symbolic Communicating Machines Library              *)
(*                                                                     *)
(*                Bertrand Jeannet and Tristan Le Gall                 *)
(*                                                                     *)
(*  Copyright 2008 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file LICENSE.        *)
(*                                                                     *)
(***********************************************************************)

(** SCM analyzer: Standard Semantics module *)


(* ####################################################################
######################## Module Standard Sem ##########################
#################################################################### *)


open Format
open Scm_syn
open Scm_manager

(*  ********************************************************************** *)
(** 2 Module type for the alphabet *)
(*  ********************************************************************** *)

module type Alphabet = sig
  type t
  val from_string: string -> t
    (** creates a letter from a string *)
  val compare: t -> t -> int
(** [compare x y] returns 0 if [x] is equal to [y], a negative integer if [x] 
   is less than [y], and a positive integer if [x] is greater than [y] *)
  val print: Format.formatter -> t -> unit
  end


(*  ********************************************************************** *)
(** 2 Functor *)
(*  ********************************************************************** *)

module type S = sig

  module Alpha : Alphabet
    (** module of the alphabet/finite partition of the lattice *)

  module AlphaSep : PLattice.Alphabet
    (** same thing, with the separation letter *)

  module PL : (PLattice.S with module A=AlphaSep)
    (** partitionned lattice *)

  module Langage : (LAutomaton.S  with module A=AlphaSep
				   and module PL=PL )
    (** QDD-like automata *)

  module NQueues : 
    (StackQueueNRel.S  
      with module PL=PL
      and module LAuto=Langage)
    (** abstraction of the queue-contents *)

  type 'abstract t
    (** abstract value *)


  (*  ********************************************************************** *)
  (** 3 Print *)
  (*  ********************************************************************** *)

  val print : 'abstract manager -> Format.formatter -> 'abstract t -> unit

  val print_dot : 
    ?margin:int ->
    ?titlestyle:string ->
    ?vertexstyle:string ->
    ?edgestyle:string ->
    ?title:string ->
    'abstract manager -> 
    Format.formatter -> 
    'abstract t -> 
    unit
      (** prints only the automaton part of the abstract value *)

  val print_regexp : 'abstract manager -> Format.formatter -> 'abstract t -> unit

  val get_automata_size : 'abstract t -> int

  (*  ********************************************************************** *)
  (** 3 Constructors *)
  (*  ********************************************************************** *)

  val bottom : 'abstract manager -> 'abstract PL.t -> 'abstract PL.lattice -> 'abstract t
  val top : 'abstract manager -> 'abstract PL.t -> 'abstract PL.lattice -> 'abstract t
    (** contructors.  NB: the same lattice (physical equality ==) and
	the same partition must be used in order to have compatible
	automata *)

  val create_partition : 'abstract manager -> 'abstract PL.t 

  val create_lattice : 'abstract manager -> 'abstract PL.lattice
    (** create a partition and a lattice.  NB: the same lattice
	(physical equality ==) and the same partition must be used in
	order to have compatible automata *)

  val get_partition : 'abstract t -> 'abstract PL.t 

  val get_lattice : 'abstract t -> 'abstract PL.lattice
    (** returns the partition and lattice of the automaton part.  NB: the same lattice
	(physical equality ==) and the same partition must be used in
	order to have compatible automata *)


  val init_abstr : 'abstract manager -> 'abstract PL.t -> 'abstract PL.lattice -> (var * iexpr) list -> 'abstract t
    (** [init_abstr partition lattice assignment_list] creates an abstract value, all
	the [n] queues being empty.  The initial values of the variables
	are defined by the assignement list.  *)

  val init_bad_state : 'abstract manager -> 'abstract PL.t -> 'abstract PL.lattice -> bexpr -> string Regexp.t -> 'abstract t
    (** [init_bad_State partition lattice bexpr] creates an abstract
	value, the content of the queue being possibly
	everything. This function helps to create a set of bad
	configurations. *)

  val elim_queues : 'abstract t -> 'abstract

  (*  ********************************************************************** *)
  (** 3 Comparison operators *)
  (*  ********************************************************************** *)

  val is_bottom :  'abstract manager -> 'abstract t -> bool

  val is_top :  'abstract manager -> 'abstract t -> bool

  val is_leq : 'abstract manager -> 'abstract t  -> 'abstract t -> bool
    (** comparision operators: NB: is_bottom is a test on the physical
	representation, others are tests involving language inclusion *)

  val is_prefix : 'abstract manager -> string list -> 'abstract t -> int -> bool 
    (** [is_prefix man w auto i] checks whether the word [w] is
	a prefix of a word in the queue [i] of [auto]. *)

  (*  ********************************************************************** *)
  (** 3 Operations on the queue content *)
  (*  ********************************************************************** *)

  (* operations on the queue only *)
  val complementary :  'abstract manager -> 'abstract t  -> 'abstract t
    (** complements the lattice automaton : ONLY WORKS WITH FINITE AUTOMATA *)

  val determinise : 'abstract manager -> 'abstract t -> 'abstract t
    (** determinise the lattice automaton *)
  val minimise : 'abstract manager -> 'abstract t -> 'abstract t
    (** minimise the lattice automaton *)

  (*  ********************************************************************** *)
  (** 3 Lattice operations *)
  (*  ********************************************************************** *)

  val join : 'abstract manager -> 'abstract t -> 'abstract t -> 'abstract t 
    (** join on the queues part is the language union, with merging *)

  val meet : 'abstract manager -> 'abstract t -> 'abstract t -> 'abstract t 
    (** meet on the queues part is the language intersection *)

  val widening : ?initial:bool ->  ?final:bool -> 'abstract manager ->  'abstract t -> 'abstract t -> 'abstract t 
    (**  widening of the abstract values, using coloured bisimulation. The user can decide whether [initial] states and [final] ones should differ from ordinary states of the automaton *)


  val widening_alt : ?initial:bool ->  ?final:bool -> 'abstract manager ->  'abstract t -> 'abstract t -> 'abstract t  
    (** widening of the abstract values, using bounded
	languages. [initial] and [final] are useless, but are included so
	this function has the same signature as widening *)

  (*  ********************************************************************** *)
  (** 3 extrapolation operations *)
  (*  ********************************************************************** *) 
  (** extrapolation operaor are similar to widening operator, but are applied on a single abstract value *) 

  val extrapolation : ?initial:bool ->  ?final:bool -> k_forward:int -> k_backward:int -> 'abstract t -> 'abstract t

  val extrapolation_alt : ?initial:bool ->  ?final:bool -> k_forward:int -> k_backward:int -> 'abstract t -> 'abstract t

  (*  ********************************************************************** *)
  (** 3 Operational Semantics *)
  (*  ********************************************************************** *)

  val output_trans : 
    'abstract manager -> 
    bexpr -> 
    Alpha.t -> 
    int -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [ouput_trans guard message channel assignment_list
	 old_abstract_value] computes the new abstract value after an output *)

  val pre_output_trans : 
    'abstract manager -> 
    bexpr -> 
    Alpha.t -> 
    int -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [pre_ouput_trans guard message channel assignment_list
	 old_abstract_value] computes the pre image of an output *)

  val input_trans : 
    'abstract manager -> 
    bexpr -> 
    Alpha.t -> 
    int -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [input_trans guard message channel assignment_list
	 old_abstract_value] computes the new abstract value after an input *)

  val pre_input_trans : 
    'abstract manager -> 
    bexpr -> 
    Alpha.t -> 
    int -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [pre_input_trans guard message channel assignment_list
	 old_abstract_value] computes the pre image of an input *)

  val epsilon_trans : 
    'abstract manager -> 
    bexpr -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [epsilon_trans guard assignment_list old_abstract_value]
	 computes the new abstract value after a transition without input nor output *)

  val pre_epsilon_trans : 
    'abstract manager -> 
    bexpr -> 
    (var * iexpr) list ->
    'abstract t -> 
    'abstract t
      (* [pre_epsilon_trans guard assignment_list old_abstract_value]
	 computes the pre image of an epsilon transition *)


end



module Make (Alpha1 : Alphabet) : (S with module Alpha = Alpha1)
 
