(* $Id: myocamlbuild.ml 1164 2011-10-12 10:18:43Z gsutre $ *)

(*
 * Copyright (c) 2010, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Alexander Heußner <alexander.heussner@labri.fr>
 *            Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(* Configuration **************************************************************)

(* Set to the list of directories containing *.mlpack files. *)
let mlpack_dirs = ["src"]

(* Set to the list of extra ocamldoc options for doc generation. *)
let ocamldoc_extra_options = ["-stars"; "-colorize-code"; "-hide"; "Pervasives"]


(* Ocamlbuild Plugin Code *****************************************************)

open Ocamlbuild_plugin

(*
 * This module fixes link problems with OCaml packs.
 *
 * Warning: this implementation interprets paths in .mlpack files as relative
 * paths w.r.t. the location of the .mlpack file.
 *
 * Generates a .ml.depends file for each .mlpack file.  This fixes byte-code
 * link problems with packs containing modules that depend on modules that are
 * outside of the pack.
 *
 * Adds a for-pack tag to each .cmx file listed in a .mlpack file.  This is
 * necessary for native compilation of packs.  The profile variant (.p.cmx) is
 * not supported since ocamlopt fails to link packages when module object files
 * have been renamed (unless they are independent from each other).
 *)
module PackageLinkFix =
struct
  let packages_in_dir dir =
    Array.fold_right
      (fun f l ->
	 if (Pathname.check_extension f "mlpack") then
	   (dir / (Pathname.remove_extension f)) :: l
	 else
	   l)
      (Pathname.readdir dir)
      []

  let byte_dep_mlpack arg out env _ =
    let arg = env arg and out = env out in
      Echo (([arg; ":"] @
	       (List.map
		  (fun s -> " "^s)
		  (string_list_of_file (arg))) @ ["\n"]),
	    out)

  let add_dependency pack =
    let file = Pathname.add_extension "ml.depends" pack
    and tags = ["ocaml"; "byte"; "pack"; "extension:cmo";
		"file:"^(Pathname.add_extension "cmo" pack)]
    in
      Ocamlbuild_pack.Log.dprintf 10
	"plugin:@ PackageLinkFix:@ dep:@ %S@ when@ %a"
	file Ocamlbuild_pack.Tools.pp_l tags ;
      dep tags [file]

  let add_tags pack =
    List.iter
      (fun s ->
	 let file = (Pathname.add_extension "cmx"
		       ((Pathname.dirname pack) / (Pathname.mk s)))
	 and tags = ["for-pack("^(Pathname.basename pack)^")"]
	 in
	   Ocamlbuild_pack.Log.dprintf 10
	     "plugin:@ PackageLinkFix:@ tag:@ %S@ with@ %a"
	     file Ocamlbuild_pack.Tools.pp_l tags ;
	   tag_file file tags)
      (string_list_of_file (Pathname.add_extension "mlpack" pack))

  let after_rules () =
    begin
      rule "ocaml dependencies mlpack"
	~prod:"%.ml.depends"
	~dep:"%.mlpack"
	(byte_dep_mlpack "%.mlpack" "%.ml.depends")
      ;
      List.iter
	(fun p -> add_dependency p ; add_tags p)
	(List.concat (List.map packages_in_dir (List.map Pathname.mk mlpack_dirs)))
    end
end

(*
 * Adds a rule to produce an .odocl file from a .mlpack file.  The .odocl is
 * simply a symbolic link to the .mlpack file.
 *)
module PackageOcamldoc =
struct
  let make_symlink arg out env _ =
    let arg = env arg and out = env out in
      ln_s (Pathname.basename arg) out

  let after_rules () =
    begin
      rule "ocaml: mlpack -> odocl"
	~prod:"%.odocl"
	~dep:"%.mlpack"
	(make_symlink "%.mlpack" "%.odocl")
    end
end

(*
 * Adds unsafe, noassert and static tags with obvious meaning.
 *)
module CustomTags =
struct
  let after_rules () =
    begin
      flag ["ocaml"; "compile"; "unsafe"] (A "-unsafe") ;
      flag ["ocaml"; "compile"; "noassert"] (A "-noassert") ;
      flag ["ocaml"; "link"; "native"; "program"; "static"]
	(Command.atomize ["-ccopt"; "-static"])
    end
end

(*
 * Adds a rule to display the system architecture.
 *)
module SystemArchitecture =
struct
  let after_rules () =
    begin
      rule "ocamlbuild: show architecture"
	~prods:["__show_architecture__"]
	~deps:[]
	(fun _ _ ->
	  Printf.printf "%s\n"
	    Ocamlbuild_pack.Ocamlbuild_Myocamlbuild_config.arch ;
	  Nop)
    end
end

(*
 * Adds a rule to populate missing files in 3rd-party/scm/NoApron from the
 * files in 3rd-party/scm.  The rule creates symbolic links.
 *)
module ScmNoApronSymlinks =
struct
  let make_symlink arg out env _ =
    let arg = env arg and out = env out in
      ln_s arg out

  let after_rules () =
    begin
      rule "Scm/NoApron symbolic links"
	~prod:"3rd-party/scm/NoApron/%"
	~deps:["3rd-party/scm/NoApron/README"; "3rd-party/scm/%"]
	(make_symlink "../%" "3rd-party/scm/NoApron/%")
    end
end

(*
 * Adds a rule to generate config.ml.
 *)
module Config =
struct
  let make_config _ _ =
    let version = with_input_file "VERSION" input_line in
    let substitutions = [("%%VERSION%%", version)]
    in
      Cmd (S [A "sed";
	      S (List.map
		   (fun (s, r) -> S [A "-e"; A ("s,"^s^","^r^",")])
		   substitutions);
	      P "config.mlp"; Sh ">"; Px "config.ml"])

  let after_rules () =
    begin
      rule "config.ml"
	~prod:"config.ml"
	~deps:["config.mlp"; "VERSION"]
	make_config
    end
end

;;

dispatch
  begin
    function
	After_rules ->
	  begin
	    PackageLinkFix.after_rules () ;
	    PackageOcamldoc.after_rules () ;
	    CustomTags.after_rules () ;
	    SystemArchitecture.after_rules () ;
	    ScmNoApronSymlinks.after_rules () ;
	    Config.after_rules () ;
	    if Sys.ocaml_version >= "3.12.1" then
	      flag ["ocaml"; "doc"] (Command.atomize ["-charset"; "UTF-8"]) ;
	    (* Beautify ocamldoc output. *)
	    flag ["ocaml"; "doc"] (Command.atomize ocamldoc_extra_options)
	  end
      | _ -> ()
  end
