(* $Id: APInv.ml 1180 2011-10-13 16:17:16Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util


module MakeBwd (A : Model.SEMANTICS_CLOSURE) =
struct
  open A

  let k_min = (Option.Main.k_min ())

  include TraceCheckerUtil.Common (A)

  let split r1 r2 =
    assert (Region.leq (Region.inter r1 r2) Region.bot) ;
    Logger.debug "@[<v 3>Split:@,@[%a@]@,@[%a@]@]@."
      Region.print r1 Region.print r2 ;
    let rec loop k =
      let r1' = clo k r1
      in
        if Region.leq (Region.inter r1' r2) Region.bot
        then (k, r1')
        else loop (k+1)
    in
      loop k_min

  let co_split r1 r2 =
    assert (Region.leq (Region.inter r1 r2) Region.bot) ;
    Logger.debug "@[<v 3>Co-Split:@,@[%a@]@,@[%a@]@]@."
      Region.print r1 Region.print r2 ;
    let rec loop k =
      let r2' = clo k r2
      in
        if Region.leq (Region.inter r1 r2') Region.bot
        then (k, Region.compl r2')
        else loop (k+1)
    in
      loop k_min

  (* Set safe over-approximation function based on selectable option. *)
  let approximate =
    match Option.APInv.safe_approximation () with
        Option.APInv.Split -> split
      | Option.APInv.CoSplit -> co_split


  (*
   * [fwd_img]:
   *
   *	Performs forward (exact) post computation along the given [trace].
   *	Returns the array of post regions along with the index of the first
   *	empty region in the array, or None if [trace] is feasible.
   *)
  let fwd_img trace =
    let img = Array.create (Array.length trace) Region.bot in
    let rec fwd_img_aux current i =
      assert (not (Region.leq current Region.bot)) ;
      if i < Array.length trace
      then
        let postimg = atom_strong_post current trace.(i)
        in
          if Region.leq postimg Region.bot
          then i
          else
            begin
              img.(i) <- postimg ;
              fwd_img_aux postimg (i+1)
            end
      else
        -1
    in
      let i = fwd_img_aux Region.top 0
      in
        if i < 0 then None else Some (img, i)

  (*
   * [bwd_trace_inv]:
   *
   *	Performs backward approximated pre computation along the given [trace].
   *	Approximation is applied only for [`State] atoms, and it is performed by
   *	splitting the obtained pre region with respect to the corresponding post
   *	region in [img].  This function returns a backward trace invariant.
   *
   *	The [trace] is assumed to be unfeasible, and the [current] region is
   *	assumed to be non-empty.
   *)
  let rec bwd_trace_inv trace img max_k inv_l inv_n current i =
    assert (not (Region.leq current Region.bot)) ;
    assert (i >= 0) ;
    let preimg = atom_strong_pre current trace.(i)
    in
      if Region.leq preimg Region.bot
      then
        (max_k, i, (inv_l, Sequence.Forward, inv_n))
      else
        match trace.(i) with
            `State (s, r) ->
              (*
               * An alternative here would be to over-approximate preimg
               * instead of current.  We choose to approximate current to
               * conform with the APInv algorithm of the SPIN'09 paper.
               *)
              let (k, preimg') = approximate current img.(i)
              in
                bwd_trace_inv trace img (max k max_k)
                  (((s, r), preimg') :: inv_l) (inv_n + 1)
                  (Region.inter preimg' r) (i-1)
          | _ ->
              bwd_trace_inv trace img max_k inv_l inv_n preimg (i-1)

  (*
   * [fwd_next_state]:
   *
   *	Returns the index of the next [`State] atom in the given [trace].
   *)
  let rec fwd_next_state trace i =
    match trace.(i) with
        `State _ -> i
      | _ ->
          if i = (Array.length trace) - 1
          then i
          else fwd_next_state trace (i+1)

  let bwd_trace_check trace =
    let trace = Sequence.to_array trace
    in
      match fwd_img trace with
          None ->
            `Feasible `Strong
        | Some (img, i) ->
            (*
             * We shift [i] to the next [`State] atom in the trace, in order
             * to conform with the APInv algorithm of the SPIN'09 paper.  But
             * this choice is arbitrary...  Worse, it often adversely impacts
             * Cegar performance in practice (according to some experiments).
             *
             * XXX Add an option to control this?  For instance:
             * XXX	-extraction-shift {none|state|max}
             * XXX		Position shift before invariant extraction
             *)
            let j = fwd_next_state trace i in
            let (max_k, i, s) =
              let preimg = atom_strong_pre Region.top trace.(j)
              in
                if Region.leq preimg Region.bot
                then
                  (min_int, j-1, ([], Sequence.Forward, 0))
                else
                  bwd_trace_inv trace img min_int [] 0 preimg (j-1)
            in
              (* Display max_k as -∞ if closure hasn't been used. *)
              if (max_k == min_int) then
                begin
                  Logger.slog Logger.Notice "@[ /       -∞@]@?" ;
                  Logger.info "@[Closure parameter: -∞@]@."
                end
              else
                begin
                  Logger.slog Logger.Notice "@[ / %8d@]@?" max_k ;
                  Logger.info "@[Closure parameter: %d@]@." max_k
                end ;
              `Invariant_backward
                (num_state_atoms trace 0 i,
                 s,
                 num_state_atoms trace j ((Array.length trace) - 1))

  let trace_check = bwd_trace_check
end

module MakeFwd = TraceCheckerUtil.Reverse (MakeBwd)
