(* $Id: Armc.ml 1269 2011-10-19 23:13:22Z gsutre $ *)

(*
 * Copyright (c) 2009, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util

(* In the RegLocMap module, leq and disjoint should be optimized by stopping
   fold as soon as the boolean b becomes false *)


(* Exception for errors that are internal to this module, such as cases in
   in pattern matching that should not arise. *)
exception InternalError

module Make (A : Model.S) =
struct
  (* We open A and define shortcuts to simplify notations *)
  open A
  module Location = Automaton.Location
  module Transition = Automaton.Transition
  module Region = Semantics.Region

  (* Maps from locations to regions *)
  module RegLocMap =
  struct
    include Map.Make(Location)

    (* XXX Use simplification function(s) from REGION signature. *)
    let simplify rlm =
      iter (fun l r -> ignore (Region.leq r r)) rlm

    let to_state_list rlm =
      fold
        (fun l r sl -> (l, r) :: sl)
        rlm []

    let print fmt rlm =
      (* simplify rlm ; *)
      (Util.Print.list_printer_from_printer "@;&@ " State.print) fmt (to_state_list rlm)

    let reg_of_loc rlm l =
      try find l rlm with
          Not_found -> Region.bot

    let leq rlm1 rlm2 =
      fold
	(fun l r b ->
	   b &&
             (Region.leq r (reg_of_loc rlm2 l)))
	rlm1
        true

    let are_disjoint rlm1 rlm2 =
      fold
	(fun l r b ->
	   b &&
             (Region.leq (Region.inter r (reg_of_loc rlm2 l)) Region.bot))
	rlm1
        true

    let is_empty rlm =
      fold
	(fun l r b ->
	   b &&
             (Region.leq r Region.bot))
	rlm
        true

    let inter rlm1 rlm2 =
      fold
	(fun l r rlm ->
	   let r' = Region.inter r (reg_of_loc rlm2 l)
	   in
	     if Region.leq r' Region.bot
	     then
	       rlm
	     else
	       add l r' rlm)
	rlm1
        empty

    let add_with_union rlm l r =
      if Region.leq r Region.bot
      then
	rlm
      else
	add l (Region.union (reg_of_loc rlm l) r) rlm
  end

  let init_reg_loc_map =
    let result =
      fold_init
        (fun (l, r) rlm -> RegLocMap.add_with_union rlm l r)
        RegLocMap.empty
    in
      (* RegLocMap.simplify result ; *)
      result

  let error_reg_loc_map =
    let result =
      fold_error
        (fun (l, r) rlm -> RegLocMap.add_with_union rlm l r)
        RegLocMap.empty
    in
      (* RegLocMap.simplify result ; *)
      result


  (* This function computes X U post(X) *)
  let reflexive_post reg_loc_map =
    let fwd_update reg trans rlm =
      let cmd = Transition.command trans
      and tgt = Transition.target trans in
      let post = Semantics.post reg cmd
      in
	if Region.leq post Region.bot
	then
	  rlm
	else
	  RegLocMap.add_with_union rlm tgt post
    in
      RegLocMap.fold
	(fun loc reg map ->
	   Automaton.fold_out_transitions (fwd_update reg) loc map)
	reg_loc_map
	reg_loc_map

  (* This function computes X U pre(X) *)
  let reflexive_pre reg_loc_map =
    let bwd_update reg trans rlm =
      let cmd = Transition.command trans
      and src = Transition.source trans in
      let pre = Semantics.pre reg cmd
      in
	if Region.leq pre Region.bot
	then
	  rlm
	else
	  RegLocMap.add_with_union rlm src pre
    in
      RegLocMap.fold
	(fun loc reg map ->
	   Automaton.fold_in_transitions (bwd_update reg) loc map)
	reg_loc_map
	reg_loc_map


  (*
  let pick_non_empty reg_loc_map =
    RegLocMap.fold
      (fun loc reg result ->
         match result with
             Some _ -> result
           | None ->
               if Region.leq reg Region.bot
               then
                 None
               else
                 Some (loc, reg))
      reg_loc_map
      None
  *)

  let find_succ loc reg rlm =
    (*
    Logger.debug "@[<v>In find_succ@ loc=@[%a@]@ reg=@[%a@]@]@."
      Location.print loc Region.print reg ;
    *)
    Automaton.fold_out_transitions
      (fun trans result ->
         match result with
             Some _ -> result
           | None ->
               let cmd = Transition.command trans
               and tgt = Transition.target trans in
               let post = Semantics.post reg cmd in
               let post_inter = Region.inter post (RegLocMap.reg_of_loc rlm tgt)
               in
                 assert (Location.equal loc (Transition.source trans)) ;
                 (* assert (Region.leq post_inter post) ; (* remark: this forces determinization of post_inter *) *)
                 if Region.leq post_inter Region.bot
                 then
                   None
                 else
                   Some (trans, tgt, post_inter))
      loc None

  let find_init rlm =
    A.fold_init
      (fun (loc, reg) result ->
         match result with
             Some _ -> result
           | None ->
               if Region.leq (Region.inter reg (RegLocMap.reg_of_loc rlm loc)) Region.bot
               then
                 None
               else
                 Some (loc, reg))
      None

  let extract_counter_example path =
    (*
    Logger.debug "@[<v>In extract_counter_example:@ path=@[%a@]@]@."
      (Util.Print.list_printer_from_printer "@ @ ;@ @ " RegLocMap.print) path ;
    *)
    let rec do_it loc reg suffix acc =
      (*
      Logger.debug "@[<v>In do_it@ acc=[@[%a@]]@ loc=@[%a@]@ reg=@[%a@]@]@."
        (Util.Print.list_printer_from_printer "@ ;@ " Transition.print) acc
        Location.print loc
        Region.print reg
      ;
      Logger.debug "@[Suffix: @[<v>%a@]@]@."
        (Util.Print.list_printer_from_printer "@ @ ;@ @ " RegLocMap.print) suffix
        ;
      *)
      match suffix with
          [] -> acc
        | rlm :: tail ->
            match find_succ loc reg rlm with
                None ->
                  (* This should not happen *)
                  raise InternalError
              | Some (trans, tgt, reg') ->
                  do_it tgt reg' tail (trans::acc)
    in
      match path with
          [] -> []
        | rlm :: tail ->
            match find_init rlm with
                None ->
                  (* This should not happen *)
                  raise InternalError
              | Some (loc, reg) ->
                  (List.rev (do_it loc reg tail []))


  (*****                        M A I N    L O O P                        *****)

  type result =
      Invariant of ((A.State.t list) Lazy.t) option * ((A.State.t list) Lazy.t) option
    | Counterexample of Transition.t list
    | Abort of string

  let armc_limit = (Option.Main.limit ())
  let k_min = (Option.Main.k_min ())
  let k_max = max_int

  (* For beautiful header with dynamic statistics. *)
  let backspace = ref 0

  (*
   * Function implementing the Abstract Regular Model Checking loop.  The input
   * map [reg_loc_map] (from locations to regions) is the current region map
   * (M_i in [BHV04]), either the initial one (M_0) or the one obtained with the
   * last (reflexive) post (M_i = post(M^a_{i-1})). The input [path] stores the
   * previously computed region maps as a list of pairs (rlm, abs_rlm) where
   * [abs_rlm] (M^a_i) is the approximation [rlm] (M_i). The input integer [k]
   * is the current extrapolation parameter.  Thus [k] implicitly defines the
   * so-called (see [BHV04]) the automata abstraction function [Closure.clo k].
   *)
  let rec armc_loop cpt_i cpt_r reg_loc_map path k =
    if (cpt_i > armc_limit) then
      (* XXX Avoid (repeated) computation of the path length. *)
      (cpt_i - 1, (List.length path) + 1, cpt_r,
       Abort ("ARMC: loop iteration " ^ (string_of_int cpt_i) ^
                " larger than limit " ^ (string_of_int armc_limit)))
    else
      begin
        Logger.slog Logger.Notice "@[%s%8d   ( %8d@]@?"
          (String.make !backspace '\b')
          cpt_i ((List.length path) + 1) ;
        backspace := 21 ;
        Logger.info "@.@[ARMC loop: %d@]@." cpt_i ;
        Logger.info "@[Abstract path:@ %d nodes@]@." ((List.length path) + 1) ;

        (*
          Logger.debug "@[<v>Current reg loc map:@ @[%a@]@]@."
          RegLocMap.print reg_loc_map ;
          Logger.debug "@[<v>Current Path (abs only):@ @[%a@]@]@."
          (Util.Print.list_printer_from_printer "@ @ ;@ @ "
          (fun fmt (a, b, c) ->
          RegLocMap.print fmt b)) path ;
        *)
        (* intersection of the current region map with the error region map *)
        let err_inter = RegLocMap.inter error_reg_loc_map reg_loc_map
        in
          (* RegLocMap.simplify err_inter ; *)
          (* let us check whether this intersection is empty or not *)
          if RegLocMap.is_empty err_inter
          then
	    (* abstraction of the current region map *)
	    let abs_reg_loc_map = RegLocMap.map (Closure.clo k) reg_loc_map in
            (* RegLocMap.simplify abs_reg_loc_map ; *)
	    (* reflexive post of the abstraction of the current region map *)
	    let post = reflexive_post abs_reg_loc_map
	    in
              (* RegLocMap.simplify post ; *)
	      assert (RegLocMap.leq reg_loc_map abs_reg_loc_map) ;
	      assert (RegLocMap.leq abs_reg_loc_map post) ;
	      (* let us check whether the abstraction is a post-fixpoint *)
	      if RegLocMap.leq post abs_reg_loc_map
	      then
                begin
	          (* safe post-fixpoint reached, return it *)
                  Logger.slog Logger.Notice "@[ )@]@?" ;
                  Logger.info "@[Safe abstract fix point.@]@." ;
                  (cpt_i, (List.length path) + 1, cpt_r,
	           Invariant
                     (Some
                        (lazy
                           (RegLocMap.fold
                              (fun l r sl -> (l, r) :: sl)
                              abs_reg_loc_map [])),
                      None))
                end
	      else
                begin
	          (* not there yet, continue loop *)
                  Logger.slog Logger.Notice "@[ )@]@?" ;
                  backspace := !backspace + 2 ;
	          armc_loop (cpt_i+1) cpt_r post ((reg_loc_map, abs_reg_loc_map, k) :: path) k
                end
          else
            refine cpt_i (cpt_r+1) err_inter path []
      end
  and refine cpt_i cpt_r err path acc =
    Logger.slog Logger.Notice "@[ / %8d@]@?" cpt_r ;
    backspace := !backspace + 11 ;
    Logger.info "@[Abstract error found.  Performing exact backward computation...@]@." ;
    backward cpt_i cpt_r err path []

  (*
   * Function to perform backward exact pre computation from a non-empty region
   * map [err] (X_i in [BHV04]) contained in the error region map.  The input
   * [path] is the same as the one given to armc_loop.  The list [acc] stores
   * the error region maps (X_j with j > i) computed previously.  This function
   * either returns a feasible counterexample (given by the list error region
   * maps) or calls back armc_loop from the failure point with an extrapolation
   * parameter larger than the one used previously at this failure point.
   *)
  and backward cpt_i cpt_r err path acc =
    match path with
	[] ->
          Logger.slog Logger.Notice "@[ )@]@?" ;
          Logger.info "@[Valid error.  Extracting counter-example...@]@." ;
          RegLocMap.simplify err ;
          List.iter RegLocMap.simplify acc ;
          (cpt_i, (List.length path) + 1, cpt_r,
           Counterexample (extract_counter_example (err :: acc)))
      | (rlm, abs_rlm, k) :: tail ->
	  let pre_inter = RegLocMap.inter (reflexive_pre err) abs_rlm
	  in
	    assert (not (RegLocMap.is_empty pre_inter)) ;
	    if (RegLocMap.are_disjoint rlm pre_inter)
	    then
              begin
                Logger.info "@[Spurious error.  Refining closure parameter...@]@." ;
	        let rec find_extrapol_param k =
		  let new_abs_rlm = RegLocMap.map (Closure.clo k) rlm
		  in
		    if k > k_max then (false, k)
		    else
		      if RegLocMap.are_disjoint new_abs_rlm pre_inter
		      then (true, k)
		      else find_extrapol_param (k + 1)
	        in
		  (* Possible optimization: save new_abs_rlm corresponding to the
		     new k instead of recomputing it in armc_loop *)
		  assert (not (RegLocMap.are_disjoint
				 (RegLocMap.map (Closure.clo k) rlm)
				 pre_inter)) ;
                  let (status, k) = find_extrapol_param (k+1)
                  in
                    if status then
                      begin
                        Logger.slog Logger.Notice "@[ / %8d )@]@?" k ;
                        Logger.info "@[Closure parameter: %d@]@." k ;
                        backspace := !backspace + 13 ;
		        armc_loop (cpt_i+1) cpt_r rlm tail k
                      end
                    else
                      (cpt_i, (List.length path) + 1, cpt_r,
                       Abort ("ARMC: closure parameter " ^ (string_of_int k) ^
                                " larger than limit " ^ (string_of_int k_max)))
              end
	    else
	      backward cpt_i cpt_r pre_inter tail (err :: acc)

  let model_check () =
    Logger.slog Logger.Notice "@[%8s %10s   %10s %10s %10s@]@."
      "" "Loops" "Nodes" "Refmnts" "k" ;
    Logger.slog Logger.Notice "@[ARMC loop: %8d   ( %8d / %8d / %8d )@]@?" 0 0 0 k_min ;
    backspace := 45 ;
    Logger.info "@[ARMC loop: %d@]@." 0 ;
    let (cpt_i, len, cpt_r, result) =
      armc_loop 1 0 init_reg_loc_map [] k_min
    in
      Logger.notice "@." ;
      Logger.report "@[ARMC:@ %d loops,@ %d nodes,@ %d refinements@]@."
        cpt_i len cpt_r ;
      result
end
