(* $Id: Cegar.ml 1245 2011-10-18 21:15:14Z gsutre $ *)

(*
 * Copyright (c) 2009, 2010, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util


(**
 * This module implements a CEGAR model-checker for models that conform to the
 * Model.S signature. The approach used in the original CEGAR approach by Clarke
 * et al. (CAV'00): the abstract graphs are partition-based and explicitly
 * computed.  Hence the set of abstract nodes forms a partition of the set of
 * concrete configurations of the model.  The initial abstract graph is just a
 * copy of the (control flow) automaton of the model.  In other words, the
 * initial partition distinguishes configurations that have different control
 * locations.  Refinement is based on the upper closure operator provided with
 * the model.
 *)

(*
 * Edges in the abstract graph are labeled with commands, just as in the
 * original (control flow) automaton of the model.  Nodes in the abstract graph
 * are labeled with:
 * - a symbolic state (loc, reg) giving the concretization of the node,
 * - two regions init and error that respectively give the subset of initial or
 *   error configurations contained in this node.
 *
 * Edges in the abstract graph follow the existential conservative semantics.
 *)

(*
 * In this first implementation, we keep all nodes in the abstract graph, even
 * the ones that are not reachable or co-reachable.  Therefore it is always true
 * that the set of abstract nodes forms a partition of the set of concrete
 * configurations.
 *
 * It is assumed that the computed abstractions will be small, and therefore it
 * shall not be costly to iterate over all nodes in the abstract graph.
 *
 * It might be more efficient to delete abstract nodes that are not reachable or
 * co-reachable.  This is something to try.
 *)



(* Exception for errors that are internal to this module, such as cases in
   in pattern matching that should not arise. *)
exception InternalError

module Make (E : TraceChecker.S) (A : Model.S) =
struct
  (* We open A and define shortcuts to simplify notations *)
  open A
  module Location = Automaton.Location
  module Transition = Automaton.Transition
  module Region = Semantics.Region

  (* Maps over locations *)
  module M = Map.Make(Location)

  (* The maps for initial and error symbolic states *)
  let (init_region_map, error_region_map) =
    let add_with_union (l, r) m =
      if (M.mem l m) then
        M.add l (Region.union (M.find l m) r) m
      else
        M.add l r m
    in
      (fold_init  add_with_union M.empty,
       fold_error add_with_union M.empty)

  (* Helper function to return the initial region of a given location *)
  let init_region_of_location l =
    try M.find l init_region_map with
        Not_found -> Region.bot

  (* Helper function to return the error region of a given location *)
  let error_region_of_location l =
    try M.find l error_region_map with
        Not_found -> Region.bot

  (* Helper function to check that the init symbolic states and error symbolic
     states of the model are disjoint.  This uses the above maps.  The returned
     value is either None or Some (l, r) where l is a location and r is the
     non-empty intersection of init and error regions for r. *)
  let check_init_error_disjointness () =
    M.fold
      (fun li ri res ->
         match res with
             Some _ -> res
           | None ->
               let r = Region.inter (error_region_of_location li) ri
               in
                 if (Region.eq Region.bot r)
                 then None
                 else Some (li, r))
      init_region_map None


  (* Labels of abstract nodes *)
  type vertex_attr = {
    loc : Location.t ;
    reg : Region.t ;
    ini : Region.t ;   (* this region shall be contained in reg *)
    err : Region.t ;   (* this region shall be contained in reg *)
  }

  (* Abstract edges are labeled with transitions.  We actually only care about
     the command labeling the transition but it is useful to have the transition
     in order to convert an abstract edge into a transition. *)
  type edge_attr = {
    transition : Transition.t ;
  }

  (* Abstract graphs use the imperative implementation of directed labeled multi-graphs *)
  module VL =
  struct
    type t = vertex_attr
    let print fmt attr =
      Format.fprintf fmt "{ @[<v>" ;
      Format.fprintf fmt "loc = @[%a@]@," Location.print attr.loc ;
      Format.fprintf fmt "reg = @[%a@]@," Region.print attr.reg ;
      Format.fprintf fmt "ini = @[%a@]@," Region.print attr.ini ;
      Format.fprintf fmt "err = @[%a@]" Region.print attr.err ;
      Format.fprintf fmt "@] }"
  end
  module EL =
  struct
    type t = edge_attr
    let print fmt attr =
      Command.print fmt (Transition.command attr.transition)
  end
  module G =
    (*
    MultiGraph.Imperative.MakeMutable (
      MultiGraph.Persistent.Make
    *)
    MultiGraph.Imperative.Make2
      (VL)
      (EL)

  (* Helper function to get the command labeling a given edge *)
  let command_of_edge e = (Transition.command (G.Edge.lbl e).transition)

  (* Helper function to get the region labeling a given node *)
  let reg_of_node v = (G.Vertex.lbl v).reg

  (* Helper function to get the location labeling a given node *)
  let loc_of_node v = (G.Vertex.lbl v).loc

  (* Include graph exploration algorithms *)
  include Util.Search.Make (G)


  (* Select graph exploration function according to user specified option *)
  let explore =
    match Option.Cegar.graph_exploration () with
        Option.Cegar.Fwd_dfs -> fwd_dfs
      | Option.Cegar.Fwd_bfs -> fwd_bfs
      | Option.Cegar.Bwd_dfs -> bwd_dfs
      | Option.Cegar.Bwd_bfs -> bwd_bfs
      | Option.Cegar.Mixed ->
          (* Use only breadth-first search and select direction so that we
             start with the smallest set of vertices.  Possible optimization:
             switch to depth-first search when the graph is very large. *)
          fun g i f ->
            if (List.length i) <= (List.length f)
            then
              fwd_bfs g i f
            else
              bwd_bfs g i f

  (* Sets of vertices *)
  module VSet =
    Collection.Set.Imperative.MakeMutable
      (Collection.Set.Persistent.MakeTreeSet (G.Vertex))

  (* Function to create an initial abstract graph *)
  let create_initial_graph () =
    (* The graph to be returned *)
    let graph = G.create Automaton.nb_locations
    (* Hash table to map locations to vertices *)
    and vertex_table = Hashtbl.create (Automaton.nb_locations) in
    (* Function to create vertices and fill vertex table from locations *)
    let create_vertices_and_fill_vertex_table () =
      Automaton.iter_locations
        (function l ->
           assert (not (Hashtbl.mem vertex_table l)) ;
           let new_vertex = G.Vertex.create { loc = l ;
                                              reg = Region.top ;
                                              ini = init_region_of_location  l ;
                                              err = error_region_of_location l }
           in
             Hashtbl.add vertex_table l new_vertex ;
             G.add_vertex graph new_vertex)
    (* Function to return the vertex corresponding to a given location *)
    and find_vertex l = Hashtbl.find vertex_table l
    in
    (* Function to add, as an edge, the given transition to the graph *)
    let add_transition_as_edge trans =
      let src = find_vertex (Transition.source trans)
      and tgt = find_vertex (Transition.target trans)
      in
        (* We only add the edge if the relation of the command is non-empty. *)
        if Semantics.rel (Transition.command trans) Region.top Region.top
        then
          let new_edge = G.Edge.create src { transition = trans } tgt
          in
            G.add_edge graph new_edge
        else
          ()
    in
      begin
        create_vertices_and_fill_vertex_table () ;
        Automaton.iter_transitions add_transition_as_edge ;
        graph
      end

  (* Function to extract the set of initial nodes from an abstract graph *)
  let init_set g =
    let s = VSet.create A.nb_init
    in
      G.iter_vertices
        (fun v ->
           if (not (Region.eq Region.bot (G.Vertex.lbl v).ini))
           then VSet.add s v)
        g ;
      s

  (* Function to extract the set of error nodes from an abstract graph *)
  let error_set g =
    let s = VSet.create A.nb_error
    in
      G.iter_vertices
        (fun v ->
           if (not (Region.eq Region.bot (G.Vertex.lbl v).err))
           then VSet.add s v)
        g ;
      s



(*****         A B S T R A C T   G R A P H    R E F I N E M E N T         *****)

  (* Function to add an edge between two abstract nodes as required by the
     existential conservative semantics.  This function uses the rel predicate
     of the model's semantics.  This function does not check whether the edge
     already exists. *)
  let add_existential_edge_with_rel graph v v' trans =
    assert (Location.equal (loc_of_node v)  (Transition.source trans)) ;
    assert (Location.equal (loc_of_node v') (Transition.target trans)) ;
    if Semantics.rel (Transition.command trans) (reg_of_node v) (reg_of_node v')
    then
      G.add_edge graph (G.Edge.create v { transition = trans } v')
    else
      ()

  (* Function to refine an edge with target v and with source distinct from v.
     Nodes v0 and v1 are the nodes that refine v: the regions labeling v0 and v1
     form a partition of the region labeling v.  This function does the same as:
     begin
       add_existential_edge_with_rel graph src v1 trans ;
       add_existential_edge_with_rel graph src v0 trans
     end
     but tries to be smart by using the fact that we previously had an edge from
     src to v. *)
  let refine_in_edge_with_rel graph src v0 v1 trans =
    assert (Location.equal (loc_of_node src) (Transition.source trans)) ;
    assert (Location.equal (loc_of_node v0)  (Transition.target trans)) ;
    assert (Location.equal (loc_of_node v1)  (Transition.target trans)) ;
    (* let us check whether there is an edge from src to v1 *)
    if Semantics.rel (Transition.command trans) (reg_of_node src) (reg_of_node v1)
    then
      begin
        (* there is an edge from src to v1 *)
        G.add_edge graph (G.Edge.create src { transition = trans } v1) ;
        (* there might still be an edge from src to v0 *)
        if Semantics.rel (Transition.command trans) (reg_of_node src) (reg_of_node v0)
        then
          G.add_edge graph (G.Edge.create src { transition = trans } v0)
        else
          ()
      end
    else
      begin
        (* there is no edge from src to v1 *)
        (* since there was an edge from src to v, there must be an edge to v0 *)
        assert (Semantics.rel (Transition.command trans) (reg_of_node src) (reg_of_node v0)) ;
        G.add_edge graph (G.Edge.create src { transition = trans } v0)
      end

  (* Same as refine_in_edge_with_rel but uses explicitly post instead of rel to
     avoid recomputation of post: if the implementation of rel uses post then
     refine_in_edge_with_rel performs twice the same post computation. *)
  let refine_in_edge_with_post graph src v0 v1 trans =
    assert (Location.equal (loc_of_node src) (Transition.source trans)) ;
    assert (Location.equal (loc_of_node v0)  (Transition.target trans)) ;
    assert (Location.equal (loc_of_node v1)  (Transition.target trans)) ;
    let post = Semantics.post (reg_of_node src) (Transition.command trans)
    in
      (* let us check whether there is an edge from src to v1 *)
      if Region.leq (Region.inter (reg_of_node v1) post) Region.bot
      then
        begin
          (* there is no edge from src to v1 *)
          (* since there was an edge from src to v, there must be an edge to v0 *)
          assert (not (Region.leq (Region.inter (reg_of_node v0) post) Region.bot)) ;
          G.add_edge graph (G.Edge.create src { transition = trans } v0)
        end
      else
        begin
          (* there is an edge from src to v1 *)
          G.add_edge graph (G.Edge.create src { transition = trans } v1) ;
          (* there might still be an edge from src to v0 *)
          if Region.leq (Region.inter (reg_of_node v0) post) Region.bot
          then
            ()
          else
            G.add_edge graph (G.Edge.create src { transition = trans } v0)
        end

  (* Function to refine an edge with source v and with target distinct from v.
     Nodes v0 and v1 are the nodes that refine v: the regions labeling v0 and v1
     form a partition of the region labeling v.  This function does the same as:
     begin
       add_existential_edge_with_rel graph v1 tgt trans ;
       add_existential_edge_with_rel graph v0 tgt trans
     end
     but tries to be smart by using the fact that we previously had an edge from
     v to tgt. *)
  let refine_out_edge_with_rel graph v0 v1 tgt trans =
    assert (Location.equal (loc_of_node tgt) (Transition.target trans)) ;
    assert (Location.equal (loc_of_node v0)  (Transition.source trans)) ;
    assert (Location.equal (loc_of_node v1)  (Transition.source trans)) ;
    (* let us check whether there is an edge from v1 to tgt *)
    if Semantics.rel (Transition.command trans) (reg_of_node v1) (reg_of_node tgt)
    then
      begin
        (* there is an edge from v1 to tgt *)
        G.add_edge graph (G.Edge.create v1 { transition = trans } tgt) ;
        (* there might still be an edge from v0 to tgt *)
        if Semantics.rel (Transition.command trans) (reg_of_node v0) (reg_of_node tgt)
        then
          G.add_edge graph (G.Edge.create v0 { transition = trans } tgt)
        else
          ()
      end
    else
      begin
        (* there is no edge from v1 to tgt *)
        (* since there was an edge from v to tgt, there must be an edge from v0 *)
        assert (Semantics.rel (Transition.command trans) (reg_of_node v0) (reg_of_node tgt)) ;
        G.add_edge graph (G.Edge.create v0 { transition = trans } tgt)
      end

  (* Same as refine_out_edge_with_rel but uses explicitly pre instead of rel to
     avoid recomputation of pre: if the implementation of rel uses pre then
     refine_in_edge_with_rel performs twice the same pre computation. *)
  let refine_out_edge_with_pre graph v0 v1 tgt trans =
    assert (Location.equal (loc_of_node tgt) (Transition.target trans)) ;
    assert (Location.equal (loc_of_node v0)  (Transition.source trans)) ;
    assert (Location.equal (loc_of_node v1)  (Transition.source trans)) ;
    let pre = Semantics.pre (reg_of_node tgt) (Transition.command trans)
    in
      (* let us check whether there is an edge from v1 to tgt *)
      if Region.leq (Region.inter (reg_of_node v1) pre) Region.bot
      then
        begin
          (* there is no edge from v1 to tgt *)
          (* since there was an edge from v to tgt, there must be an edge from v0 *)
          assert (not (Region.leq (Region.inter (reg_of_node v0) pre) Region.bot)) ;
          G.add_edge graph (G.Edge.create v0 { transition = trans } tgt)
        end
      else
        begin
          (* there is an edge from v1 to tgt *)
          G.add_edge graph (G.Edge.create v1 { transition = trans } tgt) ;
          (* there might still be an edge from v0 to tgt *)
          if Region.leq (Region.inter (reg_of_node v0) pre) Region.bot
          then
            ()
          else
            G.add_edge graph (G.Edge.create v0 { transition = trans } tgt)
        end

  (* Function to refine an edge with both source v and target v.  Nodes v0 and
     v1 are the nodes that refine v: the regions labeling v0 and v1 form a
     partition of the region labeling v. *)
  let refine_self_loop_with_rel graph v0 v1 trans =
    assert (Location.equal (loc_of_node v0) (loc_of_node v1)) ;
    assert (Location.equal (loc_of_node v0) (Transition.source trans)) ;
    assert (Location.equal (loc_of_node v0) (Transition.target trans)) ;
    add_existential_edge_with_rel graph v0 v0 trans ;
    add_existential_edge_with_rel graph v0 v1 trans ;
    add_existential_edge_with_rel graph v1 v0 trans ;
    add_existential_edge_with_rel graph v1 v1 trans

  (* Select edge refinement function according to user specified option *)
  let (refine_in_edge, refine_out_edge, refine_self_loop) =
    match Option.Cegar.graph_refinement () with
        Option.Cegar.Rel ->
          (refine_in_edge_with_rel,
           refine_out_edge_with_rel,
           refine_self_loop_with_rel)
      | Option.Cegar.PostPre ->
          (refine_in_edge_with_post,
           refine_out_edge_with_pre,
           refine_self_loop_with_rel)

  (*
   * Function to split a given node v w.r.t. a given pair of split regions.  The
   * two split regions are assumed to split the node's current region into two
   * non-empty regions.  The split regions are not required to be contained in
   * the node's current region.  However the node's current region is assumed to
   * be contained in the union of the two split regions.
   *
   * The first supplied region, sr1, shall contain v's init region, and the
   * second supplied region, sr0, shall contain v's error region.  Informally,
   * sr1 and sr0 respectively represent the good and the bad regions obtained
   * from the path invariant.
   *)
  let split_node graph init error v sr1 sr0 =
    assert (G.mem_vertex graph v) ;
    assert (Region.leq (G.Vertex.lbl v).ini sr1) ;
    assert (Region.leq (G.Vertex.lbl v).err sr0) ;
    let attr = G.Vertex.lbl v in
    let attr1 = { loc = attr.loc ;
                  reg = Region.inter attr.reg sr1 ;
                  ini = Region.inter attr.ini sr1 ;
                  err = Region.inter attr.err sr1 }
    and attr0 = { loc = attr.loc ;
                  reg = Region.inter attr.reg sr0 ;
                  ini = Region.inter attr.ini sr0 ;
                  err = Region.inter attr.err sr0 } in
    let v1 = G.Vertex.create attr1
    and v0 = G.Vertex.create attr0
    in
(*
      Logger.debug "@[<v 3>Splitting node:@," ;
      Logger.debug "@[vertex: %a@]@," G.Vertex.print v ;
      Logger.debug "@[new vertex 0: %a@]@," G.Vertex.print v0 ;
      Logger.debug "@[new vertex 1: %a@]" G.Vertex.print v1 ;
      Logger.debug "@]@." ;
*)

      (* The regions of the two new vertices v0 and v1 should form a partition
         of the region of the original vertex v *)
      assert (not (Region.eq Region.bot attr0.reg)) ;
      assert (not (Region.eq Region.bot attr1.reg)) ;
      assert (Region.eq attr.reg (Region.union attr0.reg attr1.reg)) ;
      assert (Region.eq Region.bot (Region.inter attr0.reg attr1.reg)) ;
      (* This partition should not split v's init region nor v's error region *)
      assert (Region.eq attr.ini attr1.ini) ;
      assert (Region.eq attr.err attr0.err) ;
      assert (Region.eq Region.bot attr1.err) ;
      assert (Region.eq Region.bot attr0.ini) ;

      (* We add both new vertices to the graph.  Otherwise, it could be the case
         that one of them is not added to the graph (if it is isolated), and if
         it contains initial or error configurations, these configurations would
         be missing in the abstract graph.  This would lead to a faulty invariant
         if the model is safe. *)
      (* These two lines also fix the order of vertices v0 and v1 in the graph
         independently from the order in which the refined edges are added to
         the graph. *)
      G.add_vertex graph v0 ;
      G.add_vertex graph v1 ;

      (* We update the sets of initial and error nodes. *)
      if not (Region.eq Region.bot attr.ini)
      then
        begin
          assert (VSet.mem init v) ;
          VSet.remove init v ;
          VSet.add init v1
        end ;
      if not (Region.eq Region.bot attr.err)
      then
        begin
          assert (VSet.mem error v) ;
          VSet.remove error v ;
          VSet.add error v0
        end ;

      (* We first process strict successors, then strict predecessors, and then
         self-loops, however this order is somewhat arbitrary. *)
      G.iter_succ
        (fun e ->
           let src = G.Edge.src e
           and tgt = G.Edge.tgt e
           and trans = (G.Edge.lbl e).transition
           in
             assert (G.Vertex.equal src v) ;
             if not (G.Vertex.equal tgt v)
             then refine_out_edge graph v0 v1 tgt trans)
        graph v ;
      G.iter_pred
        (fun e ->
           let src = G.Edge.src e
           and tgt = G.Edge.tgt e
           and trans = (G.Edge.lbl e).transition
           in
             assert (G.Vertex.equal tgt v) ;
             if not (G.Vertex.equal src v)
             then refine_in_edge graph src v0 v1 trans)
        graph v ;
      G.iter_succ
        (fun e ->
           let src = G.Edge.src e
           and tgt = G.Edge.tgt e
           and trans = (G.Edge.lbl e).transition
           in
             assert (G.Vertex.equal src v) ;
             if (G.Vertex.equal tgt v)
             then refine_self_loop graph v0 v1 trans)
        graph v ;

      G.del_vertex graph v



(*****         C O U N T E R E X A M P L E    R E F I N E M E N T         *****)

  (*
   * Function to check whether a list of edges is an abstract counterexample,
   * i.e. the list must be non-empty, source and target nodes must match for
   * consecutive edges, the init region must be non-empty for the first node and
   * empty for all other nodes, and the error region must be non-empty for the
   * last node and empty for all other nodes.
   *
   * XXX We should also check that nodes are pairwise distinct.
   *)
  let check_abstract_counterexample path =
    let rec do_it src path =
       match path with
           [] ->
             (* This should not happen *)
             raise InternalError
         | e::tail ->
             if G.Vertex.equal (G.Edge.src e) src
             then
               let tgt = (G.Edge.tgt e)
               in
                 if tail = []
                 then
                   ((not (Region.eq Region.bot (G.Vertex.lbl tgt).err)) &&
                      (Region.eq Region.bot (G.Vertex.lbl tgt).ini))
                 else
                   ((Region.eq Region.bot (G.Vertex.lbl tgt).ini) &&
                    (Region.eq Region.bot (G.Vertex.lbl tgt).err)) &&
                   (do_it tgt tail)
             else
               false
    in
      match path with
          [] ->
            false
        | e::tail ->
            let src = (G.Edge.src e)
            in
              if ((not (Region.eq Region.bot (G.Vertex.lbl src).ini)) &&
                    (Region.eq Region.bot (G.Vertex.lbl src).err))
              then
                do_it src path
              else
                false



  let rec trace_of_fwd_path_aux path acc n last =
    match path with
        [] ->
          (acc, n, last)
      | e :: tail ->
          let src = G.Edge.src e
          and cmd = command_of_edge e
          in
            trace_of_fwd_path_aux
              tail
              (`Command cmd :: `State (src, (G.Vertex.lbl src).reg) :: acc)
              (n+2) e

  let trace_of_fwd_path path =
    match path with
        [] ->
          ([], Sequence.Backward, 0)
      | first :: _ ->
          let src = G.Edge.src first in
          let acc = [`Guard (G.Vertex.lbl src).ini] in
          let (acc, n, last) = trace_of_fwd_path_aux path acc 1 first in
          let tgt = G.Edge.tgt last in
          let acc =
            `Guard (G.Vertex.lbl tgt).err ::
              `State (tgt, (G.Vertex.lbl tgt).reg) :: acc
          and n = n+2
          in
            assert (n = List.length acc) ;
            (acc, Sequence.Backward, n)


  module SemanticsClosure =
  struct
    include A.Semantics
    let clo = A.Closure.clo
  end

  (* Create the trace-checker module. *)
  module TC = E (SemanticsClosure)


  let do_refinement_with_trace_invariant graph init error trinv =
    Logger.debug "@[%a@]" (TC.Result.print G.Vertex.print) trinv ;
    let (path_invariant, do_node_split) =
      match trinv with
          `Feasible _ ->
            (* This should not happen (this case is already dealt with). *)
            raise InternalError
        | `Invariant_forward (_, (pinv, _, _), _) ->
            (pinv,
             fun ((v, _), r) -> split_node graph init error v r (Region.compl r))
        | `Invariant_backward (_, (pinv, _, _), _) ->
            (pinv,
             fun ((v, _), r) -> split_node graph init error v (Region.compl r) r)
    in
      assert (path_invariant != []) ;
      List.iter do_node_split path_invariant

  (*
   * In the Cegar implementation of the SPIN'09 paper, abstract nodes were split
   * in reversed order with respect to the abstract counterexample.  The compat
   * functions below are used to preserve this behavior.  This amounts to a
   * simple list reversal since, in the current implementation:
   *
   * - APInvBwd and UPInvFwd analyze, in fine, a prefix of the counterexample,
   *   and they return a trace invariant where the sequence is forward.
   *
   * - APInvFwd and UPInvBwd analyze, in fine, a suffix of the counterexample,
   *   and they return a trace invariant where the sequence is backward.
   *
   * XXX Experiment to see whether this reversed order yields better results.
   *)
  let compat_force_backward_sequence = function
      (l, Sequence.Forward, n)  -> (List.rev l, Sequence.Backward, n)
    | (l, Sequence.Backward, n) -> (List.rev l, Sequence.Forward, n)

  let compat_trace_invariant = function
      `Feasible _ as result -> result
    | `Invariant_forward (m, s, n) -> `Invariant_forward (m, compat_force_backward_sequence s, n)
    | `Invariant_backward (m, s, n) -> `Invariant_backward (m, compat_force_backward_sequence s, n)



  (*
   * The cegar_loop function implements the CEGAR loop to model-check the model
   * given as input to this functor.  This function returns None if the model is
   * safe or a pair composed of a path in the automaton and a list of regions
   * that are bot reachable and coreachable along the path.  It is assumed that
   * all nodes in the given abstract graph have either an empty init region or
   * an empty error region.
   *)



  type result =
      Invariant of ((A.State.t list) Lazy.t) option * ((A.State.t list) Lazy.t) option
    | Counterexample of Transition.t list
    | Abort of string

  let cegar_limit = (Option.Main.limit ())

  let state_of_node v = ((G.Vertex.lbl v).loc, (G.Vertex.lbl v).reg)

  let fwd_reach graph init =
    match fwd_dfs graph (VSet.elements init) [] with
        Forward_reach fwd_reachset ->
          fwd_reachset
      | _ ->
          raise InternalError

  let bwd_reach graph error =
    match bwd_dfs graph [] (VSet.elements error) with
        Backward_reach bwd_reachset ->
          bwd_reachset
      | _ ->
          raise InternalError

  let rec cegar_loop cpt graph init error =
    (* Assertions to check that init and error are consistent w.r.t. graph *)
    assert (VSet.equal init (init_set graph)) ;
    assert (VSet.equal error (error_set graph)) ;
    assert (VSet.fold
              (fun v b -> b && (not (VSet.mem error v)))
              init true) ;

    if (cpt > cegar_limit) then
      (cpt-1,
       Abort ("CEGAR: loop iteration " ^ (string_of_int cpt) ^
                " larger than limit " ^ (string_of_int cegar_limit)))
    else
      begin
        Logger.slog Logger.Notice "@[%s%8d   ( %8d / %8d@]@?" (String.make 56 '\b') cpt (G.nb_vertices graph) (G.nb_edges graph) ;
        Logger.info "@.@[CEGAR loop: %d@]@.@[Abstract graph:@ %d nodes,@ %d edges@]@." cpt (G.nb_vertices graph) (G.nb_edges graph) ;
        Logger.debug "@[<v 3>Abstract graph:@,@[%a@]@]@." G.print graph ;
        match
          explore
            graph
            (VSet.elements init)
            (VSet.elements error)
        with
            Forward_reach fwd_reachset ->
              Logger.slog Logger.Notice "@[ )@]@?" ;
              Logger.info
                "@[No counterexample, abstract forward reachability set: %d nodes@]@."
                (List.length (fwd_reachset)) ;
              (cpt,
               Invariant
                 (Some (lazy
                          (List.map
                             state_of_node
                             fwd_reachset)),
                  Some (lazy
                          (List.map
                             state_of_node
                             (bwd_reach graph error)))))
          | Backward_reach bwd_reachset ->
              Logger.slog Logger.Notice "@[ )@]@?" ;
              Logger.info
                "@[No counterexample, abstract backward reachability set: %d nodes@]@."
                (List.length (bwd_reachset)) ;
              (cpt,
               Invariant
                 (Some (lazy
                          (List.map
                             state_of_node
                             (fwd_reach graph init))),
                  Some (lazy
                          (List.map
                             state_of_node
                             bwd_reachset))))
          | Forward_path [] | Backward_path [] ->
              (* This should not happen since the sets init_nodes and error_nodes
                 are supposed to be disjoint *)
              raise InternalError
          | Forward_path p ->
              counterexample_refinement cpt graph init error p
          | Backward_path p ->
              counterexample_refinement cpt graph init error (List.rev p)
      end
  and counterexample_refinement cpt graph init error path =
    Logger.slog Logger.Notice "@[ / %8d@]@?" (List.length path) ;
    Logger.info "@[<v 3>Counterexample trace (%d):@,@[%a@]@]@."
      (List.length path)
      (Util.Print.list_printer_from_printer ",@ "
         (fun fmt e -> Command.print fmt (command_of_edge e)))
      path ;
    Logger.debug "@[<v 3>Counterexample path:@,@[%a@]@]@."
      (Util.Print.list_printer_from_printer ",@ " G.Edge.print)
      path ;
    assert (check_abstract_counterexample path) ;
    let trinv = compat_trace_invariant (TC.trace_check (trace_of_fwd_path path))
    in
      match trinv with
          `Feasible _ ->
            (* The abstract path is a valid counterexample *)
            Logger.slog Logger.Notice "@[ )@]@?" ;
            Logger.info "@[Counterexample valid@]@." ;
            (cpt,
             Counterexample (List.map (fun e -> (G.Edge.lbl e).transition) path))
        | `Invariant_forward _ | `Invariant_backward _ ->
            (* The abstract path is a spurious counterexample *)
            do_refinement_with_trace_invariant graph init error trinv ;
            Logger.slog Logger.Notice "@[ )@]@?" ;
            cegar_loop (cpt+1) graph init error

  (* Refinement function for empty paths, i.e. abstract nodes where the init
     and error regions are both non-empty. *)
  let do_refinement_empty_path graph init error v =
    let attr = G.Vertex.lbl v in
    let trace =
      ([ `Guard attr.ini ; `State (v, attr.reg) ; `Guard attr.err ],
       Sequence.Forward, 3)
    in
      do_refinement_with_trace_invariant graph init error (TC.trace_check trace)

  let model_check () =
    Logger.slog Logger.Notice "@[%9s %10s   %10s %10s %10s %10s@]@."
      "" "Loops" "Nodes" "Edges" "Cex len" "k" ;
    Logger.slog Logger.Notice "@[CEGAR loop: %8d@]@?" 0 ;
    Logger.info "@[CEGAR loop: %d@]@." 0 ;
    match check_init_error_disjointness () with
        Some (l, r) -> Counterexample []
      | None ->
          let graph = create_initial_graph () in
          let init = init_set graph
          and error = error_set graph
          in
            Logger.slog Logger.Notice "@[   ( %8d / %8d / %8s / %8s )@]@?" (G.nb_vertices graph) (G.nb_edges graph) "" "" ;
            Logger.info "@[Abstract graph:@ %d nodes,@ %d edges@]@." (G.nb_vertices graph) (G.nb_edges graph) ;
            Logger.debug "@[<v 3>Abstract graph (initial):@,@[%a@]@]@." G.print graph ;
            (* We split nodes for which both the init region and the error region
               are non empty *)
            VSet.iter
              (fun v ->
                 let attr = (G.Vertex.lbl v)
                 in
                   assert (not (Region.eq Region.bot attr.ini)) ;
                   if (not (Region.eq Region.bot attr.err))
                   then
                     begin
                       Logger.slog Logger.Notice "@[%s%8d@]@?" (String.make 21 '\b') 0 ;
                       do_refinement_empty_path graph init error v ;
                       Logger.slog Logger.Notice "@[ )@]@?"
                     end)
              init ;
            Logger.debug "@[<v 3>Abstract graph (after init/error split):@,@[%a@]@]@." G.print graph ;
            let (cpt, result) = cegar_loop 1 graph init error in
              Logger.notice "@." ;
              Logger.report "@[CEGAR:@ %d loops,@ %d nodes,@ %d edges@]@."
                cpt (G.nb_vertices graph) (G.nb_edges graph) ;
              result
end
