(* $Id: Control.ml 1145 2011-09-13 10:55:16Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Tristan Le Gall <tristan.le-gall@cea.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util

(* Exception for errors that are internal to this module, such as cases in
   in pattern matching that should not arise. *)
exception InternalError


module Make (A : Model.S) =
struct
  (* We open A and define shortcuts to simplify notations *)
  open A
  module Owner = Automaton.Owner
  module Location = Automaton.Location
  module Transition = Automaton.Transition
  module Region = Semantics.Region

  (* reachability functions *)
  module RE = ReachabilityEngine.Make(A)

  (* Maps over locations *)
  module M = Map.Make(Location)

  (* a state estimate is given by a map over locations *)
  type estimate = Region.t M.t

  let k = Option.Main.k ()
  let display_estimate_option = Option.Control.show_estimate ()
  let simulation_only_option = Option.Control.simulation_only ()


  (** [list_other_trans o] gives the list of all the transition of the
      automaton that does not belong to [o] *)
  let list_other_trans (owner:Owner.t) : (Transition.t list)=
    Automaton.fold_transitions 
      (fun t l -> if Transition.owner t = owner then l else t::l)
      []

  (* initial states *)
  let initial_state_list : State.t list =
    A.fold_init (fun s l -> s::l) []


  let initial_estimate () : (Owner.t*estimate) list =
    Automaton.fold_owners
      (fun owner lres -> 
	(* computes the state estimate for process [owner] *)
	let reachable_states : State.t list = RE.reach ~k:k initial_state_list (list_other_trans owner)
	(* add this estimate to the map *)
	in 
	let estimate : estimate=
	  List.fold_left 
	    (fun m (loc,reg) -> M.add loc reg m)
	    M.empty
	    reachable_states
	in
	(owner,estimate)::lres
      )
      []


  let rec find_estimate owner li =
    match li with
	[] -> raise Not_found
      | (o,e)::_ when Owner.equal o owner -> e
      |_::qq ->find_estimate owner qq
	

(* assign number to owners for the display of vector clocks *)

  let owners_number_map = 
    let cmpt = ref 0 in
    let mappe_res = ref Mappe.empty in
    Automaton.iter_owners 
      (fun o -> mappe_res := Mappe.add o !cmpt !mappe_res ; incr cmpt)
      ;
    !mappe_res

  let number (owner: Owner.t) : int =
    try Mappe.find owner owners_number_map with
	Not_found -> failwith "Bug: unknown owner in functon number"

  let owners_array = 
    let res =
      Array.make Automaton.nb_owners (fst (Mappe.choose owners_number_map))
    in
    Mappe.iter (fun owner i -> res.(i) <- owner) owners_number_map;
    res


(* Vector clock operations *)

  (* instead of a matrix, we represent set of vector clocks by maps *)
  type vector_clock = (Owner.t,int) Mappe.t
  type global_vector_clock = (Owner.t,vector_clock) Mappe.t

  (*  at the begining all clocks are set to 0 *) 
  let initialize_global_vector_clock () : global_vector_clock =
    let vector_init =
      Automaton.fold_owners (fun o m -> Mappe.add o 0 m) Mappe.empty
    in 
    Automaton.fold_owners (fun o m -> Mappe.add o vector_init m) Mappe.empty
      
  (* update after an output (or internal) transition *)
  let update_vector_clock_output
      (global_vc : global_vector_clock)
      (owner:Owner.t) 
      : global_vector_clock 
      = 
    let local_vector = 
      try Mappe.find owner global_vc 
      with Not_found -> failwith ("bug in update_vector_clock_output: unknown owner in global vector clock")
    in
    let local_time =
      try Mappe.find owner local_vector
      with Not_found -> failwith ("bug in update_vector_clock_output: unknown owner in local vector clock")
    in
    Mappe.add owner (Mappe.add owner (local_time +1) local_vector) global_vc

  (* update after an input transition *)
  let update_vector_clock_input
      (global_vc : global_vector_clock)
      (owner:Owner.t) 
      (external_vc : vector_clock)
      : global_vector_clock 
      = 
    let local_vector = 
      try Mappe.find owner global_vc 
      with Not_found -> failwith ("bug in update_vector_clock_input: unknown owner in global vector clock")
    in
    (* we take the maximal value *)
    Mappe.add owner (Mappe.merge max local_vector external_vc) global_vc
    
(* display with explicit name *)
  let print_vector_clock_owners fmt (vc:vector_clock) =
    Format.fprintf fmt "@[[";
    Mappe.iter (fun owner tps -> Format.fprintf fmt "@,(%a,%d)@," Owner.print owner tps) vc;
   Format.fprintf fmt "]@]"

(* display with only numbers *)
 let print_vector_clock fmt (vc:vector_clock) =
    Format.fprintf fmt "@[[";
   Array.iteri
     (fun i owner ->
       let tps = Mappe.find owner vc in
       Format.fprintf fmt " %d " tps
     )
     owners_array;
   Format.fprintf fmt "]@]"



  (* queue operations *)

  type queue  = (string*vector_clock*estimate) list

  let queues_array : queue array =
    let nb_queues =
      Automaton.fold_transitions
	(fun trans x ->
	  max x (Command.give_queue (Transition.command trans))
	)
	0
    in
    Array.make (nb_queues+1) []

  let push_queues (cmd:Command.t) (vc:vector_clock) (est:estimate) : unit =
    let i = Command.give_queue cmd in
    let m = Command.give_message cmd in
    assert (Command.give_type cmd = Command.Output);
    queues_array.(i) <- queues_array.(i)@[(m,vc,est)]

  let look_queues  (cmd:Command.t) : (string*vector_clock*estimate) =
    let i = Command.give_queue cmd in
    assert (Command.give_type cmd = Command.Input);
    List.hd queues_array.(i)

  let pop_queues  (cmd:Command.t): unit =
    let i = Command.give_queue cmd in
    assert (Command.give_type cmd = Command.Input);
    queues_array.(i) <- List.tl queues_array.(i)


  let rec print_list print_elt fmt li =
    match li with
	[] -> ()
      | el::[] -> print_elt fmt el
      | el::qq -> (print_elt fmt el; Format.fprintf fmt "@,"; print_list print_elt fmt qq)
	

  let print_queues fmt (cmd: Command.t) : unit =
    let i = Command.give_queue cmd in
    let content = queues_array.(i) in
    print_list
      (fun fmt (msg,vc,est) ->
	Format.fprintf fmt "(%s,@[<hov>%a@],<est.>)" msg print_vector_clock vc
      )
      fmt
      content

  let print_queues_current_state fmt () : unit =
    Array.iteri
      (fun i li -> 
	Logger.notice "Channel %d: @[<hov>%a@]@."
	  i
	  (print_list 
	     (fun fmt (msg,vc,est) ->
	       Format.pp_print_string fmt msg
	     )
	  )
	  li
      )
      queues_array


  let sender_receiver_array =
    let array_aux : (Owner.t option * Owner.t option) array = 
      Array.make (Array.length queues_array) (None,None) 
    in
    Automaton.iter_transitions
      (fun trans ->
	let cmd = Transition.command trans
	and owner = Transition.owner trans
	in
	let i = Command.give_queue cmd in
	if Command.give_type cmd = Command.Input
	then
	  begin
	    match array_aux.(i) with
		(x,None) -> array_aux.(i) <- (x,Some owner)
	      | (x,Some o) -> assert (Owner.equal o owner)
	  end
	else
	  if Command.give_type cmd = Command.Output
	  then
	    begin
	      match array_aux.(i) with
		  (None,x) -> array_aux.(i) <- (Some owner,x)
		| (Some o,x) -> assert (Owner.equal o owner)
	    end
      );
    array_aux


  (* print, for each queue, the sender and receiver process *)
  let display_queues_initial () =
   
    Logger.notice "List of active channels:@." ;

    let print_owner_option fmt (x: Owner.t option) :unit =
      match x with
	  None -> Format.pp_print_string fmt "None"
	| Some x -> Owner.print fmt x
    in
    Array.iteri
      (fun i (sender,receiver) ->
	Logger.notice "Channel %d : %a to %a @." i print_owner_option sender print_owner_option receiver
      )
      sender_receiver_array
    ;
    Logger.notice "@."




  (* estimate update functions *)


  (* give the size of an estimate*)
  let size (m:estimate) : int = 
    M.fold
      (fun loc reg res -> res + Region.size reg)
      m
      0
      



  let estimate_to_state_list (est: estimate) : State.t list =
    M.fold (fun loc reg l -> (loc,reg)::l) est []

  let estimate_from_state_list (slist : State.t list) :estimate =
    let map_all_empty =
      Automaton.fold_locations
	(fun loc m -> M.add loc Region.bot m)
	M.empty
    in
    let f_fold m (loc,reg) =
      try 
	let old_reg = M.find loc m in 
	let new_reg = Region.union old_reg reg in
	M.add loc new_reg m
      with
	  Not_found ->
	    M.add loc reg m
    in
    List.fold_left f_fold map_all_empty slist


  let transitions_of_command (cmd: Command.t) : Transition.t list =
    Automaton.fold_transitions
      (fun trans l -> if Transition.command trans = cmd then trans::l else l)
      []

 (* result is a list of state, since we then do a reachability operation *)
 let post_estimate (cmd: Command.t) (est:estimate) : State.t list =
   let lest = estimate_to_state_list est in
   let ltrans = transitions_of_command cmd in
   let rec find_reg (loc: Location.t) (lstate: State.t list): Region.t =
     match lstate with
	 [] -> raise Not_found
       | (l,r)::_ when Location.equal l loc ->r
       | _::qq -> find_reg loc qq
   in 
   let list_new_states =
     List.map 
       (fun trans -> 
	 let old_reg = find_reg (Transition.source trans) lest in
	 let new_reg = Semantics.post old_reg cmd in
	 (Transition.target trans,new_reg)
       )
       ltrans
   in
   list_new_states


 let update_estimate_output 
     (current_estimate:estimate) 
     (trans: Transition.t) 
     :
     estimate
     =
   let states_after_transition : State.t list = post_estimate (Transition.command trans) current_estimate 
   in
   let reachable_states : State.t list = 
     RE.reach ~k:k states_after_transition (list_other_trans (Transition.owner trans))
   in
   estimate_from_state_list reachable_states


 (* 
    code for out_of_date variable - algorithm inputTransition of the paper
    0 : irrelevant
    1 : V_j[j] = V_i[j]
    2 : V_j[j] > V_i[j]
    3 : V_j[j] < V_i[j]
 *)

 let update_estimate_input 
     (est:estimate) 
     (trans: Transition.t)
     (external_est: estimate) 
     (date:int) :estimate
     =
   let cmd = Transition.command trans in
   let queue_number = Command.give_queue cmd in
   let (sender,receiver) = sender_receiver_array.(queue_number) in
   let (sender,receiver) =
     match (sender,receiver)
     with
	 (Some s, Some r) -> (s,r)
       | _ -> failwith ""
   in
   assert (Owner.equal receiver (Transition.owner trans));
   (* find the output corresponding to the command input *)
   let cmd_prime =
     let res = ref cmd in
     Automaton.iter_transitions
       (fun tr -> 
	 let cmd_prime = Transition.command tr in
	 if
	   (Command.give_type cmd_prime = Command.Output)
	   &&
	     (Command.give_queue cmd_prime = Command.give_queue cmd)
	   &&
	     (Command.give_message cmd_prime = Command.give_message cmd)
	 then
	   res := cmd_prime
       );
     assert (!res <> cmd);
     !res
   in
   (* CS_j <- Post_delta' (CS_j) *)
   let cs_j = post_estimate cmd_prime external_est in


   let tmp : State.t list = 
     match date with
	 1 ->
	   (* tmp <- Post_delta (Reach_Delta\Delta_i (CS_j)) *)
	   let reach = RE.reach ~k:k cs_j (list_other_trans receiver)
	   in post_estimate cmd (estimate_from_state_list reach)
       | 2 ->
	 (* tmp <- Post_delta (Reach_Delta\Delta_i (Reach_Delta\Delta_j (CS_j))) *)
	 let external_est_list =  
	   RE.reach ~k:k cs_j (list_other_trans sender)
	 in
	 let reach = RE.reach ~k:k external_est_list (list_other_trans receiver)
	 in post_estimate cmd (estimate_from_state_list reach)
       | 3 -> 
	 (* tmp <- Post_delta (Reach_Delta (CS_j)) *)
	 let list_all_transitions =
	   Automaton.fold_transitions (fun trans l -> trans::l) []
	 in
	 let reach =RE.reach ~k:k cs_j list_all_transitions
	 in  post_estimate cmd (estimate_from_state_list reach)
       | _ -> failwith ""
   in
   (* CS_i <- Post_delta (CS_i) *)
   let local_est = post_estimate (Transition.command trans) est
   in
   let new_estimate : estimate =
     (* CS_i <- CS_i inter tmp *)
     let f_merge loc reg1 reg2 =
       match (reg1,reg2) with
	   (Some r1, Some r2) -> Some (Region.inter r1 r2)
	 | _ -> None
     in
     M.merge
       f_merge
       (estimate_from_state_list tmp)
       (estimate_from_state_list local_est)
   in
   new_estimate


 let update_estimate 
     (est:estimate) 
     (trans: Transition.t) 
     (external_est: estimate option) 
     (date:int) 
     :estimate =
   let cmd_type = Command.give_type (Transition.command trans) in
   match cmd_type with
       Command.Output ->
	 begin
	   assert (date = 0);
	   update_estimate_output est trans
	 end
     | Command.Input ->
       begin
	 assert (date >0) ;
	 assert (date <4) ;
	 match external_est with
	     None -> failwith ""
	   | Some ext_est -> update_estimate_input est trans ext_est date 
       end
     | Command.Internal -> failwith "NOT IMPLEMENTED"
  (* WARNING: not implemented *)  




(* check that the current state belongs to the estimate *)
 let check_estimate (state: State.t) (m:estimate) : bool =
   let loc = fst state
   and reg = snd state 
   in 
   let reg_est =
     try M.find loc m
     with Not_found -> Region.bot
   in
   Region.leq reg reg_est


(* display functions *)

  (* [print_and_assign_number ltrans] displays the list of allowed
     actions, i.e; actions of a transition from [ltrans] and assign to
     each transition a number *)
  let display_list_and_assign_number ?(manual=true) (transition_list:Transition.t list) : (int *Transition.t) list =
    let cmpt = ref 0 in
    let l_res = ref [] in
    let print_action fmt (trans:Transition.t) : unit =
      Format.fprintf 
	fmt 
	"[%d] in process %a: %a @." 
	!cmpt
	Owner.print (Transition.owner trans)
	Command.print (Transition.command trans)
      ;
      l_res := (!cmpt,trans)::!l_res;
      incr cmpt
    in
    if manual then 
      begin
    Logger.notice "List of allowed transitions:@.@[<hov>%a@]"
      (fun fmt _ -> List.iter (print_action fmt) transition_list)
      ()
      end
    else
      List.iter (fun t -> l_res := (!cmpt,t)::!l_res;incr cmpt) transition_list
    ;
    !l_res

  let display_current_state (loc : Location.t) : unit =
    Logger.notice 
      "Current location:%a @.Current content of the queues:@.@[<hov>%a@]@."
      Location.print loc
      print_queues_current_state ()

  let display_estimate (owner:Owner.t) (m : estimate) : unit=
    Logger.notice 
      "State estimate of the controller of process %a:@.@[<hov>%a@]@." 
      Owner.print owner
      (fun fmt m ->
	M.iter 
	  (fun loc reg -> Format.fprintf fmt "{Location@[%a@]: @[<hov>%a@]}@," Location.print loc Region.dump reg)
	  m
      )
      m

let display_estimate_size (owner:Owner.t) (m : estimate) : unit=
    Logger.notice 
      "SIZE of state estimate - process %a:@.@[<hov>%d@]@." 
      Owner.print owner
      (
	M.fold
	  (fun loc reg res -> res + Region.size reg)
	  m
	  0
      )
     



		   
  (* Computation of the set BAD = coreach_uc (B) *)

  let bad_states : Region.t M.t = 
    if simulation_only_option
    then
      estimate_from_state_list []
    else
      begin
    let list_initial_bad_states : State.t list =
     fold_error (fun s l -> s::l) []
    in
    let list_uc_transitions : Transition.t list=
      Automaton.fold_transitions (fun trans l -> if Transition.is_controllable trans then l else trans::l) []
    in
    let coreach : State.t list= RE.reach ~forward:false ~k:k list_initial_bad_states list_uc_transitions
    in
    estimate_from_state_list coreach
      end

  let give_bad_state (loc: Location.t) : Region.t =
    try M.find loc bad_states
    with Not_found -> Region.bot












    (* selection functions *)

  (* pick up a random initial state *)
  let choose_initial () : State.t =
    let random_nb = Random.int (List.length initial_state_list) in
    List.nth initial_state_list random_nb


    (* select the possible transitions, ie transition firable in the current state *)

  let select_possible_transitions (s:State.t) : Transition.t list =
    let current_reg = snd s in
    Automaton.fold_transitions
      (fun trans l_res ->
	let is_possible : bool =
	  (Transition.source trans = fst s)
	  (* checks that the source of the transition is the current location *)
	  &&
	    begin
	    (* checks that we can fire the transition *)
	      let new_reg = Semantics.post current_reg (Transition.command trans) in
	      not (Region.leq new_reg Region.bot)
	     end
	in
	Logger.debug "In location %a transition @[<hov> %a@] : %b@." Location.print (fst s) Transition.print trans is_possible;
	if is_possible
	then trans::l_res
	else l_res
      )
      []
      
  (* the result of the following function is a couple of lists. The
     first list is the allowed transitions, the second list is the
     forbiden transitions *)
  let select_allowed_transitions 
      (location:Location.t) 
      (current_l_estimate: (Owner.t * estimate) list)
      (possible_transitions: Transition.t list) 
      : (Transition.t list) * (Transition.t list) 
      =
    let transition_is_allowed (trans:Transition.t) : bool =
      let estimate = find_estimate (Transition.owner trans) current_l_estimate
      in
      (* checks that if we fire the transition, we will not intersect bad states *)
      not(Semantics.rel 
	    (Transition.command trans) 
	    (M.find location estimate) 
	    (give_bad_state (Transition.target trans))
      )	      
    in List.partition transition_is_allowed possible_transitions

 


(* **************************************************************************** *)
(* **************************************************************************** *)
(* **************************************************************************** *)
(* **************************************************************************** *)
(* **************************************************************************** *)

     
    (* simulation function *)

  let simulate (manual:bool) (max_step:int) =
    
    (* current state of the system/estimators *)
    let current_global_vector_clock = ref (initialize_global_vector_clock ()) in
    let current_state = ref (choose_initial ()) in
    let current_estimate = ref (initial_estimate ()) in

    (* record all estimates and their size *)
    let record_all_estimates = ref [] in
    let last_step_record = ref 0 in
    let size_all_estimates = ref 0 in

    let step_counter = ref 0 in

    (* boolean variables to control internal loops *)
    let continue = ref true in
    Logger.notice "*********** simulation is starting *****************@.";

    display_queues_initial ();
    Logger.notice "Order of processes in vector clock display: @[<hov>[";
    Array.iter
      (fun owner -> Logger.notice " %a " Owner.print owner)
      owners_array;
    Logger.notice "]@]@.@.";

    Logger.notice " initial state estimates:@.";
    if display_estimate_option 
    then
      begin
	List.iter (fun (o,e) -> display_estimate o e) !current_estimate
      end
    else
      begin
	List.iter (fun (o,e) -> display_estimate_size o e) !current_estimate
      end;

      (* record the size of the estimates *)
      size_all_estimates := List.fold_left (fun res (o,e) -> res + size e) 0 !current_estimate;


    (* record the initial estimates *)
    record_all_estimates := List.map (fun (o,e) -> (o,[e])) !current_estimate;
     

    Logger.notice "To END the simulation, just type any number which is not an allowed transition@.";

    if manual
    then     
      Logger.notice "Simulation mode: MANUAL @."
    else
      Logger.notice"Simulation mode: AUTOMATIC (at most %d steps) @." max_step;

    Format.printf "@. Press ENTER to continue.";
    Format.print_flush();
    ignore (read_line ());
    Logger.notice "@.@.@.";
    
   (*Statistics.register_start "simulation"; *)

    while !continue do
      incr step_counter;
      if !step_counter >= max_step then continue := false ;
      if manual then
	Logger.notice "*********** iteration %d is beginning *****************@." !step_counter;
      (* display current state, estimate and allowed transitions *)
      if manual then
	display_current_state (fst !current_state);

      let list_of_possible_transitions=  select_possible_transitions (!current_state)
      in
      if manual then
	Logger.debug "%d transitions out of %d are possible@." (List.length list_of_possible_transitions) Automaton.nb_transitions;
      let (allowed_transitions,forbidden_transitions) = 
	select_allowed_transitions 
	  (fst !current_state) (* current location *)
	  !current_estimate (* current state estimates*)
	  list_of_possible_transitions 
      in
      if manual && forbidden_transitions <> []
      then
	begin
	  Logger.notice "Transitions forbidden by the controller:@.";
	  List.iter (fun t -> Logger.notice "@[<hov>%a@]@." Transition.print t) forbidden_transitions
	end;
      let list_of_transitions_with_numbers = display_list_and_assign_number ~manual:manual allowed_transitions
      in

      let choice = 
	if manual then
	  begin      
	    Format.printf "Choice?(num+ENTER):";
	    Format.print_flush ();
	    (* ask for a transition to be taken *)
	    Scanf.scanf "%d\n" (fun x->x)
	  end
	else
	  let nb_choices = List.length list_of_transitions_with_numbers in
	  if nb_choices = 0 
	  then 
	    -1 (* stops the simulation, and avoids an exception raised by Random.int 0 *)
	  else
	    Random.int nb_choices
      in
      (* [find_transiton ltrans] looks into list [ltrans] for a pair (k,t) where k=[choice] and returns t *)
      let rec find_transition ltrans =
	match ltrans with
	    [] -> raise Not_found
	  | (k,t)::_ when k=choice -> t
	  | _::qq -> find_transition qq
      in
      try 
	begin
	  let chosen_transition = find_transition list_of_transitions_with_numbers
	  in
	  let new_location = Transition.target chosen_transition
	  and new_region = Semantics.post (snd !current_state) (Transition.command chosen_transition)
	  in
	  (* update current state *)
	  current_state := (new_location,new_region);

	  (* update queues array and vector clock *)
	  let cmd = Transition.command chosen_transition in
	  let owner = Transition.owner chosen_transition in
	  let local_vc = Mappe.find owner !current_global_vector_clock in
	  let is_out_of_date = ref 0 in
	  let external_estimate =
	    match Command.give_type cmd with
		Command.Internal -> 
		  begin
		    current_global_vector_clock := update_vector_clock_output!current_global_vector_clock  owner;
		    None
		  end
	      | Command.Output ->
		begin
		  current_global_vector_clock :=  update_vector_clock_output!current_global_vector_clock  owner;
		  
		  push_queues 
		    cmd 
		    (Mappe.find owner !current_global_vector_clock) 
		    (find_estimate owner !current_estimate);
		  None
		end
	      | Command.Input -> 
		begin
		  let (msg,external_vc,external_estimate) = look_queues cmd in
		  (* test whether the external estimate is out of date *)
		  let queue_number = Command.give_queue cmd in
		  let (sender,receiver) = sender_receiver_array.(queue_number) in
		  begin
		    match (sender, receiver) with
			(Some sender, Some receiver) ->
			  begin
			    assert (Owner.equal receiver owner);
			    let local_tps_sender = Mappe.find sender local_vc 
			    and external_tps_sender = Mappe.find sender external_vc
			    in 

			    (* 
			       code for out_of_date variable - algorithm inputTransition of the paper
			       0 : irrelevant
			       1 : V_j[j] = V_i[j]
			       2 : V_j[j] > V_i[j]
			       3 : V_j[j] < V_i[j]
			    *)
			    if local_tps_sender = external_tps_sender
			    then 
			      is_out_of_date := 1
			    else
			      if local_tps_sender < external_tps_sender
			      then
				is_out_of_date := 2
			      else
				is_out_of_date := 3  
			  end
		      | _ -> failwith ("Bug: queue "^(string_of_int queue_number)^" has no sender or no receiver")
		  end;
		  
		  
		  
		  (* adds 1 to current local time of the local vector clock*)
		  current_global_vector_clock :=  update_vector_clock_output !current_global_vector_clock owner;
		  (* take the maximum between the two vector clocks *)
		  current_global_vector_clock :=  update_vector_clock_input !current_global_vector_clock owner external_vc;

		  (* update queue *)
		  pop_queues cmd ;
		  (* return the result estimate *)
		  Some external_estimate
		    
		end
	  in

	  (* update and display state estimate *)
	  let new_estimate = update_estimate (find_estimate owner !current_estimate) chosen_transition external_estimate !is_out_of_date
	  in
	  size_all_estimates := !size_all_estimates + size new_estimate;
	  (* canonical form
	  let new_estimate = M.map (fun r -> Region.simplify r) new_estimate 

	  in *)

	  (* checks that the current state belongs to the new estimate *)
	  if not (check_estimate !current_state new_estimate) then failwith "CHECK ESTIMATE FAILED !!!!";



	  let rec insert_new_estimate lest =
	    match lest with
		[] -> raise Not_found
	      | (k,e)::qq when k= Transition.owner chosen_transition -> 
		(k, new_estimate)::qq
	      | tt::qq -> tt::(insert_new_estimate qq)
	  in
	  current_estimate := insert_new_estimate !current_estimate;


	  (* equality function on estimates *)
	  let eq_estimate (m1:estimate) (m2:estimate) :bool =
	    M.fold 
	      (fun loc reg1 res -> 
		res && 
		  begin
		    try let reg2 = M.find loc m2 in Region.eq reg1 reg2
		    with Not_found -> false
		  end
	      )
	      m1
	      true
	  in
	  (* list.mem with the previous function *)
	  let rec is_in_list_estimate e li =
	    match li with
		[] -> false
	      | tt::qq -> if eq_estimate e tt then true else is_in_list_estimate e qq
	  in
	  
	  let rec record_new_estimate lest =
	    match lest with
		[] -> raise Not_found
	      | (k,le)::qq when k= Transition.owner chosen_transition -> 
		if is_in_list_estimate new_estimate le (* checks that the new estimate was already computed *)
		then
		  lest
		else (* otherwise, record it *)
		  begin
		    last_step_record := !step_counter ;
		    (k, new_estimate::le)::qq
		  end
	      | tt::qq -> tt::(record_new_estimate qq)
	  in
	  record_all_estimates := record_new_estimate !record_all_estimates;



	  if manual then
	    begin
	      if display_estimate_option 
	      then
		display_estimate owner new_estimate
	      else
		display_estimate_size owner new_estimate  ;
	  (* display other information *)

	      Logger.notice 
		"new content of channel %d :@.@[%a@]@." 
		(Command.give_queue cmd)
		print_queues cmd;
	      Logger.notice
		"new vector clock of process %a : @[<hov>%a@]@." 
		Owner.print owner 
		print_vector_clock (Mappe.find owner !current_global_vector_clock);

	      Format.printf "Press ENTER to continue.";
	      Format.print_flush();
	      ignore (read_line ());
	      Logger.notice "@.@.@."
	    end
	end
      with Not_found -> continue := false 
    done;

    Logger.notice "END of SIMULATION @.@.";

    (* display the statistics *)
    let f_iter (o,le) =
 
      let f_fold_left (length_res,max_size_res,sum_size_res) m = (1+length_res, max max_size_res (size m),sum_size_res + size m)
      in 
      let (length,max_size,sum_size) = List.fold_left f_fold_left (0,0,0) le in
      Logger.notice "Process %a : number of different estimates %d, maximal size of estimates %d@." Owner.print o length max_size
    (*
      Logger.notice "Process %a : number of different  estimates %d, maximal size of estimates %d  sum of all estimate sizes %d @." Owner.print o length max_size sum_size
    *)
    in
    List.iter f_iter !record_all_estimates;
    (*Logger.notice "Last estimate was recorded at step %d out of %d steps @.@." !last_step_record !step_counter*)

  (* display the everage size of the estimate computed *)
    let nb_est = !step_counter-1 + A.Automaton.nb_owners in
    Logger.notice "Average size of the estimates %.1f@." ((float_of_int !size_all_estimates)/. (float_of_int nb_est))



    (*Statistics.register_stop "simulation" *)


(* main function *)
  let control () =

    (* debug: show the bad states *)
    Logger.info "Coreach of the bad states:@.";
    begin
      if not simulation_only_option
      then
	M.iter
	  (fun loc reg -> Logger.debug "In Location %a: @[<hov>%a@] @." Location.print loc Region.dump reg)
	  bad_states
    end;

    Random.init (int_of_float (Sys.time ()));

    let manual = 
      match Option.Control.simulation_mode () with
	  Option.Control.Manual -> true
	| _ -> false

    in
    simulate manual (Option.Main.limit ())
      
end
