(* $Id: LazyAbstraction.ml 1245 2011-10-18 21:15:14Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util
open Collection


(* Exception for errors that are internal to this module, such as cases in
   in pattern matching that should not arise. *)
exception InternalError

module Make (E : TraceChecker.S) (A : Model.S) =
struct
  (* We open A and define shortcuts to simplify notations. *)
  open A
  module Location = Automaton.Location
  module Transition = Automaton.Transition
  module Region = Semantics.Region

  (* Maps over locations. *)
  module LocMap = Map.Imperative.MakeHashMap (Location)

  (* The map for error symbolic states. *)
  let error_region_map =
    let em = LocMap.create (min Automaton.nb_locations nb_error)
    and add_with_union m (l, r) =
      if (LocMap.mem m l) then
        LocMap.add m l (Region.union (LocMap.find m l) r)
      else
        LocMap.add m l r
    in
      iter_error (add_with_union em) ;
      em

  (* Helper function to return the error region of a given location. *)
  let error_region_of_location l =
    try LocMap.find error_region_map l with
        Not_found -> Region.bot


  (*****                   V E R T E X    C L O S U R E                   *****)

  (* The mark is [true] iff the vertex has been expanded. *)
  module Mark =
  struct
    type t = bool ref
    let print fmt m =
      Format.pp_print_string fmt (if !m then "expanded" else "not expanded")
  end

  include ReachabilityTree.Make (A) (Mark)

  (* Additional shortcuts. *)
  let location = Vertex.location
  let region = Vertex.region
  let is_expanded u = !(Vertex.mark u)
  let set_expanded u = (Vertex.mark u) := true

  (*
   * [cover t v w] implements the procedure Cover of McMillan's CAV'06 paper,
   * except that we test covering even if [v] is already active (see below).
   *
   * Observe that we only consider vertices [w] that are active, to preserve
   * well-labeledness.  This condition was not present in McMillan's paper, but
   * it seems pretty obvious that we need it.
   *)
  let cover t v w =
    Logger.debug "@[<v 3>cover@,@[<h>%a@]@,@[<h>%a@]@]@." Vertex.print v Vertex.print w ;
    assert (Location.equal (location v) (location w)) ;
    (* Since [w] was created before [v], [w] cannot be a descendant of [v]. *)
    assert (not (ancestor t v w)) ;
    if (active t w) && (Region.leq (region v) (region w)) then
      begin
        add_covering t v w ;
        iter_vertices (clr_coverees t) (fun _ -> ()) t (Some v)
      end

  (*
   * [close t v] implements the procedure Close of McMillan's CAV'06 paper,
   * except that we test whether [v] is active here instead of [cover].  This
   * way, we only try to close a vertex if it is inactive.  But when we close
   * a vertex, we do it with respect to all vertices created before.  Note that
   * with this approach, a vertex may have several coverers.
   *
   * An alternative would be stop closure as soon as a coverer is found, as is
   * done in the paper.  There would be (at least) two possible choices here,
   * dependending on the order of the candidates (oldest or youngest first).
   *)
  let close t v =
    Logger.debug "@[<v 3>close@,@[<h>%a@]@]@." Vertex.print v ;
    assert (active t v) ;
    (* XXX We could stop the iteration as soon as [w] is equal to [v]. *)
    iter_location
      (fun w -> if (Vertex.compare w v < 0) then cover t v w)
      t (location v)

  (*
   * [really_active t v] closes all ancestors of [v] (including [v]), starting
   * from the root up to [v], and stopping as soon as an ancestor becomes
   * inactive (which means that this ancestor now has a coverer).  However, if
   * some ancestor already has a coverer (meaning that [v] is inactive), then no
   * ancestor is closed.  Thus, [really_active t v] behaves as [active t v] when
   * the node [v] is inactive.
   *
   * Returns [true] if and only if [v] is active (when the function returns).
   *)
  let rec really_active t v =
    Logger.debug "@[<v 3>really_active@,@[<h>%a@]@]@." Vertex.print v ;
    if has_coverers t v then
      false
    else
      let pact =
        match edge t v with
            Edge.Init _ -> true
          | Edge.Child (p, _) -> really_active t p
      in
        if pact then
          begin
            close t v ;
            not (has_coverers t v)
          end
        else
          false


  (*****                V E R T E X    R E F I N E M E N T                *****)

  (*
   * [refine_vertex t u r] implements a part (viz., only the for-loop body) of
   * the procedure Refine of McMillan's CAV'06 paper.  The non-containment
   * condition is transformed into an assertion since irreducibility of trace
   * invariants guarantees it.
   *)
  let refine_vertex t u r =
    Logger.debug "@[<v 3>refine_vertex@,@[<h>%a@]@ with@,@[<h>%a@]@]@." Vertex.print u Region.print r ;
    assert (not (Region.leq (region u) r)) ;
    clr_coverees t u ;
    Vertex.refine u (Region.inter (region u) r)

  (* Create the trace-checker module. *)
  module SemanticsClosure =
  struct
    include A.Semantics
    let clo = A.Closure.clo
  end
  module TC = E (SemanticsClosure)

  let rec trace_of_vertex_aux t acc n u =
    match edge t u with
        Edge.Init init ->
          ((`Guard init) :: (`State (u, region u)) :: acc, n+2)
      | Edge.Child (p, trans) ->
          let cmd = Transition.command trans
          in
            trace_of_vertex_aux t
              (`Command cmd :: `State (u, region u) :: acc)
              (n+2) p

  let trace_of_vertex t u error =
    let acc = [`Guard error] in
    let (acc, n) = trace_of_vertex_aux t acc 1 u
    in
      assert (n = List.length acc) ;
      (acc, Sequence.Forward, n)

  let rec refine_empty_suffix_prune t u n =
    assert (n > 0) ;
    if n = 1 then
      del_vertex t u
    else
      match edge t u with
          Edge.Init _ ->
            (* This should not happen (since n > 0). *)
            raise InternalError
        | Edge.Child (p, _) ->
            refine_empty_suffix_prune t p (n-1)

  let rec refine_empty_suffix_keep t u n =
    assert (n > 0) ;
    if not ((region u) == Region.bot) then
      refine_vertex t u Region.bot ;
    if not (n = 1) then
      match edge t u with
          Edge.Init _ ->
            (* This should not happen (since n > 0). *)
            raise InternalError
        | Edge.Child (p, _) ->
            refine_empty_suffix_keep t p (n-1)

  (* Choose how to handle empty suffixes according to user specified option. *)
  let refine_empty_suffix =
    if Option.LazyAbstraction.tree_pruning () then
      refine_empty_suffix_prune
    else
      refine_empty_suffix_keep

  let refine_branch t u trinv =
    Logger.debug "@[%a@]" (TC.Result.print Vertex.print) trinv ;
    let (path_invariant, do_vertex_refinement, n) =
      match trinv with
          `Feasible _ ->
            (* This should not happen (this case is already dealt with). *)
            raise InternalError
        | `Invariant_forward (_, (pinv, _, _), n) ->
            (pinv,
             (fun ((v, _), r) ->
                assert (not (Region.eq Region.bot (Region.inter (region v) r))) ;
                refine_vertex t v r),
             n)
        | `Invariant_backward (_, (pinv, _, _), n) ->
            (pinv,
             (fun ((v, _), r) ->
                let r = Region.compl r
                in
                  assert (not (Region.eq Region.bot (Region.inter (region v) r))) ;
                  refine_vertex t v r),
             n)
    in
      List.iter do_vertex_refinement path_invariant ;
      if n > 0 then
        refine_empty_suffix t u n


  (*****                U T I L I T Y    F U N C T I O N S                *****)

  (*
   * [path t u] return the branch from the root of [t] to the vertex [u], as a
   * list of transitions.
   *)
  let path t u =
    let rec aux u acc =
      match edge t u with
          Edge.Init _ -> acc
        | Edge.Child (p, trans) -> aux p (trans :: acc)
    in
      aux u []

  (*
   * [branch t u] return the branch from the root of [t] to the vertex [u], as a
   * list of vertices.
   *)
  let branch t u =
    let rec aux u acc =
      match parent t u with
          None -> u::acc
        | Some p -> aux p (u::acc)
    in
      aux u []

  let print_branch t fmt u =
    let print_vertex fmt u =
      match edge t u with
          Edge.Init init ->
            Format.fprintf fmt "@[<h>%a@]" Vertex.print u
        | Edge.Child (_, trans) ->
            let cmd = Transition.command trans
            in
              Format.fprintf fmt "@,@[»@;@[%a@]@;»@]@,@[<h>%a@]"
                Command.print cmd
                Vertex.print u
    in
      Format.fprintf fmt "@[<v>" ;
      List.iter (print_vertex fmt) (branch t u) ;
      Format.fprintf fmt "@]"

  let print_trace_atom fmt atom =
    match atom with
        `State (u, _) ->
          Format.fprintf fmt "@[%a@]" Location.print (location u)
      | `Guard _ -> ()
      | `Command cmd ->
          Format.fprintf fmt "@ »@;@[%a@]@;»@ " Command.print cmd

  let print_trace = Sequence.print "" print_trace_atom

  type result =
      Invariant of ((A.State.t list) Lazy.t) option * ((A.State.t list) Lazy.t) option
    | Counterexample of Transition.t list
    | Abort of string


  (*****                 T R E E    E X P L O R A T I O N                 *****)

  module type WORKLIST =
  sig
    type 'a t
    val create : unit -> 'a t
    val push : 'a -> 'a t -> unit
    val pop : 'a t -> 'a
    val is_empty : 'a t -> bool
    val iter : ('a -> unit) -> 'a t -> unit
    val length : 'a t -> int
    val push_collection : (('a -> unit) -> unit) -> 'a t -> unit
  end

  module Queue =
  struct
    include Queue
    let push_collection iter work =
      iter (fun x -> push x work)
  end

  module Stack =
  struct
    (* XXX The standard library's [Stack.length] operation runs in O(n). *)
    include Stack
    (*
     * The collection is pushed in reversed order, so that pop returns the
     * elements in the same order as the collection itself.
     *)
    let push_collection iter work =
      let revl = ref []
      in
        iter (fun x -> revl := x::!revl) ;
        List.iter (fun x -> push x work) !revl
  end

  (*
   * Generic exploration functor parametrized by a work-list implementation.
   *)
  module Explore (W : WORKLIST) =
  struct
    (* Pretty-printer for work-lists used in the main exploration function. *)
    let print_work_list fmt work =
      let print_work_element (l, e) =
        Format.fprintf fmt "@,@[<3>(%a,@ %a)@]"
          Location.print l
          Edge.print e
      in
        W.iter print_work_element work

    (*
     * [push_children t x work] adds the children of [x] in the tree [t] to the
     * work-list [work].
     *)
    let push_children t x work =
      W.push_collection (fun f -> iter_children f t x) work

    (*
     * [find_active_vertex p t] searches the tree [t] for an active vertex [u]
     * statisfying the predicate [p], and returns it (if any).
     *)
    let find_active_vertex p t =
      let rec aux work =
        if W.is_empty work then
          None
        else
          let u = W.pop work
          in
            (* The parent of [u], if any, should be active. *)
            assert (match parent t u with None -> true | Some p -> active t p) ;
            (* If [u] is covered, then its whole subtree is inactive. *)
            if has_coverers t u then
              aux work
            else
              if p u then
                Some u
              else
                begin
                  push_children t (Some u) work ;
                  aux work
                end
      in
      let work = W.create ()
      in
        push_children t None work ;
        aux work

    (*
     * [fold_active_prefix f t a] returns [(f uN ... (f u2 (f u1 a))...)], where
     * [u1, ..., uN] are the active vertices of [t], in the order obtained by a
     * depth-first traversal of the tree.
     *)
    let fold_active_prefix f t a =
      let rec aux u acc =
        (* The parent of [u], if any, should be active. *)
        assert (match parent t u with None -> true | Some p -> active t p) ;
        (* If [u] is covered, then its whole subtree is inactive. *)
        if has_coverers t u then
          acc
        else
          fold_children aux t (Some u) (f u acc)
      in
        fold_children aux t None a


    (*****                       M A I N    L O O P                       *****)

    let lart_limit = (Option.Main.limit ())

    (* For beautiful header with dynamic statistics. *)
    let backspace = ref 0

    (*
     * The following functions iteratively build an abstract reachability tree of
     * the model.  Arguments passed to these functions are:
     *
     *	[cpt_i]		loop iteration count
     *	[cpt_r]		refinement count
     *	[t]		the contructed tree
     *	[work]		the work list
     *)
    let rec explore cpt_i cpt_r t work =
      if (cpt_i > lart_limit) then
        (cpt_i - 1, cpt_r,
         Abort ("LART: loop iteration " ^ (string_of_int cpt_i) ^
                  " larger than limit " ^ (string_of_int lart_limit)))
      else
        begin
          Logger.slog Logger.Notice "@[%s%8d   ( %8d / %8d@]@?"
            (String.make !backspace '\b')
            cpt_i (nb_vertices t) (W.length work) ;
          backspace := 32 ;
          Logger.info "@.@[LART loop: %d@]@." cpt_i ;
          Logger.info "@[Abstract reachability tree:@ %d vertices@]@." (nb_vertices t) ;
          Logger.info "@[Work list:@ %d elements@]@." (W.length work) ;
          Logger.debug "@[<v 3>Abstract reachability tree:@,@[%a@]@]@." print t ;
          Logger.debug "@[<v 3>Work list:@,@[%a@]@]@." print_work_list work ;

          if W.is_empty work then
            (* Find an active leaf. *)
            match find_active_vertex (fun u -> not (is_expanded u)) t with
                None ->
                  Logger.slog Logger.Notice "@[ )@]@?" ;
                  (cpt_i, cpt_r,
                   Invariant
                     (Some
                        (lazy
                           (fold_active_prefix
                              (fun u l -> (location u, region u) :: l) t [])),
                      None))
              | Some u ->
                  assert (active t u) ;
                  process cpt_i cpt_r t u work
          else
            let (l, e) = W.pop work
            in
              match e with
                  Edge.Init _ ->
                    let u = add_vertex t l Region.top (ref false) e
                    in
                      process cpt_i cpt_r t u work
                | Edge.Child (p, _) ->
                    if not (mem_vertex t p) then
                      begin
                        Logger.slog Logger.Notice "@[ )@]@?" ;
                        backspace := !backspace + 2 ;
                        explore (cpt_i+1) cpt_r t work
                      end
                    else
                      let u = add_vertex t l Region.top (ref false) e
                      in
                        process cpt_i cpt_r t u work
        end
    and process cpt_i cpt_r t leaf work =
      Logger.debug "@[<v 3>Processing leaf:@,@[<h>%a@]@]@." Vertex.print leaf ;

      (* Sanity checks. *)
      assert (not (is_expanded leaf)) ;
      assert (not (has_children t leaf)) ;
      assert (not (has_coverers t leaf)) ;

      if not (really_active t leaf) then
        begin
          Logger.slog Logger.Notice "@[ )@]@?" ;
          backspace := !backspace + 2 ;
          explore (cpt_i+1) cpt_r t work
        end
      else
        let error = error_region_of_location (location leaf)
        in
          if Region.eq Region.bot error then
            begin
              expand t leaf work ;
              Logger.slog Logger.Notice "@[ )@]@?" ;
              backspace := !backspace + 2 ;
              explore (cpt_i+1) cpt_r t work
            end
          else
            refine cpt_i (cpt_r+1) t leaf error work
    and refine cpt_i cpt_r t leaf error work =
      let tr = trace_of_vertex t leaf error
      in
        (* Sanity check. *)
        assert ((Sequence.length tr) = 2 * (depth t leaf) + 3) ;

        Logger.slog Logger.Notice "@[ / %8d / %8d@]@?" cpt_r (depth t leaf) ;
        backspace := !backspace + 22 ;
        Logger.info "@[<v 3>Counterexample trace (%d):@,@[%a@]@]@."
          (depth t leaf)
          print_trace tr ;
        Logger.debug "@[<v 3>Counterexample branch:@,@[%a@]@]@."
          (print_branch t) leaf ;

        let trinv = TC.trace_check tr
        in
          (* XXX The trace-checker appends its statistics. *)
          backspace := !backspace + 11 ;
          match trinv with
              `Feasible _ ->
                (* The counterexample trace is feasible. *)
                Logger.slog Logger.Notice "@[ )@]@?" ;
                backspace := !backspace + 2 ;
                Logger.info "@[Counterexample valid@]@." ;
                (cpt_i, cpt_r,
                 Counterexample (path t leaf))
            | `Invariant_forward _ | `Invariant_backward _ ->
                (* The counterexample trace is spurious. *)
                refine_branch t leaf trinv ;
                if (mem_vertex t leaf) && (really_active t leaf) then
                  expand t leaf work ;
                Logger.slog Logger.Notice "@[ )@]@?" ;
                backspace := !backspace + 2 ;
                explore (cpt_i+1) cpt_r t work
    (*
     * [expand t leaf work] adds [leaf]'s location outbound edges to [work].
     *)
    and expand t leaf work =
      Logger.debug "@[<v 3>Expanding leaf:@,@[<h>%a@]@]@." Vertex.print leaf ;
      W.push_collection
        (fun f ->
          Automaton.iter_out_transitions
            (fun trans ->
              f (Transition.target trans, Edge.Child (leaf, trans)))
            (location leaf))
        work ;
      set_expanded leaf

  let model_check () =
    Logger.slog Logger.Notice "@[%8s %10s   %10s %10s %10s %10s %10s@]@."
      "" "Loops" "Nodes" "Work" "Refmnts" "Cex len" "k" ;
    Logger.slog Logger.Notice "@[LART loop: %8d   ( %8d / %8s / %8s / %8s / %8s )@]@?" 0 0 "" "" "" "" ;
    backspace := 67 ;
    Logger.info "@[LART loop: %d@]@." 0 ;
    let work = W.create ()
    and t = create ()
    in
      W.push_collection
        (fun f -> iter_init (fun (l, r) -> f (l, Edge.Init r)))
        work ;
      let (cpt_i, cpt_r, result) = explore 1 0 t work in
        Logger.notice "@." ;
        Logger.report "@[LART:@ %d loops,@ %d nodes,@ %d refinements@]@."
          cpt_i (nb_vertices t) cpt_r ;
        result
  end

  module Dfs = Explore (Stack)
  module Bfs = Explore (Queue)

  let model_check =
    match Option.LazyAbstraction.tree_exploration () with
        Option.LazyAbstraction.Bfs -> Bfs.model_check
      | Option.LazyAbstraction.Dfs -> Dfs.model_check
end
