(* $Id: Model.mli 1109 2011-08-10 12:47:22Z gsutre $ *)

(*
 * Copyright (c) 2009, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * This module defines the input interface of the model-checking engine,
 * which is given by the signature S below.  This signature specifies models
 * given by:
 * - a control-flow automaton where transitions are labeled by commands
 * - a region algebra to represent sets of environments
 * - post/pre operators over regions giving the semantics of the commands
 * - an extrapolation operator over regions for abstraction refinement
 * Intuitively, an environment is a valuation of the variables of the model.
 * Variables of the model are not explicit in this interface: one may assume
 * that there is only one variable over some domain whose value is modified
 * by commands.  A region represents a subset of this domain, i.e. a set of
 * values of this implicit variable.  Thus commands glue together the syntax
 * (control-flow automaton) and the semantics (post/pre operators).
 *)


open Util


(**
 * The COMMAND signature simply specifies an abstract data-type for commands.
 *)
module type COMMAND =
sig
  include Print.PRINTABLE_TYPE

  type t_command = Internal | Input | Output

  val give_type : t -> t_command
  val give_queue : t -> int
  val give_message : t -> string
end


(**
 * The AUTOMATON signature specifies the syntactic representation of models.
 * We consider models that are finite-state control-flow automata whose
 * Transitions are labeled with commands.  The syntax of commands is left
 * unspecified as an abstract type.  Instead of accessors to outbound and
 * inbound transitions as lists, which would potentially enforce conversions
 * to lists, we provide iterators over outbound and inbound transitions.
 *)
module type AUTOMATON =
sig
  (* Owners are values of type Owner.t *)
  module Owner :
  sig
    include Print.PRINTABLE_TYPE
    val compare : t -> t -> int
    val equal : t -> t -> bool
    val hash : t -> int
  end

  (* Locations are values of type Location.t *)
  module Location :
  sig
    include Print.PRINTABLE_TYPE
    val compare : t -> t -> int
    val equal : t -> t -> bool
    val hash : t -> int
  end

  (* Commands are values of type Command.t *)
  module Command : COMMAND

  (* Transitions are values of type Transition.t *)
  module Transition :
  sig
    include Print.PRINTABLE_TYPE

    (* Accessors for transitions *)
    val owner : t -> Owner.t
    val is_controllable : t -> bool
    val source : t -> Location.t
    val target : t -> Location.t
    val command : t -> Command.t
  end

  (* Number of locations and of transitions *)
  val nb_locations : int
  val nb_transitions : int
  val nb_owners : int

  (* Iterators over all locations and over all transitions *)
  val iter_locations :   (Location.t   -> unit) -> unit
  val iter_transitions : (Transition.t -> unit) -> unit
  val iter_owners :      (Owner.t      -> unit) -> unit
  val fold_locations :   (Location.t   -> 'a -> 'a) -> 'a -> 'a
  val fold_transitions : (Transition.t -> 'a -> 'a) -> 'a -> 'a
  val fold_owners :      (Owner.t      -> 'a -> 'a) -> 'a -> 'a

  (* Iterators over inbound and outbound transitions of a given location *)
  val iter_out_transitions : (Transition.t -> unit) -> Location.t -> unit
  val iter_in_transitions :  (Transition.t -> unit) -> Location.t -> unit
  val fold_out_transitions : (Transition.t -> 'a -> 'a) -> Location.t ->
                             'a -> 'a
  val fold_in_transitions :  (Transition.t -> 'a -> 'a) -> Location.t ->
                             'a -> 'a

  (* Pretty-printer for this automaton *)
  val print_automaton : Format.formatter -> unit -> unit
end


(**
 * The REGION signature specifies a boolean algebra that provides a symbolic
 * representation for sets of environments.  Elements in this algebra are called
 * regions.  Operators over regions (union, intersection and difference) shall
 * be exact.
 *)

(*
 * To keep the interface simple, there is no dedicated function to test whether
 * a region represents the empty set or the whole set.  Theses tests can be
 * performed using the eq or leq functions, and these two functions should be
 * optimized accordingly (by checking whether some argument is physically equal
 * to bot or top).
 *)
module type REGION =
sig
  include Print.PRINTABLE_TYPE
  (* region representing the emptyset *)
  val bot : t

  (* region representing the whole set *)
  val top : t

  (* equivalence relation over regions corresponding to set equality: eq(a, b)
     is true when a and b represent the same set *)
  val eq : t -> t -> bool

  (* partial order over regions corresponding to set inclusion: leq(a, b) is
     true when the set represented by a is contained in the set represented b *)
  val leq : t -> t -> bool

  (* operator corresponding to set union: union(a, b) is a region
     representing the set of all elements in a or b *)
  val union : t -> t -> t

  (* operator corresponding to set intersection: inter(a, b) is a region
     representing the set of all elements in a and b *)
  val inter : t -> t -> t

  (* operator corresponding to set complementation: compl(a) is a region
     representing the set of all elements (in top) not in a *)
  val compl : t -> t

  (* operator corresponding to set difference: minus(a, b) is a region
     representing the set of all elements in a and not in b *)
  val minus : t -> t -> t

  (* simplify the region, or at least normalize it *)
  val simplify : t -> t

  (* estimate the size of the internal representation of the region *)
  val size : t -> int

  (* print the internal representation of the region *)
  val dump : Format.formatter -> t -> unit
end


(**
 * The SEMANTICS signature specifies how the semantics of the model shall be
 * given to the model-checking engine.
 *)
module type SEMANTICS =
sig
  (* Region module used to give the semantics *)
  module Region : REGION

  (* Commands whose semantics is given by this module *)
  module Command : COMMAND

  (* Initial region *)
  (* val initial_region : Region.t *)

  (* symbolic post function *)
  val post : Region.t -> Command.t -> Region.t

  (* symbolic pre function *)
  val pre  : Region.t -> Command.t -> Region.t

  (* symbolic relation predicate: [rel cmd r1 r2] is true when post(r1, cmd)
     intersects r2, or equivalently when pre(r2, cmd) intersects r1 *)
  val rel  : Command.t -> Region.t -> Region.t -> bool
end


(**
 * The CLOSURE signature specifies a hierarchy of extensive idempotent operators
 * on a Region module.
 *)
module type CLOSURE =
sig
  (* Region module *)
  module Region : REGION

  (* the extensive idempotent closure operator *)
  val clo : int -> Region.t -> Region.t
end


(**
 * Aggregation of SEMANTICS and CLOSURE over the same region algebra.
 *)
module type SEMANTICS_CLOSURE =
sig
  include SEMANTICS
  include CLOSURE with module Region := Region
end


(**
 * The main signature of this module, defining the input interface of the
 * model-checking engine.  This signature basically is a container for the
 * above signatures.  Note that the module equality constraints given in
 * this signature really means that the modules should be the same.  However
 * the OCaml typing system only checks that type equality constraints are
 * satified.
 *)
module type S =
sig
  module Command : COMMAND

  module Automaton : AUTOMATON
  with module Command = Command

  module Semantics : SEMANTICS
  with module Command = Command

  module Closure : CLOSURE
  with module Region = Semantics.Region

  (* Data-type for symbolic states *)
  module State : Print.PRINTABLE_TYPE
  with type t = Automaton.Location.t * Semantics.Region.t

  (* Number of initial symbolic states and of error symbolic states *)
  val nb_init : int
  val nb_error : int

  (* Iterators over initial symbolic states and error symbolic states *)
  val iter_init :  (State.t -> unit) -> unit
  val iter_error : (State.t -> unit) -> unit
  val fold_init :  (State.t -> 'a -> 'a) -> 'a -> 'a
  val fold_error : (State.t -> 'a -> 'a) -> 'a -> 'a

  (* Pretty-printer for this model *)
  val print_model : Format.formatter -> unit -> unit
end
