(* $Id: ModelChecker.mli 1092 2011-08-08 12:50:05Z gsutre $ *)

(*
 * Copyright (c) 2009 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * Model Checker.
 *
 * This module defines the interface of model-checking engines, as a functor
 * taking as argument a module for the model to be model-checked, and returning
 * a module with a model-checking function.  The interface used for models is
 * defined in the {! Model} module.
 *)


module type S = functor (A : Model.S) ->
sig
  (**
   * Specification of (and checking functions for) the result of model-checking
   * functions is provided in the {! Validator} module.  See the documentation
   * of the {! Validator} module for more information on the [result] type
   * below.
   *)

  (**
   * Type of possible results of the model-checking function.
   *)
  type result =
      Invariant of ((A.State.t list) Lazy.t) option * ((A.State.t list) Lazy.t) option
        (** Safety certificate:
          forward inductive invariant disjoint from error symbolic states, and
          backward inductive invariant disjoint from initial symbolic states. *)
    | Counterexample of A.Automaton.Transition.t list
        (** Safety counterexample:
          path in the model's automaton that is feasible (i.e. corresponds to a
          run of the model) with respect to (1) the model's semantics and (2)
          the model's initial and error symbolic states. *)
    | Abort of string
        (** Model checking abortion, with a message string. *)

  (**
   * The model-checking function.
   *)
  val model_check : unit -> result
end
