(* $Id: Option.ml 1240 2011-10-17 21:44:06Z gsutre $ *)

(*
 * Copyright (c) 2009, 2010, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * This module manages global options and performs command-line processing.
 *)


open Util


(* Exception for errors that are internal to this module, such as cases in
   pattern matching that should not arise. *)
exception InternalError


(* Helper functions for association lists. *)
let rec value_of_key key =
  function
      [] ->
        raise InternalError
    | (k, v) :: l ->
        if compare k key = 0 then v else value_of_key key l

let rec key_of_value value =
  function
      [] ->
        raise InternalError
    | (k, v) :: l ->
        if compare v value = 0 then k else key_of_value value l

(* Helper function for symbol argument specifications. *)
let arg_spec_of_symbols opt_name symbols opt_ref doc =
  (opt_name,
   Arg.Symbol (snd (List.split symbols),
               (fun s -> opt_ref := (key_of_value s symbols))),
   " " ^ doc ^ "   (default "
   ^ (value_of_key !opt_ref symbols) ^ ")")

(* Helper function for integer argument specifications. *)
let arg_spec_of_integer opt_name minimum opt_ref doc =
  (opt_name,
   Arg.Int
     (fun n ->
        if n >= minimum then
          opt_ref := n
        else
          let s =
            if minimum = 0 then "negative"
            else "lesser than " ^ (string_of_int minimum)
          in
            raise (Arg.Bad ("Incorrect " ^ s ^ " value `" ^ (string_of_int n) ^
                               "' for option " ^ opt_name))),
   "<int> " ^ doc ^ "   (default "
   ^ (string_of_int !opt_ref) ^ ")")


(* Options for the ScmWrapper module *)
module ScmWrapper =
struct
  type rel_cmd_impl_type =
      Post
    | Pre
    | Mixed

  let rel_cmd_impl_names = [
    Post,  "post" ;
    Pre,   "pre" ;
    Mixed, "mixed"
  ]

  type extrapolation_type =
      Bisim_fwd
    | Bisim_bwd
    | Bisim_both
    | Lang_fwd
    | Lang_bwd
    | Lang_both
    | Identity

  let extrapolation_names = [
    Bisim_fwd,  "bisim-fwd" ;
    Bisim_bwd,  "bisim-bwd" ;
    Bisim_both, "bisim-both" ;
    Lang_fwd,   "lang-fwd" ;
    Lang_bwd,   "lang-bwd" ;
    Lang_both,  "lang-both" ;
    Identity,   "identity"
  ]

  type extrapol_coloring_type =
      Final
    | Init
    | Both
    | None
    | Auto

  let extrapol_coloring_names = [
    Final, "final" ;
    Init,  "init" ;
    Both,  "both" ;
    None,  "none" ;
    Auto,  "auto"
  ]

  type channel_policy_type =
      Lossy
    | Perfect
    | Normal

  let channel_policy_names = [
    Lossy,   "all-lossy" ;
    Perfect, "all-perfect" ;
    Normal,  "normal"
  ]

  let rel_cmd_impl_opt      = ref Mixed
  let extrapolation_opt     = ref Bisim_fwd
  let extrapol_coloring_opt = ref Auto
  let channel_policy_opt    = ref Normal

  let rel_cmd_impl ()      = !rel_cmd_impl_opt
  let extrapolation ()     = !extrapolation_opt
  let extrapol_coloring () = !extrapol_coloring_opt
  let channel_policy ()    = !channel_policy_opt

  (* Specification of command-line options used by this module. *)
  let arg_spec () = [
    (arg_spec_of_symbols
       "-scm-rel"
       rel_cmd_impl_names
       rel_cmd_impl_opt
       "Backend semantics operator for SCM command relation")
    ;
    (arg_spec_of_symbols
       "-extrapolation"
       extrapolation_names
       extrapolation_opt
       "Extrapolation operator")
    ;
    (arg_spec_of_symbols
       "-extrapol-coloring"
       extrapol_coloring_names
       extrapol_coloring_opt
       "Extrapolation initial coloring")
    ;
    (arg_spec_of_symbols
       "-channel-policy"
       channel_policy_names
       channel_policy_opt
       "Channel reliability policy")
  ]
end

(* Options for the Static Analyzer module*)
module StaticAnalyzer =
struct
  type forward_backward_type =
      Forward
    | Backward

  let forward_backward_names = [
    Forward,  "forward" ;
    Backward,  "backward"
  ]

  let direction_opt = ref Forward

  let direction () = !direction_opt

  (* Specification of command-line options used by this module. *)
  let arg_spec () = [
    (arg_spec_of_symbols
       "-direction"
       forward_backward_names
       direction_opt
       "forward or backward analysis")
  ]
end

(* Options for the Control module *)
module Control =
struct
  type simulation_mode_type =
      Manual
    | Automatic

  let simulation_mode_names = [
    Manual,  "manual" ;
    Automatic, "auto"
  ]

  let show_estimate_opt   = ref false
  let simulation_only_opt = ref false
  let simulation_mode_opt = ref Manual

  let show_estimate ()   = !show_estimate_opt
  let simulation_only () = !simulation_only_opt
  let simulation_mode () = !simulation_mode_opt

  (* Specification of command-line options used by this module. *)
  let arg_spec () = [
    (arg_spec_of_symbols
       "-simulation-mode"
       simulation_mode_names
       simulation_mode_opt
       "Simulation mode")
    ;
    ("-display-estimate",
     Arg.Set show_estimate_opt,
     " Displays the current estimate")
    ;
    ("-simulation-only",
     Arg.Set simulation_only_opt,
     " Simulate the system, but do not control it (version of FORTE paper)")
  ]
end

(* Options for the Cegar module *)
module Cegar =
struct
  type graph_exploration_type =
      Fwd_dfs
    | Fwd_bfs
    | Bwd_dfs
    | Bwd_bfs
    | Mixed

  let graph_exploration_names = [
    Fwd_dfs, "fwd-dfs" ;
    Fwd_bfs, "fwd-bfs" ;
    Bwd_dfs, "bwd-dfs" ;
    Bwd_bfs, "bwd-bfs" ;
    Mixed,   "mixed"
  ]

  type graph_refinement_type =
      Rel
    | PostPre

  let graph_refinement_names = [
    Rel,     "rel" ;
    PostPre, "post-pre"
  ]

  let graph_exploration_opt    = ref Mixed
  let graph_refinement_opt     = ref PostPre

  let graph_exploration ()    = !graph_exploration_opt
  let graph_refinement ()     = !graph_refinement_opt

  (* Specification of command-line options used by this module. *)
  let arg_spec () = [
    (arg_spec_of_symbols
       "-graph-exploration"
       graph_exploration_names
       graph_exploration_opt
       "Abstract graph exploration method")
    ;
    (arg_spec_of_symbols
       "-graph-refinement"
       graph_refinement_names
       graph_refinement_opt
       "Abstract graph refinement method")
  ]
end

(* Options for the LazyAbstraction module *)
module LazyAbstraction =
struct
  type tree_exploration_type =
      Bfs
    | Dfs

  let tree_exploration_names = [
    Bfs, "bfs" ;
    Dfs, "dfs"
  ]

  let tree_exploration_opt = ref Bfs
  let tree_pruning_opt     = ref true

  let tree_exploration () = !tree_exploration_opt
  let tree_pruning ()     = !tree_pruning_opt

  (* Specification of command-line options used by this module. *)
  let arg_spec () = [
    (arg_spec_of_symbols
       "-tree-exploration"
       tree_exploration_names
       tree_exploration_opt
       "Abstract tree exploration strategy")
    ;
    ("-no-tree-pruning",
     Arg.Clear tree_pruning_opt,
     " Do not prune the sub-tree of empty vertices")
  ]
end

(* Options for the APInv module *)
module APInv =
struct
  type safe_approximation_type =
      Split
    | CoSplit

  let safe_approximation_names = [
    Split,   "split" ;
    CoSplit, "co-split"
  ]

  let safe_approximation_opt = ref Split

  let safe_approximation () = !safe_approximation_opt

  (* Specification of command-line options used by this module. *)
  let arg_spec () = [
    (arg_spec_of_symbols
       "-safe-approximation"
       safe_approximation_names
       safe_approximation_opt
       "Safe over-approximation algorithm")
  ]
end

(* Options for the Main module *)
module Main =
struct
  type mc_engine_type =
      Absint
    | Armc
    | Cegar
    | Lart

  let mc_engine_names = [
    Absint, "absint" ;
    Armc,   "armc" ;
    Cegar,  "cegar" ;
    Lart,   "lart"
  ]

  type tc_engine_type =
      APInv_bwd
    | APInv_fwd
    | UPInv_bwd
    | UPInv_fwd

  let tc_engine_names = [
    APInv_bwd,    "apinv-bwd" ;
    APInv_fwd,    "apinv-fwd" ;
    UPInv_bwd,    "upinv-bwd" ;
    UPInv_fwd,    "upinv-fwd"
  ]

  let mc_engine_opt             = ref Cegar
  let tc_engine_opt             = ref APInv_bwd
  let k_opt                     = ref 1
  let k_min_opt                 = ref 0
  let k_max_opt                 = ref 2
  let tc_discard_states_opt     = ref false
  let tc_validate_opt           = ref false
  let limit_opt                 = ref max_int
  let scm_source_file_opt       = ref None
  let scm_consistency_check_opt = ref true
  let result_validation_opt     = ref true
  let show_statistics_opt       = ref false

  let mc_engine ()             = !mc_engine_opt
  let tc_engine ()             = !tc_engine_opt
  let k ()                     = !k_opt
  let k_min ()                 = !k_min_opt
  let k_max ()                 = !k_max_opt
  let tc_discard_states ()     = !tc_discard_states_opt
  let tc_validate ()           = !tc_validate_opt
  let limit ()                 = !limit_opt
  let scm_source_file ()       = !scm_source_file_opt
  let scm_consistency_check () = !scm_consistency_check_opt
  let result_validation ()     = !result_validation_opt
  let show_statistics ()       = !show_statistics_opt

  (* Main command-line options for control(1). *)
  let control_arg_spec () = [
    (arg_spec_of_integer
       "-k"
       0
       k_opt
       "Value of extrapolation parameter")
    ;
    (arg_spec_of_integer
       "-limit"
       0
       limit_opt
       "Number of simulation steps (automatic mode only)")
    ;
    ("-no-scm-consistency-check",
     Arg.Clear scm_consistency_check_opt,
     " Disable consistency check of input SCM description")
    ;
    ("-statistics",
     Arg.Set show_statistics_opt,
     " Show performance statistics")
  ]

  (* Main command-line options for verify(1). *)
  let verify_arg_spec () = [
    (arg_spec_of_symbols
       "-mc-engine"
       mc_engine_names
       mc_engine_opt
       "Model-checking engine")
    ;
    (arg_spec_of_symbols
       "-tc-engine"
       tc_engine_names
       tc_engine_opt
       "Trace-checking engine")
    ;
    (arg_spec_of_integer
       "-k-min"
       (-1)
       k_min_opt
       "Minimum value of extrapolation parameter")
    ;
    (arg_spec_of_integer
       "-k-max"
       0
       k_max_opt
       "Maximum value of extrapolation parameter")
    ;
    ("-tc-discard-states",
     Arg.Set tc_discard_states_opt,
     " Ignore abstract states for trace checking")
    ;
    ("-tc-validate",
     Arg.Set tc_validate_opt,
     " Validate all trace-checking results")
    ;
    (arg_spec_of_integer
       "-limit"
       0
       limit_opt
       "Maximum number of model-checker iterations")
    ;
    ("-no-scm-consistency-check",
     Arg.Clear scm_consistency_check_opt,
     " Disable consistency check of input SCM description")
    ;
    ("-no-validation",
     Arg.Clear result_validation_opt,
     " Disable validation of the model-checking result")
    ;
    ("-statistics",
     Arg.Set show_statistics_opt,
     " Show performance statistics")
  ]
end

(* Logging options *)
module Logging =
struct
  (* Default number of columns for pretty-printing. *)
  let () =
    try
      Format.set_margin (int_of_string (Sys.getenv ("COLUMNS")))
    with
        Not_found | Failure _ -> Format.set_margin 78

  (* Default maximum number of boxes for pretty-printing. *)
  let () =
    Format.set_ellipsis_text "..." ;
    Format.set_max_boxes max_int

  (* Specification of command-line options used by this module. *)
  let arg_spec () = [
    ("-mute",
     Arg.Unit (fun () -> Logger.set_priority Logger.Error),
     " No output unless an error occurs")
    ;
    ("-quiet",
     Arg.Unit (fun () -> Logger.set_priority Logger.Report),
     " Suppress non-essential messages")
    ;
    ("-normal",
     Arg.Unit (fun () -> Logger.set_priority Logger.Notice),
     " Restore the default output of messages")
    ;
    ("-verbose",
     Arg.Unit (fun () -> Logger.set_priority Logger.Info),
     " More informational messages")
    ;
    ("-debug",
     Arg.Unit (fun () -> Logger.set_priority Logger.Debug),
     " All messages, including debugging ones")
    ;
    ("-columns",
     Arg.Int (fun n -> Format.set_margin (max 1 n)),
     "<int> Use <int> columns for pretty-printing")
    ;
    ("-box-depth",
     Arg.Int (fun n -> Format.set_max_boxes (max 1 n)),
     "<int> Use at most <int> nested boxes for pretty-printing")
  ]
end

(* Miscellaneous options *)
module Miscellaneous =
struct
  (* Specification of command-line options used by this module. *)
  let arg_spec () = [
    ("-version",
     Arg.Unit (fun () -> print_endline Config.version ; exit 0),
     " Print version and exit")
  ]
end


(* Main part of Option module *)
let usage_msg = "Usage: "^(Sys.argv.(0))^" [options] [<SCM source file>]\n"

let control_arg_spec () =
    (("\nMain Options:", Arg.Unit (fun () -> ()), "") :: (Main.control_arg_spec ()))
  @ (("\nSCM Wrapper Options:", Arg.Unit (fun () -> ()), "") :: (ScmWrapper.arg_spec ()))
  @ (("\nSupervisory Control Options:", Arg.Unit (fun () -> ()), "") :: (Control.arg_spec ()))
  @ (("\nLogging Options:", Arg.Unit (fun () -> ()), "") :: (Logging.arg_spec ()))
  @ (("\nMiscellaneous Options:", Arg.Unit (fun () -> ()), "") :: (Miscellaneous.arg_spec ()))

let verify_arg_spec () =
    (("\nMain Options:", Arg.Unit (fun () -> ()), "") :: (Main.verify_arg_spec ()))
  @ (("\nSCM Wrapper Options:", Arg.Unit (fun () -> ()), "") :: (ScmWrapper.arg_spec ()))
  @ (("\nCegar Options:", Arg.Unit (fun () -> ()), "") :: (Cegar.arg_spec ()))
  @ (("\nLazy Abstraction Options:", Arg.Unit (fun () -> ()), "") :: (LazyAbstraction.arg_spec ()))
  @ (("\nAPInv Options:", Arg.Unit (fun () -> ()), "") :: (APInv.arg_spec ()))
  @ (("\nStatic Analyzer Options:", Arg.Unit (fun () -> ()), "") :: (StaticAnalyzer.arg_spec ()))
  @ (("\nLogging Options:", Arg.Unit (fun () -> ()), "") :: (Logging.arg_spec ()))
  @ (("\nMiscellaneous Options:", Arg.Unit (fun () -> ()), "") :: (Miscellaneous.arg_spec ()))

let initialize arg_spec =
  Arg.parse
    (Arg.align arg_spec)
    (fun a ->
       match !Main.scm_source_file_opt
       with
           None ->
             Main.scm_source_file_opt := Some a
         | Some _ ->
             raise (Arg.Bad ("unexpected argument `" ^ a ^ "' (multiple input files are not allowed)")))
    usage_msg
