(* $Id: ReachabilityEngine.ml 1102 2011-08-09 17:49:45Z gsutre $ *)

(*
 * Copyright (c) 2010, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Tristan Le Gall <tlegall@ulb.ac.be>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util


exception InternalError

module Make (A : Model.S) =
struct
  (* We open A and define shortcuts to simplify notations *)
  open A
  module Location = Automaton.Location
  module Transition = Automaton.Transition
  module Region = Semantics.Region


  (* internal types for the construction of the cfg *)
  type t_vertex = DummyVertex | Vertex of Location.t
  type t_hedge = DummyHedge | Hedge of Transition.t
  (* printing functions for those types *)
  let print_vertex fmt v =
    match v with
	Vertex x -> Location.print fmt x
      | DummyVertex -> Format.fprintf fmt "DummyV"

  let print_hedge fmt h =
    match h with
	Hedge x -> Transition.print fmt x
      | DummyHedge -> Format.fprintf fmt "DummyH"


  (*  prints a moving clock in normal mode *)
  let cmpt = ref 0

  let succ_cmpt () = 
    begin
      if !cmpt = 3
      then
	cmpt := 0
      else
	incr cmpt
    end


  let display_clock () =
 
    let aff =
    match !cmpt with
	0 ->  "|"
      | 1 ->  "/"
      | 2 ->  "-"
      | 3 ->  "\\"
      | _ -> failwith "bug in Display_clock (ReachbilityEngine.ml)"
    in
    Logger.slog Logger.Notice "@[\b%s@]@?" aff;
    succ_cmpt()


  (** 
      Optional parameters :
      the highest [k] is, the more precise the analysis is (in general)
      if [forward], then a forward analysis is done, otherwise a backward analysis is done 
  *)

  let reach 
      ?(forward=true) 
      ?(k=1)
      ?(debug=false)
      (initial_states: State.t list) 
      (transition_list:Transition.t list) : State.t list 
      =
    (* trivial case *)
    if (transition_list = [] ) || (initial_states = [])
    then 
      initial_states
    else
      begin
	(* creation of the graph *)      
	let (g:(t_vertex,t_hedge,unit,unit,unit) PSHGraph.t) = 
	  PSHGraph.create 
	    {
	      PSHGraph.hashv = Hashhe.stdcompare;
	      PSHGraph.hashh = Hashhe.stdcompare;
	      PSHGraph.comparev = compare;
	      PSHGraph.compareh = compare;
	    } 
	    Automaton.nb_locations 
	    () 
	in

	let empty_set : t_vertex PSette.t = PSette.empty g.PSHGraph.compare.PSHGraph.comparev
	in
	(* construction of the graph *)
	Automaton.iter_locations (fun l ->  PSHGraph.add_vertex g (Vertex l) ()) ;
	

	(* IMPORTANT NOTE: if we perform a backward analysis we need to
	   reverse the graph     
	*)

	let f_iter_trans (t:Transition.t) =
	  let idx = Transition.source t in
	  let idy = Transition.target t in
	  if List.mem t transition_list
	  (* tests whether the transition is active and should count in the reachability computation *)
	  then
	    begin
	      if forward 
	      then (* ***forward analysis*** *)
		PSHGraph.add_hedge g (Hedge t) () ~pred:[|Vertex idx|] ~succ:[|Vertex idy|]
	      else (* ***backward analysis*** *)
		PSHGraph.add_hedge g (Hedge t) () ~pred:[|Vertex idy|] ~succ:[|Vertex idx|]
	    end
	in
	Automaton.iter_transitions f_iter_trans ;
	(* HERE: graph [g] is completely built *)


	(* MANAGER *)
	let manager :(t_vertex,t_hedge,Region.t,unit) Fixpoint.manager = 
	  {
	    Fixpoint.bottom = (fun v -> Region.bot) ;
	    Fixpoint.canonical = (fun v abs -> ());
	    (* no canonical function *)

	    Fixpoint.is_bottom = (fun v a ->  Region.leq a Region.bot);
	    Fixpoint.is_leq = (fun v a1 a2 -> Region.leq a1 a2) ;
	    Fixpoint.join = (fun v a1 a2 -> Region.union a1 a2) ;
	    Fixpoint.join_list = 
	      begin fun v la -> 
		Region.simplify (
		  List.fold_left 
		    Region.union
		    Region.bot 
		    la
		)
	      end;

	    Fixpoint.widening = 
	      (begin fun vertex x y -> 	
		if Region.leq x y
		then
		  (* XXX put widening *)
		  Closure.clo k y 
		else 
		  begin
		    Logger.debug
		      "vertex %a \n abs1 = @. @[ %a @] @."
		      print_vertex vertex
		      Region.print x ;
		    Logger.debug
		      "abs2 = @. @[ %a @] @."
		      Region.print y ;
		    
		    raise (Invalid_argument "Fixpoint.widening")
		  end
	       end);

	    Fixpoint.apply=
	      (begin fun hedge tx ->
		Logger.debug  "@[ hedge : %a @]@." print_hedge hedge ;
		assert ((Array.length tx) = 1);
		let x = tx.(0) in
		((), 
		 begin
		   match hedge with
		       DummyHedge -> failwith "DummyHegde"
		     | Hedge hedge ->
		       if forward
		       then (* forward analysis *)
			 Semantics.post x (Transition.command hedge)
		       else (* backward analysis *)
			 Semantics.pre x (Transition.command hedge)
		 end)
	       end);

	    Fixpoint.arc_init=(fun v -> ());

	    Fixpoint.abstract_init =
	      (fun loc ->
		match loc with
		    DummyVertex -> failwith ""
		  | Vertex loc ->
		    begin
		      let res = ref None in
		      List.iter
			begin 
			  fun (l_init,r_init) ->
			    if Location.equal l_init loc
			    then 
			      res := 
				begin
				  match !res with
				      None -> 
					Some r_init
				    | Some r ->
				      Some (Region.union r r_init)
				end
			end
			initial_states
		      ;
		      match !res with
			  Some r -> r
			| None -> failwith ""
			  
		    end);
	    Fixpoint.accumulate=true;

	    Fixpoint.print_abstract= Region.print;
	    Fixpoint.print_arc =(fun fmt _ -> ()) ;
	    
	    Fixpoint.print_vertex=print_vertex;
	    Fixpoint.print_hedge=print_hedge;

	    Fixpoint.widening_start=2 ;
	    Fixpoint.widening_descend=2;
	    
	    Fixpoint.print_fmt = Format.std_formatter;
	    Fixpoint.print_analysis= false;
	    Fixpoint.print_component = false;
	    Fixpoint.print_step= debug ;
	    Fixpoint.print_state= debug ;
	    Fixpoint.print_postpre= debug;
	    Fixpoint.print_workingsets= false;

	    Fixpoint.dot_fmt = None;
	    Fixpoint.dot_vertex = print_vertex; 
	    Fixpoint.dot_hedge = print_hedge;
	    
	    Fixpoint.dot_attrvertex = (fun x y -> ());
	    Fixpoint.dot_attrhedge = (fun x y -> ());
	  }

	in

	let sette_init : t_vertex PSette.t =
	  List.fold_left (fun res s -> PSette.add (Vertex (fst s)) res) empty_set initial_states
	in
	assert (initial_states <> []);
	assert (transition_list <> []);
	let strat = 
	  Fixpoint.make_strategy_default 
	    DummyVertex
	    DummyHedge
	    g 
	    sette_init 
	in
	Logger.debug "strategy=%a@." (Fixpoint.print_strategy manager) strat;


	let output = 
	  Fixpoint.analysis_std
	    manager 
	    g 
	    sette_init
	    strat
	in
	Logger.info "@[postfixpt=%a@]@." (Fixpoint.print_output manager) output;
	
	let list_output = ref [] in
	let f_iter_output_to_mappe vertex reach ~pred ~succ =
	  match vertex with
	      DummyVertex -> ()
	    | Vertex v -> list_output := (v, reach)::!list_output
	in
	PSHGraph.iter_vertex output f_iter_output_to_mappe ;
	!list_output
	  
      end

end
