(* $Id: ReachabilityTree.ml 1183 2011-10-13 22:02:51Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util
open Collection


(* Exception for errors that are internal to this module. *)
exception InternalError

module Make (A : Model.S) (Mark : Print.PRINTABLE_TYPE) =
struct
  (* Open A and define shortcuts to essential submodules. *)
  open A
  module Location = Automaton.Location
  module Transition = Automaton.Transition
  module Region = Semantics.Region

  (* Maps over locations. *)
  module LMap = Map.Persistent.MakeTreeMap (Location)

  module type VERTEX =
  sig
    type vertex_set
    type edge
    type tree

    type t_data = {
      tree : tree ;
      depth : int ;
      location : Location.t ;
      mutable region : Region.t ;
      mark : Mark.t ;
      parent : edge ;
      mutable children : vertex_set ;
      mutable coverees : vertex_set ;
      mutable coverers : vertex_set ;
      mutable zombie : bool ;
    }

    include Print.PRINTABLE_TYPE

    val create : tree -> int -> Location.t -> Region.t -> Mark.t -> edge -> t

    val data : t -> t_data
    val id : t -> int

    val location : t -> Location.t
    val region : t -> Region.t
    val mark : t -> Mark.t
    val refine : t -> Region.t -> unit

    val compare : t -> t -> int
    val equal : t -> t -> bool
    val hash : t -> int
  end

  module type EDGE =
  sig
    type vertex

    type t =
        Init of Region.t
      | Child of (vertex * Transition.t)

    include Print.PRINTABLE_TYPE with type t := t
  end

  module type TREE =
  sig
    type vertex
    type vertex_set

    type t = {
      mutable init : vertex_set ;
      mutable lmap : vertex_set LMap.t ;
      mutable size : int ;
    }

    val create : unit -> t

    val nb_vertices : t -> int
  end

  module rec Vertex : (VERTEX with type vertex_set = VSet.t
                              and  type edge = Edge.t
                              and  type tree = Tree.t) =
  struct
    type vertex_set = VSet.t
    type edge = Edge.t
    type tree = Tree.t

    type t_data = {
      tree : tree ;
      depth : int ;
      location : Location.t ;
      mutable region : Region.t ;
      mark : Mark.t ;
      parent : edge ;
      mutable children : vertex_set ;
      mutable coverees : vertex_set ;
      mutable coverers : vertex_set ;
      mutable zombie : bool ;
    }

    let print_data fmt data =
      Format.fprintf fmt "{ @[" ;
      Format.fprintf fmt "@[<3>location@ =@ @[%a@]@],@ " Location.print data.location ;
      Format.fprintf fmt "@[<3>region@ =@ @[%a@]@],@ " Region.print data.region ;
      Format.fprintf fmt "@[<3>mark@ =@ @[%a@]@]" Mark.print data.mark ;
      Format.fprintf fmt "@] }"

    include Uid.Make (
      struct
        type t = t_data
        let print = print_data
      end)

    let create t d loc reg m e =
      create
        { tree = t ;
          depth = d ;
          location = loc ;
          region = reg ;
          mark = m ;
          parent = e ;
          children = VSet.empty ;
          coverees = VSet.empty ;
          coverers = VSet.empty ;
          zombie = false }

    let data = value

    let location u = (data u).location
    let region u = (data u).region
    let mark u = (data u).mark

    let refine u reg =
      assert (Region.leq reg (region u)) ;
      (data u).region <- reg
  end
  and Edge : (EDGE with type vertex = Vertex.t) =
  struct
    type vertex = Vertex.t

    type t =
        Init of Region.t
      | Child of (vertex * Transition.t)

    let print fmt e =
      match e with
          Init init ->
            Format.fprintf fmt "Init@ @[%a@]" Region.print init
        | Child (p, trans) ->
            Format.fprintf fmt "Child@ of@ %d@ via@ @[%a@]" (Vertex.id p) Transition.print trans
  end
  and Tree : (TREE with type vertex = Vertex.t
                   and  type vertex_set = VSet.t) =
  struct
    type vertex = Vertex.t
    type vertex_set = VSet.t

    type t = {
      mutable init : vertex_set ;          (* children of the virtual root *)
      mutable lmap : vertex_set LMap.t ;   (* vertices in a given location *)
      mutable size : int ;                 (* total number of vertices *)
    }

    let create () =
      { init = VSet.empty ;
        lmap = LMap.empty ;
        size = 0 }

    let nb_vertices t = t.size
  end
  (* Sets of vertices. *)
  and VSet : (Set.PS with type elt = Vertex.t) = Set.Persistent.MakeTreeSet (Vertex)

  include Tree

  let vset_of_loc t loc =
    try LMap.find t.lmap loc with
        Not_found -> VSet.empty

  let mem_vertex t u =
    (t == (Vertex.data u).Vertex.tree) && (not (Vertex.data u).Vertex.zombie)

  let depth t u =
    assert (mem_vertex t u) ;
    (Vertex.data u).Vertex.depth

  let has_children t u =
    assert (mem_vertex t u) ;
    not (VSet.is_empty (Vertex.data u).Vertex.children)

  let has_coverees t u =
    assert (mem_vertex t u) ;
    not (VSet.is_empty (Vertex.data u).Vertex.coverees)

  let has_coverers t u =
    assert (mem_vertex t u) ;
    not (VSet.is_empty (Vertex.data u).Vertex.coverers)

  let edge t u =
    assert (mem_vertex t u) ;
    (Vertex.data u).Vertex.parent

  let parent t u =
    assert (mem_vertex t u) ;
    match (Vertex.data u).Vertex.parent with
        Edge.Init _ -> None
      | Edge.Child (p, _) -> Some p

  let rec ancestor t u v =
    assert (mem_vertex t u) ;
    assert (mem_vertex t v) ;
    if Vertex.equal u v then
      true
    else
      if (Vertex.data v).Vertex.depth > (Vertex.data u).Vertex.depth then
        match (Vertex.data v).Vertex.parent with
            Edge.Init _ ->
              (* This should not happen since v has non-zero depth. *)
              raise InternalError
          | Edge.Child (p, _) ->
              ancestor t u p
      else
        false

  let add_vertex t loc reg m e =
    let d = match e with
        Edge.Init _ -> 0
      | Edge.Child (p, _) -> (Vertex.data p).Vertex.depth + 1
    in
    let u = Vertex.create t d loc reg m e
    in
      begin
        match e with
            Edge.Init _ ->
              t.init <- VSet.add t.init u
          | Edge.Child (p, _) ->
              let p_data = Vertex.data p
              in
                p_data.Vertex.children <- VSet.add p_data.Vertex.children u
      end ;
      t.lmap <- LMap.add t.lmap loc (VSet.add (vset_of_loc t loc) u) ;
      t.size <- t.size + 1 ;
      u

  let add_covering t v w =
    assert (mem_vertex t v) ;
    assert (mem_vertex t w) ;
    let v_data = Vertex.data v
    and w_data = Vertex.data w
    in
      v_data.Vertex.coverers <- VSet.add v_data.Vertex.coverers w ;
      w_data.Vertex.coverees <- VSet.add w_data.Vertex.coverees v

  let del_covering t v w =
    assert (mem_vertex t v) ;
    assert (mem_vertex t w) ;
    let v_data = Vertex.data v
    and w_data = Vertex.data w
    in
      v_data.Vertex.coverers <- VSet.remove v_data.Vertex.coverers w ;
      w_data.Vertex.coverees <- VSet.remove w_data.Vertex.coverees v

  let clr_coverees t w =
    assert (mem_vertex t w) ;
    let w_data = Vertex.data w
    in
      VSet.iter
        (fun v ->
          let v_data = Vertex.data v
          in
            v_data.Vertex.coverers <- VSet.remove v_data.Vertex.coverers w)
        w_data.Vertex.coverees ;
      w_data.Vertex.coverees <- VSet.empty

  let clr_coverers t v =
    assert (mem_vertex t v) ;
    let v_data = Vertex.data v
    in
      VSet.iter
        (fun w ->
          let w_data = Vertex.data w
          in
            w_data.Vertex.coverees <- VSet.remove w_data.Vertex.coverees v)
        v_data.Vertex.coverers ;
      v_data.Vertex.coverers <- VSet.empty

  let rec active t u =
    assert (mem_vertex t u) ;
    if VSet.is_empty (Vertex.data u).Vertex.coverers then
      match edge t u with
          Edge.Init _ -> true
        | Edge.Child (p, _) -> active t p
    else
      false

  let iter_children_real f t u =
    assert (mem_vertex t u) ;
    VSet.iter f (Vertex.data u).Vertex.children

  let fold_children_real f t u a =
    assert (mem_vertex t u) ;
    VSet.fold f (Vertex.data u).Vertex.children a

  let iter_children f t o =
    match o with
        None ->
          VSet.iter f t.init
      | Some u ->
          iter_children_real f t u

  let fold_children f t o a =
    match o with
        None ->
          VSet.fold f t.init a
      | Some u ->
          fold_children_real f t u a

  let rec iter_vertices_real pre post t u =
    assert (mem_vertex t u) ;
    pre u ;
    iter_children_real (iter_vertices_real pre post t) t u ;
    post u

  let rec fold_vertices_real pre post t u a =
    assert (mem_vertex t u) ;
    post u
      (fold_children_real
         (fold_vertices_real pre post t) t u
         (pre u a))

  let iter_vertices pre post t o =
    match o with
        None ->
          VSet.iter (iter_vertices_real pre post t) t.init
      | Some u ->
          iter_vertices_real pre post t u

  let fold_vertices pre post t o a =
    match o with
        None ->
          VSet.fold (fold_vertices_real pre post t) t.init a
      | Some u ->
          fold_vertices_real pre post t u a

  let iter_location f t loc =
    VSet.iter f (vset_of_loc t loc)

  let fold_location f t loc a =
    VSet.fold f (vset_of_loc t loc) a

  let iter_coverees f t u =
    assert (mem_vertex t u) ;
    VSet.iter f (Vertex.data u).Vertex.coverees

  let fold_coverees f t u a =
    assert (mem_vertex t u) ;
    VSet.fold f (Vertex.data u).Vertex.coverees a

  let iter_coverers f t u =
    assert (mem_vertex t u) ;
    VSet.iter f (Vertex.data u).Vertex.coverers

  let fold_coverers f t u a =
    assert (mem_vertex t u) ;
    VSet.fold f (Vertex.data u).Vertex.coverers a

  let del_vertex t u =
    assert (mem_vertex t u) ;
    begin
      match (Vertex.data u).Vertex.parent with
          Edge.Init _ ->
            t.init <- VSet.remove t.init u
        | Edge.Child (p, _) ->
            let p_data = Vertex.data p
            in
              p_data.Vertex.children <- VSet.remove p_data.Vertex.children u
    end ;
    iter_vertices_real
      (fun v ->
        clr_coverees t v ;
        clr_coverers t v)
      (fun v ->
        let loc = (Vertex.data v).Vertex.location
        in
          (Vertex.data v).Vertex.zombie <- true ;
          t.lmap <- LMap.add t.lmap loc (VSet.remove (vset_of_loc t loc) v) ;
          t.size <- t.size - 1)
      t u


  let print_edge_short fmt e =
    match e with
        Edge.Init init ->
          Format.fprintf fmt "Init@ %a" Region.print init
      | Edge.Child (p, trans) ->
          Format.fprintf fmt "%a" Transition.print trans

  let print_vertex_set fmt s =
    if VSet.is_empty s then
      Format.fprintf fmt "@ ∅"
    else
      VSet.iter (fun u -> Format.fprintf fmt "@ %d" (Vertex.id u)) s

  let print_vertex_data_full fmt u =
    let u_data = Vertex.data u
    in
      Format.fprintf fmt "{ @[" ;
      Format.fprintf fmt "@[<3>depth@ =@ %d@],@ " u_data.Vertex.depth ;
      Format.fprintf fmt "@[<3>location@ =@ @[%a@]@],@ " Location.print u_data.Vertex.location ;
      Format.fprintf fmt "@[<3>region@ =@ @[%a@]@],@ " Region.dump u_data.Vertex.region ;
      Format.fprintf fmt "@[<3>mark@ =@ @[%a@]@],@ " Mark.print u_data.Vertex.mark ;
      Format.fprintf fmt "@[<3>%a@],@ " Edge.print u_data.Vertex.parent ;
      Format.fprintf fmt "@[<3>children@ =%a@],@ " print_vertex_set u_data.Vertex.children ;
      Format.fprintf fmt "@[<3>coverees@ =%a@],@ " print_vertex_set u_data.Vertex.coverees ;
      Format.fprintf fmt "@[<3>coverers@ =%a@],@ " print_vertex_set u_data.Vertex.coverers ;
      Format.fprintf fmt "@[<3>zombie@ =@ %B@]" u_data.Vertex.zombie ;
      Format.fprintf fmt "@] }"

  let print_vertex_full fmt u =
    Format.fprintf fmt "[%d]@;@[%a@]" (Vertex.id u) print_vertex_data_full u

  let print_vertex_short fmt u =
    let u_data = Vertex.data u
    in
      Format.fprintf fmt "[%d]@;@[" (Vertex.id u) ;
      Format.fprintf fmt "@[%a@]@ ×@ @[%a@],@ @[%a@]"
        Location.print u_data.Vertex.location
        Region.print u_data.Vertex.region
        Mark.print u_data.Vertex.mark ;
      if not (VSet.is_empty u_data.Vertex.coverees) then
        Format.fprintf fmt ",@ %d@ coverees" (VSet.cardinal u_data.Vertex.coverees) ;
      if not (VSet.is_empty u_data.Vertex.coverers) then
        Format.fprintf fmt ",@ %d@ coverers" (VSet.cardinal u_data.Vertex.coverers) ;
      Format.fprintf fmt ",@ %s"
        (if active u_data.Vertex.tree u then "active" else "inactive") ;
      Format.fprintf fmt "@]"

  let print fmt t =
    Format.fprintf fmt "@[<v>" ;
    Format.fprintf fmt "@[<v 3>%d vertices:" (nb_vertices t) ;
    iter_vertices
      (Format.fprintf fmt "@,@[<h>%a@]" print_vertex_full)
      (fun _ -> ())
      t None ;
    Format.fprintf fmt "@]@," ;
    Format.fprintf fmt "@[<v 3>Tree structure:" ;
    iter_vertices
      (fun u ->
        Format.fprintf fmt "@,@[<v 3>@[%a@]"
          print_edge_short (edge t u) ;
        Format.fprintf fmt "@,@[<v 3>@[<h>%a@]"
          print_vertex_short u)
      (fun _ ->
        Format.fprintf fmt "@]@]")
      t None ;
    Format.fprintf fmt "@]@]"
end
