(* $Id: ReachabilityTree.mli 1183 2011-10-13 22:02:51Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * Reachability Trees.
 *
 * This module provides an implementation for abstract reachability trees.  Such
 * trees are at the core of {e lazy abstraction refinement} approaches (some
 * references are given below).  The functor implemented in this module takes as
 * arguments a model (see the {! Model} module) and a printable type of marks.
 *
 * A {e reachability tree} is a directed rooted tree where each {e vertex} is
 * labeled by a location and by a region, and each {e edge} is labeled by a
 * transition of the model's automaton.  As an exception to the latter rule, the
 * edges from the root to its children are labeled by initial regions.  Labels
 * of edges remain constant after creation, and the same applies to vertex
 * locations.  However, the regions labeling vertices may evolve, typically due
 * to refinement.
 *
 * The root of a reachability tree is a {e virtual} vertex, that, essentially,
 * binds the subtrees associated with each initial symbolic state into a single
 * tree.  In particular, the root has no associated location nor region.  In the
 * following, all considered vertices are distinct from the root (i.e., the root
 * is not a vertex).
 *
 * To be useful, a reachability tree usually must respect some “correctness”
 * conditions, such as:
 * {ul
 * {li for every edge {i (u, v)} labeled by a transition {i t}, the location of
 *     {i u} is the source of {i t}, the location of {i v} is the target of
 *     {i t}, and the region of {i v} contains the post image of the region of
 *     {i u} w.r.t. the command of {i t}.}
 * {li for each vertex {i u} that is not a leaf (i.e., {i u} has a child), there
 *     exists, for each transition {i t} with source equal to the location of
 *     {i u}, a child {i v} of u such that the edge {i (u, v)} is labeled by
 *     {i t}.}
 * }
 *
 * {b Remark.} This module does not enforce any “correctness” condition on the
 * reachability trees.
 *
 * Termination of tree-based algorithms often rely on cover properties.  To this
 * end, a covering relation is associated to each tree.  More details are given
 * below regarding covering relations.  In addition, vertices are decored with
 * additional attributes that are typically needed by tree-based model-checking
 * algorithms.
 *
 * @see <http://dx.doi.org/10.1145/503272.503279> Lazy Abstraction
 * @see <http://dx.doi.org/10.1007/11817963_14> Lazy Abstraction with Interpolants
 *)

(**
 * {2 Covering Relations}
 *
 * The vertex covering information is given by a relation, written ◅, that is
 * interpreted as follows: {i v ◅ w} means “{i v} is covered by {i w}” or,
 * equivalently, “{i w} covers {i v}”.
 *
 * A pair {i (v, w)} in ◅ is called a {e covering pair}.  When {i v ◅ w}, we say
 * that {i v} is a {e coveree} of {i w}, and that {i w} is a {e coverer} of
 * {i v}.  A vertex {i v} is called {e covered} when it has a {e coverer} (i.e.,
 * there exists w such that v ◅ w).
 *
 * In the reachability tree constructed by a model-checking algorithm, vertices
 * that are covered should be “discarded”, and in particular they should not be
 * expanded (since their descendants shall be explored from their coverers).
 * However, due to refinement, a vertex may become covered {e after} expansion
 * of (part of) its subtree.  In such a case, descendants of the covered vertex
 * should, essentially, inherit this coveredness.  To prevent confusion between
 * direct covering and inherited covering, a different term is used here for the
 * inherited version.
 *
 * Recall that an ancestor of a vertex {i u} is any vertex on the branch from
 * the root to {i u} (in particular, {i u} is an ancestor of {i u}).
 * Given a vertex {i u}, the {e subtree rooted in {i u}} contains all vertices
 * for which {i u} is an ancestor.
 *
 * A vertex {i u} is called {e active} when none of its ancestors is covered.
 * Conversely, {i u} is called {e inactive} when it has an ancestor that is
 * covered.  Observe that the set of active vertices forms a prefix of the tree.
 *
 * All of the above only introduced vocabulary associated with a given covering
 * relation, which is, a priori, an arbitrary binary relation over vertices.
 * But as for reachability trees, to be useful, a covering relation usually must
 * respect some “correctness” conditions, such as:
 * {ul
 * {li if {i v ◅ w}, then {i v} and {i w} have the same location, and the region
 *     of {i v} is contained in the region of {i w}.}
 * {li if {i u} is an ancestor of {i v}, then {i u} cannot be covered by {i v}.}
 * {li if {i w} covers some vertex, then {i w} is active.}
 * }
 *
 * {b Remark.} This module does not enforce any “correctness” condition on the
 * covering relations.
 *)


open Util

module Make (A : Model.S) (Mark : Print.PRINTABLE_TYPE) :
sig
  (** {2 Reachability Tree Structure} *) (*************************************)

  (**
   * The printable type of reachability trees.
   *)
  include Print.PRINTABLE_TYPE

  (**
   * Module to manage vertices.  Each vertex may belong to at most one tree.
   *)
  module Vertex :
  sig
    (**
     * The printable type of vertices.
     *)
    include Print.PRINTABLE_TYPE

    (**
     * [location u] returns the location of the vertex [u].
     *)
    val location : t -> A.Automaton.Location.t

    (**
     * [region u] returns the region of the vertex [u].
     *)
    val region : t -> A.Semantics.Region.t

    (**
     * [mark u] returns the mark of the vertex [u].
     *)
    val mark : t -> Mark.t

    (**
     * [refine u reg] replaces the region of [u] by [reg].
     * {L {b Precondition} [reg] is contained in the region of [u].}
     *)
    val refine : t -> A.Semantics.Region.t -> unit

    (**
     * Vertices are ordered (see [Set.OrderedType] or [Map.OrderedType]) and
     * hashable (see [Hashtbl.HashedType]).
     *)

    (**
     * Total ordering over vertices (creation order).
     *)
    val compare : t -> t -> int

    (**
     * Equality predicate on vertices.
     *)
    val equal : t -> t -> bool

    (**
     * Hashing function on vertices.
     *)
    val hash : t -> int
  end

  (**
   * The printable type of possible inbound edges for a vertex.
   *)
  module Edge :
  sig
    type t =
        Init of A.Semantics.Region.t
      | Child of (Vertex.t * A.Automaton.Transition.t)

    include Print.PRINTABLE_TYPE with type t := t
  end

  (**
   * [edge t u] returns the inbound edge of the vertex [u] in the tree [t].
   * {L {b Precondition} [u] is contained in [t].}
   *)
  val edge : t -> Vertex.t -> Edge.t

  (**
   * [parent t u] returns the parent of the vertex [u] in the tree [t] (if any).
   * {L {b Precondition} [u] is contained in [t].}
   *)
  val parent : t -> Vertex.t -> Vertex.t option

  (**
   * [ancestor t u v] is [true] iff [u] is an ancestor of [v] in the tree [t].
   * {L {b Precondition} [u] and [v] are contained in [t].}
   *)
  val ancestor : t -> Vertex.t -> Vertex.t -> bool

  (**
   * [create ()] creates a new, empty reachability tree.  The tree initially
   * contains no vertex (it only has a virtual root).
   *)
  val create : unit -> t

  (**
   * [add_vertex t loc reg m e] creates a new vertex [u] with location [loc],
   * region [reg] and mark [m], and adds [u] to the tree [t], with inbound edge
   * [e].  If the edge [e] is an initial region, then [u] is added as a child of
   * the root.  Returns this new vertex [u].
   * {L {b Precondition} If the inbound edge [e] is [Child (p, tra)] then [p] is
   *    contained in [t], [tra]'s source is equal to [p]'s location, and [tra]'s
   *    target is equal to [loc].}
   *)
  val add_vertex : t -> A.Automaton.Location.t -> A.Semantics.Region.t ->
    Mark.t -> Edge.t -> Vertex.t

  (**
   * [del_vertex t u] removes the vertex [u], as well as its subtree, from the
   * tree [t].  All covering pairs involving a vertex in the subtree rooted in
   * [u] are removed from the covering relation of [t].  Deletion of a vertex is
   * permanent (i.e., vertices cannot be undeleted).
   * {L {b Precondition} [u] is contained in [t].}
   *)
  val del_vertex : t -> Vertex.t -> unit

  (**
   * [mem_vertex t u] tests whether the tree [t] contains the vertex [u].  Note
   * that each vertex may belong to at most one tree.
   *)
  val mem_vertex : t -> Vertex.t -> bool

  (**
   * [depth t u] returns the depth of [u] in the tree [t].
   * {L {b Precondition} [u] is contained in [t].}
   *)
  val depth : t -> Vertex.t -> int

  (**
   * [has_children t u] returns [true] iff the vertex [u] has at least one child
   * in the tree [t].
   * {L {b Precondition} [u] is contained in [t].}
   *)
  val has_children : t -> Vertex.t -> bool

  (** {2 Size Functions} *) (**************************************************)

  (**
   * [nb_vertices t] returns the number of vertices in the tree [t].
   *)
  val nb_vertices : t -> int

  (** {2 Covering Structure} *) (**********************************************)

  (**
   * The following functions manage covering relations.  Each function has an
   * implicit {b precondition} over its arguments requiring that the given
   * reachability tree contains the given vertices.
   *)

  (**
   * [add_covering t v w] adds the covering pair [v] ◅ [w] to the tree [t].
   *)
  val add_covering : t -> Vertex.t -> Vertex.t -> unit

  (**
   * [del_covering t v w] removes the covering pair [v] ◅ [w] from the tree [t].
   *)
  val del_covering : t -> Vertex.t -> Vertex.t -> unit

  (**
   * [clr_coverees t w] removes all covering pairs · ◅ [w] from the tree [t].
   *)
  val clr_coverees : t -> Vertex.t -> unit

  (**
   * [clr_coverers t v] removes all covering pairs [v] ◅ · from the tree [t].
   *)
  val clr_coverers : t -> Vertex.t -> unit

  (**
   * [has_coverees t w] returns [true] iff there exists a covering pair · ◅ [w]
   * in the tree [t].
   *)
  val has_coverees : t -> Vertex.t -> bool

  (**
   * [has_coverers t v] returns [true] iff there exists a covering pair [v] ◅ ·
   * in the tree [t].
   *)
  val has_coverers : t -> Vertex.t -> bool

  (**
   * [active t u] returns [true] iff the vertex [u] is active in the tree [t].
   *)
  val active : t -> Vertex.t -> bool

  (** {2 Iterators} *)  (******************************************************)

  (**
   * The following iterators apply a given function [f] to some vertices of a
   * given reachability tree.  Recall that the root is {e not} considered to be
   * a vertex.  Each vertex is passed at most once to the function [f].
   * For added flexibility, tree traversal iterators apply two functions instead
   * of a single one: a {e pre-order} operator and a {e post-order} operator.
   *)

  (** {4 Global Iterators} *) (*----------------------------------------------*)

  (**
   * [iter_vertices pre post t u] traverses the subtree of [t] rooted in [u],
   * applying the pre-order operation [pre] and the post-order operation [post]
   * to all visited vertices.  More precisely, [iter_vertices pre post t u] is
   * equivalent to [begin
   *   pre u ;
   *   iter_children (iter_vertices pre post t) t u ;
   *   post u
   * end].
   * If [u] is [None], then [u] stands for the (virtual) root of the tree.
   *)
  val iter_vertices : (Vertex.t -> unit) -> (Vertex.t -> unit) -> t ->
    Vertex.t option -> unit

  (**
   * [fold_vertices pre post t u a] is similar to [iter_vertices pre post t u],
   * but returns a value.  More precisely, [fold_vertices pre post t u a] is
   * equivalent to [begin
   *   post u
   *     (fold_children
   *        (fold_vertices pre post t) t u
   *        (pre u a))
   * end].
   * If [u] is [None], then [u] stands for the (virtual) root of the tree.
   *)
  val fold_vertices : (Vertex.t -> 'a -> 'a) -> (Vertex.t -> 'a -> 'a) -> t ->
    Vertex.t option -> 'a -> 'a

  (**
   * [iter_location f t loc] applies [f] in turn to all vertices of [t] with
   * location [loc], in increasing order.
   *)
  val iter_location : (Vertex.t -> unit) -> t -> A.Automaton.Location.t -> unit

  (**
   * [fold_location f t loc a] returns [(f uN ... (f u2 (f u1 a))...)], where
   * [u1, ..., uN] are the vertices of [t] with location [loc], in increasing
   * order.
   *)
  val fold_location : (Vertex.t -> 'a -> 'a) -> t -> A.Automaton.Location.t ->
    'a -> 'a

  (** {4 Local Iterators} *) (*-----------------------------------------------*)

  (**
   * [iter_children f t u] applies [f] in turn to all children of [u] in [t], in
   * increasing order.
   * If [u] is [None], then [u] stands for the (virtual) root of the tree.
   * {L {b Precondition} [u] is a vertex contained in [t].}
   *)
  val iter_children : (Vertex.t -> unit) -> t -> Vertex.t option -> unit

  (**
   * [fold_children f t u a] returns [(f vN ... (f v2 (f v1 a))...)], where [v1,
   * ..., vN] are the children of [u] in [t], in increasing order.
   * If [u] is [None], then [u] stands for the (virtual) root of the tree.
   * {L {b Precondition} [u] is a vertex contained in [t].}
   *)
  val fold_children : (Vertex.t -> 'a -> 'a) -> t -> Vertex.t option -> 'a -> 'a

  (** {4 Covering Iterators} *) (*--------------------------------------------*)

  (**
   * [iter_coverees f t u] applies [f] in turn to all coverees of [u] in [t], in
   * increasing order.
   * {L {b Precondition} [u] is a vertex contained in [t].}
   *)
  val iter_coverees : (Vertex.t -> unit) -> t -> Vertex.t -> unit

  (**
   * [fold_coverees f t u a] returns [(f vN ... (f v2 (f v1 a))...)], where [v1,
   * ..., vN] are the coverees of [u] in [t], in increasing order.
   * {L {b Precondition} [u] is a vertex contained in [t].}
   *)
  val fold_coverees : (Vertex.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a

  (**
   * [iter_coverers f t u] applies [f] in turn to all coverers of [u] in [t], in
   * increasing order.
   * {L {b Precondition} [u] is a vertex contained in [t].}
   *)
  val iter_coverers : (Vertex.t -> unit) -> t -> Vertex.t -> unit

  (**
   * [fold_coverers f t u a] returns [(f vN ... (f v2 (f v1 a))...)], where [v1,
   * ..., vN] are the coverers of [u] in [t], in increasing order.
   * {L {b Precondition} [u] is a vertex contained in [t].}
   *)
  val fold_coverers : (Vertex.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a
end
