(* $Id: ScmWrapper.ml 1116 2011-08-11 08:36:09Z gsutre $ *)

(*
 * Copyright (c) 2009, 2010 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Scm_syn
open Util

module Option = Option.ScmWrapper


(*
 * Important note on lossy channels.
 *
 * This module implements the lossy channel semantics by losses on emissions.
 * In other words, when a channel is lossy, every send transition into the
 * channel has an implicit copy that acts as an epsilon transition with respect
 * to the channel.  This ``simulation'' of the classical lossy semantics is
 * correct for reachability since:
 *
 *      (q_0, empty, ..., empty)       (-- lossy -->)^*        config
 *
 *                             if and only if
 *
 *      (q_0, empty, ..., empty)  (-- loss on emission -->)^*  config
 *
 * Correctness relies on the property that all initial configurations of an scm
 * model have empty channels (see the definition of init_region).
 *)


(* Exception for errors that are internal to this module, such as cases in
   in pattern matching that should not arise. *)
exception InternalError


(* Commands in systems of communicating machines.  This signature provides the
   interface of the module ScmCommand defined below. *)
module type SCM_COMMAND =
sig
  type t

  (* Helper types for components of a command. *)
  type guard = Scm_syn.bexpr
  type action = (int * Scm_syn.taction * Scm_syn.var) option
  type assignment = (Scm_syn.var * Scm_syn.iexpr)

  (* Constructor to extract a command from an scm transition. *)
  val extract : Scm_syn.transition -> t

  (* Accessors to get the individual components of a command. *)
  val guard : t -> guard
  val action : t -> action
  val assignments : t -> assignment list

  (* Helper functions for pretty-printing. *)
  val print_guard : Format.formatter -> guard -> unit
  val print_action : Format.formatter -> action -> unit
  val print_assignment : Format.formatter -> assignment -> unit

  (* Pretty-printer. *)
  val print : Format.formatter -> t -> unit

  (* Access to the action. *)
  type t_command = Internal | Input | Output

  val give_type : t -> t_command
  val give_queue : t -> int
  val give_message : t -> string
end

(* Commands in systems of communicating machines.  This module provides the
   implementation of the signature SCM_COMMAND defined above. *)
module ScmCommand : SCM_COMMAND =
struct
  (* Helper types for components of a command. *)
  type guard = Scm_syn.bexpr
  type action = (int * Scm_syn.taction * Scm_syn.var) option
  type assignment = (Scm_syn.var * Scm_syn.iexpr)

  (* Definition of the type t of commands. *)
  type t = {
    guard : guard ;
    action : action ;
    assignments : assignment list ;
  }

  type t_command = Internal | Input | Output

  let give_type c =
    match c.action with
        None -> Internal
      | Some (_,Scm_syn.SEND,_) -> Output
      | Some (_,Scm_syn.RECEIVE,_) -> Input

  let give_queue c =
    match c.action with
        None ->
          failwith "Command.give_queue is not defined with internal commands"
      | Some (x,_,_) -> x

  let give_message c =
    match c.action with
        None ->
          failwith "Command.give_message is not defined with internal commands"
      | Some (_,_,x) -> x

  (* Constructor to extract a command from an scm transition. *)
  let extract trans = { guard = trans.Scm_syn.guard ;
                        action = trans.Scm_syn.action ;
                        assignments = trans.Scm_syn.assignement }

  (* Accessors to get the individual components of a command. *)
  let guard       cmd = cmd.guard
  let action      cmd = cmd.action
  let assignments cmd = cmd.assignments

  (* Helper functions for pretty-printing. *)
  let print_guard = Scm_print.print_bexpr
  let print_action = Scm_print.print_action
  let print_assignment fmt (v, e) =
    Format.fprintf fmt "%s@ :=@ @[%a@]" v Scm_print.print_iexpr e

  (* Pretty-printer. *)
  let print fmt cmd =
    if not (cmd.guard = TRUE) then
      Format.fprintf fmt "@[%a@]@ "
        print_guard cmd.guard ;
    Format.fprintf fmt "|-@;@[%a@]@;-|"
      print_action cmd.action ;
    if not (cmd.assignments = []) then
      Format.fprintf fmt "@ @[%a@]"
        (Print.list_printer_from_printer "@ ;@ " print_assignment) cmd.assignments
end


(* Functor providing an implementation of the Model.AUTOMATON interface for a
   given global SCM. *)
module MakeAutomaton (A : sig val scm : Scm_syn.global_scm end) =
struct

  exception InconsistentGlobalScm of string

  type location_attr = {
    name : string ;
  }

  type transition_attr = {
    owner : string ;
    controllable : bool ;
    command : ScmCommand.t ;
  }

  module G = MultiGraph.Persistent.Make
    (struct
       type t = location_attr
       let print fmt attr =
         Format.fprintf fmt "%s" attr.name
     end)
    (struct
       type t = transition_attr
       let print fmt attr =
         ScmCommand.print fmt attr.command
     end)

  (* Translation of the scm data structure into a graph.  Note that the lists
     init_locations and error_locations respectively respect the order of the
     lists gb_init_states and gb_bad_states in A.scm.  All states of the scm are
     explicitly added to the graph, with add_vertex and not only with add_edge,
     as otherwise isolated scm states, which could be initial or bad states,
     would not belong to the set of vertices of the graph. *)
  let (control_flow_graph, init_locations, error_locations) =
    (* Hash table to map scm state ids to vertices. *)
    let vertex_table = Hashtbl.create (List.length A.scm.gb_lstates) in
    (* Function to create vertices and fill vertex table from the list of all
       scm states.  This function returns a new empty graph will all vertices
       added to it. *)
    let create_vertices_and_fill_vertex_table () =
      List.fold_left
        (fun g s ->
           if Hashtbl.mem vertex_table s.id_state
           then
             raise (InconsistentGlobalScm
                      ("Duplicate state in scm.gb_lstates: " ^ s.id_state))
           else
             let new_vertex = G.Vertex.create { name = s.id_state }
             in
               Hashtbl.add vertex_table s.id_state new_vertex ;
               G.add_vertex g new_vertex)
        G.empty
        A.scm.gb_lstates
    (* Function to find the vertex corresponding to a given scm state id. *)
    and find_vertex id =
      try Hashtbl.find vertex_table id
      with Not_found ->
        raise (InconsistentGlobalScm ("Missing state in scm.gb_lstates: " ^ id))
    in
    (* Function adding to a given graph all outbound transitions listed in a
       given scm state. *)
    let add_outbound_transitions g s =
      let src = find_vertex s.id_state
      in
      let add_transition g t =
        let tgt =
          try Hashtbl.find vertex_table t.new_state
          with Not_found ->
            raise (InconsistentGlobalScm
                     ("Missing state in scm.gb_lstates: " ^ s.id_state))
        in
        let new_edge =
          G.Edge.create
            src
            { owner = t.Scm_syn.owner ;
              controllable = t.Scm_syn.controllable ;
              command = ScmCommand.extract t }
            tgt
        in
          G.add_edge g new_edge
      in
        List.fold_left add_transition g s.ltrans
    in
    let graph = create_vertices_and_fill_vertex_table ()
    in
      (List.fold_left add_outbound_transitions graph A.scm.gb_lstates,
       List.map find_vertex A.scm.gb_init_states,
       List.map (fun (s, _, _) -> find_vertex s) A.scm.gb_bad_states)

  let owners =
    G.fold_edges
      (fun e s -> Sette.add (G.Edge.lbl e).owner s)
      control_flow_graph
      Sette.empty

  (* Implementation of the signature Model.AUTOMATON. *)
  module Owner =
  struct
    type t = string
    let print = Format.pp_print_string
    let compare = Pervasives.compare
    let equal = Pervasives.(=)
    let hash = Hashtbl.hash
  end

  module Location =
  struct
    include G.Vertex

    let print fmt l =
      Format.fprintf fmt "%s" (G.Vertex.lbl l).name
  end

  module Command = ScmCommand

  module Transition =
  struct
    include G.Edge

    let source = src
    let target = tgt
    let command e = (lbl e).command
    let owner e = (lbl e).owner
    let is_controllable e = (lbl e).controllable

    let print fmt e =
      Format.fprintf fmt "%a@ »@;@[%a@]@;»@ %a@ [owner=%s,%s]"
        Location.print (source e)
        Command.print  (command e)
        Location.print (target e)
        (owner e)
        (if is_controllable e then "C" else "U")
  end

  let nb_locations   = G.nb_vertices control_flow_graph
  let nb_transitions = G.nb_edges    control_flow_graph
  let nb_owners      = Sette.cardinal owners

  let iter_locations   f = G.iter_vertices  f control_flow_graph
  let iter_transitions f = G.iter_edges f control_flow_graph
  let iter_owners      f = Sette.iter f owners

  let fold_locations   f = G.fold_vertices  f control_flow_graph
  let fold_transitions f = G.fold_edges f control_flow_graph
  let fold_owners      f = Sette.fold f owners

  let iter_out_transitions f = G.iter_succ f control_flow_graph
  let iter_in_transitions  f = G.iter_pred f control_flow_graph

  let fold_out_transitions f = G.fold_succ f control_flow_graph
  let fold_in_transitions  f = G.fold_pred f control_flow_graph

  let print_automaton fmt () =
    let print_out_transition fmt t =
      Format.fprintf fmt "@[%a@]@;»@ %a"
        Command.print  (Transition.command t)
        Location.print (Transition.target t)
    in
      Format.fprintf fmt "@[<v>" ;
      Format.fprintf fmt "@[<v 3>All %d locations:" nb_locations ;
      iter_locations (Format.fprintf fmt "@,%a" Location.print) ;
      Format.fprintf fmt "@]@," ;
      Format.fprintf fmt "@[<3>Initial locations:@ %a@]@,"
        (Print.list_printer_from_printer ",@ " Location.print) init_locations ;
      Format.fprintf fmt "@[<3>Error locations:@ %a@]@,"
        (Print.list_printer_from_printer ",@ " Location.print) error_locations ;
      Format.fprintf fmt "@[<v 3>All %d transitions:" nb_transitions ;
      iter_transitions (Format.fprintf fmt "@,@[<3>%a@]" Transition.print) ;
      Format.fprintf fmt "@]@," ;
      Format.fprintf fmt "@[<v 3>Outbound transitions of each location:" ;
      iter_locations
        (fun l ->
           Format.fprintf fmt "@,@[<v 3>%a :" Location.print l ;
           iter_out_transitions
             (Format.fprintf fmt "@,» @[<3>%a@]" print_out_transition) l ;
           Format.fprintf fmt "@]") ;
      Format.fprintf fmt "@]@]"
end


(* Functor providing an implementation of the Model.SEMANTICS interface for a
   given global SCM. *)
module MakeSemantics (A : sig val scm : Scm_syn.global_scm end) =
struct

  (* Alphabet module required by StandardSem.Make functor. *)
  module Alpha =
  struct
    type t = string
    let compare = compare
    let from_string x = x
    let print = Format.pp_print_string
  end

  module Sem = StandardSem.Make (Alpha)

  (* We abstract away all numerical information for the time being, i.e. the
     abstract numerical lattice is the trivial Bidon lattice with two elements.
     This is explicited in the following line as well as in the definition of
     the Region.t type.  However, we do not rely on this assumption in the rest
     of the code, hence it might still work with a real numerical lattice. *)
  let manager = Scm_manager.finite_manager () A.scm
  let partition = Sem.create_partition manager
  let lattice = Sem.create_lattice manager

  (* Implementation of the signature Model.SEMANTICS. *)
  module Region =
  struct
    type t = Scm_manager.bidon Sem.t

    let bot = Sem.bottom manager partition lattice
    let top = Sem.top manager partition lattice

    let print fmt r =
      if r == bot
      then (Format.pp_print_string fmt "⊥")
      else
        if r == top
        then (Format.pp_print_string fmt "⊤")
        else
          Sem.print_regexp manager fmt r

    let dump = Sem.print manager

    let leq r1 r2 =
      if (r1 == bot) || (r2 == top)
      then true
      else
        if r2 == bot
        then Sem.is_bottom manager r1
        else Sem.is_leq manager r1 r2

    let eq r1 r2 =
      (leq r1 r2) && (leq r2 r1)

    let union r1 r2 =
      if (r1 == top || r2 == top)
      then top
      else
        if r1 == bot
        then r2
        else
          if r2 == bot
          then r1
          else
            Sem.join manager r1 r2

    let inter r1 r2 =
      if (r1 == bot || r2 == bot)
      then bot
      else
        if r1 == top
        then r2
        else
          if r2 == top
          then r1
          else
            Sem.meet manager r1 r2

    let compl r =
      if r == bot
      then top
      else
        if r == top
        then bot
        else
          Sem.complementary manager r

    let minus r1 r2 =
      inter r1 (compl r2)

    let simplify = Sem.minimise manager

    let size = Sem.get_automata_size
  end

  module Command = ScmCommand

  let init_region =
    let scm_decl_list_to_init_list dl =
      List.fold_right
        (fun (v, _, e) il ->
           match e with
               Some ie -> (v, ie)::il
             | None -> il)
        dl []
    in
      (* Creates an abstract value with all queues being empty. *)
      Sem.init_abstr manager partition lattice
        (scm_decl_list_to_init_list A.scm.gb_vars)

  (* The following list of error regions is ordered exactly as the respective
     list gb_bad_states in A.scm. *)
  let error_regions =
    List.map
      (fun (_, b, r) -> Sem.init_bad_state manager partition lattice b r)
      A.scm.gb_bad_states

  (* The list of lossy channels is transformed into the array channels of
     booleans: the channels are 0, ..., (Array.length channels - 1) and for each
     channel c, channels.(c) is true iff c is lossy. *)
  let channels =
    match Option.channel_policy () with
        Option.Normal ->
          let a = Array.make A.scm.Scm_syn.gb_nb_channels false
          in
            List.iter (fun c -> a.(c) <- true) A.scm.gb_lossy_list ;
            a
      | Option.Lossy   -> Array.make A.scm.Scm_syn.gb_nb_channels true
      | Option.Perfect -> Array.make A.scm.Scm_syn.gb_nb_channels false

  let post r cmd =
    let guard = ScmCommand.guard cmd
    and action = ScmCommand.action cmd
    and assign = ScmCommand.assignments cmd
    in
      match action with
          None -> Sem.epsilon_trans manager guard assign r
        | Some (n, a, p) ->
            match a with
                SEND ->
                  if channels.(n) then
                    Region.union
                      (Sem.epsilon_trans manager guard assign r)
                      (Sem.output_trans manager guard p n assign r)
                  else
                    Sem.output_trans manager guard p n assign r
              | RECEIVE ->
                  Sem.input_trans  manager guard p n assign r

  let pre r cmd =
    let guard = ScmCommand.guard cmd
    and action = ScmCommand.action cmd
    and assign = ScmCommand.assignments cmd
    in
      match action with
          None -> Sem.pre_epsilon_trans manager guard assign r
        | Some (n, a, p) ->
            match a with
                SEND ->
                  if channels.(n) then
                    Region.union
                      (Sem.pre_epsilon_trans manager guard assign r)
                      (Sem.pre_output_trans manager guard p n assign r)
                  else
                    Sem.pre_output_trans manager guard p n assign r
              | RECEIVE ->
                  Sem.pre_input_trans  manager guard p n assign r

  let rel cmd r1 r2 =
    match Option.rel_cmd_impl () with
        Option.Post ->
          (* Use post only *)
          not (Region.leq (Region.inter (post r1 cmd) r2) Region.bot)
      | Option.Pre ->
          (* Use pre only *)
          not (Region.leq (Region.inter (pre r2 cmd) r1) Region.bot)
      | Option.Mixed ->
          (* Use post for send actions and pre for receive actions *)
          match ScmCommand.action cmd with
              None
            | Some (_, SEND, _) ->
                not (Region.leq (Region.inter (post r1 cmd) r2) Region.bot)
            | Some (_, RECEIVE, _) ->
                not (Region.leq (Region.inter (pre r2 cmd) r1) Region.bot)

  (* Select SCM widening variant according to user specified options *)
  let scm_extrapolation =
    let (auto_col_i, auto_col_f, check_col) =
      match Option.extrapolation () with
          Option.Bisim_fwd  | Option.Lang_fwd  -> (false, true, fun (i, f) -> f)
        | Option.Bisim_bwd  | Option.Lang_bwd  -> (true,  false, fun (i, f) -> i)
        | Option.Bisim_both | Option.Lang_both -> (true,  true, fun (i, f) -> i || f)
        | Option.Identity                      -> (false, false, fun (i, f) -> true) in
    (* Coloring of initial and final states *)
    let (col_i, col_f) =
      match Option.extrapol_coloring () with
          Option.Both  -> (true,  true)
        | Option.Final -> (false, true)
        | Option.Init  -> (true,  false)
        | Option.None  -> (false, false)
        | Option.Auto  -> (auto_col_i, auto_col_f)
    in
      if not (check_col (col_i, col_f)) then
        Logger.warning "@[WARNING: State coloring does not guarantee extrapolation convergence to identity.@]@." ;
      match Option.extrapolation () with
          Option.Bisim_fwd  -> (fun k r -> Sem.extrapolation ~initial:col_i ~final:col_f ~k_forward:k ~k_backward:0 r)
        | Option.Bisim_bwd  -> (fun k r -> Sem.extrapolation ~initial:col_i ~final:col_f ~k_forward:0 ~k_backward:k r)
        | Option.Bisim_both -> (fun k r -> Sem.extrapolation ~initial:col_i ~final:col_f ~k_forward:k ~k_backward:k r)
        | Option.Lang_fwd   -> (fun k r -> Sem.extrapolation_alt ~initial:col_i ~final:col_f ~k_forward:k ~k_backward:0 r)
        | Option.Lang_bwd   -> (fun k r -> Sem.extrapolation_alt ~initial:col_i ~final:col_f ~k_forward:0 ~k_backward:k r)
        | Option.Lang_both  -> (fun k r -> Sem.extrapolation_alt ~initial:col_i ~final:col_f ~k_forward:k ~k_backward:k r)
        | Option.Identity   -> (fun k r -> r)

  (* Helper function for the implementation of the signature Model.CLOSURE. *)
  let widening_to_clo k r =
    if k < 0 then
      Sem.extrapolation ~initial:false ~final:false ~k_forward:0 ~k_backward:0 r
    else
      scm_extrapolation k r
end


(* Main functor of this module, providing an implementation of the Model.S
   interface for a given global SCM. *)
module Make (A : sig val scm : Scm_syn.global_scm end) =
struct
  module Command = ScmCommand
  module Automaton = MakeAutomaton (A)
  module Semantics = MakeSemantics (A)
  module Closure =
  struct
    module Region = Semantics.Region
    let clo = Semantics.widening_to_clo
  end

  let name =        A.scm.Scm_syn.gb_id_scm
  let variables =   A.scm.Scm_syn.gb_vars
  let parameters =  A.scm.Scm_syn.gb_params

  (* Data-type for symbolic states. *)
  module State =
  struct
    type t = Automaton.Location.t * Semantics.Region.t

    let print fmt (l, r) =
      Format.fprintf fmt "%a@ |->@;@[%a@]"
        Automaton.Location.print l
        Semantics.Region.print r
  end

  let nb_init  = List.length Automaton.init_locations
  let nb_error = List.length Automaton.error_locations

  (* Iterators over initial symbolic states and error symbolic states. *)
  let iter_init f =
    List.iter (fun l -> f (l, Semantics.init_region))
      Automaton.init_locations
  let fold_init f a =
    List.fold_left (fun x l -> f (l, Semantics.init_region) x) a
      Automaton.init_locations
  let iter_error f =
    List.iter2 (fun l r -> f (l, r))
      Automaton.error_locations
      Semantics.error_regions
  let fold_error f a =
    List.fold_left2 (fun x l r -> f (l, r) x) a
      Automaton.error_locations
      Semantics.error_regions

  let print_model fmt () =
    let print_channel fmt (c, b) =
      Format.fprintf fmt "%d" c ;
      if b then Format.fprintf fmt "↓"
    and print_scm_decl fmt (v, _, e) =
      match e with
          None -> Format.pp_print_string fmt v
        | Some ie -> ScmCommand.print_assignment fmt (v, ie)
    in
      Format.fprintf fmt "@[<v>" ;
      Format.fprintf fmt "@[<3>Model:@ Global SCM with name:@ %s@]@," name ;
      Format.fprintf fmt "@[<v 3>Channels:@,@[%a@]@]@,"
        (Print.array_printer_from_printer ",@ " print_channel)
        Semantics.channels ;
      Format.fprintf fmt "@[<v 3>Variables:@,@[%a@]@]@,"
        (Print.list_printer_from_printer ",@ " print_scm_decl) variables ;
      Format.fprintf fmt "@[<v 3>Parameters:@,@[%a@]@]@,"
        (Print.list_printer_from_printer ",@ " print_scm_decl) parameters ;
      Format.fprintf fmt "@[<v 3>Automaton:@,@[%a@]@]@,"
        Automaton.print_automaton () ;
      Format.fprintf fmt "@[<v 3>Initial symbolic states:" ;
      iter_init (Format.fprintf fmt "@,@[<3>%a@]" State.print) ;
      Format.fprintf fmt "@]@," ;
      Format.fprintf fmt "@[<v 3>Error symbolic states:" ;
      iter_error (Format.fprintf fmt "@,@[<3>%a@]" State.print) ;
      Format.fprintf fmt "@]" ;
      Format.fprintf fmt "@]"
end

(* Module providing functions to fix the order of lists after SCM parsing. *)
module OrderFix =
struct
  let fix_transition t =
    { owner = t.owner ;
      controllable = t.controllable ;
      new_state = t.new_state ;
      guard = t.guard ;
      action = t.action ;
      assignement = List.rev t.assignement }

  let fix_state s =
    { id_state = s.id_state ;
      ltrans = List.fold_left (fun l t -> (fix_transition t)::l) [] s.ltrans }

  let fix_automaton a =
    { id_auto = a.id_auto ;
      ldec = List.rev a.ldec ;
      init_states = List.rev a.init_states ;
      lstate = List.fold_left (fun l s -> (fix_state s)::l) [] a.lstate }

  let fix_parsed_scm s =
    { id_scm = s.id_scm ;
      nb_channels = s.nb_channels ;
      lossy_list = List.rev s.lossy_list ;
      params = List.rev s.params ;
      lauto = List.fold_left (fun l a -> (fix_automaton a)::l) [] s.lauto ;
      threshold = List.rev s.threshold ;
      bad_states = List.rev s.bad_states }
end

let parse = Frontend.parse_lexbuf Format.std_formatter

let make_global s = Scm_prod.from_syntree (OrderFix.fix_parsed_scm s)

let check_global = Scm_syn.check_global_scm
