(* $Id: ScmWrapper.mli 836 2010-07-05 22:53:30Z sutre $ *)

(*
 * Copyright (c) 2009 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * SCM Model Wrapper.
 *
 * This module provides a wrapper around Tristan Le Gall's syntax and semantics
 * modules for systems of communicating machines (SCMs).  This permits access to
 * the SCM library through an implementation that respects the input interface
 * of the model-checking engines, defined in the {! Model} module.
 *
 * The current implementation abstracts away all numerical information, i.e. the
 * abstract numerical lattice used for lattice automata is the trivial lattice
 * with two elements.
 *
 * The semantics of lossy channels in this model corresponds to a (potential)
 * loss of messages on emissions.  Formally, if a channel c is declared lossy,
 * then the semantics of send operations in this channel is given by:
 *
 * {C w -- c!a --> w' if w' = w or w' = w·a}
 *
 * If initial configurations have empty queues, then the reachability set with
 * this {e loss on emissions} semantics is the same as the reachability set with
 * the classical lossy semantics (where channels can lose messages at any time
 * non-deterministically).
 *)


(**
 * Functor providing an implementation of the {! Model.S} interface for a given
 * global SCM.  Such a global SCM may be obtained with the following code, which
 * reads the SCM from the standard input:
 *
 * {[
 *   let scm_lu = Frontend.parse_lexbuf
 *                  Format.std_formatter (Lexing.from_channel stdin)
 *   let scm_glob = Scm_prod.from_syntree scm_lu
 * ]}
 *)
module Make (A : sig val scm : Scm_syn.global_scm end) : Model.S

val parse : Lexing.lexbuf -> Scm_syn.scm
val make_global : Scm_syn.scm -> Scm_syn.global_scm
val check_global : Scm_syn.global_scm -> unit
