(* $Id: StaticAnalyzer.ml 1245 2011-10-18 21:15:14Z gsutre $ *)

(*
 * Copyright (c) 2010, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Tristan Le Gall <tlegall@ulb.ac.be>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util


(**
 * This module implements a static analyzer for models that conform to the
 * Model.S signature.
 *)

exception InternalError

module Make (A : Model.S) =
struct
  (* We open A and define shortcuts to simplify notations *)
  open A
  module Location = Automaton.Location
  module Transition = Automaton.Transition
  module Region = Semantics.Region
  module RE = ReachabilityEngine.Make(A)


  type result =
      Invariant of ((A.State.t list) Lazy.t) option * ((A.State.t list) Lazy.t) option
    | Counterexample of Transition.t list
    | Abort of string

  let forward = Option.StaticAnalyzer.direction () == Option.StaticAnalyzer.Forward
  (* if [forward], then a forward analysis is done, otherwise a backward analysis is done *)



let empty_intersection (reachability_set: State.t list) =
  (* checks whether there is a nonempty intersection between the
     reachable states (resp. coreachable states) and the error
     (resp. initial) states *)

  let rec find_region x li =
    match li with
	[] -> raise Not_found
      | (loc,reg)::qq when Location.equal loc x -> reg
      | _::qq -> find_region x qq
  in
  if forward
  then
    let f_fold_error (l_err,r_err) res =
      res
      &&
	(let r_reach = find_region l_err reachability_set in Region.leq (Region.inter r_err r_reach) Region.bot)
	(*  check for a non-empty intersection with an error state *)
    in
      A.fold_error f_fold_error true
  else
    let f_fold_init (l_init,r_init) res =
      res
      &&
	(let r_coreach = find_region l_init reachability_set in Region.leq (Region.inter r_init r_coreach) Region.bot)
	(*  check for a non-empty intersection with an initial state *)
    in
      A.fold_init f_fold_init true


let cons x l = x::l

let initial_states : State.t list =
  if forward
  then (* list of initial states *)
    A.fold_init cons []
  else (* list of error states *)
    A.fold_error cons []

let transition_list : Transition.t list =
  A.Automaton.fold_transitions cons []

let cpt_max = (Option.Main.limit ())
let k_min = (Option.Main.k_min ())
let k_max = (Option.Main.k_max ())

let rec analysis_loop cpt k =
  if (cpt > cpt_max) then
    (cpt - 1,
     Abort ("ABSINT: loop iteration " ^ (string_of_int cpt) ^
              " larger than limit " ^ (string_of_int cpt_max)))
  else
    begin
      Logger.slog Logger.Notice "@[%s%8d  @]@?" (String.make 23 '\b') cpt ;
      Logger.info "@.@[ABSINT loop: %d@]@." cpt ;
      let reachability_set =
        RE.reach ~forward:forward ~k:k initial_states transition_list
      in
        if empty_intersection reachability_set then
          begin
            Logger.slog Logger.Notice "@[ )@]@?" ;
            Logger.info "@[Safe abstract fix point.@]@." ;
            (cpt,
             if forward
             then Invariant (Some (Lazy.lazy_from_val reachability_set), None)
             else Invariant (None, Some (Lazy.lazy_from_val reachability_set)))
          end
        else
          begin
            Logger.info "@[Abstract error found.@]@." ;
            let k = k + 1
            in
	      if k > k_max then
                (cpt,
                 Abort ("ABSINT: closure parameter " ^ (string_of_int k) ^
                          " larger than limit " ^ (string_of_int k_max)))
              else
                begin
                  Logger.slog Logger.Notice "@[ ( %8d )@]@?" k ;
                  Logger.info "@[Closure parameter: %d@]@." k ;
                  analysis_loop (cpt+1) k
                end
          end
    end

let model_check () =
  Logger.slog Logger.Notice "@[%10s %10s   %10s@]@."
    "" "Loops" "k" ;
  Logger.slog Logger.Notice "@[ABSINT loop: %8d   ( %8d )@]@?" 0 k_min ;
  Logger.info "@[ABSINT loop: %d@]@." 0 ;
  let (cpt, result) =
    analysis_loop 1 k_min
  in
    Logger.notice "@." ;
    Logger.report "@[ABSINT:@ %d loops@]@." cpt ;
    result
end
