(* $Id: TraceChecker.mli 1178 2011-10-13 09:20:22Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * Trace Checker.
 *
 * This module defines the interface of trace-checking engines, as a functor
 * taking as argument a semantics+closure module, and returning a module with
 * a forward/backward trace-checking function.  A trace-checking function checks
 * whether a given trace is feasible, and provides a trace invariant when it is
 * not feasible.
 *
 * The following presents in detail the expected behavior of trace-checking
 * engines.  Given a transition system {i (S, A, →)}, a {i trace} is a (finite)
 * sequence {i X{_ 0}, ..., X{_ n}} where each {i X{_ i}} is either a subset of
 * {i S} or an action in {i A}.  Such a trace may, equivalently, be seen as the
 * alternating sequence {i σ{_ 0}, U{_ 1}, σ{_ 1}, ..., U{_ m}, σ{_ m}} where:
 * {ul
 * {li {i U{_ 1}, ..., U{_ m}} is the subsequence of {i X{_ 0}, ..., X{_ n}}
 *     composed of subsets of {i S}, and}
 * {li {i σ{_ 0}, ..., σ{_ m} ∈ A{^ *}} are the (potentially empty) action
 *     sequences in between.}
 * }
 * In particular, {i σ{_ 0}} (resp. {i σ{_ m}}) is empty if {i T} starts (resp.
 * ends) with a subset of {i S}.
 *
 * A trace {i T = X{_ 0}, ..., X{_ n}} is {i feasible} if there exists {i m+2}
 * states {i s{_ 0} ∈ S, s{_ 1} ∈ U{_ 1}, ..., s{_ m} ∈ U{_ m}, s{_ m+1} ∈ S}
 * such that, for each {i 0 ≤ i ≤ m}, the transition system contains a path from
 * {i s{_ i}} to {i s{_ i+1}} labeled by {i σ{_ i}}.  Otherwise, the trace is
 * called {i unfeasible}.
 *
 * A {i forward trace invariant} for {i T} is a sequence {i H{_ 1}, ..., H{_ m}}
 * of subsets of {i S}, such that, for every {i i } with {i 0 ≤ i ≤ m}, it holds
 * that:
 * {C {i
 * H{_ i+1} ⊇ post(H{_ i} ∩ U{_ i}, σ{_ i}) ∩ U{_ i+1}
 * }}
 * where {i U{_ 0} = U{_ m+1} = S},  {i H{_ 0} = S} and {i H{_ m+1} = ∅}.
 *
 * A {i backward trace invariant} for {i T} is a sequence {i H{_ 1}, ..., H{_ m}}
 * of subsets of {i S}, such that, for every {i i } with {i 0 ≤ i ≤ m}, it holds
 * that:
 * {C {i
 * H{_ i} ⊇ pre(H{_ i+1} ∩ U{_ i+1}, σ{_ i}) ∩ U{_ i}
 * }}
 * where {i U{_ 0} = U{_ m+1} = S}, {i H{_ 0} = ∅} and {i H{_ m+1} = S}.
 *
 * Remark that every forward trace invariant may be converted into a backward
 * one, and vice versa.  Indeed, {i H{_ 1}, ..., H{_ m}} is a forward trace
 * invariant if and only if {i (S∖H{_ 1}), ..., (S∖H{_ m})} is a backward trace
 * invariant.  Notice also that a trace is unfeasible if and only if it admits a
 * (forward or backward) trace invariant.
 *
 * A forward trace invariant {i H{_ 1}, ..., H{_ m}} is called {i irreducible}
 * if there exists {i k, l}, with {i 0 ≤ k ≤ l ≤ m+1}, such that, for every
 * {i 1 ≤ i ≤ m}, the following conditions hold:
 * {C
 * {i (i ≤ k) ⇒ (H{_ i} = S)}
 * and
 * {i (i ≥ l) ⇒ (H{_ i} = ∅)}
 * and
 * {i (k < i < l) ⇒ (U{_ i}∖H{_ i} ∉ \{∅, U{_ i}\})}
 * }
 *
 * A backward trace invariant {i H{_ 1}, ..., H{_ m}} is called {i irreducible}
 * if there exists {i k, l}, with {i 0 ≤ k ≤ l ≤ m+1}, such that, for every
 * {i 1 ≤ i ≤ m}, the following conditions hold:
 * {C
 * {i (i ≤ k) ⇒ (H{_ i} = ∅)}
 * and
 * {i (i ≥ l) ⇒ (H{_ i} = S)}
 * and
 * {i (k < i < l) ⇒ (U{_ i}∖H{_ i} ∉ \{∅, U{_ i}\})}
 * }
 *
 * Remark that irreducibility is preserved under the above-mentioned conversion
 * of trace invariants (from forward to backward, and vice versa).
 *
 * For more flexibility, trace-checking engines may use a relaxed variant of
 * feasibility, where the sets {i U{_ i}} are discarded.  Formally, a trace
 * {i T = X{_ 0}, ..., X{_ n}} is {i weakly feasible} if there exists {i m+2}
 * states {i s{_ 0}, ..., s{_ m+1} ∈ S} such that, for each {i 0 ≤ i ≤ m}, the
 * transition system contains a path from {i s{_ i}} to {i s{_ i+1}} labeled by
 * {i σ{_ i}}.  To prevent confusion, standard feasibility shall be called
 * {i strong feasibility}.
 *)


open Util


(**
 * Common interface to all trace-checking engines.
 *)
module T (A : Model.SEMANTICS_CLOSURE) :
sig
  (**
   * Union type for the possible basic elements composing a trace.  This type
   * defines two kinds of actions: commands or guards.  The latter are useful
   * to specify initial and final states.
   *)
  module Atom :
  sig
    type 'a t = [
        `State of 'a * A.Region.t
          (** Subset of states. *)
      | `Guard of A.Region.t
          (** Guard action. *)
      | `Command of A.Command.t
          (** Command action. *)
    ]

    val print : (Format.formatter -> 'a -> unit) ->
      Format.formatter -> 'a t -> unit
  end

  (**
   * A trace is a finite sequence of atoms.
   *)
  module Trace :
  sig
    type 'a t = 'a Atom.t Sequence.t

    val print : (Format.formatter -> 'a -> unit) ->
      Format.formatter -> 'a t -> unit
  end

  (**
   * Type of possible results for the trace-checking function.  The meaning of
   * these values is detailed in the documentation of the {! TraceChecker}
   * module.  Trace invariants shall be {e irreducible}, and encoded as triples
   * {i (m, H, n)} that expose the interesting factor in the sequence:
   * {ul
   * {li [`Invariant_forward (m, H, n)] denotes the irreducible forward trace
   *     invariant {i (S){^ m}, H, (∅){^ n}}.}
   * {li [`Invariant_backward (m, H, n)] denotes the irreducible backward trace
   *     invariant {i (∅){^ m}, H, (S){^ n}}.}
   * }
   * In both cases, the sequence {i H} consists of pairs [((s, r), h)] where
   * [(s, r)] is the original [`State] atom in the trace, and [h] is the
   * corresponding {i invariant region}, which must be distinct from {i ∅} and
   * {i S}.
   *)
  module Result :
  sig
    type 'a invariant = int * (('a * A.Region.t) * A.Region.t) Sequence.t * int

    type 'a t = [
        `Feasible of [ `Strong | `Weak ]
      | `Invariant_forward of 'a invariant
      | `Invariant_backward of 'a invariant
    ]

    val print : (Format.formatter -> 'a -> unit) ->
      Format.formatter -> 'a t -> unit
  end
end

module type S = functor (A : Model.SEMANTICS_CLOSURE) ->
sig
  include module type of T(A)

  (**
   * The trace-checking function.
   *)
  val trace_check : 'a Trace.t -> 'a Result.t
end
