(* $Id: TraceCheckerUtil.ml 1178 2011-10-13 09:20:22Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util


module Common (A : Model.SEMANTICS_CLOSURE) =
struct
  include TraceChecker.T(A)

  let atom_strong_post reg = function
      `State (_, r) | `Guard r -> A.Region.inter reg r
    | `Command cmd -> A.post reg cmd

  let atom_strong_pre reg = function
      `State (_, r) | `Guard r -> A.Region.inter reg r
    | `Command cmd -> A.pre reg cmd

  let atom_weak_post reg = function
      `State (_, r) -> reg
    | `Guard r -> A.Region.inter reg r
    | `Command cmd -> A.post reg cmd

  let atom_weak_pre reg = function
      `State (_, r) -> reg
    | `Guard r -> A.Region.inter reg r
    | `Command cmd -> A.pre reg cmd

  let num_state_atoms a m n =
    let r = ref 0
    in
      for i = m to n do
        match a.(i) with `State _ -> incr r | _ -> ()
      done ;
      !r

  module Reduction =
  struct
    module Region = A.Region

    (*
     * [prefix p l] returns [(h, n)] where [List.rev h] is the longest prefix
     * of the list [l] whose elements all satisfy the predicate [p], and [n] is
     * the length of [h].
     *)
    let prefix p l =
      let rec aux l acc len =
        match l with
            [] -> (acc, len)
          | x :: tail ->
            if p x then
              aux tail (x::acc) (len+1)
            else
              (acc, len)
      in
        aux l [] 0

    let check_non_disjointness ((_, r_t), r_i) =
      not (Region.eq Region.bot (Region.inter r_t r_i))

    let check_non_containment ((_, r_t), r_i) =
      not (Region.leq r_t r_i)

    (*
     * [reduce_trace_invariant p1 p2 (m, inv, n)] reduces [inv] by extracting a
     * prefix w.r.t. [p1], and then a suffix w.r.t. [p2].  This function assumes
     * that the direction was adjusted by the caller.
     *)
    let reduce_trace_invariant p1 p2 (m, (s_l, s_d, s_n), n) =
      let (pre1, len1) = prefix p1 s_l in
      let (pre2, len2) = prefix p2 pre1
      in
      let inv' = (pre2, s_d, len2)
      and (m', n') =
        match s_d with
            Sequence.Forward  -> (m + len1 - len2,  n + (s_n - len1))
          | Sequence.Backward -> (m + (s_n - len1), n + len1 - len2)
      in
        (m', inv', n')

    (*
     * [tweak (m, s, n)] transforms the sequence [s] so that its direction
     * maintains the respective order of [m] and [n] after reduction.
     *)
    let tweak (m, s, n) =
      if m < n then
        (m, Sequence.forward s, n)
      else if m > n then
        (m, Sequence.backward s, n)
      else
        (m, s, n)

    type invariant_direction = Fwd | Bwd

    let predicates d (_, (_, s_d, _), _) =
      match (d, s_d) with
          (Fwd, Sequence.Forward) | (Bwd, Sequence.Backward) ->
            (check_non_disjointness, check_non_containment)
        | (Fwd, Sequence.Backward) | (Bwd, Sequence.Forward) ->
            (check_non_containment, check_non_disjointness)

    let reduce d inv =
      let inv = tweak inv in
      let (p1, p2) = predicates d inv
      in
        reduce_trace_invariant p1 p2 inv
  end

  let reduce_fwd_invariant inv = Reduction.reduce Reduction.Fwd inv
  let reduce_bwd_invariant inv = Reduction.reduce Reduction.Bwd inv
end

module Reverse (E : TraceChecker.S) (A : Model.SEMANTICS_CLOSURE) =
struct
  module B =
  struct
    module Region = A.Region
    module Command = A.Command
    let post = A.pre
    let pre = A.post
    let rel cmd r1 r2 = A.rel cmd r2 r1
    let clo = A.clo
  end

  include E (B)

  let trace_check trace =
    match trace_check (Sequence.rev trace) with
        `Feasible _ as result -> result
      | `Invariant_forward (m, s, n) ->
          `Invariant_backward (n, Sequence.rev s, m)
      | `Invariant_backward (m, s, n) ->
          `Invariant_forward (n, Sequence.rev s, m)
end

module DiscardStates (E : TraceChecker.S) (A : Model.SEMANTICS_CLOSURE) =
struct
  include Common (A)
  module TC = E (A)

  let unplug = function
      `State (s, r) -> `State ((s, r), A.Region.top)
    | `Guard _ as atom -> atom
    | `Command _ as atom -> atom

  let replug (((s, r), t), i) =
    assert (t == A.Region.top) ;
    ((s, r), i)

  let trace_check trace =
    match TC.trace_check (Sequence.map unplug trace) with
        `Feasible _ -> `Feasible `Weak
      | `Invariant_forward (m, s, n) ->
          `Invariant_forward
            (reduce_fwd_invariant (m, Sequence.map replug s, n))
      | `Invariant_backward (m, s, n) ->
          `Invariant_backward
            (reduce_bwd_invariant (m, Sequence.map replug s, n))
end

(*
 * XXX Now that the ['a result] type includes original [`State] regions, the
 * XXX implementation of this functor could be simplified.
 *)
module Validate (E : TraceChecker.S) (A : Model.SEMANTICS_CLOSURE) =
struct
  include Common (A)
  module TC = E (A)

  let is_feasible atom_post atom_pre (l, d, n) =
    let r =
      match d with
          Sequence.Forward ->  List.fold_left atom_post A.Region.top l
        | Sequence.Backward -> List.fold_left atom_pre  A.Region.top l
    in
      not (A.Region.leq r A.Region.bot)

  let is_invariant_forward trace (m, s, n) =
    let trace = Sequence.to_array trace
    and inv = Sequence.to_array (Sequence.map (fun (_, r) -> r) s)
    in
      (*
       * The trace invariant [(m, inv, n)] must have one, and only one, region
       * for each [`State] atom in [trace].
       *)
      if num_state_atoms trace 0 ((Array.length trace) - 1) =
        m + (Array.length inv) + n
      then
        (*
         * Functions to check that [(m, inv, n)] is an irreducible forward
         * invariant.
         *)
        let rec precheck m t =
          assert (m > 0) ;
          assert (t < Array.length trace) ;
          match trace.(t) with
              `State (_, r) ->
                if m = 1 then check r (t+1) 0
                else precheck (m-1) (t+1)
            | _ -> precheck m (t+1)
        and check img t i =
          if i < Array.length inv then
            begin
              assert (t < Array.length trace) ;
              let img' = atom_strong_post img trace.(t)
              in
                match trace.(t) with
                    `State (_, r) ->
                      (*
                       * Check that [img'] is contained in [inv.(i)].  Further,
                       * irreducibility requires that [inv.(i)] intersects [r]
                       * but does not contain [r].
                       *)
                      if (A.Region.leq img' inv.(i)) &&
                        (not (A.Region.leq r inv.(i))) &&
                        (not (A.Region.leq (A.Region.inter r inv.(i)) A.Region.bot))
                      then
                        check (atom_strong_post inv.(i) trace.(t)) (t+1) (i+1)
                      else
                        false
                  | _ -> check img' (t+1) i
            end
          else
            postcheck img t
        and postcheck img t =
          if t < Array.length trace then
            let img' = atom_strong_post img trace.(t)
            in
              match trace.(t) with
                  `State (_, r) ->
                    (*
                     * Check that [img'] is empty.
                     *)
                    A.Region.leq img' A.Region.bot
                | _ -> postcheck img' (t+1)
          else
            A.Region.leq img A.Region.bot
        in
          if m > 0 then
            precheck m 0
          else
            check A.Region.top 0 0
      else
        false

  let is_invariant_backward trace (m, s, n) =
    let trace = Sequence.to_array trace
    and inv = Sequence.to_array (Sequence.map (fun (_, r) -> r) s)
    in
      (*
       * The trace invariant [(m, inv, n)] must have one, and only one, region
       * for each [`State] atom in [trace].
       *)
      if num_state_atoms trace 0 ((Array.length trace) - 1) =
        m + (Array.length inv) + n
      then
        (*
         * Functions to check that [(m, inv, n)] is an irreducible backward
         * invariant.
         *)
        let rec precheck n t =
          assert (n > 0) ;
          assert (t >= 0) ;
          match trace.(t) with
              `State (_, r) ->
                if n = 1 then check r (t-1) ((Array.length inv) - 1)
                else precheck (n-1) (t-1)
            | _ -> precheck n (t-1)
        and check img t i =
          if i >= 0 then
            begin
              assert (t >= 0) ;
              let img' = atom_strong_pre img trace.(t)
              in
                match trace.(t) with
                    `State (_, r) ->
                      (*
                       * Check that [img'] is contained in [inv.(i)].  Further,
                       * irreducibility requires that [inv.(i)] intersects [r]
                       * but does not contain [r].
                       *)
                      if (A.Region.leq img' inv.(i)) &&
                        (not (A.Region.leq r inv.(i))) &&
                        (not (A.Region.leq (A.Region.inter r inv.(i)) A.Region.bot))
                      then
                        check (atom_strong_pre inv.(i) trace.(t)) (t-1) (i-1)
                      else
                        false
                  | _ -> check img' (t-1) i
            end
          else
            postcheck img t
        and postcheck img t =
          if t >= 0 then
            let img' = atom_strong_pre img trace.(t)
            in
              match trace.(t) with
                  `State (_, r) ->
                    (*
                     * Check that [img'] is empty.
                     *)
                    A.Region.leq img' A.Region.bot
                | _ -> postcheck img' (t-1)
          else
            A.Region.leq img A.Region.bot
        in
          if n > 0 then
            precheck n ((Array.length trace) - 1)
          else
            check A.Region.top ((Array.length trace) - 1) ((Array.length inv) - 1)
      else
        false

  let trace_check trace =
    let result = TC.trace_check trace in
    let valid =
      match result with
          `Feasible `Strong -> is_feasible atom_strong_post atom_strong_pre trace
        | `Feasible `Weak -> is_feasible atom_weak_post atom_weak_pre trace
        | `Invariant_forward  inv -> is_invariant_forward  trace inv
        | `Invariant_backward inv -> is_invariant_backward trace inv
    in
      if not valid then
        Logger.error "@[ERROR: Trace-checker result is invalid.  Please report.@]@." ;
      result
end
