(* $Id: TraceCheckerUtil.mli 1178 2011-10-13 09:20:22Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util


(**
 * Common utility functor for trace-checking engines.
 *)
module Common : functor (A : Model.SEMANTICS_CLOSURE) ->
sig
  include module type of TraceChecker.T(A)

  val atom_strong_post : A.Region.t -> 'a Atom.t -> A.Region.t
  val atom_strong_pre  : A.Region.t -> 'a Atom.t -> A.Region.t
  val atom_weak_post : A.Region.t -> 'a Atom.t -> A.Region.t
  val atom_weak_pre  : A.Region.t -> 'a Atom.t -> A.Region.t

  (**
   * [num_state_atoms a m n] returns the number of [`State] atoms in the array
   * [a], between the indexes [m] and [n] (inclusive).
   *)
  val num_state_atoms : 'a Atom.t array -> int -> int -> int

  (**
   * [reduce_fwd_invariant inv] takes a forward trace invariant (that might be
   * reducible), reduces it and returns the resulting {e irreducible} forward
   * trace invariant.
   *)
  val reduce_fwd_invariant : 'a Result.invariant -> 'a Result.invariant

  (**
   * [reduce_bwd_invariant inv] takes a backward trace invariant (that might be
   * reducible), reduces it and returns the resulting {e irreducible} backward
   * trace invariant.
   *)
  val reduce_bwd_invariant : 'a Result.invariant -> 'a Result.invariant
end

(**
 * Functor reversing a given trace-checking engine.  Conceptually, reversion
 * means that the given trace-checking engine is applied to the reversed model,
 * where:
 * {ul
 * {li the source and target of each transition are swapped, and}
 * {li the semantics of each command is inversed.}
 * }
 *)
module Reverse : functor (E : TraceChecker.S) -> TraceChecker.S

(**
 * Functor transforming a given trace-checking engine by discarding [`State]
 * regions and then reducing the obtained trace invariant.
 *)
module DiscardStates : functor (E : TraceChecker.S) -> TraceChecker.S

(**
 * Functor wrapping a given trace-checking engine with result validation.  When
 * validation fails, a [Failure] exception is raised.
 *)
module Validate : functor (E : TraceChecker.S) -> TraceChecker.S
