(* $Id: UPInv.ml 1166 2011-10-12 13:06:21Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Util


module MakeFwd (A : Model.SEMANTICS_CLOSURE) =
struct
  open A

  let k_min = (Option.Main.k_min ())

  include TraceCheckerUtil.Common (A)

  let rec fwd_is_feasible trace reg i =
    if Region.leq reg Region.bot
    then
      false
    else
      if i < Array.length trace then
        fwd_is_feasible trace (atom_strong_post reg trace.(i)) (i+1)
      else
        true

  let rec fwd_simplify_trace_inv inv_l simpl_l simpl_n =
    match inv_l with
        [] ->
          (simpl_l, simpl_n)
      | (s, r, cloimg) :: tail ->
          if Region.leq r cloimg
          then
            (*
             * The region [cloimg] from the trace invariant contains the
             * corresponding [`State] region [r] of the input trace, so the
             * prefix [inv_l] may be completely ignored.
             *)
            (simpl_l, simpl_n)
          else
            fwd_simplify_trace_inv tail (((s, r), cloimg) :: simpl_l) (simpl_n + 1)

  (*
   * [fwd_trace_inv]:
   *
   *	Performs forward approximated post computation along the given [trace].
   *	Returns a forward trace invariant, or [`Feasible] if the approximation
   *	was too coarse (or [trace] is feasible).
   *
   *	The [current] region is assumed to be non-empty.
   *)
  let rec fwd_trace_inv approx trace inv_l inv_n current i =
    assert (not (Region.leq current Region.bot)) ;
    if i < Array.length trace then
      let img = atom_strong_post current trace.(i)
      in
        if Region.leq img Region.bot
        then
          let (simpl_l, simpl_n) = fwd_simplify_trace_inv inv_l [] 0
          in
            `Invariant_forward
              (inv_n - simpl_n,
               (simpl_l, Sequence.Forward, simpl_n),
               num_state_atoms trace i ((Array.length trace) - 1))
        else
          (*
           * XXX For the last one (i.e., i = (Array.length trace) - 1),
           * XXX we should simply return [`Feasible] (no extra computation).
           *)
          match trace.(i) with
              `State (s, r) ->
                (*
                 * An alternative here would be to apply approx on img
                 * instead of current.  We choose to apply on current to
                 * conform with the UPInv algorithm of the SPIN'09 paper.
                 *)
                let cloimg = approx current (* non empty *)
                in
                  fwd_trace_inv approx trace
                    ((s, r, cloimg) :: inv_l) (inv_n + 1)
                    (Region.inter cloimg r) (i+1)
            | _ ->
                fwd_trace_inv approx trace inv_l inv_n img (i+1)
    else
      `Feasible `Strong

  let fwd_trace_check trace =
    let trace = Sequence.to_array trace
    in
      if fwd_is_feasible trace Region.top 0
      then
        `Feasible `Strong
      else
        let rec loop k =
          let result = fwd_trace_inv (clo k) trace [] 0 Region.top 0
          in
            match result with
                `Feasible `Strong ->
                  loop (k+1)
              | _ ->
                  Logger.slog Logger.Notice "@[ / %8d@]@?" k ;
                  Logger.info "@[Closure parameter: %d@]@." k ;
                  result
        in
          loop k_min

  let trace_check = fwd_trace_check
end

module MakeBwd = TraceCheckerUtil.Reverse (MakeFwd)
