(* $Id: Validator.ml 1092 2011-08-08 12:50:05Z gsutre $ *)

(*
 * Copyright (c) 2009, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(* Inductive invariant checking and path feasibility. *)


open Util

module Make (A : Model.S) =
struct
  (* We open A and define shortcuts to simplify notations *)
  open A
  module Location = Automaton.Location
  module Transition = Automaton.Transition
  module Region = Semantics.Region

  (* Maps over locations *)
  module LMap = Map.Make(Location)

  (* Converts a list of symbolic states into a map from locations to regions. *)
  let symb_state_list_to_map states =
    let add_with_union m (l, r) =
      if (LMap.mem l m) then
        LMap.add l (Region.union (LMap.find l m) r) m
      else
        LMap.add l r m
    in
      List.fold_left add_with_union LMap.empty states

  (* Funtion to check that a given list of symbolic states is a safe forward
     inductive invariant. *)
  let is_safe_fwd_invariant inv_list =
    let inv_map = symb_state_list_to_map inv_list in
    let reg_of_loc l =
      try LMap.find l inv_map with
          Not_found -> Region.bot
    in
    let check_post_fixpoint_transition t =
      let src = Transition.source t
      and tgt = Transition.target t
      and cmd = Transition.command t
      in
        Region.leq (Semantics.post (reg_of_loc src) cmd) (reg_of_loc tgt)
    in
      (fold_init
         (fun (l, r) b ->
            b &&
            (Region.leq r (reg_of_loc l)))
         true)
      &&
      (Automaton.fold_transitions
         (fun t b ->
            b &&
            (check_post_fixpoint_transition t))
         true)
      &&
      (fold_error
         (fun (l, r) b ->
            b &&
            (Region.leq (Region.inter r (reg_of_loc l)) Region.bot))
         true)

  (* Funtion to check that a given list of symbolic states is a safe backward
     inductive invariant. *)
  let is_safe_bwd_invariant inv_list =
    let inv_map = symb_state_list_to_map inv_list in
    let reg_of_loc l =
      try LMap.find l inv_map with
          Not_found -> Region.bot
    in
    let check_post_fixpoint_transition t =
      let src = Transition.source t
      and tgt = Transition.target t
      and cmd = Transition.command t
      in
        Region.leq (Semantics.pre (reg_of_loc tgt) cmd) (reg_of_loc src)
    in
      (fold_error
         (fun (l, r) b ->
            b &&
            (Region.leq r (reg_of_loc l)))
         true)
      &&
      (Automaton.fold_transitions
         (fun t b ->
            b &&
            (check_post_fixpoint_transition t))
         true)
      &&
      (fold_init
         (fun (l, r) b ->
            b &&
            (Region.leq (Region.inter r (reg_of_loc l)) Region.bot))
         true)

  (* Iterated post computation along a given path p. *)
  let rec path_post (l, r) p =
    match p with
        [] ->
          (l, r)
      | t :: p' ->
          let src = Transition.source t
          and tgt = Transition.target t
          and cmd = Transition.command t
          in
            if Location.equal l src
            then
              path_post (tgt, (Semantics.post r cmd)) p'
            else
              path_post (tgt, Region.bot) p'

  (* Iterated pre computation along a given path p.  The path p is assumed to be
     in reverse order. *)
  let rec path_pre (l, r) p =
    match p with
        [] ->
          (l, r)
      | t :: p' ->
          let src = Transition.source t
          and tgt = Transition.target t
          and cmd = Transition.command t
          in
            if Location.equal l tgt
            then
              path_pre (src, (Semantics.pre r cmd)) p'
            else
              path_pre (src, Region.bot) p'

  (* Function to check whether the empty path is a feasible counterexample. *)
  let is_empty_path_feasible () =
    fold_init
      (fun (li, ri) bi ->
         bi ||
         (fold_error
            (fun (le, re) be ->
               be ||
               ((Location.equal li le) &&
                (not (Region.leq (Region.inter ri re) Region.bot))))
            false))
      false

  (* Returns the init region of a given location. *)
  let init_reg_of_loc loc =
    fold_init
      (fun (l, r) reg ->
         if (Location.equal l loc)
         then
           Region.union reg r
         else
           reg)
      Region.bot

  (* Returns the error region of a given location. *)
  let error_reg_of_loc loc =
    fold_error
      (fun (l, r) reg ->
         if (Location.equal l loc)
         then
           Region.union reg r
         else
           reg)
      Region.bot

  (* Function to check, using forward post iteration, whether a given non-empty
     empty path is a feasible counterexample. *)
  let is_non_empty_path_fwd_feasible path =
    match path with
        [] -> invalid_arg "is_non_empty_path_fwd_feasible"
      | t :: _ ->
          let src = Transition.source t in
          let (tgt, post) = path_post (src, init_reg_of_loc src) path
          in
            fold_error
              (fun (le, re) b ->
                 b ||
                 ((Location.equal tgt le) &&
                  (not (Region.leq (Region.inter post re) Region.bot))))
              false

  (* Function to check, using backward pre iteration, whether a given non-empty
     empty path is a feasible counterexample. *)
  let is_non_empty_path_bwd_feasible path =
    match List.rev path with
        [] -> invalid_arg "is_non_empty_path_bwd_feasible"
      | t :: _ as path' ->
          let tgt = Transition.target t in
          let (src, pre) = path_pre (tgt, error_reg_of_loc tgt) path'
          in
            fold_init
              (fun (li, ri) b ->
                 b ||
                 ((Location.equal src li) &&
                  (not (Region.leq (Region.inter pre ri) Region.bot))))
              false
end
