(* $Id: Validator.mli 1092 2011-08-08 12:50:05Z gsutre $ *)

(*
 * Copyright (c) 2009 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * Inductive invariant checking and path feasibility.
 *
 * This module provides simple functions to validate the results returned by
 * model-checkers.  The implementation of this module is intentionally separated
 * from that of model-checkers.  The goal here is to provide functions that are
 * simple and (hopefully) easy to read, with little regard for efficiency.
 *)

(**
 * Functor providing inductive invariant checking and path feasibility checking
 * functions for a given model.
 *)
module Make (A : Model.S) :
sig
  (** {2 Invariant Checking } *) (*********************************************)

  (**
   * Potential invariants are represented as lists of symbolic states of the
   * model.  Such a list stands for the union of its elements.  Thus if a
   * location does not appear in the list, then its associated region is the
   * empty region [Region.bot].
   *)

  (**
   * [is_safe_fwd_invariant l] returns true if [l] is a safe forward inductive
   * invariant:
   * {ul
   * {li [l] contains all initial symbolic states of the model, and}
   * {li [l] is a post-fixpoint of the model's post image operator, and}
   * {li [l] is disjoint from all error symbolic states of the model.}
   * }
   *)
  val is_safe_fwd_invariant : A.State.t list -> bool

  (**
   * [is_safe_bwd_invariant l] returns true if [l] is a safe backward inductive
   * invariant:
   * {ul
   * {li [l] contains all error symbolic states of the model, and}
   * {li [l] is a post-fixpoint of the model's pre image operator, and}
   * {li [l] is disjoint from all initial symbolic states of the model.}
   * }
   *)
  val is_safe_bwd_invariant : A.State.t list -> bool

  (** {2 Path Feasibility Checking } *) (**************************************)

  (**
   * A {e path} is a list of contiguous transitions of the model's automaton.  A
   * path [[t1; ...; tN]] is {e feasible} if we can insert (N+1) environments
   * between the transitions of the path such that these environments respect
   * the semantics of the commands labeling the transitions of the path.
   *)

  (**
   * [is_empty_path_feasible ()] returns true if there is a non empty
   * intersection between initial and error symbolic states of the model.
   *)
  val is_empty_path_feasible : unit -> bool

  (**
   * [is_non_empty_path_fwd_feasible p] returns true if [p] is a non empty path
   * that is feasible.  Feasibility checking is performed with a forward post
   * computation along the path.
   *)
  val is_non_empty_path_fwd_feasible : A.Automaton.Transition.t list -> bool

  (**
   * [is_non_empty_path_bwd_feasible p] returns true if [p] is a non empty path
   * that is feasible.  Feasibility checking is performed with a backward pre
   * computation along the path.
   *)
  val is_non_empty_path_bwd_feasible : A.Automaton.Transition.t list -> bool
end
