(* $Id: control.ml 1146 2011-09-13 12:33:20Z gsutre $ *)

(*
 * Copyright (c) 2009, 2010, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * This module contains the main code of the McScM controller.  It basically
 * glues together all other modules.
 *)

open Svar

module Logger = Util.Logger
module Statistics = Util.Statistics


let _ = Statistics.register_start "total"

(* Set default logging priority. *)
let _ = Logger.set_priority Logger.Notice

(* Set default limit (number of simulation steps). *)
let _ = Option.Main.limit_opt := 100

(* Initialize global options.  This also parses command-line arguments. *)
let _ = Option.initialize (Option.control_arg_spec ())

(* Display tool header. *)
let _ =
  Logger.report "@[McScM %s - Supervisory Control Tool@]@."
    Config.version

(* Separator string of length equal to the screen width (number of columns). *)
let separator = String.make (Format.get_margin ()) '-'

(* Parse the SCM description from source file or standard input. *)
let _ = Statistics.register_start "parsing + product"
let _ = Logger.info "@[Parsing input SCM description...@]@?"

let scm_lu =
  let input_channel =
    match Option.Main.scm_source_file () with
        None -> stdin
      | Some f -> open_in f
  in
    ScmWrapper.parse (Lexing.from_channel input_channel)

let _ = Logger.info "@[ done.@]@."

(* Build global SCM by cartesian product. *)
let _ = Logger.info "@[Computing global SCM by cartesian product of local systems...@]@?"

let scm_glob = ScmWrapper.make_global scm_lu

let _ = Logger.info "@[ done.@]@."
let _ = Statistics.register_stop "parsing + product"

(* Check consistency of global SCM. *)
let _ =
  if Option.Main.scm_consistency_check ()
  then
    begin
      Statistics.register_start "consistency check" ;
      Logger.info "@[Checking consistency of global SCM...@]@?" ;
      ScmWrapper.check_global scm_glob ;
      Logger.info "@[ done.@]@." ;
      Statistics.register_stop "consistency check"
    end


(* Create model module using the SCM wrapper. *)
let _ =
  Statistics.register_start "SCM wrapping" ;
  Logger.info "@[Building SCM wrapper module (with control flow automaton)...@]@?"

module M = ScmWrapper.Make (struct let scm = scm_glob end)

let _ =
  Logger.info "@[ done.@]@." ;
  Statistics.register_stop "SCM wrapping"

let _ =
  Logger.report "@[Model:@ %d locations,@ %d transitions,@ %d/%d init/error symbolic states@]@."
    M.Automaton.nb_locations
    M.Automaton.nb_transitions
    M.nb_init
    M.nb_error

(* Display the model. *)
let _ =
  Logger.info "@[<v>%s@,%a@,%s@]@."
    separator
    M.print_model ()
    separator

(* Create the controller module *)
module C = Svar.Control.Make (M)

(* Perform supervisory control *)
let _ = Statistics.register_start "supervisory control"
let (res, msg) =
  try C.control () ; (true, "") with
      Scanf.Scan_failure msg -> (false, msg)
let _ = Statistics.register_stop  "supervisory control"

let _ =
  Logger.info "@[<v>%s@]@."
    separator

(* Display result of supervisory control *)
let _ =
  Logger.report "@[Result: %s@]@."
    (if res then "Supervisory control complete."
     else "Aborted: " ^ msg ^ ".")

let _ = Statistics.register_stop "total"

let _ =
  if Option.Main.show_statistics ()
  then
    Logger.report "@[%a@]@." Statistics.print ()

(* Flush logger *)
let _ = Logger.flush ()

(* Exit *)
let _ = exit 0
