(* $Id: verify.ml 1240 2011-10-17 21:44:06Z gsutre $ *)

(*
 * Copyright (c) 2009, 2010, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * This module contains the main code of the McScM model-checker.  It basically
 * glues together all other modules.
 *)

open Svar

module Logger = Util.Logger
module Statistics = Util.Statistics


let _ = Statistics.register_start "total"

(* Set default logging priority. *)
let _ = Logger.set_priority Logger.Notice

(* Initialize global options.  This also parses command-line arguments. *)
let _ = Option.initialize (Option.verify_arg_spec ())

(* Display tool header. *)
let _ =
  Logger.report "@[McScM %s - Model Checking Tool@]@."
    Config.version

(* Separator string of length equal to the screen width (number of columns). *)
let separator = String.make (Format.get_margin ()) '-'

(* Parse the SCM description from source file or standard input. *)
let _ = Statistics.register_start "parsing + product"
let _ = Logger.info "@[Parsing input SCM description...@]@?"

let scm_lu =
  let input_channel =
    match Option.Main.scm_source_file () with
        None -> stdin
      | Some f -> open_in f
  in
    ScmWrapper.parse (Lexing.from_channel input_channel)

let _ = Logger.info "@[ done.@]@."

(* Build global SCM by cartesian product. *)
let _ = Logger.info "@[Computing global SCM by cartesian product of local systems...@]@?"

let scm_glob = ScmWrapper.make_global scm_lu

let _ = Logger.info "@[ done.@]@."
let _ = Statistics.register_stop "parsing + product"

(* Check consistency of global SCM. *)
let _ =
  if Option.Main.scm_consistency_check ()
  then
    begin
      Statistics.register_start "consistency check" ;
      Logger.info "@[Checking consistency of global SCM...@]@?" ;
      ScmWrapper.check_global scm_glob ;
      Logger.info "@[ done.@]@." ;
      Statistics.register_stop "consistency check"
    end


(* Create model module using the SCM wrapper. *)
let _ =
  Statistics.register_start "SCM wrapping" ;
  Logger.info "@[Building SCM wrapper module (with control flow automaton)...@]@?"

module M = ScmWrapper.Make (struct let scm = scm_glob end)

let _ =
  Logger.info "@[ done.@]@." ;
  Statistics.register_stop "SCM wrapping"

let _ =
  Logger.report "@[Model:@ %d locations,@ %d transitions,@ %d/%d init/error symbolic states@]@."
    M.Automaton.nb_locations
    M.Automaton.nb_transitions
    M.nb_init
    M.nb_error

(* Display the model. *)
let _ =
  Logger.info "@[<v>%s@,%a@,%s@]@."
    separator
    M.print_model ()
    separator

(* Use trace-cheking engine specified by selectable option. *)
let e =
  match Option.Main.tc_engine () with
      Option.Main.APInv_bwd ->
        (module Svar.APInv.MakeBwd : TraceChecker.S)
    | Option.Main.APInv_fwd ->
        (module Svar.APInv.MakeFwd : TraceChecker.S)
    | Option.Main.UPInv_bwd ->
        (module Svar.UPInv.MakeBwd : TraceChecker.S)
    | Option.Main.UPInv_fwd ->
        (module Svar.UPInv.MakeFwd : TraceChecker.S)

let e =
  if Option.Main.tc_discard_states () then
    (module Svar.TraceCheckerUtil.DiscardStates
        ((val e : TraceChecker.S)) : TraceChecker.S)
  else e

let e =
  if Option.Main.tc_validate () then
    (module Svar.TraceCheckerUtil.Validate
        ((val e : TraceChecker.S)) : TraceChecker.S)
  else e

module E = (val e : TraceChecker.S)

(* Create the model-checker module *)
module MC =
  (val (match Option.Main.mc_engine () with
            Option.Main.Absint ->
              (module Svar.StaticAnalyzer.Make : ModelChecker.S)
          | Option.Main.Armc ->
              (module Svar.Armc.Make : ModelChecker.S)
          | Option.Main.Cegar ->
              (module Svar.Cegar.Make (E) : ModelChecker.S)
          | Option.Main.Lart ->
              (module Svar.LazyAbstraction.Make (E) : ModelChecker.S)
   ) : ModelChecker.S)

module C = MC (M)

(* Perform model-checking *)
let _ = Statistics.register_start "model-checking"
let res = C.model_check ()
let _ = Statistics.register_stop  "model-checking"

let _ =
  Logger.info "@[<v>%s@]@."
    separator

(* Display result of model-checking *)
let _ =
  Logger.report "@[Result: %s@]@."
    (match res with
         C.Invariant _ ->
           "Model is safe."
       | C.Counterexample path ->
           "Model is unsafe (" ^ (string_of_int (List.length path)) ^ ")."
       | C.Abort msg ->
           "Aborted: " ^ msg ^ ".")

(* If model is unsafe then display abstract path *)
let _ =
  match res with
      C.Counterexample path ->
        Logger.notice "@[<v 3>Counterexample:" ;
        List.iter
          (Logger.notice "@,@[<3>%a@]" M.Automaton.Transition.print)
          path ;
        Logger.notice "@]@."
    | _ ->
        ()

(* Create validator module *)
module V = Validator.Make(M)

(* Check result of model-checking with validator *)
let _ =
  if Option.Main.result_validation ()
  then
    begin
      Statistics.register_start "validation" ;
      Logger.report "@[Result validation:@]@?" ;
      let (res_check_ok, status) =
        match res with
            C.Invariant (None, None) ->
              (false, "not available")
          | C.Invariant (lazy_fwd_inv_opt, lazy_bwd_inv_opt) ->
              let fwd_inv_status =
                match lazy_fwd_inv_opt with
                    None -> true
                  | Some lazy_fwd_inv ->
                      Logger.notice "@.@[   Validation of safe forward  invariant:@]@?" ;
                      let result = V.is_safe_fwd_invariant (Lazy.force lazy_fwd_inv) in
                        Logger.notice "@[ %s@]@." (if result then "passed." else "failed!") ;
                        result
              and bwd_inv_status =
                match lazy_bwd_inv_opt with
                    None -> true
                  | Some lazy_bwd_inv ->
                      Logger.notice "@[   Validation of safe backward invariant:@]@?" ;
                      let result = V.is_safe_bwd_invariant (Lazy.force lazy_bwd_inv) in
                        Logger.notice "@[ %s@]@." (if result then "passed." else "failed!") ;
                        result
              in
                (fwd_inv_status && bwd_inv_status, "")
          | C.Counterexample [] ->
              let cex_status =
                Logger.notice "@.@[   Validation of feasibility of empty counterexample:@]@?" ;
                let result = V.is_empty_path_feasible () in
                  Logger.notice "@[ %s@]@." (if result then "passed." else "failed!") ;
                  result
              in
                (cex_status, "")
          | C.Counterexample path ->
              let fwd_cex_status =
                Logger.notice "@.@[   Validation of forward  feasibility of counterexample:@]@?" ;
                let result = V.is_non_empty_path_fwd_feasible path in
                  Logger.notice "@[ %s@]@." (if result then "passed." else "failed!") ;
                  result
              and bwd_cex_status =
                Logger.notice "@[   Validation of backward feasibility of counterexample:@]@?" ;
                let result = V.is_non_empty_path_bwd_feasible path in
                  Logger.notice "@[ %s@]@." (if result then "passed." else "failed!") ;
                  result
              in
                (fwd_cex_status && bwd_cex_status, "")
          | C.Abort _ ->
              (false, "aborted")
      in
        if (not (status = ""))
        then
          Logger.report "@[ %s.@]@." status
        else
          begin
            Logger.notice "@[Result validation:@]" ;
            if res_check_ok
            then
              Logger.report "@[ passed.@]@."
            else
              begin
                Logger.report "@[ failed!@]@." ;
                Logger.error "@[ERROR: Model-checker result is invalid.  Please report.@]@."
              end
          end ;
        Statistics.register_stop "validation"
    end

let _ = Statistics.register_stop "total"

let _ =
  if Option.Main.show_statistics ()
  then
    Logger.report "@[%a@]@." Statistics.print ()

(* Flush logger *)
let _ = Logger.flush ()

(* Exit *)
let _ = exit 0
