(* $Id: BenchGraph.ml 802 2010-05-13 09:09:20Z sutre $ *)

(*
 * Copyright (c) 2009 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(* Performance Test Module for Labeled Directed Multi-Graphs. *)


open Print


let _ = Format.set_margin 120


module SearchTest (G : MultiGraph.I with type Vertex.Label.t = string and type Edge.Label.t = string) =
struct
  include Search.Make (G)

  let list_of_vertices_printer =
    Print.list_printer_from_printer ";@ " G.Vertex.print

  let print_traversal fmt =
    function
        Forward_path p ->
          Format.fprintf Format.std_formatter "Forward Path:@ {@[%a@]}"
          Path.print
          p
      | Backward_path p ->
          Format.fprintf Format.std_formatter "Backward Path:@ {@[%a@]}"
          Path.print
          p
      | Forward_reach s ->
          Format.fprintf Format.std_formatter "No path,@ forward reach:@ %d@ vertices."
          (List.length s)
      | Backward_reach s ->
          Format.fprintf Format.std_formatter "No path,@ backward reach:@ %d@ vertices."
          (List.length s)

  (* Function to build a tree with the given width and depth, and with dum dummy
     edges for each real edge.  No dummy edge is created if dum < 1. *)
  let build_tree width depth dum =
    let graph = G.create 1 in
    let rec do_it d v =
      if d <= depth
      then
        for i = 1 to width do
          let child = G.Vertex.create ((G.Vertex.lbl v) ^ "_" ^ (string_of_int i))
          in
            G.add_edge graph (G.Edge.create v (string_of_int i) child) ;
            for i = 1 to dum do
              G.add_edge graph
                (G.Edge.create (G.Vertex.create "dummy") ("dummy") (G.Vertex.create "dummy"))
            done ;
            do_it (d+1) child
        done
    in
    let root = G.Vertex.create "r"
    in
      G.add_vertex graph root ;
      do_it 1 root ;
      ([root], graph)

  (* Similar to build_tree, but constructs a kind of mesh consisting of layers
     of bi-partite graphs. *)
  let build_mesh width depth dum =
    let create_row d =
      let rec do_it d i acc =
        if i > width
        then
          acc
        else
          do_it d (i+1)
            ((G.Vertex.create ((string_of_int d) ^ "_" ^ (string_of_int i))) :: acc)
      in
        do_it d 1 []
    and graph = G.create 1 in
    let rec do_it d l =
      if d <= depth
      then
        let l' = create_row d
        in
          List.iter
            (fun v' ->
               List.iter
                 (fun v ->
                    G.add_edge graph
                      (G.Edge.create v ((G.Vertex.lbl v) ^ " -> " ^ (G.Vertex.lbl v')) v'))
                 l ;
               for i = 1 to dum do
                 G.add_edge graph
                   (G.Edge.create (G.Vertex.create "dummy") ("dummy") (G.Vertex.create "dummy"))
               done)
            l' ;
          do_it (d+1) l'
    in
    let first_row = create_row 0
    in
      do_it 1 first_row ;
      (first_row, graph)

  let bench_expl f msg graph initial final =
    Format.printf "@[Graph exploration (%s):@]@?" msg ;
    Statistics.register_start msg ;
    let trav = f graph initial final
    in
      Statistics.register_stop msg ;
      Format.printf "@[ done.@]@." ;
      Format.printf "@[Result of %s:@ @[%a@]@]@."
        msg
        print_traversal
        trav

  let bench gen width depth dum =
    Statistics.register_start "total" ;
    Statistics.register_start "graph generation" ;
    Format.printf "@[Graph generation:@]@?" ;
    let (init, graph) = gen width depth dum
    in
      Statistics.register_stop "graph generation" ;
      Format.printf "@[ done.@]@." ;
      let initial = init
      and final = []
      in
        (*
        Format.printf "@.@[<v>Graph:@,{@[<v 3>@,@[%a@]@]@,}@]@."
          G.print
          graph ;
        *)
        Format.printf "@[Graph has@ %d vertices@ and@ %d egdes.@]@."
          (G.nb_vertices graph) (G.nb_edges graph) ;
        Format.printf "@[Initial vertices:@ [@[%a@]]@]@."
          list_of_vertices_printer
          initial ;
        Format.printf "@[Final vertices:@ [@[%a@]]@]@."
          list_of_vertices_printer
          final ;

        bench_expl fwd_dfs "forward dfs" graph initial final ;
        bench_expl fwd_bfs "forward bfs" graph initial final ;
        bench_expl bwd_dfs "backward dfs" graph initial final ;
        bench_expl bwd_bfs "backward bfs" graph initial final ;

        Statistics.register_stop "total" ;
        Logger.notice "@[%a@]@." Statistics.print ()

  let bench_tree = bench build_tree
  let bench_mesh = bench build_mesh
end

module Str =
struct
  type t = string
  let print = Format.pp_print_string
end

module T = SearchTest(MultiGraph.Imperative.Make (Str) (Str))

let _ =
  T.bench_tree 3 12 1 ;
  Format.printf "@.@[Garbage collection...@]@?" ;
  Statistics.clear () ;
  Gc.compact () ;
  Format.printf "@[ done.@]@.@.@[%s@]@.@."
    (String.make (Format.get_margin ()) '-') ;
  T.bench_mesh 100 300 4 ;
