(* $Id: Collection.ml 1265 2011-10-19 22:51:55Z gsutre $ *)

(*
 * Copyright (c) 2009, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


module Set =
struct
  module type S =
  sig
    type elt
    type t
    val is_empty : t -> bool
    val mem : t -> elt -> bool
    val cardinal : t -> int
    val iter : (elt -> unit) -> t -> unit
    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
    val elements : t -> elt list
    val equal : t -> t -> bool
  end

  module type IS =
  sig
    include S
    val create : int -> t
    val clear : t -> unit
    val copy: t -> t
    val add : t -> elt -> unit
    val remove : t -> elt -> unit
  end

  module type PS =
  sig
    include S
    val empty : t
    val add : t -> elt -> t
    val remove : t -> elt -> t
    val filter : (elt -> bool) -> t -> t
  end

  module Imperative =
  struct
    module MakeHashSet (T : Hashtbl.HashedType) =
    struct
      module HT = Hashtbl.Make (T)
      type elt = T.t
      type t = unit HT.t
      let create x = HT.create x
      let clear = HT.clear
      let copy = HT.copy
      let add s e = HT.replace s e ()
      let remove = HT.remove
      let mem = HT.mem
      let cardinal = HT.length
      let iter f s = HT.iter (fun e () -> f e) s
      let fold f s = HT.fold (fun e () b -> f e b) s
      let is_empty s = (cardinal s) = 0
      let elements s =
        (List.rev (fold (fun e l -> e::l) s []))

      let subset s1 s2 = (* tests whether s1 is contained in s2 *)
        try
          iter (HT.find s2) s1 ;
          true
        with Not_found -> false
      let equal s1 s2 =
        if (cardinal s1) < (cardinal s2)
        then
          (subset s1 s2) && (subset s2 s1)
        else
          (subset s2 s1) && (subset s1 s2)
    end

    module MakeMutable (Set : PS) =
    struct
      type elt = Set.elt
      type t = Set.t ref
      let create _ = ref Set.empty
      let clear s = s := Set.empty
      let copy s = ref !s
      let add s e = s := (Set.add !s e)
      let remove s e = s := (Set.remove !s e)
      let mem s = Set.mem !s
      let cardinal s = Set.cardinal !s
      let iter f s = Set.iter f !s
      let fold f s = Set.fold f !s
      let is_empty s = Set.is_empty !s
      let elements s = Set.elements !s
      let equal s1 s2 = Set.equal !s1 !s2
    end
  end

  module Persistent =
  struct
    module MakeTreeSet (T : Set.OrderedType) =
    struct
      module ST = Set.Make (T)
      type elt = T.t
      type t = ST.t
      let empty = ST.empty
      let add s e = ST.add e s
      let remove s e = ST.remove e s
      let filter = ST.filter
      let is_empty = ST.is_empty
      let mem s e = ST.mem e s
      let cardinal = ST.cardinal (* warning: this function is in O(n) *)
      let iter = ST.iter
      let fold = ST.fold
      let elements = ST.elements
      let equal = ST.equal
    end
  end
end

module Map =
struct
  module type M =
  sig
    type key
    type 'a t
    val is_empty: 'a t -> bool
    val mem:  'a t -> key -> bool
    val find: 'a t -> key -> 'a
    val size : 'a t -> int
    val iter: (key -> 'a -> unit) -> 'a t -> unit
    val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  end

  module type IM =
  sig
    include M
    val create: int -> 'a t
    val clear: 'a t -> unit
    val copy: 'a t -> 'a t
    val add: 'a t -> key -> 'a -> unit
    val remove: 'a t -> key -> unit
  end

  module type PM =
  sig
    include M
    val empty : 'a t
    val add: 'a t -> key -> 'a -> 'a t
    val remove: 'a t -> key -> 'a t
  end

  module Imperative =
  struct
    module MakeHashMap (T : Hashtbl.HashedType) =
    struct
      module HT = Hashtbl.Make (T)
      type key = T.t
      type 'a t = 'a HT.t
      let create x = HT.create x
      let clear = HT.clear
      let copy = HT.copy
      let add m k a = HT.replace m k a
      let remove = HT.remove
      let mem = HT.mem
      let find = HT.find
      let size = HT.length
      let iter = HT.iter
      let fold = HT.fold
      let is_empty m = (size m) = 0
    end

    module MakeMutable (Map : PM) =
    struct
      type key = Map.key
      type 'a t = 'a Map.t ref
      let create _ = ref Map.empty
      let clear m = m := Map.empty
      let copy m = ref !m
      let add m k a = m := (Map.add !m k a)
      let remove m k = m := (Map.remove !m k)
      let is_empty m = Map.is_empty !m
      let mem m = Map.mem !m
      let find m = Map.find !m
      let size m = Map.size !m
      let iter f m = Map.iter f !m
      let fold f m = Map.fold f !m
    end
  end

  module Persistent =
  struct
    module MakeTreeMap (T : Map.OrderedType) =
    struct
      module MT = Map.Make (T)
      type key = T.t
      type 'a t = 'a MT.t
      let empty = MT.empty
      let add m k a = MT.add k a m
      let remove m k = MT.remove k m
      let is_empty = MT.is_empty
      let mem m k = MT.mem k m
      let find m k = MT.find k m
      let iter = MT.iter
      let fold = MT.fold
      let size m = MT.fold (fun _ _ n -> n+1) m 0
    end
  end
end
