(* $Id: Logger.ml 738 2009-07-18 16:18:23Z sutre $ *)

(*
 * Copyright (c) 2009 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(* Logging of messages. *)


type log_level =
    Error
  | Report
  | Warning
  | Notice
  | Info
  | Trace
  | Debug

let stderr_printer = function msg -> Format.fprintf Format.err_formatter msg
let stdout_printer = function msg -> Format.fprintf Format.std_formatter msg

let dummy_formatter = (Format.make_formatter (fun _ _ _ -> ()) (fun _ -> ()))
let ignore_printer = function msg -> Format.ifprintf dummy_formatter msg

let priority = ref Notice
let error_logging   = ref true
let report_logging  = ref true
let warning_logging = ref true
let notice_logging  = ref true
let info_logging    = ref false
let trace_logging   = ref false
let debug_logging   = ref false

let error   msg = if !error_logging   then stderr_printer msg else ignore_printer msg
let report  msg = if !report_logging  then stdout_printer msg else ignore_printer msg
let warning msg = if !warning_logging then stdout_printer msg else ignore_printer msg
let notice  msg = if !notice_logging  then stdout_printer msg else ignore_printer msg
let info    msg = if !info_logging    then stdout_printer msg else ignore_printer msg
let trace   msg = if !trace_logging   then stdout_printer msg else ignore_printer msg
let debug   msg = if !debug_logging   then stdout_printer msg else ignore_printer msg

let log =
  function
      Error   -> error
    | Report  -> report
    | Warning -> warning
    | Notice  -> notice
    | Info    -> info
    | Trace   -> trace
    | Debug   -> debug

let slog p = if p = !priority then stdout_printer else ignore_printer

let set_priority p =
  priority := p ;
  match p with
      Error ->
        error_logging   := true ;
        report_logging  := false ;
        warning_logging := false ;
        notice_logging  := false ;
        info_logging    := false ;
        trace_logging   := false ;
        debug_logging   := false
    | Report ->
        error_logging   := true ;
        report_logging  := true ;
        warning_logging := false ;
        notice_logging  := false ;
        info_logging    := false ;
        trace_logging   := false ;
        debug_logging   := false
    | Warning ->
        error_logging   := true ;
        report_logging  := true ;
        warning_logging := true ;
        notice_logging  := false ;
        info_logging    := false ;
        trace_logging   := false ;
        debug_logging   := false
    | Notice ->
        error_logging   := true ;
        report_logging  := true ;
        warning_logging := true ;
        notice_logging  := true ;
        info_logging    := false ;
        trace_logging   := false ;
        debug_logging   := false
    | Info ->
        error_logging   := true ;
        report_logging  := true ;
        warning_logging := true ;
        notice_logging  := true ;
        info_logging    := true ;
        trace_logging   := false ;
        debug_logging   := false
    | Trace ->
        error_logging   := true ;
        report_logging  := true ;
        warning_logging := true ;
        notice_logging  := true ;
        info_logging    := true ;
        trace_logging   := true ;
        debug_logging   := false
    | Debug ->
        error_logging   := true ;
        report_logging  := true ;
        warning_logging := true ;
        notice_logging  := true ;
        info_logging    := true ;
        trace_logging   := true ;
        debug_logging   := true

let flush () =
  Format.pp_print_flush Format.err_formatter () ;
  Format.pp_print_flush Format.std_formatter ()
