(* $Id: MultiGraph.ml 1265 2011-10-19 22:51:55Z gsutre $ *)

(*
 * Copyright (c) 2009, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(* Labeled Directed Multi-Graphs. *)


open Print
open Collection


module type VERTEX =
sig
  include PRINTABLE_TYPE
  module Label : PRINTABLE_TYPE
  val create : Label.t -> t
  val lbl : t -> Label.t
  val compare : t -> t -> int
  val equal : t -> t -> bool
  val hash : t -> int
end

module type G =
sig
  include PRINTABLE_TYPE
  module Vertex : VERTEX
  module Edge :
  sig
    include PRINTABLE_TYPE
    module Label : PRINTABLE_TYPE
    val create : Vertex.t -> Label.t -> Vertex.t -> t
    val src : t -> Vertex.t
    val tgt : t -> Vertex.t
    val lbl : t -> Label.t
    val compare : t -> t -> int
    val equal : t -> t -> bool
    val hash : t -> int
  end
  val nb_vertices : t -> int
  val nb_edges : t -> int
  val mem_vertex : t -> Vertex.t -> bool
  val mem_edge : t -> Edge.t -> bool
  val iter_vertices : (Vertex.t -> unit) -> t -> unit
  val fold_vertices : (Vertex.t -> 'a -> 'a) -> t -> 'a -> 'a
  val iter_edges : (Edge.t -> unit) -> t -> unit
  val fold_edges : (Edge.t -> 'a -> 'a) -> t -> 'a -> 'a
  val iter_succ : (Edge.t -> unit) -> t -> Vertex.t -> unit
  val fold_succ : (Edge.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a
  val iter_pred : (Edge.t -> unit) -> t -> Vertex.t -> unit
  val fold_pred : (Edge.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a
end

module type P =
sig
  include G
  val empty : t
  val add_vertex : t -> Vertex.t -> t
  val del_vertex : t -> Vertex.t -> t
  val add_edge : t -> Edge.t -> t
  val del_edge : t -> Edge.t -> t
end

module type I =
sig
  include G
  val create : int -> t
  val add_vertex : t -> Vertex.t -> unit
  val del_vertex : t -> Vertex.t -> unit
  val add_edge : t -> Edge.t -> unit
  val del_edge : t -> Edge.t -> unit
end


(* Implementation of the Vertex module that is common to some persistent and
   imperative multi-graph implementations. *)
module MakeVertex (VLabel : PRINTABLE_TYPE) =
struct
  module Label = VLabel
  include Uid.Make (Label)
  let lbl = value
end

(* Implementation of the Edge module that is common to some persistent and
   imperative multi-graph implementations. *)
module MakeEdge (Vertex : sig include VERTEX val id : t -> int end) (ELabel : PRINTABLE_TYPE) =
struct
  module Label = ELabel
  include Uid.Make (
    struct
      type t = Vertex.t * ELabel.t * Vertex.t
      let print fmt (v1, l, v2) =
        Format.fprintf fmt "%d@ »@;@[%a@]@;»@ %d"
          (Vertex.id v1)
          Label.print l
          (Vertex.id v2)
    end)
  let create v1 l v2 = create (v1, l, v2)
  let src e = let (v, _, _) = value e in v
  let tgt e = let (_, _, v) = value e in v
  let lbl e = let (_, l, _) = value e in l
end


module Persistent =
struct
  module Make (VLabel : PRINTABLE_TYPE) (ELabel : PRINTABLE_TYPE) =
  struct
    module Vertex = MakeVertex (VLabel)
    module Edge = MakeEdge (Vertex) (ELabel)
    module ESet = Set.Persistent.MakeTreeSet (Edge)
    module VMap = Map.Persistent.MakeTreeMap (Vertex)

    type t = {
      vertices : (ESet.t * ESet.t) VMap.t ; (* (succs, preds) of each vertex *)
      nb_vertices : int ;
      nb_edges : int ;
    }

    let nb_vertices g =
      assert (g.nb_vertices = VMap.size g.vertices) ;
      g.nb_vertices
    let nb_edges g =
      assert (
        g.nb_edges =
          VMap.fold
            (fun _ (succs, _) n -> n + (ESet.cardinal succs))
            g.vertices 0) ;
      g.nb_edges

    let mem_vertex g v = VMap.mem g.vertices v
    let mem_edge g e =
      try
        let (succs, _) = VMap.find g.vertices (Edge.src e)
        in
          ESet.mem succs e
      with Not_found -> false

    let iter_vertices f g = VMap.iter (fun v _ -> f v) g.vertices
    let fold_vertices f g = VMap.fold (fun v _ -> f v) g.vertices

    let iter_edges f g =
      VMap.iter (fun _ (succs, _) -> ESet.iter f succs) g.vertices
    let fold_edges f g =
      VMap.fold (fun _ (succs, _) -> ESet.fold f succs) g.vertices

    let iter_succ f g v =
      assert (mem_vertex g v) ;
      let (succs, _) = VMap.find g.vertices v
      in
        ESet.iter f succs

    let fold_succ f g v =
      assert (mem_vertex g v) ;
      let (succs, _) = VMap.find g.vertices v
      in
        ESet.fold f succs

    let iter_pred f g v =
      assert (mem_vertex g v) ;
      let (_, preds) = VMap.find g.vertices v
      in
        ESet.iter f preds

    let fold_pred f g v =
      assert (mem_vertex g v) ;
      let (_, preds) = VMap.find g.vertices v
      in
        ESet.fold f preds

    let empty = { vertices = VMap.empty ;
                  nb_vertices = 0 ;
                  nb_edges = 0 }

    let add_vertex g v =
      assert (not (mem_vertex g v)) ;
      { vertices = VMap.add g.vertices v (ESet.empty, ESet.empty) ;
        nb_vertices = g.nb_vertices + 1 ;
        nb_edges = g.nb_edges }

    let del_vertex g v = (* requires VMap.find and then VMap.add: not optimal! *)
      assert (mem_vertex g v) ;
      let update_succ e (g, n) =
        assert (Vertex.equal v (Edge.src e)) ;
        let tgt = Edge.tgt e
        in
          if Vertex.equal v tgt
          then
            (g, n + 1)
          else
            let (tgt_succs, tgt_preds) = VMap.find g.vertices tgt
            in
              assert (ESet.mem tgt_preds e) ;
              let g' =
                { vertices =
                    VMap.add
                      g.vertices
                      tgt
                      (tgt_succs, ESet.remove tgt_preds e) ;
                  nb_vertices = g.nb_vertices ;
                  nb_edges = g.nb_edges - 1 }
              in
                (g', n)
      and update_pred e (g, n) =
        assert (Vertex.equal v (Edge.tgt e)) ;
        let src = Edge.src e
        in
          if Vertex.equal v src
          then
            (g, n + 1)
          else
            let (src_succs, src_preds) = VMap.find g.vertices src
            in
              assert (ESet.mem src_succs e) ;
              let g' =
                { vertices =
                    VMap.add
                      g.vertices
                      src
                      (ESet.remove src_succs e, src_preds) ;
                  nb_vertices = g.nb_vertices ;
                  nb_edges = g.nb_edges - 1 }
              in
                (g', n)
      and (succs, preds) = VMap.find g.vertices v
      in
      let (g, nb_selfs_from_succs) = ESet.fold update_succ succs (g, 0)
      in
      let (g, nb_selfs_from_preds) = ESet.fold update_pred preds (g, 0)
      in
        assert (nb_selfs_from_succs = nb_selfs_from_preds) ;
        { vertices = VMap.remove g.vertices v ;
          nb_vertices = g.nb_vertices - 1 ;
          nb_edges = g.nb_edges - nb_selfs_from_succs }

    let add_edge g e = (* requires VMap.find and then VMap.add: not optimal! *)
      assert (not (mem_edge g e)) ;
      let src = Edge.src e
      and tgt = Edge.tgt e
      in
        if Vertex.equal src tgt
        then
          let ((succs, preds), n) =
            try (VMap.find g.vertices src, 0)
            with Not_found -> ((ESet.empty, ESet.empty), 1)
          in
            { vertices =
                VMap.add
                  g.vertices
                  src
                  (ESet.add succs e, ESet.add preds e) ;
              nb_vertices = g.nb_vertices + n ;
              nb_edges = g.nb_edges + 1 }
        else
          let ((src_succs, src_preds), src_n) =
            try (VMap.find g.vertices src, 0)
            with Not_found -> ((ESet.empty, ESet.empty), 1)
          and ((tgt_succs, tgt_preds), tgt_n) =
            try (VMap.find g.vertices tgt, 0)
            with Not_found -> ((ESet.empty, ESet.empty), 1)
          in
            { vertices =
                VMap.add
                  (VMap.add
                     g.vertices
                     src
                     (ESet.add src_succs e, src_preds))
                  tgt
                  (tgt_succs, ESet.add tgt_preds e) ;
              nb_vertices = g.nb_vertices + src_n + tgt_n ;
              nb_edges = g.nb_edges + 1 }

    let del_edge g e = (* requires VMap.find and then VMap.add: not optimal! *)
      assert (mem_edge g e) ;
      let src = Edge.src e
      and tgt = Edge.tgt e
      in
        if Vertex.equal src tgt
        then
          let (succs, preds) = VMap.find g.vertices src
          in
            { vertices =
                VMap.add
                  g.vertices
                  src
                  (ESet.remove succs e, ESet.remove preds e) ;
              nb_vertices = g.nb_vertices ;
              nb_edges = g.nb_edges - 1 }
        else
          let (src_succs, src_preds) = VMap.find g.vertices src
          and (tgt_succs, tgt_preds) = VMap.find g.vertices tgt
          in
            { vertices =
                VMap.add
                  (VMap.add
                     g.vertices
                     src
                     (ESet.remove src_succs e, src_preds))
                  tgt
                  (tgt_succs, ESet.remove tgt_preds e) ;
              nb_vertices = g.nb_vertices ;
              nb_edges = g.nb_edges - 1 }

    let print fmt g =
      let print_succ_edge fmt e =
        Format.fprintf fmt "@[%a@]@;»@ %d"
          Edge.Label.print (Edge.lbl e)
          (Vertex.id (Edge.tgt e))
      and print_pred_edge fmt e =
        Format.fprintf fmt "@[%a@]@;«@ %d"
          Edge.Label.print (Edge.lbl e)
          (Vertex.id (Edge.src e))
      in
        Format.fprintf fmt "@[<v>" ;
        Format.fprintf fmt "@[<v 3>%d vertices:" (nb_vertices g) ;
        iter_vertices (Format.fprintf fmt "@,@[<h>%a@]" Vertex.print) g ;
        Format.fprintf fmt "@]@," ;
        Format.fprintf fmt "@[<v 3>%d edges:" (nb_edges g) ;
        iter_edges (Format.fprintf fmt "@,@[<h>%a@]" Edge.print) g ;
        Format.fprintf fmt "@]@," ;
        Format.fprintf fmt "@[<v 3>Outbound edges of each vertex:" ;
        iter_vertices
          (fun v ->
             Format.fprintf fmt "@,@[<v 3>%d :" (Vertex.id v) ;
             iter_succ
               (Format.fprintf fmt "@,» @[<3>%a@]" print_succ_edge) g v ;
             Format.fprintf fmt "@]")
          g ;
        Format.fprintf fmt "@]@," ;
        Format.fprintf fmt "@[<v 3>Inbound edges of each vertex:" ;
        iter_vertices
          (fun v ->
             Format.fprintf fmt "@,@[<v 3>%d :" (Vertex.id v) ;
             iter_pred
               (Format.fprintf fmt "@,« @[<3>%a@]" print_pred_edge) g v ;
             Format.fprintf fmt "@]")
          g ;
        Format.fprintf fmt "@]@]"
  end
end


module Imperative =
struct
  module Make (VLabel : PRINTABLE_TYPE) (ELabel : PRINTABLE_TYPE) =
  struct
    module Vertex = MakeVertex (VLabel)
    module Edge = MakeEdge (Vertex) (ELabel)
    module ESet = Set.Imperative.MakeHashSet (Edge)
    module VMap = Map.Imperative.MakeHashMap (Vertex)

    type t = {
      vertices : (ESet.t * ESet.t) VMap.t ; (* (succs, preds) of each vertex *)
      mutable nb_vertices : int ;
      mutable nb_edges : int ;
    }

    let nb_vertices g =
      assert (g.nb_vertices = VMap.size g.vertices) ;
      g.nb_vertices
    let nb_edges g =
      assert (
        g.nb_edges =
          VMap.fold
            (fun _ (succs, _) n -> n + (ESet.cardinal succs))
            g.vertices 0) ;
      g.nb_edges

    let mem_vertex g v = VMap.mem g.vertices v
    let mem_edge g e =
      try
        let (succs, _) = VMap.find g.vertices (Edge.src e)
        in
          ESet.mem succs e
      with Not_found -> false

    let iter_vertices f g = VMap.iter (fun v _ -> f v) g.vertices
    let fold_vertices f g = VMap.fold (fun v _ -> f v) g.vertices

    let iter_edges f g =
      VMap.iter (fun _ (succs, _) -> ESet.iter f succs) g.vertices
    let fold_edges f g =
      VMap.fold (fun _ (succs, _) -> ESet.fold f succs) g.vertices

    let iter_succ f g v =
      assert (mem_vertex g v) ;
      let (succs, _) = VMap.find g.vertices v
      in
        ESet.iter f succs

    let fold_succ f g v =
      assert (mem_vertex g v) ;
      let (succs, _) = VMap.find g.vertices v
      in
        ESet.fold f succs

    let iter_pred f g v =
      assert (mem_vertex g v) ;
      let (_, preds) = VMap.find g.vertices v
      in
        ESet.iter f preds

    let fold_pred f g v =
      assert (mem_vertex g v) ;
      let (_, preds) = VMap.find g.vertices v
      in
        ESet.fold f preds

    let average_degree g =
      if (nb_vertices g) = 0
      then 0
      else (nb_edges g) / (nb_vertices g)

    let create n = { vertices = VMap.create n ;
                     nb_vertices = 0 ;
                     nb_edges = 0 }

    let add_vertex g v =
      assert (not (mem_vertex g v)) ;
      let degree = average_degree g
      in
        VMap.add g.vertices v (ESet.create degree, ESet.create degree) ;
        g.nb_vertices <- g.nb_vertices + 1

    let del_vertex g v =
      assert (mem_vertex g v) ;
      let update_succ e n =
        assert (Vertex.equal v (Edge.src e)) ;
        let tgt = Edge.tgt e
        in
          if Vertex.equal v tgt
          then
            n + 1
          else
            let (_, tgt_preds) = VMap.find g.vertices tgt
            in
              assert (ESet.mem tgt_preds e) ;
              ESet.remove tgt_preds e ;
              g.nb_edges <- g.nb_edges - 1 ;
              n
      and update_pred e n =
        assert (Vertex.equal v (Edge.tgt e)) ;
        let src = Edge.src e
        in
          if Vertex.equal v src
          then
            n + 1
          else
            let (src_succs, _) = VMap.find g.vertices src
            in
              assert (ESet.mem src_succs e) ;
              ESet.remove src_succs e ;
              g.nb_edges <- g.nb_edges - 1 ;
              n
      and (succs, preds) = VMap.find g.vertices v
      in
      let nb_selfs_from_succs = ESet.fold update_succ succs 0
      and nb_selfs_from_preds = ESet.fold update_pred preds 0
      in
        assert (nb_selfs_from_succs = nb_selfs_from_preds) ;
        VMap.remove g.vertices v ;
        g.nb_vertices <- g.nb_vertices - 1 ;
        g.nb_edges <- g.nb_edges - nb_selfs_from_succs

    let add_edge g e =
      assert (not (mem_edge g e)) ;
      let add_vertex_and_get_succs_preds v =
        try VMap.find g.vertices v
        with Not_found ->
          let degree = average_degree g
          in
          let (succs, preds) = (ESet.create degree, ESet.create degree)
          in
            VMap.add g.vertices v (succs, preds) ;
            g.nb_vertices <- g.nb_vertices + 1 ;
            (succs, preds)
      and src = Edge.src e
      and tgt = Edge.tgt e
      in
        begin
          if Vertex.equal src tgt
          then
            let (succs, preds) = add_vertex_and_get_succs_preds src
            in
              ESet.add succs e ;
              ESet.add preds e
          else
            let (src_succs, _) = add_vertex_and_get_succs_preds src
            and (_, tgt_preds) = add_vertex_and_get_succs_preds tgt
            in
              ESet.add src_succs e ;
              ESet.add tgt_preds e
        end ;
        g.nb_edges <- g.nb_edges + 1

    let del_edge g e =
      assert (mem_edge g e) ;
      let src = Edge.src e
      and tgt = Edge.tgt e
      in
        begin
          if Vertex.equal src tgt
          then
            let (succs, preds) = VMap.find g.vertices src
            in
              ESet.remove succs e ;
              ESet.remove preds e
          else
            let (src_succs, _) = VMap.find g.vertices src
            and (_, tgt_preds) = VMap.find g.vertices tgt
            in
              ESet.remove src_succs e ;
              ESet.remove tgt_preds e
        end ;
        g.nb_edges <- g.nb_edges - 1

    let print fmt g =
      let print_succ_edge fmt e =
        Format.fprintf fmt "@[%a@]@;»@ %d"
          Edge.Label.print (Edge.lbl e)
          (Vertex.id (Edge.tgt e))
      and print_pred_edge fmt e =
        Format.fprintf fmt "@[%a@]@;«@ %d"
          Edge.Label.print (Edge.lbl e)
          (Vertex.id (Edge.src e))
      in
        Format.fprintf fmt "@[<v>" ;
        Format.fprintf fmt "@[<v 3>%d vertices:" (nb_vertices g) ;
        iter_vertices (Format.fprintf fmt "@,@[<h>%a@]" Vertex.print) g ;
        Format.fprintf fmt "@]@," ;
        Format.fprintf fmt "@[<v 3>%d edges:" (nb_edges g) ;
        iter_edges (Format.fprintf fmt "@,@[<h>%a@]" Edge.print) g ;
        Format.fprintf fmt "@]@," ;
        Format.fprintf fmt "@[<v 3>Outbound edges of each vertex:" ;
        iter_vertices
          (fun v ->
             Format.fprintf fmt "@,@[<v 3>%d :" (Vertex.id v) ;
             iter_succ
               (Format.fprintf fmt "@,» @[<3>%a@]" print_succ_edge) g v ;
             Format.fprintf fmt "@]")
          g ;
        Format.fprintf fmt "@]@," ;
        Format.fprintf fmt "@[<v 3>Inbound edges of each vertex:" ;
        iter_vertices
          (fun v ->
             Format.fprintf fmt "@,@[<v 3>%d :" (Vertex.id v) ;
             iter_pred
               (Format.fprintf fmt "@,« @[<3>%a@]" print_pred_edge) g v ;
             Format.fprintf fmt "@]")
          g ;
        Format.fprintf fmt "@]@]"
  end


  (* Optimized for a small number of graphs with a large number of vertices *)
  (* Remark: an alternative data structure for succs and preds would be a map
     from vertices to edges (or from vertices to triples (id, label, vertex)) in
     order to quickly access successor vertices, and the corresponding edges
     (this could lead to an optimization in del_vertex). *)
  module Make2 (VLabel : PRINTABLE_TYPE) (ELabel : PRINTABLE_TYPE) =
  struct

    module type VERTEX =
    sig
      module Label : Print.PRINTABLE_TYPE
      type edge
      type graph
      type edge_set
      type t
      val print : Format.formatter -> t -> unit
      val create : Label.t -> t
      val id : t -> int
      val lbl : t -> Label.t
      val mem_graph : t -> graph -> bool
      val add_graph : t -> graph -> unit
      val del_graph : t -> graph -> unit
      val edges : t -> graph -> (edge_set * edge_set) (* returns (succs, preds) *)
      val compare : t -> t -> int
      val equal : t -> t -> bool
      val hash : t -> int
    end

    module type EDGE =
    sig
      module Label : Print.PRINTABLE_TYPE
      type vertex
      type t
      val print : Format.formatter -> t -> unit
      val create : vertex -> Label.t -> vertex -> t
      val id : t -> int
      val src : t -> vertex
      val tgt : t -> vertex
      val lbl : t -> Label.t
      val compare : t -> t -> int
      val equal : t -> t -> bool
      val hash : t -> int
    end

    module type GRAPH =
    sig
      type vertex
      type vertex_set
      type t
      val create : int -> t
      val id : t -> int
      val vertices : t -> vertex_set
      val nb_vertices : t -> int
      val nb_edges : t -> int
      val add_vertex : t -> vertex -> unit
      val del_vertex : t -> vertex -> unit
      val delta_nb_edges : t -> int -> unit
      val compare : t -> t -> int
      val equal : t -> t -> bool
      val hash : t -> int
    end

    module rec Vertex : (VERTEX with module Label = VLabel
                                and type edge = Edge.t
                                and type graph = Graph.t
                                and type edge_set = ESet.t) =
    struct
      module Label = VLabel
      type edge = Edge.t
      type graph = Graph.t
      type edge_set = ESet.t

      module GMap = Map.Imperative.MakeHashMap (Graph)

      include Uid.Make (
        struct
          type t = Label.t * ((edge_set * edge_set) GMap.t)
          let print fmt (l, _) = Label.print fmt l
        end)

      let create l = create (l, GMap.create 1)
      let lbl v = let (l, _) = value v in l
      let gmap v = let (_, m) = value v in m

      let mem_graph v g = GMap.mem (gmap v) g
      let add_graph v g =
        assert (not (mem_graph v g)) ;
        let degree =
          if (Graph.nb_vertices g) = 0
          then 0
          else (Graph.nb_edges g) / (Graph.nb_vertices g)
        in
          GMap.add (gmap v) g (ESet.create degree, ESet.create degree)
      let del_graph v g =
        assert (mem_graph v g) ;
        GMap.remove (gmap v) g
      let edges v g =
        assert (mem_graph v g) ;
        GMap.find (gmap v) g
    end

    and Edge : (EDGE with module Label = ELabel
                     and type vertex = Vertex.t) =
    struct
      module Label = ELabel
      type vertex = Vertex.t

      include Uid.Make (
        struct
          type t = Vertex.t * ELabel.t * Vertex.t
          let print fmt (v1, l, v2) =
            Format.fprintf fmt "%d@ »@;@[%a@]@;»@ %d"
              (Vertex.id v1)
              Label.print l
              (Vertex.id v2)
        end)

      let create v1 l v2 = create (v1, l, v2)
      let src e = let (v, _, _) = value e in v
      let tgt e = let (_, _, v) = value e in v
      let lbl e = let (_, l, _) = value e in l
    end

    and Graph : (GRAPH with type vertex = Vertex.t
                       and type vertex_set = VSet.t) =
    struct
      type vertex = Vertex.t
      type vertex_set = VSet.t

      include Uid.Make (
        struct
          type t = vertex_set * (int ref) * (int ref)
          let print fmt g = () (* not used *)
        end)

      let create n = create (VSet.create n, ref 0, ref 0)
      let vertices g =
        let (vertices, _, _) = value g
        in
          vertices
      let nb_vertices g =
        let (_, nb_vertices_ref, _) = value g
        in
          !nb_vertices_ref
      let nb_edges g =
        let (_, _, nb_edges_ref) = value g
        in
          !nb_edges_ref

      let add_vertex g v =
        assert (not (VSet.mem (vertices g) v)) ;
        let (vertices, nb_vertices_ref, _) = value g
        in
          VSet.add vertices v ;
          incr nb_vertices_ref
      let del_vertex g v =
        assert (VSet.mem (vertices g) v) ;
        let (vertices, nb_vertices_ref, _) = value g
        in
          VSet.remove vertices v ;
          decr nb_vertices_ref
      let delta_nb_edges g n =
        let (_, _, nb_edges_ref) = value g
        in
          nb_edges_ref := !nb_edges_ref + n
    end

    and VSet : (Set.IS with type elt = Vertex.t) = Set.Imperative.MakeHashSet (Vertex)
    and ESet : (Set.IS with type elt = Edge.t)   = Set.Imperative.MakeHashSet (Edge)
(*
    and VSet : (Set.IS with type elt = Vertex.t) = Set.Imperative.MakeMutable (Set.Persistent.MakeTreeSet (Vertex))
    and ESet : (Set.IS with type elt = Edge.t)   = Set.Imperative.MakeMutable (Set.Persistent.MakeTreeSet (Edge))
*)

    type t = Graph.t
    let create = Graph.create

    let nb_vertices g =
      assert (Graph.nb_vertices g = VSet.cardinal (Graph.vertices g)) ;
      Graph.nb_vertices g
    let nb_edges g =
      assert (
        Graph.nb_edges g =
          VSet.fold
            (fun v n ->
               let (succs, _) = Vertex.edges v g
               in
                 n + (ESet.cardinal succs))
            (Graph.vertices g) 0) ;
      Graph.nb_edges g

    let mem_vertex g v =
      let result = Vertex.mem_graph v g
      in
        assert (let alt_result = VSet.mem (Graph.vertices g) v
                in (result or (not alt_result)) && (alt_result or (not result))) ;
        result
    let mem_edge g e =
      let src = Edge.src e
      and tgt = Edge.tgt e
      in
        (* As mem_vertex is cheap, we first check whether g contains both the
           source and target vertices of e. *)
        if (mem_vertex g src) && (mem_vertex g tgt)
        then
          (* We check whether e is among the successor edges of the source of e. *)
          let (src_succs, _) = Vertex.edges src g in
            ESet.mem src_succs e
        else
          false

    let iter_vertices f g = VSet.iter f (Graph.vertices g)
    let fold_vertices f g = VSet.fold f (Graph.vertices g)
    let iter_edges f g =
      VSet.iter
        (fun v ->
           let (succs, _) = Vertex.edges v g
           in
             ESet.iter f succs)
        (Graph.vertices g)
    let fold_edges f g =
      VSet.fold
        (fun v ->
           let (succs, _) = Vertex.edges v g
           in
             ESet.fold f succs)
        (Graph.vertices g)

    let iter_succ f g v =
      assert (mem_vertex g v) ;
      let (succs, _) = Vertex.edges v g
      in
        ESet.iter f succs

    let fold_succ f g v =
      assert (mem_vertex g v) ;
      let (succs, _) = Vertex.edges v g
      in
        ESet.fold f succs

    let iter_pred f g v =
      assert (mem_vertex g v) ;
      let (_, preds) = Vertex.edges v g
      in
        ESet.iter f preds

    let fold_pred f g v =
      assert (mem_vertex g v) ;
      let (_, preds) = Vertex.edges v g
      in
        ESet.fold f preds

    let add_vertex g v =
      assert (not (mem_vertex g v)) ;
      Vertex.add_graph v g ;
      Graph.add_vertex g v

    let del_vertex g v =
      assert (mem_vertex g v) ;
      let update_succ e n =
        assert (Vertex.equal v (Edge.src e)) ;
        let tgt = Edge.tgt e
        in
          if Vertex.equal v tgt
          then
            n + 1
          else
            let (_, tgt_preds) = Vertex.edges tgt g
            in
              assert (ESet.mem tgt_preds e) ;
              ESet.remove tgt_preds e ;
              Graph.delta_nb_edges g (- 1) ;
              n
      and update_pred e n =
        assert (Vertex.equal v (Edge.tgt e)) ;
        let src = Edge.src e
        in
          if Vertex.equal v src
          then
            n + 1
          else
            let (src_succs, _) = Vertex.edges src g
            in
              assert (ESet.mem src_succs e) ;
              ESet.remove src_succs e ;
              Graph.delta_nb_edges g (- 1) ;
              n
      and (succs, preds) = Vertex.edges v g
      in
      let nb_selfs_from_succs = ESet.fold update_succ succs 0
      and nb_selfs_from_preds = ESet.fold update_pred preds 0
      in
        assert (nb_selfs_from_succs = nb_selfs_from_preds) ;
        Graph.del_vertex g v ;
        Graph.delta_nb_edges g (- nb_selfs_from_succs) ;
        Vertex.del_graph v g

    let add_edge g e =
      assert (not (mem_edge g e)) ;
      let src = Edge.src e
      and tgt = Edge.tgt e
      in
        if not (mem_vertex g src) then add_vertex g src ;
        if not (mem_vertex g tgt) then add_vertex g tgt ;
        Graph.delta_nb_edges g 1 ;
        let (src_succs, _) = Vertex.edges src g
        and (_, tgt_preds) = Vertex.edges tgt g
        in
          ESet.add src_succs e ;
          ESet.add tgt_preds e

    let del_edge g e =
      assert (mem_edge g e) ;
      let (src_succs, _) = Vertex.edges (Edge.src e) g
      and (_, tgt_preds) = Vertex.edges (Edge.tgt e) g
      in
        Graph.delta_nb_edges g (- 1) ;
        ESet.remove src_succs e ;
        ESet.remove tgt_preds e

    let print fmt g =
      let print_succ_edge fmt e =
        Format.fprintf fmt "@[%a@]@;»@ %d"
          Edge.Label.print (Edge.lbl e)
          (Vertex.id (Edge.tgt e))
      and print_pred_edge fmt e =
        Format.fprintf fmt "@[%a@]@;«@ %d"
          Edge.Label.print (Edge.lbl e)
          (Vertex.id (Edge.src e))
      in
        Format.fprintf fmt "@[<v>" ;
        Format.fprintf fmt "@[<v 3>%d vertices:" (nb_vertices g) ;
        iter_vertices (Format.fprintf fmt "@,@[<h>%a@]" Vertex.print) g ;
        Format.fprintf fmt "@]@," ;
        Format.fprintf fmt "@[<v 3>%d edges:" (nb_edges g) ;
        iter_edges (Format.fprintf fmt "@,@[<h>%a@]" Edge.print) g ;
        Format.fprintf fmt "@]@," ;
        Format.fprintf fmt "@[<v 3>Outbound edges of each vertex:" ;
        iter_vertices
          (fun v ->
             Format.fprintf fmt "@,@[<v 3>%d :" (Vertex.id v) ;
             iter_succ
               (Format.fprintf fmt "@,» @[<3>%a@]" print_succ_edge) g v ;
             Format.fprintf fmt "@]")
          g ;
        Format.fprintf fmt "@]@," ;
        Format.fprintf fmt "@[<v 3>Inbound edges of each vertex:" ;
        iter_vertices
          (fun v ->
             Format.fprintf fmt "@,@[<v 3>%d :" (Vertex.id v) ;
             iter_pred
               (Format.fprintf fmt "@,« @[<3>%a@]" print_pred_edge) g v ;
             Format.fprintf fmt "@]")
          g ;
        Format.fprintf fmt "@]@]"
  end

  module MakeMutable (Graph : P) =
  struct
    module Vertex = Graph.Vertex
    module Edge = Graph.Edge
    type t = Graph.t ref
    let create _ = ref Graph.empty
    let add_vertex g v = g := (Graph.add_vertex !g v)
    let del_vertex g v = g := (Graph.del_vertex !g v)
    let add_edge g e = g := (Graph.add_edge !g e)
    let del_edge g e = g := (Graph.del_edge !g e)
    let nb_vertices g = Graph.nb_vertices !g
    let nb_edges g = Graph.nb_edges !g
    let mem_vertex g v = Graph.mem_vertex !g v
    let mem_edge g e = Graph.mem_edge !g e
    let iter_vertices f g = Graph.iter_vertices f !g
    let fold_vertices f g = Graph.fold_vertices f !g
    let iter_edges f g = Graph.iter_edges f !g
    let fold_edges f g = Graph.fold_edges f !g
    let iter_succ f g = Graph.iter_succ f !g
    let fold_succ f g = Graph.fold_succ f !g
    let iter_pred f g = Graph.iter_pred f !g
    let fold_pred f g = Graph.fold_pred f !g
    let print fmt g = Graph.print fmt !g
  end
end
