(* $Id: MultiGraph.mli 755 2009-12-02 17:15:54Z sutre $ *)

(*
 * Copyright (c) 2009 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * Labeled Directed Multi-Graphs.
 *
 * This module provides both imperative and persistent implementations for
 * labeled directed multi-graphs (labeled directed graphs with multiple labeled
 * edges).  The interfaces are inspired from the [Sig] module of the Ocamlgraph
 * library.
 *
 * This module is based on the following formal definition for labeled directed
 * multi-graphs.  Given a set {i L} of {e vertex labels} and a set {i K} of {e
 * edge labels}, a {e labeled directed multi-graph} (called {e multi-graph} for
 * short) is any 6-tuple {i G = (V, E, l, k, s, t)} where:
 * {ul
 * {li {i V} is a finite set of {e vertices},}
 * {li {i E} is a finite set of {e edges},}
 * {li {i l} is a {e labeling} function from {i V} to {i L},}
 * {li {i k} is a {e labeling} function from {i E} to {i K},}
 * {li {i s} and {i t} are functions from {i E} to {i V} mapping each edge to a
 *     {e source vertex} {i s(e)} and a {e target vertex} {i t(e)}.}
 * }
 *
 * Observe that there may be multiple edges with the same end nodes and the same
 * label in a labeled directed multi-graph.  This contrasts with (standard)
 * labeled directed graphs, where the set of edges is a subset of {i V} × {i K}
 * × {i V}.
 *
 * On the contrary to labeled directed graphs, operations on edges (membership
 * testing, addition and removal of edges) can be performed in labeled directed
 * multi-graphs without equality tests over edge labels.  Therefore the functors
 * in this module can be applied on edge label types that are not equipped with
 * equality or comparison functions (the Ocamlgraph library requires edge labels
 * to be comparable).
 *
 * The implementations provided in this module conform to the signature {! G}.
 * In an implementation of this signature, all multi-graphs share the same
 * global set of (possible) vertices (type [Vertex.t]) and the same global set
 * of (possible) edges (type [Edge.t]).  Equality, comparison and hash functions
 * on vertices and edges are provided in the corresponding modules [Vertex] and
 * [Edge].
 *
 * This module follows an aggressive approach for better performance.  All
 * functions perform assertion checking on their arguments, specified in this
 * documentation by the {b Precondition} tag.  This assertion checking is
 * implemented with the [assert] construct, and, hence, it is turned off by the
 * -noassert compiler option.
 *
 * @see <http://en.wikipedia.org/wiki/Multigraph> Multigraph
 * @see <http://ocamlgraph.lri.fr/> Ocamlgraph
 *)


(**
 * Signature specifying the common interface to all multi-graph implementations.
 *)
module type G =
sig
  (** {2 Multi-Graph Structure} *) (*******************************************)

  (**
   * The printable type of multi-graphs.
   *)
  include Print.PRINTABLE_TYPE

  (**
   * Module to manage vertices.  Vertices are values of type [Vertex.t] and are
   * shared by all multi-graphs.
   *)
  module Vertex :
  sig
    (**
     * The printable type of vertices.
     *)
    include Print.PRINTABLE_TYPE

    (**
     * The printable type of vertex labels.
     *)
    module Label : Print.PRINTABLE_TYPE

    (**
     * [create l] creates a new vertex labeled with [l].  The resulting vertex
     * is distinct (w.r.t. to the equality predicate below) from all other
     * vertices managed by this module.
     *)
    val create : Label.t -> t

    (**
     * [lbl v] returns the label of the vertex [v].
     *)
    val lbl : t -> Label.t

    (**
     * Vertices are ordered (see [Set.OrderedType] or [Map.OrderedType]) and
     * hashable (see [Hashtbl.HashedType]).
     *)

    (**
     * Total ordering over vertices.
     *)
    val compare : t -> t -> int

    (**
     * [equal v1 v2] tests whether v1 and v2 are the same vertex, i.e. represent
     * the same element in the set of vertices.  There may be multiple vertices
     * with the same label.
     *)
    val equal : t -> t -> bool

    (**
     * Hashing function on vertices.
     *)
    val hash : t -> int
  end

  (**
   * Module to manage edges.  Edges are values of type [Edge.t] and are shared
   * by all multi-graphs.
   *)
  module Edge :
  sig
    (**
     * The printable type of edges.
     *)
    include Print.PRINTABLE_TYPE

    (**
     * The printable type of edge labels.
     *)
    module Label : Print.PRINTABLE_TYPE

    (**
     * [create v l v'] creates a new edge from [v] to [v'] labeled with [l].
     * The resulting edge is distinct (w.r.t. to the equality predicate below)
     * from all other edges managed by this module.
     *)
    val create : Vertex.t -> Label.t -> Vertex.t -> t

    (**
     * [src e] returns the source vertex of the edge [e].
     *)
    val src : t -> Vertex.t

    (**
     * [tgt e] returns the target vertex of the edge [e].
     *)
    val tgt : t -> Vertex.t

    (**
     * [lbl e] returns the label of the edge [e].
     *)
    val lbl : t -> Label.t

    (**
     * Edges are ordered (see [Set.OrderedType] or [Map.OrderedType]) and
     * hashable (see [Hashtbl.HashedType]).
     *)

    (**
     * Total ordering over edges.
     *)
    val compare : t -> t -> int

    (**
     * [equal e1 e2] tests whether e1 and e2 are the same edge, i.e. represent
     * the same element in the set of edges.  There may be multiple edges with
     * the same end nodes and the same label.
     *)
    val equal : t -> t -> bool

    (**
     * Hashing function on edges.
     *)
    val hash : t -> int
  end

  (** {2 Size Functions} *) (**************************************************)

  (**
   * [nb_vertices g] returns the number of vertices in the multi-graph [g].
   *)
  val nb_vertices : t -> int

  (**
   * [nb_edges g] returns the number of edges in the multi-graph [g].
   *)
  val nb_edges : t -> int

  (** {2 Membership Functions} *) (********************************************)

  (**
   * [mem_vertex g v] tests whether the multi-graph [g] contains the vertex [v].
   *)
  val mem_vertex : t -> Vertex.t -> bool

  (**
   * [mem_edge g e] tests whether the multi-graph [g] contains the edge [e].
   *)
  val mem_edge : t -> Edge.t -> bool

  (** {2 Iterators} *)  (******************************************************)

  (**
   * The following iterators apply a given function [f] to some vertices (resp.
   * to some edges) of a given multi-graph.  Each vertex (resp. each edge) is
   * passed at most once to the function [f].  The order in which the vertices
   * (resp. the edges) are passed to [f] is unspecified.
   *)

  (** {4 Global Iterators} *) (*----------------------------------------------*)

  (**
   * [iter_vertices f g] applies [f] in turn to all vertices of [g].
   *)
  val iter_vertices : (Vertex.t -> unit) -> t -> unit

  (**
   * [fold_vertices f g a] returns [(f vN ... (f v2 (f v1 a))...)], where [v1,
   * ..., vN] are the vertices of [g].
   *)
  val fold_vertices : (Vertex.t -> 'a -> 'a) -> t -> 'a -> 'a

  (**
   * [iter_edges f g] applies [f] in turn to all edges of [g].
   *)
  val iter_edges : (Edge.t -> unit) -> t -> unit

  (**
   * [fold_edges f g a] returns [(f eN ... (f e2 (f e1 a))...)], where [e1, ...,
   * eN] are the edges of [g].
   *)
  val fold_edges : (Edge.t -> 'a -> 'a) -> t -> 'a -> 'a

  (** {4 Local Iterators} *) (*-----------------------------------------------*)

  (**
   * [iter_succ f g v] applies [f] in turn to all {e outbound} edges of [v] in
   * [g] (i.e. all edges in [g] with source [E.src e] equal to [v]).
   * {L {b Precondition} [v] is a vertex contained in [g].}
   *)
  val iter_succ : (Edge.t -> unit) -> t -> Vertex.t -> unit

  (**
   * [fold_succ f g v a] returns [(f eN ... (f e2 (f e1 a))...)], where [e1,
   * ..., eN] are the {e outbound} edges of [v] in [g] (i.e. the edges in [g]
   * with source [E.src e] equal to [v]).
   * {L {b Precondition} [v] is a vertex contained in [g].}
   *)
  val fold_succ : (Edge.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a

  (**
   * [iter_pred f g v] applies [f] in turn to all {e inbound} edges of [v] in
   * [g] (i.e. all edges in [g] with target [E.tgt e] equal to [v]).
   * {L {b Precondition} [v] is a vertex contained in [g].}
   *)
  val iter_pred : (Edge.t -> unit) -> t -> Vertex.t -> unit

  (**
   * [fold_pred f g v a] returns [(f eN ... (f e2 (f e1 a))...)], where [e1,
   * ..., eN] are the {e inbound} edges of [v] in [g] (i.e. the edges in [g]
   * with target [E.tgt e] equal to [v]).
   * {L {b Precondition} [v] is a vertex contained in [g].}
   *)
  val fold_pred : (Edge.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a
end


(**
 * Interface for persistent implementations of labeled directed multi-graphs.
 *)
module type P =
sig
  include G

  (**
   * The empty multi-graph.
   *)
  val empty : t

  (**
   * [add_vertex g v] adds the vertex [v] to the multi-graph [g].
   * {L {b Precondition} [v] is not already contained in [g].}
   *)
  val add_vertex : t -> Vertex.t -> t

  (**
   * [del_vertex g v] removes the vertex [v] from the multi-graph [g].  All
   * edges [e] with source [E.src e] or target [E.tgt e] equal to [v] are also
   * removed from [g].
   * {L {b Precondition} [v] is contained in [g].}
   *)
  val del_vertex : t -> Vertex.t -> t

  (**
   * [add_edge g e] adds the edge [e] to the multi-graph [g].  The source
   * [E.src e] and target [E.tgt e] vertices of [e] are also added to [g] if
   * they are not already in [g].
   * {L {b Precondition} [e] is not already contained in [g].}
   *)
  val add_edge : t -> Edge.t -> t

  (**
   * [del_edge g e] removes the edge [e] from the multi-graph [g].
   * {L {b Precondition} [e] is contained in [g].}
   *)
  val del_edge : t -> Edge.t -> t
end


(**
 * Interface for imperative implementations of labeled directed multi-graphs.
 *)
module type I =
sig
  include G

  (**
   * [create n] creates a new, empty multi-graph, with initial size [n].  For
   * best results, [n] should be on the order of the expected number of vertices
   * that will be in the multi-graph.  The multi-graph grows as needed, so [n]
   * is just an initial guess.
   *)
  val create : int -> t

  (**
   * [add_vertex g v] adds the vertex [v] to the multi-graph [g].
   * {L {b Precondition} [v] is not already contained in [g].}
   *)
  val add_vertex : t -> Vertex.t -> unit

  (**
   * [del_vertex g v] removes the vertex [v] from the multi-graph [g].  All
   * edges [e] with source [E.src e] or target [E.tgt e] equal to [v] are also
   * removed from [g].
   * {L {b Precondition} [v] is contained in [g].}
   *)
  val del_vertex : t -> Vertex.t -> unit

  (**
   * [add_edge g e] adds the edge [e] to the multi-graph [g].  The source
   * [E.src e] and target [E.tgt e] vertices of [e] are also added to [g] if
   * they are not already in [g].
   * {L {b Precondition} [e] is not already contained in [g].}
   *)
  val add_edge : t -> Edge.t -> unit

  (**
   * [del_edge g e] removes the edge [e] from the multi-graph [g].
   * {L {b Precondition} [e] is contained in [g].}
   *)
  val del_edge : t -> Edge.t -> unit
end


(**
 * Persistent implementations of labeled directed multi-graphs.  All operations
 * over multi-graphs are purely applicative (no side-effects).
 *)
module Persistent :
sig
  module Make (VLabel : Print.PRINTABLE_TYPE) (ELabel : Print.PRINTABLE_TYPE) :
    P with module Vertex.Label = VLabel
      and  module Edge.Label = ELabel
end


(**
 * Imperative implementations of labeled directed multi-graphs.  All operations
 * over multi-graphs are performed with in-place modification.
 *)
module Imperative :
sig
  module Make (VLabel : Print.PRINTABLE_TYPE) (ELabel : Print.PRINTABLE_TYPE) :
    I with module Vertex.Label = VLabel
      and  module Edge.Label = ELabel

  module Make2 (VLabel : Print.PRINTABLE_TYPE) (ELabel : Print.PRINTABLE_TYPE) :
    I with module Vertex.Label = VLabel
      and  module Edge.Label = ELabel

  module MakeMutable (Graph : P) :
    I with module Vertex = Graph.Vertex
      and  module Edge = Graph.Edge
end
