(* $Id: Search.ml 739 2009-07-18 17:09:41Z sutre $ *)

(*
 * Copyright (c) 2009 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(* Paths in Labeled Directed Multi-Graphs. *)


open Print
open Collection


(* New exception type for errors that are internal to this module, such as cases
   in pattern matching that should not arise. *)
exception InternalError


module type G =
sig
  type t
  module Vertex :
  sig
    include PRINTABLE_TYPE
    val compare : t -> t -> int
  end
  module Edge :
  sig
    include PRINTABLE_TYPE
    val src : t -> Vertex.t
    val tgt : t -> Vertex.t
  end
  val mem_vertex : t -> Vertex.t -> bool   (* used in assertions only *)
  val iter_succ : (Edge.t -> unit) -> t -> Vertex.t -> unit
  val fold_succ : (Edge.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a
  val iter_pred : (Edge.t -> unit) -> t -> Vertex.t -> unit
  val fold_pred : (Edge.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a
end


module Make (G : G) =
struct
  (* The printable type of paths. *)
  module Path =
  struct
    type t = G.Edge.t list
    let print =
      Print.list_printer_from_printer ",@ " G.Edge.print
  end

  (* Type of possible results of the graph exploration functions. *)
  type traversal =
      Forward_path   of Path.t
    | Backward_path  of Path.t
    | Forward_reach  of G.Vertex.t list
    | Backward_reach of G.Vertex.t list

  (* Sets of vertices. *)
  module VSet = Set.Persistent.MakeTreeSet (G.Vertex)

  (* Maps over vertices. *)
  module VMap = Map.Persistent.MakeTreeMap (G.Vertex)

  (* Function to convert a list of vertices into a set of vertices. *)
  let set_of_vertices l =
    List.fold_left (fun s v -> VSet.add s v) VSet.empty l

  (*
   * The following graph exploration implementations use a visit map to record
   * visited edges.  Visit maps are association tables with keys of type
   * G.Vertex.t and values of type (G.Edge.t option).  A vertex v is is the map
   * if it has been visited, and the value associated with v is either:
   * - None if v is a starting vertex, or
   * - Some e if the e is the edge that was taken to visit v.
   *)

  (* Function to reconstruct a path from a visit map. *)
  let extract_path prev e visit =
    let rec construct_path e acc =
      try
        match (VMap.find visit (prev e)) with
            None -> e :: acc
          | Some e' -> construct_path e' (e :: acc)
        with
            Not_found -> raise InternalError
    in
      construct_path e []

  (* Type of possible results of the generic graph exploration functions. *)
  type generic_traversal =
      Path  of Path.t
    | Reach of G.Vertex.t list

  module type WORKLIST =
  sig
    type 'a t
    val create : unit -> 'a t
    val push : 'a -> 'a t -> unit
    val pop : 'a t -> 'a
    val is_empty : 'a t -> bool
  end

  (* Generic functor providing a graph exploration function with a work
     list implemented by the given module [W]. *)
  module Explore (W : WORKLIST) =
  struct
    (*
     * Generic function for graph exploration, with the following arguments:
     * - g:      the graph
     * - prev:   function to get the previous vertex of a given edge
     * - next:   function to get the next vertex of a given edge
     * - iter:   function to iterate over the next edges to explore
     * - start:  the vertices that we look for, as a list
     * - goal:   the vertices that we look for, as a list
     *)
    let search g prev next iter start goal =
      assert (List.for_all (G.mem_vertex g) start) ;
      assert (List.for_all (G.mem_vertex g) goal) ;
      let goal_set = set_of_vertices goal in
      let rec do_it visit work =
        if W.is_empty work
        then
          Reach (List.rev (VMap.fold (fun v _ l -> v::l) visit []))
        else
          let e = W.pop work
          in
          (* Logger.notice "@[Visiting:@ %a@]@." G.Edge.print e ; *)
          let v = (next e) in
            if VMap.mem visit v
            then
              do_it visit work
            else
              if VSet.mem goal_set v
              then
                Path (extract_path prev e visit)
              else
                let visit' = VMap.add visit v (Some e)
                in
                  iter (fun e -> W.push e work) g v ;
                  do_it visit' work
      in
        (* There should not be any vertex that is both in start and goal. *)
        assert (not (List.exists (VSet.mem goal_set) start)) ;
        (* starting work list and visit map *)
        let work = W.create () in
        let visit =
          List.fold_left
            (fun vm v ->
               iter (fun e -> W.push e work) g v ;
               VMap.add vm v None)
            VMap.empty
            start
        in
          do_it visit work
  end

  module Dfs = Explore (Stack)
  module Bfs = Explore (Queue)

  (*
   * In the following fwd/bwd DFS exploration functions, the iterator passed to
   * Dfs.search ensures that we iterate in reversed order.  Indeed, we want the
   * first successors/predecessors to be pushed last, so that they are processed
   * first.  We also reverse the list of starting vertices for the same reason.
   *)

  (* Forward depth-first exploration. *)
  let fwd_dfs g initial final =
    match
      Dfs.search
        g
        G.Edge.src
        G.Edge.tgt
        (fun f g v ->
           List.iter f (G.fold_succ (fun e l -> e::l) g v []))
        (List.rev initial)
        final
    with
        Path p  -> Forward_path p
      | Reach s -> Forward_reach s

  (* Forward breadth-first exploration. *)
  let fwd_bfs g initial final =
    match
      Bfs.search
        g
        G.Edge.src
        G.Edge.tgt
        G.iter_succ
        initial
        final
    with
        Path p  -> Forward_path p
      | Reach s -> Forward_reach s

  (* Backward depth-first exploration. *)
  let bwd_dfs g initial final =
    match
      Dfs.search
        g
        G.Edge.tgt
        G.Edge.src
        (fun f g v ->
           List.iter f (G.fold_pred (fun e l -> e::l) g v []))
        (List.rev final)
        initial
    with
        Path p  -> Backward_path p
      | Reach s -> Backward_reach s

  (* Backward breadth-first exploration. *)
  let bwd_bfs g initial final =
    match
      Bfs.search
        g
        G.Edge.tgt
        G.Edge.src
        G.iter_pred
        final
        initial
    with
        Path p  -> Backward_path p
      | Reach s -> Backward_reach s
end
