(* $Id: Search.mli 755 2009-12-02 17:15:54Z sutre $ *)

(*
 * Copyright (c) 2009 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * Exploration of Labeled Directed Multi-Graphs.
 *
 * This module provides a printable type for paths in labeled directed
 * multi-graphs, together with graph exploration algorithms: depth-first search
 * and breadth-first search, forward and backward versions.  The exploration
 * functions in this module either extract a path from an initial node to a
 * final node when such a path exists, or return the set of reachable vertices
 * otherwise.
 *
 * Multi-graphs are described in the documentation of the {! MultiGraph} module.
 *
 * This module follows an aggressive approach for better performance.  All
 * functions perform assertion checking on their arguments, specified in this
 * documentation by the {b Precondition} tag.  This assertion checking is
 * implemented with the [assert] construct, and, hence, it is turned off by the
 * -noassert compiler option.
 *)


(**
 * Input signature of the functor {! Make}.  This minimal multi-graph interface
 * is a ``subset'' of the {! MultiGraph.G} signature.
 *)
module type G =
sig
  type t
  module Vertex :
  sig
    include Print.PRINTABLE_TYPE
    val compare : t -> t -> int
  end
  module Edge :
  sig
    include Print.PRINTABLE_TYPE
    val src : t -> Vertex.t
    val tgt : t -> Vertex.t
  end
  val mem_vertex : t -> Vertex.t -> bool   (* used in assertions only *)
  val iter_succ : (Edge.t -> unit) -> t -> Vertex.t -> unit
  val fold_succ : (Edge.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a
  val iter_pred : (Edge.t -> unit) -> t -> Vertex.t -> unit
  val fold_pred : (Edge.t -> 'a -> 'a) -> t -> Vertex.t -> 'a -> 'a
end


(**
 * Functor providing a printable type for paths together with graph exploration
 * functions for a given multi-graph implementation.
 *)
module Make (G : G) :
sig
  (**
   * The printable type of paths.  A path is a list [[e1; ...; eN]] of edges.
   *)
  module Path : Print.PRINTABLE_TYPE with type t = G.Edge.t list

  (**
   * Type of possible results of the graph exploration functions provided by
   * this module.
   * {ul
   * {li [Forward_path p] : the path [p] verifies [G.Edge.tgt e = G.Edge.src e']
   * for any two consecutive edges [e] and [e'] in [p].}
   * {li [Backward_path p] : the path [p] verifies [G.Edge.tgt e' = G.Edge.src e]
   * for any two consecutive edges [e] and [e'] in [p].}
   * {li [Forward_reach s] : the set [s] of vertices (given as a list) contains
   *     all its successors.}
   * {li [Backward_reach s] : the set [s] of vertices (given as a list) contains
   *     all its predecessors.}
   * }
   *
   * Note that if the traversal result is [Backward_path p] then the path [p]
   * must be reversed (e.g. with [List.rev p]) to get it in the proper order.
   *)
  type traversal =
      Forward_path   of Path.t
        (** forward path, verifying [G.Edge.tgt e = G.Edge.src e'] for any two
          consecutive edges [e] and [e']. *)
    | Backward_path  of Path.t
        (** backward path, verifying [G.Edge.tgt e' = G.Edge.src e] for any two
          consecutive edges [e] and [e']. *)
    | Forward_reach  of G.Vertex.t list
        (** forward reachability set, closed under successor. *)
    | Backward_reach of G.Vertex.t list
        (** backward reachability set, closed under predecessor. *)

  (**
   * The following graph exploration functions look for a path from a vertex in
   * a given list [initial] of vertices to a vertex in a given list [final] of
   * vertices.  The lists [initial] and [final] are assumed to be disjoint.
   * Each exploration function returns:
   * {ul
   * {li either [*_path p] if a path [p] was found,}
   * {li or [*_reach s] otherwise, where [s] is the list of reachable vertices
   *     ([s] is the list of all visited vertices).}
   * }
   *
   * Forward exploration functions return only [Forward_*] results and backward
   * exploration functions return only [Backward_*] results.
   *
   * To simply compute the forward (resp. backward) reachability set of a given
   * list of initial (resp. final) vertices, use an empty list of final (resp.
   * initial) vertices.
   *)

  (**
   * [fwd_dfs g initial final] looks for a path from [initial] to [final], by a
   * forward depth-first exploration of the multi-graph [g].
   * {L {b Precondition} [initial] and [final] are disjoint, and they are both
   * contained in [g].}
   *)
  val fwd_dfs : G.t -> G.Vertex.t list -> G.Vertex.t list -> traversal

  (**
   * [fwd_bfs g initial final] looks for a path from [initial] to [final], by a
   * forward breadth-first exploration of the multi-graph [g].
   * {L {b Precondition} [initial] and [final] are disjoint, and they are both
   * contained in [g].}
   *)
  val fwd_bfs : G.t -> G.Vertex.t list -> G.Vertex.t list -> traversal

  (**
   * [bwd_dfs g initial final] looks for a path from [initial] to [final], by a
   * backward depth-first exploration of the multi-graph [g].
   * {L {b Precondition} [initial] and [final] are disjoint, and they are both
   * contained in [g].}
   *)
  val bwd_dfs : G.t -> G.Vertex.t list -> G.Vertex.t list -> traversal

  (**
   * [bwd_bfs g initial final] looks for a path from [initial] to [final], by a
   * backward breadth-first exploration of the multi-graph [g].
   * {L {b Precondition} [initial] and [final] are disjoint, and they are both
   * contained in [g].}
   *)
  val bwd_bfs : G.t -> G.Vertex.t list -> G.Vertex.t list -> traversal
end
