(* $Id: Sequence.mli 1178 2011-10-13 09:20:22Z gsutre $ *)

(*
 * Copyright (c) 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


(**
 * Finite sequences.
 *
 * To avoid unnecessary list reversals in some contexts (typically, successive
 * list iterations performed in unrelated functions), this module defines a new
 * type (['a t]) that combines a list and a direction, and provides associated
 * helper functions.  This new type also contains the list's length.  Users of
 * this module should maintain the consistency accross values of type ['a t].
 * In particular, [(l, Backward, n)] and [((List.rev l), Forward, n)] denote the
 * same finite sequence.
 *)


type direction = Forward | Backward

type 'a t = 'a list * direction * int

val print : (unit, Format.formatter, unit) format ->
  (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit

val is_empty : 'a t -> bool
val length : 'a t -> int
val rev : 'a t -> 'a t
val forward : 'a t -> 'a t
val backward : 'a t -> 'a t

(**
 * [map f s] returns the sequence obtained by applying [f] on each element of
 * [s].  Its implementation is based on {!List.rev_map}.
 *)
val map : ('a -> 'b) -> 'a t -> 'b t

(**
 * [filter p s] returns the sequence obtained from [s] by keeping only the
 * elements satisfying [p].
 *)
val filter : ('a -> bool) -> 'a t -> 'a t

(**
 * [to_array s] returns a fresh array containing the elements of [s].
 *)
val to_array : 'a t -> 'a array

(**
 * [of_array a] is equivalent to [(Array.to_list a, Forward, Array.length a)].
 *)
val of_array : 'a array -> 'a t
