(* $Id: Uid.ml 1265 2011-10-19 22:51:55Z gsutre $ *)

(*
 * Copyright (c) 2009, 2011 CNRS & Université Bordeaux 1.
 *
 * Author(s): Grégoire Sutre <gregoire.sutre@labri.fr>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *)


open Print


exception Overflow

(*
 * Note: this functor ensures that created values are totally ordered by their
 * creation time.  The first value has an id equal to zero (instead of min_int,
 * for better display).  The last possible value has an id equal to max_int.
 * The exception overflow is raised by create after (max_int + 1) calls.
 *)

(*
 * The next_id counter is local to the result module of the Make functor, hence,
 * the ids are not unique across multiple applications of the functor.  If
 * globally unique ids are required, then the next_id counter should be moved
 * outside the Make functor.  However, equality and comparison functions raise
 * an assertion violation if they are applied on two physically distinct values
 * that have the same id.  This is illustrated with the following example:
 *
 * # #load "Util.cmo";;
 * # open Util;;
 * # module M = struct type t = string let print = Format.pp_print_string end;;
 * # module U1 = Uid.Make (M);;
 * # module U2 = Uid.Make (M);;
 * # let x1 = U1.create "one";;
 * # let y1 = U1.create "one";;
 * # let x2 = U2.create "one";;
 * # U1.id x1;;
 * - : int = 0
 * # U1.id y1;;
 * - : int = 1
 * # U1.id x2;;
 * - : int = 0
 * # U1.equal x1 x2;;
 * Exception: Assert_failure ("Uid.ml", 60, 4).
 * # U1.equal y1 x2;;
 * - : bool = false
 *)

module Make (T : PRINTABLE_TYPE) =
struct
  let next_id = ref 0

  type t = int * T.t

  let create x =
    let n = !next_id
    in
      if (n < 0)
      then
        (* the last created value had an id of max_int *)
        raise Overflow
      else
        begin
          incr next_id ;
          (n, x)
        end

  let id (n, _) = n
  let value (_, x) = x

  let compare (n1, x1) (n2, x2) =
    assert ((n1 != n2) or (x1 == x2)) ;
    compare n1 n2

  let equal (n1, x1) (n2, x2) =
    assert ((n1 != n2) or (x1 == x2)) ;
    n1 = n2

  let hash (n, a) = n

  let print fmt (n, a) =
    Format.fprintf fmt "[%d]@;@[%a@]" n T.print a

  let lift_una_op f x = create (f (value x))
  let lift_bin_op f x y = create (f (value x) (value y))

  let lift_una_fun f x = (f (value x))
  let lift_bin_fun f x y = (f (value x) (value y))
end
